# Basic Funcs
# Thomams Johnson III
# 07-23-2023

library(tables)

# Directory search function to safegauard against code running outside of
# the project directory. This is not foolproof. It will only search from the 
# home directory or "~" unless altered.
directory_search = function(start_directory = "~/")
{
  
  # check for the necessary files
  check_files = function(project_directory)
  {
    file_bool_rproj = file.exists(paste0(start_directory,"/SDC-Comparisons.Rproj"))
    # Prepare to search for code files
    code_dir = "/Code/"
    necessary_code_files = c("Basic-Funcs.R", "Estimator-Evaluation.R",
                        "CE-data-cleaning.R", "GADP-from-scratch.R",
                        "Iterative-Model-Gen-Code.R")
    # Prepare to search for report files
    reports_dir = "/Code/Reports/"
    necessary_report_files = c("More-tuned-SDC-GAPD-CGAPD-Comparisons-01-29-2025.qmd",
                               "CE-Real-Data-Tables-fincbtax-salaryx-tuned-sims.qmd")
    
    # check the code files are present
    code_files_bool = logical(length(necessary_code_files))
    # for loop to iterate for code files
    for(index_bool in 1:length(necessary_code_files))
    {
      if(file.exists(paste0(project_directory, code_dir, 
                            necessary_code_files[index_bool])))
      {
        code_files_bool[index_bool] = TRUE
      }
    }
    if(any(code_files_bool) == FALSE)
    {
      print("Code files were not detected.")
      print("Please manually specify the project directory (`base_dir` = path/to/SDC-Comparisons-Submission/).")
    }
    
    # logical vector for managing detection of report files
    report_files_bool = logical(length(necessary_report_files))
    
    # for loop to iterate for report files
    for(index_bool in 1:length(necessary_report_files))
    {
      if(file.exists(paste0(project_directory, reports_dir,
                            necessary_report_files[index_bool])))
      {
        report_files_bool[index_bool] = TRUE
      }
        
    }
    if(any(report_files_bool) == FALSE)
    {
      print("Code files were not detected.")
      print("Please manually specify the project directory (`base_dir` = path/to/SDC-Comparisons-Submission/).")
    }
    # boolean to determine if the files were not found
    # TRUE if all files are found, FLASE otherwise
    missing_files_bool = all(c(code_files_bool, report_files_bool, file_bool_rproj) == TRUE)
    
    return(missing_files_bool)
  }
  # Attempt to get the project directory via a recursive search
  obtain_project_directory = function(suggested_directory)
  {
    potential_directories =list.dirs(suggested_directory)
    # indices of the directories that belong to the project
    project_directory_index = which(stringr::str_detect(potential_directories, "SDC-Comparisons-Submission$") == TRUE)
   # Get the project directory
    project_directory = potential_directories[project_directory_index]
    
    # ensure that the necessary files are present
    file_presence_bool = check_files(project_directory)
    
    if(file_presence_bool == TRUE)
    {
      # Do nothing project directory will be returned
    }
    else
    {
      print("Could not find directory for the project. Please manually specify in `base_dir`.")
      quit(save = "no")
    }
    return(project_directory)
  }
  
  
  if (is.na(start_directory) == TRUE)
  {
    # If nothing specified, start in the home directory
    start_directory = "~/"
  }
  
  found_project_directory = NA
  
  # If the substring of the project directory is found
  if(stringr::str_detect(start_directory, "SDC-Comparisons-Submission"))
  {
    files_all_present_bool = check_files(start_directory)
    
    if(files_all_present_bool == TRUE)
    {
      # When starting directory is project directory
      print("Relevant files found.")
      found_project_directory = start_directory
      return(found_project_directory)
    }
    else
    {
      print("Relevant files are not found.")
      # start in home directory
      found_project_directory = obtain_project_directory("~")
      
      return(found_project_directory)
    }
  }
  # If not in anything that could be considered the project directory
  else
  {
    found_project_directory = obtain_project_directory(start_directory)
    print(found_project_directory)
    
    if(is.na(found_project_directory) == TRUE | is.character(found_project_directory) == FALSE)
    {
      found_project_directory = obtain_project_directory("~/")
    }
    
    return(found_project_directory)
  }
  print("An unexpected situation has emerged. None of the previous cases were met.")
  print("Attempting to find project directory. If the project directory is not found")
  print("it will have to be manually specified in `base_dir`.")
  found_project_directory = obtain_project_directory(start_directory)
  
  if(is.na(found_project_directory) == TRUE | is.character(found_project_directory) == FALSE)
  {
    found_project_directory = obtain_project_directory("~/")
  }
  
  return(found_project_directory)
}


# Produce a CSV from from the results
produce_csv = function(created_object, base_str, file_name_str)
{
  write_csv(created_object, paste0(base_str, file_name_str))
  return(file_name_str)
}
# Produce a CSV from the results when test set and results data is saved
produce_csv_ver_2 = 
  function(created_object, base_str, file_name_str)
  {
    tabular_results = created_object$output
    write_csv(tabular_results, paste0(base_str, file_name_str))
    return(file_name_str)
  }
# Save the test set specifically
save_test_set = 
  function(created_object, base_str, file_name_str)
  {
    test_set = created_object$testing_set
    write_csv(test_set, paste0(base_str, file_name_str))
    return(file_name_str)
  }

# Used to calculate the MSER
# Permutation type is for "gadp" or "cgadp"
calculate_MSER = function(x_df, permutation_type = "gadp")
{
  mser_dataframe = x_df %>% 
    pivot_wider(names_from = !!as.name(permutation_type), values_from = "mse") %>%
    group_by(rho, sample_size, estimator, noise, theta) %>% 
    summarise(Yes = round(mean(Yes),3), No = round(mean(No),3)) %>% ungroup() %>% # Convert the permutation MSEs to a yes/no column for calculating MSER
    mutate(MSER = round(Yes / No,3))
  return(mser_dataframe)
}

# For iteratively generated training set
calculate_MSER.iter = function(x_df, permutation_type = "gadp")
{
  mser_dataframe = x_df %>% 
    pivot_wider(names_from = !!as.name(permutation_type), values_from = "mse") %>%
    group_by(rho, training_size, testing_size, estimator, noise, theta) %>% 
    rename(training = training_size, testing = testing_size) %>%
    summarise(Yes = round(mean(Yes),3), No = round(mean(No),3)) %>% ungroup() %>% # Convert the permutation MSEs to a yes/no column for calculating MSER
    mutate(MSER = round(Yes / No,3))
  return(mser_dataframe)
}


calculate_RMSER = function (x_df,  permutation_type = "gadp")
{
  rmser_df = x_df %>%
    pivot_wider(names_from =!!as.name(permutation_type), values_from = "mse") %>%
    group_by(rho, sample_size, estimator, noise, theta) %>% 
    summarise(Yes = round(mean(Yes^(1/2)),3), No = round(mean(No^(1/2)),3)) %>% ungroup() %>% # Convert the permutation RMSEs to a yes/no column for calculating RMSER
    mutate(RMSER = round(Yes / No,3))
  return(rmser_df)
}

# For iteratively gerated training set
calculate_RMSER.iter = function (x_df,  permutation_type = "gadp")
{
  rmser_df = x_df %>%
    pivot_wider(names_from =!!as.name(permutation_type), values_from = "mse") %>%
    group_by(rho, training_size, testing_size, estimator, noise, theta) %>% 
    rename(training = training_size, testing = testing_size) %>%
    summarise(Yes = round(mean(Yes^(1/2)),3), No = round(mean(No^(1/2)),3)) %>% ungroup() %>% # Convert the permutation RMSEs to a yes/no column for calculating RMSER
    mutate(RMSER = round(Yes / No,3))
  return(rmser_df)
}


# transform the data to get the MSE
organize_mse_results = function(mse_df, permutation_method = "gadp")
{
  if(permutation_method == "gadp")
  {
    transformed_mse_df = mse_df %>% rename(SE_equal = Stacked_equal,
                                           SE_prop = Stacked_prop) %>%
      pivot_longer(cols = 6:15, names_to = "estimator", values_to = "mse") %>%
      mutate(noise = ifelse(y == 1, "0.2", "0.4"),
             gadp = ifelse(gadp == 1, "Yes", "No"))
  }
  else if (permutation_method == "cgadp")
  {
    transformed_mse_df = mse_df %>% rename(SE_equal = Stacked_equal,
                                           SE_prop = Stacked_prop) %>%
      pivot_longer(cols = 6:15, names_to = "estimator", values_to = "mse") %>%
      mutate(noise = ifelse(y == 1, "0.2", "0.4"),
             cgadp = ifelse(cgadp == 1, "Yes", "No"))
  }
  else
  {
      print("This option has not been specified or an issue emerged.")
  }
  return(transformed_mse_df)
}


graph_only_gadp = function(df = NA, title = "Blank", subtitle = "1000 Monte Carlo Simulations,  Testing set = 500")
{
  df %>% 
    filter(gadp == "Yes") %>%
    group_by(training_size, gadp, estimator, noise, theta) %>% 
    summarise(mse = mean(mse)) %>%
    ggplot(aes(x = training_size, y = mse, color = estimator)) +
    facet_grid(vars(theta),vars(noise), scales = "free_y", labeller = label_both) +
    geom_line(linetype = "dashed") +
    geom_point(aes(shape = estimator)) +
    scale_shape_manual(values=c(1:10)) +
    scale_x_continuous(breaks = c(500,1000, 2000)) +
    labs(x = "n", y = "MSE", title = title,
         subtitle = subtitle) + theme_bw() + 
    theme(axis.text.x = element_text(angle = 45, vjust = 0.5))
}


prep_for_latex_table = function(result_df, permutation_type = "gadp")
{
  prepped_df = result_df %>%
    pivot_wider(names_from = !!as.name(permutation_type), values_from = "mse") %>%
    group_by(rho, training_size, testing_size, estimator, noise, theta) %>%
    rename(training = training_size, testing = testing_size) %>%
    summarise(Yes = round(mean(Yes),3), No = round(mean(No),3)) %>% ungroup() %>% # Convert the permutation MSEs to a yes/no column for calculating MSER
    mutate(MSER = round(Yes / No,3)) %>%
    #pivot_wider(names_from = "estimator", values_from = c("Yes", "No", "MSER"))
    pivot_longer(cols=c("Yes", "No", "MSER"),names_to ="Metric" , values_to = c("Measure")) %>%
    mutate(Metric = ifelse(Metric == "Yes", "MSE Yes", Metric),
           Metric = ifelse(Metric == "No", "MSE No", Metric)) %>%
    pivot_wider(names_from = "estimator", values_from = "Measure") %>%
    select(-c(testing)) 
  return(prepped_df)
}

prep_for_latex_table_ver_2 = function(result_df, permutation_type = "gadp")
{
  if(permutation_type == "gadp")
  {
    prepped_df = result_df %>%
      pivot_wider(names_from = !!as.name(permutation_type), values_from = "mse") %>%
      group_by(rho, training_size, testing_size, estimator, noise, theta) %>%
      rename(training = training_size, testing = testing_size) %>%
      summarise(Yes = round(mean(Yes),3), No = round(mean(No),3)) %>% ungroup() %>% # Convert the permutation MSEs to a yes/no column for calculating MSER
      mutate(MSER1 = round(Yes / No,3)) %>%
    #pivot_wider(names_from = "estimator", values_from = c("Yes", "No", "MSER"))
      pivot_longer(cols=c("Yes", "No", "MSER1"),names_to ="Metric" , values_to = c("Measure")) %>%
      mutate(Metric = ifelse(Metric == "Yes", "MSE Yes", Metric),
             Metric = ifelse(Metric == "No", "MSE No", Metric)) %>%
      filter(Metric == "MSER1") %>%
      pivot_wider(names_from = "estimator", values_from = "Measure") %>%
      select(-c(testing)) 
  }
  else
  {
    prepped_df = result_df %>%
      pivot_wider(names_from = !!as.name(permutation_type), values_from = "mse") %>%
      group_by(rho, training_size, testing_size, estimator, noise, theta) %>%
      rename(training = training_size, testing = testing_size) %>%
      summarise(Yes = round(mean(Yes),3), No = round(mean(No),3)) %>% ungroup() %>% # Convert the permutation MSEs to a yes/no column for calculating MSER
      mutate(MSER2 = round(Yes / No,3)) %>%
      #pivot_wider(names_from = "estimator", values_from = c("Yes", "No", "MSER"))
      pivot_longer(cols=c("Yes", "No", "MSER2"),names_to ="Metric" , values_to = c("Measure")) %>%
      mutate(Metric = ifelse(Metric == "Yes", "MSE Yes", Metric),
             Metric = ifelse(Metric == "No", "MSE No", Metric)) %>%
      filter(Metric == "MSER2") %>%
      pivot_wider(names_from = "estimator", values_from = "Measure") %>%
      select(-c(testing))
  }
  return(prepped_df)
}


# Get all perfromance metrics
obtain_CE_perfromance_metrics = function(result_df, permutation_type = "gadp")
{
  if(permutation_type == "gadp")
  {
    prepped_df = result_df %>%
      pivot_wider(names_from = !!as.name(permutation_type), values_from = "mse") %>%
      rename(training = sample_size, testing = testing_size) %>%
      group_by(training, testing, estimator, response, theta) %>%
      summarise(Yes = round(mean(Yes),3), No = round(mean(No),3)) %>% ungroup() %>% # Convert the permutation MSEs to a yes/no column for calculating MSER
      mutate(MSER1 = round(Yes / No,3)) %>%
      #pivot_wider(names_from = "estimator", values_from = c("Yes", "No", "MSER"))
      pivot_longer(cols=c("Yes", "No", "MSER1"),names_to ="Metric" , values_to = c("Measure")) %>%
      mutate(Metric = ifelse(Metric == "Yes", "MSE Yes", Metric),
             Metric = ifelse(Metric == "No", "MSE No", Metric)) %>%
      select(-c(testing)) 
  }
  else
  {
    prepped_df = result_df %>%
      pivot_wider(names_from = !!as.name(permutation_type), values_from = "mse") %>%
      rename(training = sample_size, testing = testing_size) %>%
      group_by(training, testing, estimator, response, theta) %>%
      summarise(Yes = round(mean(Yes),3), No = round(mean(No),3)) %>% ungroup() %>% # Convert the permutation MSEs to a yes/no column for calculating MSER
      mutate(MSER2 = round(Yes / No,3)) %>%
      #pivot_wider(names_from = "estimator", values_from = c("Yes", "No", "MSER"))
      pivot_longer(cols=c("Yes", "No", "MSER2"),names_to ="Metric" , values_to = c("Measure")) %>%
      mutate(Metric = ifelse(Metric == "Yes", "MSE Yes", Metric),
             Metric = ifelse(Metric == "No", "MSE No", Metric)) %>%
      select(-c(testing))
  }
  return(prepped_df)
}


# Get all performance metrics from sims with more tuning for predictive models
obtain_CE_performance_metrics_tuned = function(result_df, permutation_type = "gadp")
{
  if(permutation_type == "gadp")
  {
    prepped_df = result_df %>%
      pivot_wider(names_from = !!as.name(permutation_type), values_from = "mse") %>%
      rename(training = training_size, testing = testing_size) %>%
      group_by(training, testing, estimator, response, theta) %>%
      summarise(Yes = round(mean(Yes),3), No = round(mean(No),3)) %>% ungroup() %>% # Convert the permutation MSEs to a yes/no column for calculating MSER
      mutate(MSER1 = round(Yes / No,3)) %>%
      #pivot_wider(names_from = "estimator", values_from = c("Yes", "No", "MSER"))
      pivot_longer(cols=c("Yes", "No", "MSER1"),names_to ="Metric" , values_to = c("Measure")) %>%
      mutate(Metric = ifelse(Metric == "Yes", "MSE Yes", Metric),
             Metric = ifelse(Metric == "No", "MSE No", Metric)) %>%
      select(-c(testing)) 
  }
  else
  {
    prepped_df = result_df %>%
      pivot_wider(names_from = !!as.name(permutation_type), values_from = "mse") %>%
      rename(training = training_size, testing = testing_size) %>%
      group_by(training, testing, estimator, response, theta) %>%
      summarise(Yes = round(mean(Yes),3), No = round(mean(No),3)) %>% ungroup() %>% # Convert the permutation MSEs to a yes/no column for calculating MSER
      mutate(MSER2 = round(Yes / No,3)) %>%
      #pivot_wider(names_from = "estimator", values_from = c("Yes", "No", "MSER"))
      pivot_longer(cols=c("Yes", "No", "MSER2"),names_to ="Metric" , values_to = c("Measure")) %>%
      mutate(Metric = ifelse(Metric == "Yes", "MSE Yes", Metric),
             Metric = ifelse(Metric == "No", "MSE No", Metric)) %>%
      select(-c(testing))
  }
  return(prepped_df)
}



# Prep the dataframe for latex table for CE data
prep_for_latex_table_CE_data_ver_1 = function(result_df, permutation_type = "gadp")
{
  if(permutation_type == "gadp")
  {
    prepped_df = result_df %>%
      pivot_wider(names_from = !!as.name(permutation_type), values_from = "mse") %>%
      rename(training = sample_size, testing = testing_size) %>%
      group_by(training, testing, estimator, response, theta) %>%
      summarise(Yes = round(mean(Yes),3), No = round(mean(No),3)) %>% ungroup() %>% # Convert the permutation MSEs to a yes/no column for calculating MSER
      mutate(MSER1 = round(Yes / No,3)) %>%
      #pivot_wider(names_from = "estimator", values_from = c("Yes", "No", "MSER"))
      pivot_longer(cols=c("Yes", "No", "MSER1"),names_to ="Metric" , values_to = c("Measure")) %>%
      mutate(Metric = ifelse(Metric == "Yes", "MSE Yes", Metric),
             Metric = ifelse(Metric == "No", "MSE No", Metric)) %>%
      filter(Metric == "MSER1") %>%
      pivot_wider(names_from = "estimator", values_from = "Measure") %>%
      dplyr::select(-c(testing)) 
  }
  else
  {
    prepped_df = result_df %>%
      pivot_wider(names_from = !!as.name(permutation_type), values_from = "mse") %>%
      rename(training = sample_size, testing = testing_size) %>%
      group_by(training, testing, estimator, response, theta) %>%
      summarise(Yes = round(mean(Yes),3), No = round(mean(No),3)) %>% ungroup() %>% # Convert the permutation MSEs to a yes/no column for calculating MSER
      mutate(MSER2 = round(Yes / No,3)) %>%
      #pivot_wider(names_from = "estimator", values_from = c("Yes", "No", "MSER"))
      pivot_longer(cols=c("Yes", "No", "MSER2"),names_to ="Metric" , values_to = c("Measure")) %>%
      mutate(Metric = ifelse(Metric == "Yes", "MSE Yes", Metric),
             Metric = ifelse(Metric == "No", "MSE No", Metric)) %>%
      filter(Metric == "MSER2") %>%
      pivot_wider(names_from = "estimator", values_from = "Measure") %>%
      dplyr::select(-c(testing))
  }
  return(prepped_df)
}

# Prep the dataframe for the Latex tables for
# CE data where the predictive models underwent more tuning
prep_for_latex_table_CE_data_ver_tuned = function(result_df, permutation_type = "gadp")
{
  if(permutation_type == "gadp")
  {
    prepped_df = result_df %>%
      pivot_wider(names_from = !!as.name(permutation_type), values_from = "mse") %>%
      rename(training = training_size, testing = testing_size) %>%
      group_by(training, testing, estimator, response, theta) %>%
      summarise(Yes = round(mean(Yes),3), No = round(mean(No),3)) %>% ungroup() %>% # Convert the permutation MSEs to a yes/no column for calculating MSER
      mutate(MSER1 = round(Yes / No,3)) %>%
      #pivot_wider(names_from = "estimator", values_from = c("Yes", "No", "MSER"))
      pivot_longer(cols=c("Yes", "No", "MSER1"),names_to ="Metric" , values_to = c("Measure")) %>%
      mutate(Metric = ifelse(Metric == "Yes", "MSE Yes", Metric),
             Metric = ifelse(Metric == "No", "MSE No", Metric)) %>%
      filter(Metric == "MSER1") %>%
      pivot_wider(names_from = "estimator", values_from = "Measure") %>%
      dplyr::select(-c(testing)) 
  }
  else
  {
    prepped_df = result_df %>%
      pivot_wider(names_from = !!as.name(permutation_type), values_from = "mse") %>%
      rename(training = training_size, testing = testing_size) %>%
      group_by(training, testing, estimator, response, theta) %>%
      summarise(Yes = round(mean(Yes),3), No = round(mean(No),3)) %>% ungroup() %>% # Convert the permutation MSEs to a yes/no column for calculating MSER
      mutate(MSER2 = round(Yes / No,3)) %>%
      #pivot_wider(names_from = "estimator", values_from = c("Yes", "No", "MSER"))
      pivot_longer(cols=c("Yes", "No", "MSER2"),names_to ="Metric" , values_to = c("Measure")) %>%
      mutate(Metric = ifelse(Metric == "Yes", "MSE Yes", Metric),
             Metric = ifelse(Metric == "No", "MSE No", Metric)) %>%
      filter(Metric == "MSER2") %>%
      pivot_wider(names_from = "estimator", values_from = "Measure") %>%
      dplyr::select(-c(testing))
  }
  return(prepped_df)
}
# Version to of the function for prepping the Latex tables
prep_for_latex_table_ver_2 = function(result_df, permutation_type = "gadp")
{
  if(permutation_type == "gadp")
  {
    prepped_df = result_df %>%
      pivot_wider(names_from = !!as.name(permutation_type), values_from = "mse") %>%
      group_by(rho, training_size, testing_size, estimator, noise, theta) %>%
      rename(training = training_size, testing = testing_size) %>%
      summarise(Yes = round(mean(Yes),3), No = round(mean(No),3)) %>% ungroup() %>% # Convert the permutation MSEs to a yes/no column for calculating MSER
      mutate(MSER1 = round(Yes / No,3)) %>%
      #pivot_wider(names_from = "estimator", values_from = c("Yes", "No", "MSER"))
      pivot_longer(cols=c("Yes", "No", "MSER1"),names_to ="Metric" , values_to = c("Measure")) %>%
      mutate(Metric = ifelse(Metric == "Yes", "MSE Yes", Metric),
             Metric = ifelse(Metric == "No", "MSE No", Metric)) %>%
      filter(Metric == "MSER1") %>%
      pivot_wider(names_from = "estimator", values_from = "Measure") %>%
      select(-c(testing)) 
  }
  else
  {
    prepped_df = result_df %>%
      pivot_wider(names_from = !!as.name(permutation_type), values_from = "mse") %>%
      group_by(rho, training_size, testing_size, estimator, noise, theta) %>%
      rename(training = training_size, testing = testing_size) %>%
      summarise(Yes = round(mean(Yes),3), No = round(mean(No),3)) %>% ungroup() %>% # Convert the permutation MSEs to a yes/no column for calculating MSER
      mutate(MSER2 = round(Yes / No,3)) %>%
      #pivot_wider(names_from = "estimator", values_from = c("Yes", "No", "MSER"))
      pivot_longer(cols=c("Yes", "No", "MSER2"),names_to ="Metric" , values_to = c("Measure")) %>%
      mutate(Metric = ifelse(Metric == "Yes", "MSE Yes", Metric),
             Metric = ifelse(Metric == "No", "MSE No", Metric)) %>%
      filter(Metric == "MSER2") %>%
      pivot_wider(names_from = "estimator", values_from = "Measure") %>%
      select(-c(testing))
  }
  return(prepped_df)
}



# Organize results table from simulation output for CE data
output_org_func = function(results_output, permutation_type = "gadp", testing_size )
{
  if(permutation_type == "gadp")
  {
    results_organized = results_output %>% rename(SE_equal = Stacked_equal,
             SE_prop = Stacked_prop) %>%
      pivot_longer(cols = 5:14, names_to = "estimator", values_to = "mse") %>%
      mutate(response = ifelse(y == 1, "y_1", "y_2"),
           gadp = ifelse(gadp == 1, "Yes", "No"),
           testing_size = testing_size) 
    return(results_organized)
  }
  else
  {
    results_organized = results_output %>% rename(SE_equal = Stacked_equal,
                                                  SE_prop = Stacked_prop) %>%
      pivot_longer(cols = 5:14, names_to = "estimator", values_to = "mse") %>%
      mutate(response = ifelse(y == 1, "y_1", "y_2"),
             cgadp = ifelse(cgadp == 1, "Yes", "No"),
             testing_size = testing_size) 
    return(results_organized)
  }
}




# craft_latex table uses the tables package to convert R dataframes to the 
# multicolumn and multirow Latex tables.
craft_latex_table = function(results_df)
{
  results_table = tabular(Factor(training) * Factor(rho) * Factor(theta) * Factor(noise) * 
            Factor(Metric) ~ # The parameters for each row, the MSER values for the estimators are printed from the following formula
            (LASSO + LR + Linear_SVM + NNET + Poly_SVM + RF + 
               Radial_SVM + SE_equal + SE_prop + XGBTree) * (plain_val), 
            data = results_df)
  return(latex.tabular(results_table))
}


craft_latex_CE_sims_table = function(results_df)
{
  results_table = tabular(Factor(training) * Factor(theta) * Factor(response) * 
                            Factor(Metric) ~ 
                            (LASSO + LR + Linear_SVM + NNET + Poly_SVM + RF + 
                               Radial_SVM + SE_equal + SE_prop + XGBTree) * (plain_val), 
                          data = results_df)
  return(latex.tabular(results_table))
}

# Return the value, nothing special
plain_val = function(x)
{
  return(x)
}


# obtain the data
get_data_func = function(data_dir, file_prefix)
{
  # example of file prefix is "model-1-gadp-.*-pred-10.csv"
  # Use regular expressions to obtain the relevant files
  accumulated_file_names = list.files(paste0(data_dir), pattern = file_prefix)
  
  # set up list for storing the dataframes
  load_files = function(file_name, data_path = data_dir)
  {
    loaded_data = read_csv(paste0(data_path, file_name), lazy = TRUE) %>%
      rename(XGBTrees = XGBTree)
    return(loaded_data)
  }
  # Iterate through and collect the data
  loaded_df_list = lapply(accumulated_file_names, load_files,
                          data_path = data_dir)
  # get the combined dataframe
  combined_df = bind_rows(loaded_df_list)
  return(combined_df)
}


get_data_func_no_rho = function(data_dir, file_prefix)
{
  # example of file prefix is "model-1-gadp-.*-pred-10.csv"
  # Use regular expressions to obtain the relevant files
  accumulated_file_names = list.files(paste0(data_dir), pattern = file_prefix)
  
  # set up list for storing the dataframes
  load_files = function(file_name, data_path = data_dir)
  {
    loaded_data = read_csv(paste0(data_path, file_name), lazy = TRUE) %>%
      mutate(rho = as.numeric(str_replace(str_extract(file_name, 
                                           pattern= "rho-[0-9]_[0-9]"), 
                               "rho-0_","0."))) %>%
      rename(XGBTrees = XGBTree)
    return(loaded_data)
  }
  # Iterate through and collect the data
  loaded_df_list = lapply(accumulated_file_names, load_files,
                          data_path = data_dir)
  # get the combined dataframe
  combined_df = bind_rows(loaded_df_list)
  return(combined_df)
}
