#' Thomas Hilton Johnson III
#' 02-08-2024
#' The Horvitz-Thompson estimator run for the CE real data application
#' 


options(scipen = 999)
library(MASS)
library(tidyverse)
library(doSNOW)
# Load the data
set.seed(100)
# Using SALARYX as the sensitive y variable
# Excluding the idenitifivation and heavily related and likely senstive financial variables.
CE_xy = rpms::CE %>% dplyr::select(-c(NEWID, PSU, CID, QINTRVMO, STATE, INCNONWK, 
                                      ARM_FORC, IN_COLL, POPSIZE, CUTENURE, FAM_TYPE,MARITAL, EDUCA,
                                      MEMBRACE, EARNER, OCCUCODE, REGION, BLS_URBN, HORIGIN, FINLWT21)) %>%
  filter(is.na(SALARYX) == FALSE) %>%
  filter(FINCBTAX <= quantile(FINCBTAX, probs = 0.75) + 3 * IQR(FINCBTAX), 
         FINCBTAX >= quantile(FINCBTAX, probs = 0.25) -3 * IQR(FINCBTAX),
         TOTXEST <= quantile(TOTXEST, probs = 0.75) + 3 * IQR(TOTXEST), 
         TOTXEST >= quantile(TOTXEST, probs = 0.25) -3 * IQR(TOTXEST),
         TOTEXPCQ <= quantile(TOTEXPCQ, prob = 0.75) + 3 * IQR(TOTEXPCQ),
         TOTEXPCQ >= quantile(TOTEXPCQ, prob = 0.25) - 3 * IQR(TOTEXPCQ),
         SALARYX <= quantile(SALARYX, probs = .75) + 3 * IQR(SALARYX),
         SALARYX >= quantile(SALARYX, probs = .25) - 3 * IQR(SALARYX))%>%
  mutate(TOTXEST = TOTXEST + abs(min(TOTXEST)) + 0.0001,
         TOTEXPCQ = TOTEXPCQ + abs(min(TOTEXPCQ)) + 0.0001,
         FINCBTAX = FINCBTAX + abs(min(FINCBTAX)) + 0.0001) %>%
  dplyr::select("AGE", "SEX", "TOTXEST", "TOTEXPCQ", "FINCBTAX", "SALARYX")

# Getting imputation values
mode_popsize = 1
avg_roomsq = as.integer(mean(CE_xy$ROOMSQ, na.rm = TRUE))
med_bedroomq = as.integer(median(CE_xy$BEDROOMQ, na.rm = TRUE))
med_bathroomq = as.integer(median(CE_xy$BATHRMQ, na.rm = TRUE))
mode_arm_forc = "N"
mode_in_coll = "No"
med_irax = as.integer(median(CE_xy$IRAX,na.rm = TRUE))
med_liquidx = as.integer(median(CE_xy$IRAX, na.rm = TRUE))
med_stockx = as.integer(median(CE_xy$STOCKX,na.rm = TRUE))
med_studentx = as.integer(median(CE_xy$STUDNTX, na.rm = TRUE))

# Reference: https://stackoverflow.com/questions/39126537/replace-na-in-a-factor-column
# if using addNA
# CE_xy = CE_xy %>%
#   mutate(ARM_FORC = as.character(ARM_FORC),
#          ARM_FORC = ifelse(is.na(ARM_FORC), mode_arm_forc, ARM_FORC),
#          ARM_FORC = factor(ARM_FORC),
#          IN_COLL = as.character(IN_COLL),
#          IN_COLL = ifelse(is.na(IN_COLL), mode_in_coll, IN_COLL),
#          IN_COLL = factor(IN_COLL),
#          POPSIZE = as.integer(POPSIZE),
#          POPSIZE = ifelse(is.na(POPSIZE), mode_popsize, POPSIZE),
#          POPSIZE = factor(POPSIZE),
#          REGION = as.integer(REGION),
#          REGION = ifelse(is.na(REGION), 3, REGION),
#          REGION = factor(REGION))
# Above is the imputation of the factors by conversion to characters or integers and then back to factors

CE_xy = CE_xy %>% replace_na(
  replace = list(
    ROOMSQ = avg_roomsq,
    BEDROOMQ = med_bedroomq,
    BATHRMQ = med_bathroomq,
    IRAX = med_irax,
    LIQUIDX = med_liquidx,
    STOCKX = med_stockx,
    STUDNTX = med_studentx)) %>% na.omit()

# Response: SALARYX, FINCBTAX
# ind. Variables: STOCKX, LIQUIDX, STUDNTX, IRAX, TOTXEST, SEX, EHOUSNGC

# Divide into two groups based on male and female, then sample from both groups then combine
set.seed(90)
CE_xy_male = CE_xy %>% filter(SEX == "M")
CE_xy_female = CE_xy %>% filter(SEX == "F")
CE_xy_dat = rbind(CE_xy_male[sample(1:nrow(CE_xy_male),1000, replace = FALSE),],
                  CE_xy_female[sample(1:nrow(CE_xy_female),1000, replace = FALSE),])


variable_rewrite = function(CE_dat)
{
  CE_dat = CE_dat %>% rename(y_2 = SALARYX, y_1 = FINCBTAX)
  CE_dat = model.matrix(~., data = CE_dat)[,-1]
  CE_dat = data.frame(CE_dat)
  return(CE_dat)
}


# Directory search function to safegauard against code running outside of
# the project directory. This is not foolproof. It will only search from the 
# home directory or "~" unless altered.
directory_search = function(start_directory = "~/")
{
  
  # check for the necessary files
  check_files = function(project_directory)
  {
    file_bool_rproj = file.exists(paste0(start_directory,"/SDC-Comparisons.Rproj"))
    # Prepare to search for code files
    code_dir = "/Code/"
    necessary_code_files = c("Basic-Funcs.R", "Estimator-Evaluation.R",
                             "CE-data-cleaning.R", "GADP-from-scratch.R",
                             "Iterative-Model-Gen-Code.R")
    # Prepare to search for report files
    reports_dir = "/Code/Reports/"
    necessary_report_files = c("More-tuned-SDC-GAPD-CGAPD-Comparisons-01-29-2025.qmd",
                               "CE-Real-Data-Tables-fincbtax-salaryx-tuned-sims.qmd")
    
    # check the code files are present
    code_files_bool = logical(length(necessary_code_files))
    # for loop to iterate for code files
    for(index_bool in 1:length(necessary_code_files))
    {
      if(file.exists(paste0(project_directory, code_dir, 
                            necessary_code_files[index_bool])))
      {
        code_files_bool[index_bool] = TRUE
      }
    }
    if(any(code_files_bool) == FALSE)
    {
      print("Code files were not detected.")
      print("Please manually specify the project directory (`base_dir` = path/to/SDC-Comparisons-Submission/).")
    }
    
    # logical vector for managing detection of report files
    report_files_bool = logical(length(necessary_report_files))
    
    # for loop to iterate for report files
    for(index_bool in 1:length(necessary_report_files))
    {
      if(file.exists(paste0(project_directory, reports_dir,
                            necessary_report_files[index_bool])))
      {
        report_files_bool[index_bool] = TRUE
      }
      
    }
    if(any(report_files_bool) == FALSE)
    {
      print("Code files were not detected.")
      print("Please manually specify the project directory (`base_dir` = path/to/SDC-Comparisons-Submission/).")
    }
    # boolean to determine if the files were not found
    # TRUE if all files are found, FLASE otherwise
    missing_files_bool = all(c(code_files_bool, report_files_bool, file_bool_rproj) == TRUE)
    
    return(missing_files_bool)
  }
  # Attempt to get the project directory via a recursive search
  obtain_project_directory = function(suggested_directory)
  {
    potential_directories =list.dirs(suggested_directory)
    # indices of the directories that belong to the project
    project_directory_index = which(stringr::str_detect(potential_directories, "SDC-Comparisons-Submission$") == TRUE)
    # Get the project directory
    project_directory = potential_directories[project_directory_index]
    
    # ensure that the necessary files are present
    file_presence_bool = check_files(project_directory)
    
    if(file_presence_bool == TRUE)
    {
      # Do nothing project directory will be returned
    }
    else
    {
      print("Could not find directory for the project. Please manually specify in `base_dir`.")
      quit(save = "no")
    }
    return(project_directory)
  }
  
  
  if (is.na(start_directory) == TRUE)
  {
    # If nothing specified, start in the home directory
    start_directory = "~/"
  }
  
  found_project_directory = NA
  
  # If the substring of the project directory is found
  if(stringr::str_detect(start_directory, "SDC-Comparisons-Submission"))
  {
    files_all_present_bool = check_files(start_directory)
    
    if(files_all_present_bool == TRUE)
    {
      # When starting directory is project directory
      print("Relevant files found.")
      found_project_directory = start_directory
      return(found_project_directory)
    }
    else
    {
      print("Relevant files are not found.")
      # start in home directory
      found_project_directory = obtain_project_directory("~")
      
      return(found_project_directory)
    }
  }
  # If not in anything that could be considered the project directory
  else
  {
    found_project_directory = obtain_project_directory(start_directory)
    print(found_project_directory)
    
    if(is.na(found_project_directory) == TRUE | is.character(found_project_directory) == FALSE)
    {
      found_project_directory = obtain_project_directory("~/")
    }
    
    return(found_project_directory)
  }
  print("An unexpected situation has emerged. None of the previous cases were met.")
  print("Attempting to find project directory. If the project directory is not found")
  print("it will have to be manually specified in `base_dir`.")
  found_project_directory = obtain_project_directory(start_directory)
  
  if(is.na(found_project_directory) == TRUE | is.character(found_project_directory) == FALSE)
  {
    found_project_directory = obtain_project_directory("~/")
  }
  
  return(found_project_directory)
}

# Assumes you are able to open the R project and use the 'here' package
# to locate files. If not, set base_dir to the R project directory (may have to start from home directory) 
# so that it can proper locate scripts and subdirectories.
base_dir = directory_search(here::here())


source(paste0(base_dir, "/Code/Basic-Funcs.R"))
source(paste0(base_dir, "/Code/Estimator-Evaluation.R"))

CE_xy_dat_preproc = variable_rewrite(CE_dat = CE_xy_dat)

model_CE_sim_runs_directory = paste0(base_dir, "/Data/tuned-sims/CE/")

if(dir.exists(model_CE_sim_runs_directory) == FALSE)
{
  dir.create(model_CE_sim_runs_directory)
}

CE_xy_dat_preproc = CE_xy_dat_preproc %>% 
  mutate(TOTXEST = TOTXEST + abs(min(TOTXEST)) + .Machine$double.eps)
## GADP Runs -----

# Focusing on SALARYX, the other response variable is to allow the gadp function to work with a matrix of variables
CE_Data_GADP_theta_0.2 = estimator_eval_stacked_ensemb_cv_stag(model_data = CE_xy_dat_preproc , # the dataset
                                                             m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                             m = 1000, # Number of Monte Carlo simulations
                                                             samp_n = c(1000), # sample sizes to be used
                                                             cores = parallel::detectCores() - 2,
                                                             response_prefix = "y",
                                                             acquire_variable_names = get_variables,
                                                             
                                                             perturbation_func = gadp_func_ver_2,
                                                             theta_value = 0.2)

CE_Data_GADP_theta_0.2_n_1200 = estimator_eval_stacked_ensemb_cv_stag(model_data = CE_xy_dat_preproc , # the dataset
                                                             m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                             m = 1000, # Number of Monte Carlo simulations
                                                             samp_n = c(1200), # sample sizes to be used
                                                             cores = parallel::detectCores() - 2,
                                                             response_prefix = "y",
                                                             acquire_variable_names = get_variables,
                                                             
                                                             perturbation_func = gadp_func,
                                                             theta_value = 0.2)

CE_Data_GADP_theta_0.4 = estimator_eval_stacked_ensemb_cv_stag(model_data = CE_xy_dat_preproc , # the dataset
                                               m_start = 1, # Primarily used for debugging determines which iteration to start from
                                               m = 1000, # Number of Monte Carlo simulations
                                               samp_n = c(1000), # sample sizes to be used
                                               cores = parallel::detectCores() - 2,
                                               response_prefix = "y",
                                               acquire_variable_names = get_variables,
                                               
                                               perturbation_func = gadp_func,
                                               theta_value = 0.4)

CE_Data_GADP_theta_0.4_n_1200 = estimator_eval_stacked_ensemb_cv_stag(model_data = CE_xy_dat_preproc, # the dataset
                                                             m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                             m = 1000, # Number of Monte Carlo simulations
                                                             samp_n = c(1200), # sample sizes to be used
                                                             cores = parallel::detectCores() - 2,
                                                             response_prefix = "y",
                                                             acquire_variable_names = get_variables,
                                                             
                                                             perturbation_func = gadp_func,
                                                             theta_value = 0.4)
  

CE_Data_GADP_theta_0.6 = estimator_eval_stacked_ensemb_cv_stag(model_data = CE_xy_dat_preproc, # the dataset
                                                             m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                             m = 1000, # Number of Monte Carlo simulations
                                                             samp_n = c(1000), # sample sizes to be used
                                                             cores = parallel::detectCores() - 2,
                                                             response_prefix = "y",
                                                             acquire_variable_names = get_variables,
                                                             
                                                             perturbation_func = gadp_func,
                                                             theta_value = 0.6)

CE_Data_GADP_theta_0.6_n_1200 = estimator_eval_stacked_ensemb_cv_stag(model_data = CE_xy_dat_preproc, # the dataset
                                                             m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                             m = 1000, # Number of Monte Carlo simulations
                                                             samp_n = c(1200), # sample sizes to be used
                                                             cores = parallel::detectCores() - 2,
                                                             response_prefix = "y",
                                                             acquire_variable_names = get_variables,
                                                             
                                                             perturbation_func = gadp_func,
                                                             theta_value = 0.6)

CE_Data_GADP_theta_0.8 = estimator_eval_stacked_ensemb_cv_stag(model_data = CE_xy_dat_preproc , # the dataset
                                                             m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                             m = 1000, # Number of Monte Carlo simulations
                                                             samp_n = c(1000), # sample sizes to be used
                                                             cores = parallel::detectCores() - 2,
                                                             response_prefix = "y",
                                                             acquire_variable_names = get_variables,
                                                             
                                                             perturbation_func = gadp_func,
                                                             theta_value = 0.8)

CE_Data_GADP_theta_0.8_n_1200 = estimator_eval_stacked_ensemb_cv_stag(model_data = CE_xy_dat_preproc, # the dataset
                                                             m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                             m = 1000, # Number of Monte Carlo simulations
                                                             samp_n = c(1200), # sample sizes to be used
                                                             cores = parallel::detectCores() - 2,
                                                             response_prefix = "y",
                                                             acquire_variable_names = get_variables,
                                                             
                                                             perturbation_func = gadp_func,
                                                             theta_value = 0.8)


gadp_sims_CE_data = bind_rows(list(CE_Data_GADP_theta_0.2,
                                   CE_Data_GADP_theta_0.2_n_1200,
                                   CE_Data_GADP_theta_0.4,
                                   CE_Data_GADP_theta_0.4_n_1200,
                                   CE_Data_GADP_theta_0.6,
                                   CE_Data_GADP_theta_0.6_n_1200,
                                   CE_Data_GADP_theta_0.8 ,
                                   CE_Data_GADP_theta_0.8_n_1200
                                   ))

write_csv(gadp_sims_CE_data, paste0(model_CE_sim_runs_directory, "gadp-sims-CE-outlier-rem-ver-1.csv"))

## CGADP Runs ------------------------------------------------------
# Need to shift TOTXEST to enable log normal distribution estimation and CDF usage due to negative values
CE_xy_dat_preproc_cgadp = CE_xy_dat_preproc %>% mutate(TOTXEST = TOTXEST + abs(min(TOTXEST)) + .Machine$double.eps)
# Theta = 0.2 run
CE_Data_CGADP_theta_0.2 = estimator_evaluation_stacked_ensemb_cgadp_cv_stag(model_data = CE_xy_dat_preproc_cgadp, # the dataset
                                                             m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                             m = 1000, # Number of Monte Carlo simulations
                                                             samp_n = c(1000), # sample sizes to be used
                                                             cores = parallel::detectCores() - 2,
                                                             response_prefix = "y",
                                                             acquire_variable_names = get_variables,
                                                             
                                                             perturbation_func =real_CE_cgadp_ver_1,
                                                             theta_value = 0.2)

CE_Data_CGADP_theta_0.4 = estimator_evaluation_stacked_ensemb_cgadp_cv_stag(model_data = CE_xy_dat_preproc_cgadp, # the dataset
                                                                    m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                                    m = 1000, # Number of Monte Carlo simulations
                                                                    samp_n = c(1000), # sample sizes to be used
                                                                    cores = parallel::detectCores() - 2,
                                                                    response_prefix = "y",
                                                                    acquire_variable_names = get_variables,
                                                                    
                                                                    perturbation_func =real_CE_cgadp_ver_1,
                                                                    theta_value = 0.4)


CE_Data_CGADP_theta_0.6 = estimator_evaluation_stacked_ensemb_cgadp_cv_stag(model_data = CE_xy_dat_preproc_cgadp, # the dataset
                                                                    m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                                    m = 1000, # Number of Monte Carlo simulations
                                                                    samp_n = c(1000), # sample sizes to be used
                                                                    cores = parallel::detectCores() - 2,
                                                                    response_prefix = "y",
                                                                    acquire_variable_names = get_variables,
                                                                    
                                                                    perturbation_func =real_CE_cgadp_ver_1,
                                                                    theta_value = 0.6)

CE_Data_CGADP_theta_0.8 = estimator_evaluation_stacked_ensemb_cgadp_cv_stag(model_data = CE_xy_dat_preproc_cgadp, # the dataset
                                                                    m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                                    m = 1000, # Number of Monte Carlo simulations
                                                                    samp_n = c(1000), # sample sizes to be used
                                                                    cores = parallel::detectCores() - 2,
                                                                    response_prefix = "y",
                                                                    acquire_variable_names = get_variables,
                                                                    
                                                                    perturbation_func =real_CE_cgadp_ver_1,
                                                                    theta_value = 0.8)

cgadp_CE_data_n.1000 = bind_rows(list(CE_Data_CGADP_theta_0.2, 
                                      CE_Data_CGADP_theta_0.4, 
                                      CE_Data_CGADP_theta_0.6, 
                                      CE_Data_CGADP_theta_0.8))

write_csv(cgadp_CE_data_n.1000,paste0(model_CE_sim_runs_directory, "cgadp-sims-CE-outlier-rem-ver-1-n-1000.csv"))

CE_Data_CGADP_theta_0.2_n1200 = estimator_evaluation_stacked_ensemb_cgadp_cv_stag(model_data = CE_xy_dat_preproc_cgadp, # the dataset
                                                                    m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                                    m = 1000, # Number of Monte Carlo simulations
                                                                    samp_n = c(1200), # sample sizes to be used
                                                                    cores = parallel::detectCores() - 2,
                                                                    response_prefix = "y",
                                                                    acquire_variable_names = get_variables,
                                                                    
                                                                    perturbation_func =real_CE_cgadp_ver_1,
                                                                    theta_value = 0.2)

CE_Data_CGADP_theta_0.4_n1200 = estimator_evaluation_stacked_ensemb_cgadp_cv_stag(model_data = CE_xy_dat_preproc_cgadp, # the dataset
                                                                          m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                                          m = 1000, # Number of Monte Carlo simulations
                                                                          samp_n = c(1200), # sample sizes to be used
                                                                          cores = parallel::detectCores() - 2,
                                                                          response_prefix = "y",
                                                                          acquire_variable_names = get_variables,
                                                                          
                                                                          perturbation_func =real_CE_cgadp_ver_1,
                                                                          theta_value = 0.4)


CE_Data_CGADP_theta_0.6_n1200 = estimator_evaluation_stacked_ensemb_cgadp_cv_stag(model_data = CE_xy_dat_preproc_cgadp, # the dataset
                                                                          m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                                          m = 1000, # Number of Monte Carlo simulations
                                                                          samp_n = c(1200), # sample sizes to be used
                                                                          cores = parallel::detectCores() - 2,
                                                                          response_prefix = "y",
                                                                          acquire_variable_names = get_variables,
                                                                          
                                                                          perturbation_func =real_CE_cgadp_ver_1,
                                                                          theta_value = 0.6)

CE_Data_CGADP_theta_0.8_n1200 = estimator_evaluation_stacked_ensemb_cgadp_cv_stag(model_data = CE_xy_dat_preproc_cgadp, # the dataset
                                                                          m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                                          m = 1000, # Number of Monte Carlo simulations
                                                                          samp_n = c(1200), # sample sizes to be used
                                                                          cores = parallel::detectCores() - 2,
                                                                          response_prefix = "y",
                                                                          acquire_variable_names = get_variables,
                                                                          
                                                                          perturbation_func =real_CE_cgadp_ver_1,
                                                                          theta_value = 0.8)

cgadp_CE_data_n.1000.1200 = bind_rows(list(CE_Data_CGADP_theta_0.2,
                                           CE_Data_CGADP_theta_0.2_n1200,
                                            CE_Data_CGADP_theta_0.4, 
                                          CE_Data_CGADP_theta_0.4_n1200,
                                          CE_Data_CGADP_theta_0.6, 
                                          CE_Data_CGADP_theta_0.6_n1200,
                                          CE_Data_CGADP_theta_0.8,
                                          CE_Data_CGADP_theta_0.8_n1200))

write_csv(cgadp_CE_data_n.1000.1200,paste0(model_CE_sim_runs_directory, "cgadp-sims-CE-outlier-rem-ver-1-n-1000-1200.csv"))

CE_data_sims_complete = bind_rows(
  list(prep_for_latex_table_CE_data_ver_1(
    output_org_func(gadp_sims_CE_data, permutation_type = "gadp", testing_size = 1000)),
     prep_for_latex_table_CE_data_ver_1(
       output_org_func(cgadp_CE_data_n.1000.1200, permutation_type = "cgadp", testing_size = 1000), permutation_type = "cgadp")))

craft_latex_CE_sims_table(CE_data_sims_complete %>% filter(training == 1000))

craft_latex_CE_sims_table(CE_data_sims_complete)
