#' Thomas Johnson III
#' 12-22-2024
#' Job Submission Script
#' Runs Model 1 Simulations for GADP for various simulation model Parameters

#' Encountering issues because they do not have the R packages I have starting with DoSNOW.

library(tidyverse)
library(doFuture)
library(future.callr)

# Test increasing number of Model predictors.


# set cores for Longleaf
# One core to handle retrieving and organizing the data
core_alloc = min(parallelly::availableCores(omit = 1), 255)


produce_csv = function(created_object, base_str, file_name_str)
{
  write_csv(created_object, paste0(base_str, file_name_str))
  return(file_name_str)
}

produce_csv_ver_2 = 
  function(created_object, base_str, file_name_str)
  {
    tabular_results = created_object$output
    write_csv(tabular_results, paste0(base_str, file_name_str))
    return(file_name_str)
  }

# Model 1 CGADP ----------------
model_1_cgadp = function(unprotected_data, target_indices, 
                         theta = 0.4, dat.seed = 12)
{
  
  # Set seed for generating random multivariate normal units
  set.seed(dat.seed)
  
  unprotected_data = as.matrix(unprotected_data)
  
  y_names = colnames(unprotected_data[,target_indices])
  
  
  
  
  # To be released as it is not sensitive.
  original_x = unprotected_data[,-target_indices]
  
  # convert to standard uniform (for sanity check)
  unprotected_data_x = apply(unprotected_data[,-target_indices],2,punif)
  # Convert to standard normal distibution
  unprotected_data_x = apply(unprotected_data_x, 2, qnorm, mean = 0, sd = 1)
  
  
  unprotected_data_y = unprotected_data[,target_indices]
  
  y_means = numeric(length(target_indices))
  
  y_sds = numeric(length(target_indices))
  
  # y looks to be closely normally distributed
  for(y_index in 1:length(target_indices))
  {
    # Calculate the empirical mean and standard deviation
    y_means[y_index] = mean(unprotected_data_y[,y_index])
    y_sds[y_index] = sd(unprotected_data_y[,y_index])
    # Use cumulative function to convert to standard uniform
    y_cumulative = pnorm(unprotected_data_y[,y_index], y_means[y_index],
                         y_sds[y_index])
    # Covert to standard normal
    y_stan_normal = qnorm(y_cumulative, mean = 0, sd = 1)
    unprotected_data_y[,y_index] = y_stan_normal
    
  }
  # Initial preprocessing finished
  unprotected_data = cbind(unprotected_data_x, unprotected_data_y)
  
  # Spearman rank correlation matrix
  spearman_rank_corr = cor(unprotected_data, method = "spearman")
  
  # Claculate the Pearson correlation matrix
  pearson_corr = matrix(rep(0, nrow(spearman_rank_corr) * nrow(spearman_rank_corr)),
                        nrow = nrow(spearman_rank_corr), byrow = TRUE)
  # Derive Pearson correlation matrix from Spearman rank correlation matrix
  for(row_index_i in 1:ncol(unprotected_data))
  {
    for(col_index_j in 1:ncol(unprotected_data))
    {
      
      pearson_corr[row_index_i, col_index_j] = 
        2 * sin((pi * spearman_rank_corr[row_index_i, col_index_j]) / 6)
    }
  }
  
  
  #print("Constructed Pearson correlation matrix")
  
  # https://blogs.sas.com/content/iml/2010/12/10/converting-between-correlation-and-covariance-matrices.html
  # By Rick Wicklin on DoLoop
  # https://stats.stackexchange.com/questions/62850/obtaining-covariance-matrix-from-correlation-matrix
  # Community Wiki
  #pearson_corr = as.matrix(Matrix::nearPD(pearson_corr)$mat)
  
  # Standard deviation should be one as the distributions are standard normal.
  #stan_dev_matrix = diag(ncol(unprotected_data))
  
  #covariance_mat = stan_dev_matrix %*% pearson_corr %*% stan_dev_matrix
  
  covariance_mat = as.matrix(nearPD(pearson_corr)$mat)
  
  # means of the target variable
  target_means = colMeans(unprotected_data[,target_indices])
  
  # means of the auxiliary variable
  auxiliary_means = colMeans(unprotected_data[,-target_indices])
  # Covariance of the target variable
  target_covariance = covariance_mat[target_indices,target_indices]
  
  # Covariance calculated 
  # Auxiliary variables covariance
  aux_covariance = covariance_mat[,-target_indices]
  
  target_aux_covariance = covariance_mat[-target_indices,target_indices]
  
  
  
  # update the unprotected data covariance
  unprotected_cov = covariance_mat
  # Get the unprotected data column means
  unprotected_means = colMeans(unprotected_data)
  
  # Using the new Pearson from the Spearman covariance matrix
  new_y_variates_covariance = unprotected_cov[target_indices,target_indices] * theta^2
  new_y_variates_covariance = cbind(unprotected_cov[target_indices,-target_indices] , 
                                    new_y_variates_covariance)
  
  right_term = new_y_variates_covariance %*%
    solve(unprotected_cov, tol=NULL) %*% t(new_y_variates_covariance)
  
  new_covariance = target_covariance - right_term
  
  # Make sure covariance matrixis near positive definite
  new_covariance = as.matrix(nearPD(new_covariance, eig.tol = 1e-10, 
                                    maxit = 200, posd.tol = 1e-09,
                                    conv.tol = 1e-09)$mat)
  
  # For storing the privatized y, or y after applying gadp
  randomized_y = matrix(rep(0, nrow(unprotected_data) * length(target_indices)), nrow = nrow(unprotected_data), byrow = TRUE)
  
  for( index in 1:nrow(unprotected_data))
  {
    new_means = target_means + new_y_variates_covariance %*%
      solve(unprotected_cov) %*% (unprotected_data[index,]- unprotected_means)
    
    new_means = t(new_means)
    
    new_rand_vars = MASS::mvrnorm(n = 1, mu = new_means, Sigma =  new_covariance)
    
    
    
    randomized_y[index,] = new_rand_vars
  }
  
  
  # Convert the y-values to uniform distribution to convert back to 
  for(y_index in 1:length(target_indices))
  {
    # Convert the distributions back to the uniform distribution
    new_randomized_mean = mean(randomized_y[,y_index])
    new_randomized_sd = sd(randomized_y[,y_index])
    randomized_y_unif = pnorm(randomized_y[,y_index], mean = new_randomized_mean,
                              sd = new_randomized_sd)
    
    # Prevent infinities, use approximations ot 0 and 1,
    randomized_y_unif = scales::rescale(randomized_y_unif, to = c(0 + .Machine$double.eps, 1 - .Machine$double.eps))
    # Convert the y-values back to their original distribution
    
    y_index_original_dist = qnorm(randomized_y_unif,
                                  mean = y_means[y_index], sd = y_sds[y_index])
    # Update the randomized values to no longer be stnadard normal, but the in the
    # same distirbution as the original y-values, but not the original y-values
    # themselves
    # print("sum of infinites")
    # print(sum(is.infinite(y_index_original_dist)))
    # print(y_means[y_index]) 
    # print(y_sds[y_index])
    
    randomized_y[,y_index] = y_index_original_dist
  }
  
  colnames(randomized_y) = y_names
  
  randomized_data = cbind(original_x, randomized_y)
  
  return(randomized_data)
}


# Function for Generating Model 1 -----------------
# Model 1 COde
correlate_uniform_m.1_p_ver_2 <- function(N = 1000, rho = 0.1, max_pred = 50, dat.seed = 90,
                                          noise_1 = 0.2, noise_2 = noise_1 * 2)
{
  set.seed(dat.seed)
  base_pred = 8
  num_pred = max_pred - base_pred # extra predictors that are noise
  
  
  # Generating the data using mu vector and variance-covariance matrix
  sigma.mat = diag(max_pred) + ifelse(diag(max_pred) == 0, rho , 0)
  x_mat_model = apply(cbind(rep(1,N), 
                            MASS::mvrnorm(N, mu = rep(0,max_pred), Sigma = sigma.mat, tol = 1)), 
                      2, pnorm)
  colnames(x_mat_model) <- paste0("X", 0:max_pred)
  # Account for extra predictors, for generating model 1's y
  coeff.mat_m.1 = matrix(c(1, 0, 1, 0, 1.5, 0, 0, 0, 1, rep(0,num_pred)), 
                         nrow =max_pred+1, ncol = 1, byrow = TRUE)
  
  # error for the model
  error = rnorm(N, 0, noise_1)
  error_2 = rnorm(N, 0, noise_2)
  
  model_1.rho =x_mat_model %*% coeff.mat_m.1
  
  # Model 1 from McConvill 2017 Dissertation
  y_1 <- model_1.rho + error
  # 
  # # Account for extra predictors generating model 2's y
  # # Have account for the lack of one predictor
  # coeff.mat_m.2 = matrix(c(1.5, 0, 4, 0, 0, 8, 0, 0, rep(0, num_pred + 1)), 
  #                    nrow =max_pred+1, ncol = 1, byrow = TRUE)
  # model_2.rho = x_mat_model %*% coeff.mat_m.2
  # # Model 2 from McConville 2017 Dissertation, modified form to get positive definite matrix
  
  # Model 1 with more noise
  y_2 <- model_1.rho + error_2
  
  m.1.2_model_rho = data.frame(x_mat_model, y_1, y_2)
  # Remove the intercept term
  m.1.2_model_rho = m.1.2_model_rho[,-1]
  return(m.1.2_model_rho)
}

# SUpport Functions for locating Variables -------------

get_variables = function(target_variable_prefix = "y", variable_names)
{
  target_variables = sapply(variable_names, str_detect, 
                            pattern = target_variable_prefix)
  target_variable_indices = which(target_variables == TRUE)
  predictor_variables = variable_names[-target_variable_indices]
  target_and_predictor_vars = list("responses" = variable_names[target_variable_indices],
                                   "predictors" = predictor_variables)
  return(target_and_predictor_vars)
}



sdc_formula_generation = function(target_variables, predictor_variables)
{
  # get the response part of the formula
  response_part = paste0(target_variables,collapse = " + ")
  
  # Get the predictor part of the formula
  predictors_part = paste0(predictor_variables,collapse = " + ")
  
  # Get the complete formula
  complete_formula = formula(paste0(response_part, " ~ ", predictors_part, collapse= ""))
  return(complete_formula)
}

# Evaluation Function -----------------------
# Code for Performing Machine Learning Evaluations With CGADP
estimator_evaluation_stacked_ensemb_cgadp_cv = function(# the dataset
  m_start = 1, # Primarily used for debugging determines which iteration to start from
  m = 1000, # Number of Monte Carlo simulations
  samp_n = c(500, 1000, 2000), # sample sizes to be used for training set
  cores = 2, # Number of CPU threads/ logical processors to use
  response_prefix = "y", # prefix for response variables
  acquire_variable_names = get_variables, #Function for getting the variables
  perturbation_func, # CGADP or GADP function
  theta_value = 0.4, # privacy parameter value
  model_generating_function, # function to generate the model
  testing_data_seed = 30000, # Seed for the testing set
  model_predictor_quant = 10, # Quantity of the predictors for the model
  model_rho = 0.2, # correlation of predictors in the model
  model_noise = 0.2, # Noise to be used in the model
  testing_n = 500) # Size of the testing set
{
  
  # Cross-Validation Funcitons -----------
  # XGBoosted Trees
  xgbt_cv = function(x_pred, y_resp, 
                     parameter_grid = list("nrounds" = c(10,20, 40),
                                           "max_depth" = c(2,4),
                                           "eta" = c(0.2, 0.4, 0.6),
                                           "gamma" = c(0, 0.5, 1),
                                           "colsample_bytree" = c(0.2, 0.5, 1),
                                           "min_child_weight" = c(0.5, 1),
                                           "subsample" = c(0.4, 0.5, 0.6)),
                     cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = expand.grid(parameter_grid)
    
    # Construct the XGBoosted trees
    xgbt_control= caret::trainControl(method = "cv", number = cv_folds)
    # Train XGBoosted trees
    xgbt_obj = caret::train(x = x_pred, y = as.vector(y_resp),
                            method = "xgbTree", tuneGrid = param_grid,
                            trControl = xgbt_control,
                            nthread = 1,
                            verbose = 0,
                            print_every_n =0)
    
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_xgbt_obj = xgbt_obj$finalModel
    return(final_xgbt_obj)
  }
  
  # Random forest
  ranger_cv = function(x_pred, y_resp, num_trees = 500,
                       frac_pred = c(1/3, 1/4), 
                       min_node_size = c(5,10),
                       cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = as.matrix(expand_grid(num_trees, frac_pred, min_node_size))
    # Set up folds to be used in cross-validation
    fold_ids = split(1:nrow(x_pred), 1:cv_folds)
    # Vector of average mse for each combination of hyperparameters
    avg_mse_vals = numeric(nrow(param_grid))
    # Loop through each set of hyperparameters
    for(param_index in 1:nrow(param_grid))
    {
      # Vector to store mses per each test fold
      cv_mse_vec = numeric(cv_folds)
      # Cycle through each potential test fold
      for(fold_index in 1:cv_folds)
      {
        # Folds that will be used for training
        cv_sample = unsplit(fold_ids[-fold_index], rep(1:(cv_folds), cv_folds-1))
        # Fold that will be used for test
        cv_test = unsplit(fold_ids[-fold_index], 1:(cv_folds))
        # COnstruct the random forest
        ranger_rf_obj = ranger(x = x_pred[cv_sample,],
                               y = y_resp[cv_sample],
                               mtry = ceiling(ncol(x_pred) * param_grid[param_index, "frac_pred"]),
                               min.node.size = param_grid[param_index, "min_node_size"],
                               num.trees = param_grid[param_index, "num_trees"],
                               num.threads = 1)
        # Get the predictions
        ranger_predictions = predict(ranger_rf_obj, x_pred[cv_sample,])$predictions
        # Calculate the MSE
        mse_inst = 1 / length(cv_sample) * sum((y_resp[cv_sample] - ranger_predictions)^2)
        # Store the MSEs across each fold
        cv_mse_vec[fold_index] = mse_inst
      }
      # Store the average of the MSEs across eeach fold
      avg_mse_vals[param_index] = mean(cv_mse_vec) 
    }
    # Taking the minimum MSE or the first occurrence of the minimum MSE
    min_mse_position = which.min(avg_mse_vals)
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_ranger_obj = ranger(x = x_pred, y = y_resp,
                              mtry = ceiling(ncol(x_pred) * param_grid[min_mse_position, "frac_pred"]),
                              min.node.size = param_grid[min_mse_position, "min_node_size"],
                              num.trees = param_grid[min_mse_position, "num_trees"],
                              num.threads = 1)
    return(final_ranger_obj)
  }
  
  # Neural Network
  nnet_cv = function(x_pred, y_resp, 
                     parameter_grid = list("size" = c(3, 5, 10, 20),
                                           "decay" = c(0.001, 0.01, 0.1) ),
                     cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = expand.grid(parameter_grid)
    
    # Construct the XGBoosted trees
    nnet_control= caret::trainControl(method = "cv", number = cv_folds)
    # Train XGBoosted trees
    nnet_obj = caret::train(x = x_pred, y = as.vector(y_resp),
                            method = "nnet", tuneGrid = param_grid,
                            trControl = nnet_control,
                            maxit = 500,
                            linout=FALSE,
                            trace=FALSE)
    
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_nnet_obj = nnet_obj$finalModel
    return(final_nnet_obj)
  }
  
  # generate the testing data
  testing_model_data = model_generating_function(N = testing_n, rho = model_rho, 
                                                 max_pred = model_predictor_quant, 
                                                 dat.seed = testing_data_seed,
                                                 noise_1 = model_noise)
  
  # Getting variable names:
  all_variable_names = colnames(testing_model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  
  # sdc_data_formula = sdc_form(target_variables = organized_vars$response, 
  #                             predictor_variables = organized_vars$predictors)
  
  # Calculate MSE function
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  # Use futures with callr to bypass R connections limits
  # plan(multisession, workers = core_alloc)
  plan(callr, workers = core_alloc)
  
  results = foreach(samp_size_index = 1:length(samp_n), .combine = "rbind", 
                    .options.future= list(packages = c( "Matrix",  "ranger", 
                                                        "nnet", "e1071", "xgboost", "methods", "glmnet",
                                                        "tidyr", "caret")),
                    .inorder = FALSE) %:%
    foreach( iteration = m_start:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",
            .inorder = FALSE) %dofuture% {
              
              # setting the seed
              dat.seed = iteration + samp_n[samp_size_index]
              set.seed(dat.seed)
              
              # Training dataset
              # Generated with the same parameters sxcluding the seed
              training_model_data = 
                model_generating_function(N = samp_n[samp_size_index], 
                                          rho = model_rho, 
                                          max_pred = model_predictor_quant, 
                                          dat.seed = dat.seed,
                                          noise_1 = model_noise)
              
              model_data = rbind(training_model_data, 
                                 testing_model_data)
              
              # Performing SDC GADP method
              privatized_dataframe = 
                perturbation_func(unprotected_data = model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              
              privatized_responses = privatized_dataframe[, target_variables_indices]
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,organized_vars$predictors]
              
              # Get the sample from privatized df
              drawing = 1:samp_n[samp_size_index]
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              # Adding intercept term for linear regression
              x_dat_holdout = cbind(1,as.matrix(privatized_dataframe)[-drawing,organized_vars$predictors])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after cgapd, but preparing for case when both x and y are randomized
              x_dat = as.matrix(cbind(1,
                                      sample_private_data[,organized_vars$predictors]))
              
              
              # Get the predictors for the training set for unprotected data
              x_dat_unprotected = 
                as.matrix(cbind(1, sample_unprotected_data[,organized_vars$predictors]))
              
              # Get the predictors for the testing set for unprotected data
              x_dat_holdout_unprotected = 
                as.matrix(cbind(1, model_data[-drawing,organized_vars$predictors]))
              
              
              
              ### CGAPD Used----
              
              # Linear Regression on data after cgapd
              betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(current_response[drawing])
              
              
              linear_reg_pred = as.vector(x_dat_holdout %*% betas_linear)
              
              # Get the predictions for the training data
              
              training_lin_reg_pred = as.vector(x_dat %*% betas_linear)
              
              
              # Get the mse for the training data
              
              
              training_lin_reg_mse = calc_mse(as.matrix(current_response[drawing]),
                                              training_lin_reg_pred)
              
              # Get the mse for the test data
              
              
              linear_reg_mse = calc_mse(orig_data_response[-drawing],
                                        linear_reg_pred)
              
              # LASSO on data after cgapd ---------------------
              
              # Build the LASSO regression model
              # using cross validation
              lasso_regress_model = cv.glmnet( 
                x = as.matrix(sample_private_data[,organized_vars$predictors]),
                y = current_response[drawing])
              
              
              lasso_pred = predict(lasso_regress_model, 
                                   as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]),
                                   s = "lambda.min")
              
              # get predictions on the training data
              training_lasso_pred = predict(lasso_regress_model, 
                                            as.matrix(privatized_dataframe[drawing,organized_vars$predictors]),
                                            s = "lambda.min")
              
              # Get the MSE for the training data
              training_lasso_mse = 
                calc_mse(orig_data_response[drawing],
                         training_lasso_pred)
              
              # # MSE for the testing data
              
              lasso_mse = calc_mse(orig_data_response[-drawing],
                                   lasso_pred)
              
              # Neural Net Regression on data after cgapd
              cgapd_nnet = nnet_cv(x_pred = x_dat[,-1], 
                                   y_resp = current_response[drawing])
              
              
              cgapd_nnet_reg_pred = predict(cgapd_nnet,x_dat_holdout[,-1])
              
              # training data predictions
              training_cgapd_nnet_reg_pred = predict(cgapd_nnet,x_dat[,-1])
              
              # training data MSE
              training_cgapd_nnet_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_nnet_reg_pred)
              
              
              # testing data MSE
              cgapd_nnet_reg_mse = calc_mse(orig_data_response[-drawing],
                                            cgapd_nnet_reg_pred)
              
              ## SVM Runs for CGADP -------
              
              # SVM Radial Regression on data after cgapd
              cgapd_radial_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                               gamma = seq(0.001,0.1,length.out = 2), 
                                               cost = seq(0.1,1,length.out = 2),
                                               
                                               kernel = c("radial"))
              
              cgapd_radial_svm = cgapd_radial_svm_tune$best.model
              
              cgapd_radial_svm_reg_pred = predict(cgapd_radial_svm,x_dat_holdout[,-1])
              
              # Get the training predictions of the radial svm
              training_cgapd_radial_svm_reg_pred =
                predict(cgapd_radial_svm, x_dat[,-1])
              
              # Get the training mse for radial svm
              training_cgapd_radial_svm_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_radial_svm_reg_pred)
              
              # Testing data MSE
              cgapd_radial_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                  cgapd_radial_svm_reg_pred)
              
              
              
              # SVM Linear Regression on data after cgapd
              cgapd_linear_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                               
                                               cost = seq(0.1,1,length.out = 2),
                                               
                                               kernel = c("linear"))
              
              cgapd_linear_svm = cgapd_linear_svm_tune$best.model
              
              # Predictions for the training data for Linear SVM
              training_cgapd_linear_svm_reg_pred =
                predict(cgapd_linear_svm,x_dat_holdout[,-1])
              
              # Training MSE for Linear SVM
              training_cgapd_linear_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_linear_svm_reg_pred)
              
              # Predictions for the test dataset
              cgapd_linear_svm_reg_pred = predict(cgapd_linear_svm,x_dat_holdout[,-1])
              
              # Testing data MSE
              cgapd_linear_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                  cgapd_linear_svm_reg_pred)
              
              # SVM Polynomial Regression on data after gapd
              cgapd_poly_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                             gamma = seq(0.001,0.1,length.out = 2), 
                                             cost = seq(0.1,1,length.out = 2),
                                             degree = seq(1,2, length.out=2),
                                             kernel = c("polynomial"),
                                             cross = 10)
              
              cgapd_poly_svm = cgapd_poly_svm_tune$best.model
              
              # Training data polynomial SVM predictions
              training_cgapd_poly_svm_reg_pred = predict(cgapd_poly_svm,x_dat[,-1])
              
              # Training data MSE polynomial SVM
              training_cgapd_poly_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_poly_svm_reg_pred)
              
              # Predictions for testing dataset
              cgapd_poly_svm_reg_pred = predict(cgapd_poly_svm,x_dat_holdout[,-1])
              
              # MSE for testing dataset for polynomial kernel SVM
              cgapd_poly_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                cgapd_poly_svm_reg_pred)
              
              ## Random Forest Run ---------------------------
              
              # Random forest on data after cgapd
              num_x_vars = length(organized_vars$predictors)
              
              ranger_rf_regress_model = ranger_cv( 
                x_pred = as.matrix(sample_private_data[,organized_vars$predictors]),
                y_resp = current_response[drawing])
              
              # Training dataset predicions for the random forest
              training_ranger_rf_pred = 
                predict(ranger_rf_regress_model, 
                        privatized_dataframe[drawing,organized_vars$predictors])$predictions
              
              # training dataset MSE for the radom forest
              training_ranger_rf_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_ranger_rf_pred)
              
              
              # Testing dataset prediction for random forest
              ranger_rf_pred = predict(ranger_rf_regress_model, 
                                       privatized_dataframe[-drawing,organized_vars$predictors])$predictions
              
              # Testing dataset MSE for random forest
              ranger_rf_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                       ranger_rf_pred)
              
              # XGBT on data after cgapd
              xgbt_regress_model = xgbt_cv( 
                x_pred = as.matrix(sample_private_data[,organized_vars$predictors]),
                y_resp = current_response[drawing])
              
              # Training data predictions for the XGB tree
              training_xgbt_pred = 
                predict(xgbt_regress_model,
                        as.matrix(privatized_dataframe[drawing,organized_vars$predictors]))
              
              # training data MSE for the XGB tree
              training_xgbt_mse = 
                calc_mse(as.matrix(current_response[drawing]) ,
                         training_xgbt_pred)
              
              # Testing data predicitons for the XGB tree
              xgbt_pred = predict(xgbt_regress_model, 
                                  as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]))
              
              # Testing data MSE for the XGB tree
              xgbt_mse = calc_mse(orig_data_response[-drawing],
                                  xgbt_pred)
              
              ##
              ## Before CGAPD (On original data) -----------------------------------
              # Linear Regression on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              
              # Beta coefficients for the original data
              orig_data_betas_linear = solve(t(x_dat_unprotected) %*% 
                                               x_dat_unprotected) %*% t(x_dat_unprotected) %*% 
                as.matrix(sample_orig_data_response)
              
              # Predictions on the training data no GADP
              training_lin_reg_unprivat_pred = 
                as.vector(x_dat_unprotected %*% orig_data_betas_linear)
              
              
              # MSE for training data for linear regression, no GADP
              training_linear_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_lin_reg_unprivat_pred)
              
              # Testing data MSE for linear regression no ADP
              linear_reg_unprivatized_pred = as.vector(x_dat_holdout_unprotected %*% 
                                                         orig_data_betas_linear)
              
              linear_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_reg_unprivatized_pred)
              
              # LASSO on data before gapd (original data)
              
              
              lasso_regress_model_orig_data =
                cv.glmnet( x = as.matrix(model_data[drawing,organized_vars$predictors]), # Get rid of intercept term
                           y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # Training data predictions for LASSO
              
              training_lasso_pred_orig_data = 
                predict(lasso_regress_model_orig_data,
                        as.matrix(model_data[drawing,organized_vars$predictors]), 
                        s = "lambda.min")
              
              # Training data MSE for LASSO
              training_lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_lasso_pred_orig_data)
              
              # Predictions for the test dataset for the LASSO model
              lasso_pred_orig_data = predict(lasso_regress_model_orig_data,
                                             as.matrix(model_data[-drawing,organized_vars$predictors]), 
                                             s = "lambda.min")
              # MSE for the LASSO model for the test data
              lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         lasso_pred_orig_data)
              
              
              # Neural Net before GAPD 
              
              no_cgapd_nnet = nnet_cv(x_pred = x_dat_unprotected[,-1], 
                                      y_resp = sample_orig_data_response)
              
              # The predictions of the training set for neural network
              training_nnet_reg_unprivatized_pred = predict(no_cgapd_nnet, 
                                                            x_dat_unprotected[,-1])
              
              # The MSE for the training set for the neural network
              training_nnet_reg_unprivatized_mse = 
                calc_mse(orig_data_response[drawing],
                         training_nnet_reg_unprivatized_pred)
              
              # Predicitons on testing set for neural net
              nnet_reg_unprivatized_pred = predict(no_cgapd_nnet, 
                                                   x_dat_holdout_unprotected[,-1])
              
              # MSE for testing set for neural network
              nnet_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         nnet_reg_unprivatized_pred)
              
              # SVM Radial before GAPD
              no_cgapd_radial_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                  y = sample_orig_data_response,
                                                  gamma = seq(0.001,0.1,length.out = 2), 
                                                  cost = seq(0.1,1,length.out = 2),
                                                  
                                                  kernel = c("radial"))
              
              # Get the best model for Radial SVM
              no_cgapd_radial_svm = no_cgapd_radial_svm_tune$best.model
              
              # Predictions on the training set for radial SVM
              training_radial_svm_reg_unprivatized_pred = 
                predict(no_cgapd_radial_svm, x_dat_unprotected[,-1])
              
              # The MSE for the training set for radial SVM
              training_radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_radial_svm_reg_unprivatized_pred)
              
              # Get the predictions of Radial SVM for the testing set
              radial_svm_reg_unprivatized_pred = 
                predict(no_cgapd_radial_svm, x_dat_holdout_unprotected[,-1])
              
              # Get the MSE of the Radial SVM for the testing set
              radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         radial_svm_reg_unprivatized_pred)
              
              # Linear SVM before GAPD
              no_cgapd_linear_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                  y = sample_orig_data_response,
                                                  cost = seq(0.1,1,length.out = 2),
                                                  
                                                  kernel = c("linear"))
              # The Linear SVM best tuned model
              no_cgapd_linear_svm = no_cgapd_linear_svm_tune$best.model
              
              # The predictions on the training set, Linear SVM
              training_linear_svm_reg_unprivatized_pred = 
                predict(no_cgapd_linear_svm, x_dat_unprotected[,-1])
              
              # the MSE on the training set, Linear SVM
              training_linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_linear_svm_reg_unprivatized_pred)
              
              # The predictions for the testing set Linear SVM
              linear_svm_reg_unprivatized_pred = 
                predict(no_cgapd_linear_svm, x_dat_holdout_unprotected[,-1])
              
              # The MSE for the testing set Linear SVM
              linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_svm_reg_unprivatized_pred)
              
              
              # Polynomial SVM before GAPD
              no_cgapd_poly_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                y = sample_orig_data_response,
                                                gamma = seq(0.001,0.1,length.out = 2), 
                                                cost = seq(0.1,1,length.out = 2),
                                                degree = seq(1,2, length.out=2),
                                                kernel = c("polynomial"),
                                                cross = 10)
              # Get the best tuned polynomial SVM
              no_cgapd_poly_svm = no_cgapd_poly_svm_tune$best.model
              
              # The training set predicitons for the polynomial SVM
              training_poly_svm_reg_unprivatized_pred = 
                predict(no_cgapd_poly_svm, x_dat_unprotected[,-1])
              
              # the training set MSE for the polynomial SVM
              training_poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_poly_svm_reg_unprivatized_pred)
              
              # The testing set predicitons using the polynomial SVM
              poly_svm_reg_unprivatized_pred = 
                predict(no_cgapd_poly_svm, x_dat_holdout_unprotected[,-1])
              
              # The testing set MSE for the polynomial SVM
              poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         poly_svm_reg_unprivatized_pred)
              
              # Random forest on data before cgapd
              ranger_rf_model_orig_data = ranger_cv( 
                x_pred = model_data[drawing,organized_vars$predictors],
                y_resp = orig_data_response[drawing])
              
              # The training set predictions from random forest
              training_ranger_rf_pred_orig_data = 
                predict(ranger_rf_model_orig_data,
                        model_data[drawing,organized_vars$predictors])$predictions
              
              # The training set MSE for the random forest
              training_ranger_rf_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_ranger_rf_pred_orig_data)
              
              # Testing set predictions for random forest
              ranger_rf_pred_orig_data = predict(ranger_rf_model_orig_data,
                                                 model_data[-drawing,organized_vars$predictors])$predictions
              # Testing set MSE for random forest
              ranger_rf_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         ranger_rf_pred_orig_data)
              
              # XGBT on data before cgapd
              xgbt_regress_model_orig_data =
                xgbt_cv(x_pred = as.matrix(sample_unprotected_data[,organized_vars$predictors]),
                        y_resp = orig_data_response[drawing])
              
              # The training set predictions for XGB tree
              training_xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(sample_unprotected_data[,organized_vars$predictors]))
              # Training set MSE for XGB tree for 
              training_xgbt_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_xgbt_pred_orig_data)
              
              # The testing set XGB tree predictions
              xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[-drawing,organized_vars$predictors]))
              
              # The testing set MSE for XGB tree
              xgbt_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         xgbt_pred_orig_data)
              
              
              #print("Constructing the stacked ensemble")
              ###
              ### Stacked Ensemble Equal Weighting ----
              # CGAPD
              avg_stacked_ensemble_predictions_cgapd = (linear_reg_pred  + 
                                                          lasso_pred  +  cgapd_nnet_reg_pred + 
                                                          cgapd_radial_svm_reg_pred  + 
                                                          cgapd_linear_svm_reg_pred  + 
                                                          cgapd_poly_svm_reg_pred  +
                                                          ranger_rf_pred  + xgbt_pred) / 8
              
              avg_stacked_ensemble_cgapd_mse =
                calc_mse(orig_data_response[-drawing],
                         avg_stacked_ensemble_predictions_cgapd)
              # No GAPD
              avg_stacked_ensemble_pred_no_cgapd = 
                (linear_reg_unprivatized_pred  + 
                   lasso_pred_orig_data  + 
                   nnet_reg_unprivatized_pred  + 
                   radial_svm_reg_unprivatized_pred  + 
                   linear_svm_reg_unprivatized_pred  + 
                   poly_svm_reg_unprivatized_pred  +
                   ranger_rf_pred_orig_data  + xgbt_pred_orig_data) / 8
              
              avg_stacked_ensemble_no_cgapd_mse =
                calc_mse(orig_data_response[-drawing],
                         avg_stacked_ensemble_pred_no_cgapd)
              
              
              ### Stacked Ensemble Equal Weighting by Proportion of variance Explained ----------------
              # under CGADP
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_cgapd = 
                c(training_lin_reg_mse, training_lasso_mse, 
                  training_cgapd_nnet_reg_mse, training_cgapd_radial_svm_reg_mse,
                  training_cgapd_linear_svm_reg_mse, 
                  training_cgapd_poly_svm_reg_mse, training_ranger_rf_mse,
                  training_xgbt_mse)
              
              # The models with the lowest MSE get the highest weight
              prop_var_exp_cgapd_weighting = 
                (training_set_mses_cgapd / sum(training_set_mses_cgapd))^(-1)
              
              prop_var_exp_cgapd_weighting = prop_var_exp_cgapd_weighting / 
                sum(prop_var_exp_cgapd_weighting)
              
              
              
              # Get the predictions for the stacked ensemble for the prop
              # of variance explained
              prop_var_stacked_ens_cgapd_pred = 
                linear_reg_pred * prop_var_exp_cgapd_weighting[1] +
                lasso_pred * prop_var_exp_cgapd_weighting[2] + 
                cgapd_nnet_reg_pred * prop_var_exp_cgapd_weighting[3] + 
                cgapd_radial_svm_reg_pred * prop_var_exp_cgapd_weighting[4] + 
                cgapd_linear_svm_reg_pred * prop_var_exp_cgapd_weighting[5] + 
                cgapd_poly_svm_reg_pred * prop_var_exp_cgapd_weighting[6] +
                ranger_rf_pred * prop_var_exp_cgapd_weighting[7] + 
                xgbt_pred * prop_var_exp_cgapd_weighting[8]
              
              
              
              # Get the MSE
              prop_var_stacked_ens_cgapd_mse = 
                calc_mse(orig_data_response[-drawing],
                         prop_var_stacked_ens_cgapd_pred)
              
              
              # Get the Stacked Ensemble (Prop. of Var. explained) with No GADP 
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_no_cgapd = 
                c(training_linear_reg_mse_orig_data, 
                  training_lasso_mse_orig_data, 
                  training_nnet_reg_unprivatized_mse, 
                  training_radial_svm_reg_mse_orig_data,
                  training_linear_svm_reg_mse_orig_data, 
                  training_poly_svm_reg_mse_orig_data, 
                  training_ranger_rf_mse_orig_data,
                  training_xgbt_mse_orig_data)
              
              # weighting for the SE
              prop_var_exp_no_cgapd_weighting = 
                (training_set_mses_no_cgapd / sum(training_set_mses_no_cgapd))^(-1)
              
              prop_var_exp_no_cgapd_weighting = prop_var_exp_no_cgapd_weighting /
                sum(prop_var_exp_no_cgapd_weighting)
              
              
              # get the predictions
              prop_var_stacked_ens_no_cgapd_pred = 
                linear_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[1] +
                lasso_pred_orig_data * prop_var_exp_no_cgapd_weighting[2] + 
                nnet_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[3] + 
                radial_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[4] + 
                linear_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[5] + 
                poly_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[6] +
                ranger_rf_pred_orig_data * prop_var_exp_no_cgapd_weighting[7] + 
                xgbt_pred_orig_data * prop_var_exp_no_cgapd_weighting[8]
              
              # Get the MSE
              prop_var_stacked_ens_no_cgapd_mse = 
                calc_mse(orig_data_response[-drawing],
                         prop_var_stacked_ens_no_cgapd_pred)
              
              
              # Compile MSEs into a vector
              cgapd_mse_vector = 
                c(linear_reg_mse, lasso_mse, cgapd_nnet_reg_mse,
                  cgapd_radial_svm_reg_mse, cgapd_linear_svm_reg_mse,
                  cgapd_poly_svm_reg_mse, ranger_rf_mse,
                  xgbt_mse, avg_stacked_ensemble_cgapd_mse, 
                  prop_var_stacked_ens_cgapd_mse)
              
              no_cgapd_mse_vector =
                c(linear_reg_mse_orig_data, lasso_mse_orig_data,
                  nnet_reg_mse_orig_data, radial_svm_reg_mse_orig_data,
                  linear_svm_reg_mse_orig_data, poly_svm_reg_mse_orig_data,
                  ranger_rf_mse_orig_data, xgbt_mse_orig_data,
                  avg_stacked_ensemble_no_cgapd_mse, 
                  prop_var_stacked_ens_no_cgapd_mse)
              
              #print("Constructing the final results")
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "training_size","testing_size", "y", 
                                   "LR", "LASSO", "NNET", "Radial_SVM",
                                   "Linear_SVM", "Poly_SVM", "RF", "XGBTree",
                                   "Stacked_equal", "Stacked_prop","cgadp", "theta")
              
              
              # construct the dataframe of results
              # With CGADP
              
              estimator_reg_df = matrix( rep(0, length(data_frame_names) * 2), nrow = 2, ncol = length(data_frame_names), byrow = TRUE)
              estimator_reg_df[1,] = c(dat.seed,iteration, samp_n[samp_size_index], testing_n, 
                                       response_index, cgapd_mse_vector, 1, theta_value)
              
              # Adding results for LASSO without CGAPD
              estimator_reg_df[2,] = c(dat.seed, iteration, 
                                       samp_n[samp_size_index], testing_n,
                                       response_index, no_cgapd_mse_vector, 0, theta_value)
              
              
              colnames(estimator_reg_df) = data_frame_names
              
              return(estimator_reg_df)
            }
  #### Stopping Cluster -------------
  
  results = data.frame(results)
  
  # Add name of the estimators
  
  return(results)
}


# Parameters for the Simualtion Models ------
#' Initializing the parameters to be able to construct the simulation models.

# Sample sizes
n_sizes = c(500, 1000, 2000)

# Correlation amongst the predictors
rho_params = seq(0.2,0.6, by=0.2)

# Privacy Parameter, lower is more privacy and less data utility
var_theta = seq(0.2,0.8, by=0.2)

# Quantity of Predictors
# Will start to increase compute time for the SVMs more and more.
# particularly the Polynomial SVM
predictor_quant = c(10, 50, 100)

# Number of simulations to run
m_sims = 1000

# Get the start and ends of the simulations
# Use in case there are memory issues
# sim_start = c(1, 201, 401, 601, 801)
# sim_end = c(200, 400, 600, 1000)

# Get all combinations of simulation parameters
# Edit as necessary as the simulations likely will not finish in one run
param_df = expand_grid(rho_params, var_theta, predictor_quant)
param_df = param_df %>% arrange(predictor_quant, var_theta, rho_params)


# Set the working Directory
#setwd(here::here())



base_dir = "~/" # Your base directory goes here (where it will reference to store results)

setwd(base_dir)
# Setup broader data directory
initial_dat_dir = paste0(base_dir,"/Data/")

if(dir.exists(initial_dat_dir) == FALSE)
{
  dir.create(initial_dat_dir)
}

# Setup data directory 
data_dir = paste0(initial_dat_dir,"tuned-sims/")

if(dir.exists(data_dir) == FALSE)
{
  dir.create(data_dir)
}

# Data Dirctory for Model 1
model_1_dir = paste0(data_dir, "Model-1/")

if(dir.exists(model_1_dir) == FALSE)
{
  dir.create(model_1_dir)
}

for(row_index in 1:nrow(param_df))
{
  results_df = estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
                                                            m = m_sims, # end of portion of simulations
                                                            samp_n = n_sizes, # sample sizes to be used for training set
                                          cores = core_alloc, # Number of CPU threads/ logical processors to use
                                          response_prefix = "y", # prefix for response variables
                                          acquire_variable_names = get_variables, #Function for getting the variables
                                          perturbation_func = model_1_cgadp, # CGADP or GADP function
                                          theta_value = param_df$var_theta[row_index], # privacy parameter value
                                          model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
                                          testing_data_seed = 30000, # Seed for the testing set
                                          model_predictor_quant = param_df$predictor_quant[row_index], # Quantity of the predictors for the model
                                          model_rho = param_df$rho_params[row_index], # correlation of predictors in the model
                                          model_noise = 0.2, # Noise to be used in the model
                                          testing_n = 500 # size of testing set
                                          )
  # string for the rho value
  rho_string = str_replace(paste0(param_df$rho_params[row_index]), "0.", "0_")
  # String for the privacy parameter value
  var_theta_string = str_replace(paste0(param_df$var_theta[row_index]), "0.", "0_")
  # String for the qunatity of predictors
  predictor_string = paste0(param_df$predictor_quant[row_index])
  # Name of the results
  results_file_name = paste0("model-1-cgadp-rho-",
                        rho_string, "-theta-",
                        var_theta_string, 
                        "-pred-", predictor_string,".csv")
  
  produce_csv(created_object= results_df, base_str = model_1_dir, file_name_str = results_file_name)
  
  remove(results_df)
  remove(results_file_name)
}
# CLose the R session when COde is done
quit(save = "no")


