#' Estimator evaluation code
#' Thomas Johnson III
#' 10-09-2023
#' 
#' 

# Run code to get GAPD function loaded
source(paste0(here::here(), "/Code/GADP-from-scratch.R"))


get_variables = function(target_variable_prefix = "y", variable_names)
{
  target_variables = sapply(variable_names, str_detect, 
                            pattern = target_variable_prefix)
  target_variable_indices = which(target_variables == TRUE)
  predictor_variables = variable_names[-target_variable_indices]
  target_and_predictor_vars = list("responses" = variable_names[target_variable_indices],
                                   "predictors" = predictor_variables)
  return(target_and_predictor_vars)
}



sdc_formula_generation = function(target_variables, predictor_variables)
{
  # get the response part of the formula
  response_part = paste0(target_variables,collapse = " + ")
  
  # Get the predictor part of the formula
  predictors_part = paste0(predictor_variables,collapse = " + ")
  
  # Get the complete formula
  complete_formula = formula(paste0(response_part, " ~ ", predictors_part, collapse= ""))
  return(complete_formula)
}

## LASSO ------------------------------------------
# Initial code for evaluating the estimator
# Focused on the LASSO
estimator_evaluation_lasso = function(model_data, # the dataset
                                       m = 1000, # Number of Monte Carlo simulations
                                       n = c(500, 800), # sample sizes to be used
                                       cores = 2, # Number of CPU threads/ logical processors to use
                                       response_prefix = "y",
                                       acquire_variable_names = get_variables,
                                       sdc_form = sdc_formula_generation,
                                       perturbation_func = gadp_func,
                                      theta_value = 0.4)  # Value to control the level of security via peturbation
{
  
  # Getting variable names:
  all_variable_names = colnames(model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  
  # sdc_data_formula = sdc_form(target_variables = organized_vars$responses, 
  #          predictor_variables = organized_vars$predictors)
  
  
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  sim_cluster = makeCluster(cores)
  registerDoSNOW(sim_cluster)
  
  results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
            .packages = c("glmnet", "sdcMicro"),
          .inorder = FALSE) %:%
    foreach( iteration = 1:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$responses), 
            .combine = "rbind",.inorder = FALSE) %dopar% {
    
    # setting the seed
    dat.seed = iteration + n[samp_size_index]
    set.seed(dat.seed)
    
    # Performing SDC GADP method
    privatized_model_response = 
      perturbation_func(unprotected_data = model_data, 
                target_indices = target_variables_indices, 
                theta = theta_value, dat.seed = dat.seed)
    
    privatized_responses = privatized_model_response
    
    current_response = privatized_responses[,response_index]
    
    privatized_dataframe = model_data[,organized_vars$predictors]
    
    privatized_dataframe[,colnames(privatized_responses)[response_index]] = 
      current_response
    
    # Get the sample from privatized df
    drawing = sample(1:nrow(privatized_dataframe), size = n[samp_size_index], replace = FALSE)
    
    # Sample data after gapd process
    sample_private_data = privatized_dataframe[drawing,]
    
    # Sample data without gapd process
    sample_unprotected_data = model_data[drawing,]
    
    # LASSO on data after gapd
    lasso_regress_model = cv.glmnet( 
      x = as.matrix(sample_private_data[,organized_vars$predictors]),
                                  y = current_response[drawing])
    
    lasso_pred = predict(lasso_regress_model, 
                         as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]),
                         s = "lambda.min")
    
    lasso_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         lasso_pred)
    
    # LASSO on data before gapd
    
    
    lasso_regress_model_orig_data =
      cv.glmnet( x = as.matrix(sample_unprotected_data[,organized_vars$predictors]),
              y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
    
    
    lasso_pred_orig_data = predict(lasso_regress_model_orig_data,
                         as.matrix(model_data[-drawing,organized_vars$predictors]), 
                         s = "lambda.min")
    
    lasso_mse_orig_data =
      calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
               lasso_pred_orig_data)
    
    # Dataframe names
    data_frame_names = c("seed", "iteration", "sample_size", "y", "mse", "gadp")
    
    
    # construct the dataframe of results
    lasso_df = data.frame(dat.seed,iteration, n[samp_size_index], 
                          response_index, lasso_mse, 1)
    
    # Adding results for LASSO without GAPD
    lasso_df[2,] = c(dat.seed,iteration, n[samp_size_index],
                     response_index, lasso_mse_orig_data, 0)
    
    colnames(lasso_df) = data_frame_names
    
    
    return(lasso_df)
            }
  
  stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  # Add name of the estimator
  results$estimator = "LASSO"
  results$perturb_value = theta_value
  return(results)
}





## Linear Regression ------------------

# Initial code for evaluating the estimator
# Focused on the Linear Regression
estimator_evaluation_linear_reg = function(model_data, # the dataset
                                      m = 1000, # Number of Monte Carlo simulations
                                      n = c(500, 800), # sample sizes to be used
                                      cores = 2,
                                      response_prefix = "y",
                                      acquire_variable_names = get_variables,
                                      sdc_form = sdc_formula_generation,
                                      perturbation_func = gadp_func,
                                      theta_value = 0.4) # Number of CPU threads/ logical processors to use
{
  
  # Getting variable names:
  all_variable_names = colnames(model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  
  # print(organized_vars)
  # sdc_data_formula = sdc_form(target_variables = organized_vars$response, 
  #                             predictor_variables = organized_vars$predictors)
  
  
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  sim_cluster = makeCluster(cores)
  registerDoSNOW(sim_cluster)
  
  results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
                    .packages = c("sdcMicro"),
                    .inorder = FALSE) %:%
    foreach( iteration = 1:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",.inorder = FALSE) %dopar% {
              
              # setting the seed
              dat.seed = iteration + n[samp_size_index]
              set.seed(dat.seed)
              
              # Performing SDC GADP method
              privatized_model_response = 
                perturbation_func(unprotected_data = model_data, 
                          target_indices = target_variables_indices, 
                          theta = theta_value, dat.seed = dat.seed)
              
              privatized_responses = privatized_model_response
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,organized_vars$predictors]
              
              privatized_dataframe[,colnames(privatized_responses)[response_index]] = 
                current_response
              
              # Get the sample from privatized df
              drawing = sample(1:nrow(privatized_dataframe), size = n[samp_size_index], replace = FALSE)
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              x_dat_holdout = cbind(1,as.matrix(model_data)[-drawing,organized_vars$predictors])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after gapd
              x_dat = as.matrix(cbind(1,
                                      sample_private_data[,organized_vars$predictors]))
                
              # Linear Regression on data after gapd
              betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(current_response[drawing])
              
              
              linear_reg_pred = x_dat_holdout %*% betas_linear
              
              
              linear_reg_mse = calc_mse(orig_data_response[-drawing],
                                        t(linear_reg_pred[,1]))
              
              # LASSO on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              

              orig_data_betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(sample_orig_data_response)
              
              
              
              linear_reg_unprivatized_pred = x_dat_holdout %*% orig_data_betas_linear
              
              linear_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         t(linear_reg_unprivatized_pred[,1]))
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "sample_size", "y", "mse", "gadp")
              
              
              # construct the dataframe of results
              linear_reg_df = data.frame(dat.seed,iteration, n[samp_size_index], 
                                    response_index, linear_reg_mse, 1)
              
              # Adding results for LASSO without GAPD
              linear_reg_df[2,] = c(dat.seed,iteration, n[samp_size_index],
                               response_index, linear_reg_mse_orig_data, 0)
              
              colnames(linear_reg_df) = data_frame_names
              
              
              return(linear_reg_df)
            }
  
  stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  # Add name of the estimator
  results$estimator = "LR"
  results$perturb_value = theta_value
  return(results)
}

## Neural Network ---------------------------------
# initial code for evaluating the estimator
# Focused on the Linear Regression
estimator_evaluation_nnet_reg = function(model_data, # the dataset
                                           m = 1000, # Number of Monte Carlo simulations
                                           n = c(500, 800), # sample sizes to be used
                                           cores = 2,
                                           response_prefix = "y",
                                           acquire_variable_names = get_variables,
                                           sdc_form = sdc_formula_generation,
                                           perturbation_func = gadp_func,
                                           theta_value = 0.4, # Number of CPU threads/ logical processors to use
                                         hidden_layer_count = 3) 
{
  
  # Getting variable names:
  all_variable_names = colnames(model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  
  # print(organized_vars)
  # sdc_data_formula = sdc_form(target_variables = organized_vars$response, 
  #                             predictor_variables = organized_vars$predictors)
  
  
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  sim_cluster = makeCluster(cores)
  registerDoSNOW(sim_cluster)
  
  results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
                    .packages = c("nnet","sdcMicro"),
                    .inorder = FALSE) %:%
    foreach( iteration = 1:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",.inorder = FALSE) %dopar% {
              
              # setting the seed
              dat.seed = iteration + n[samp_size_index]
              set.seed(dat.seed)
              
              # Performing SDC GADP method
              privatized_model_response = 
                perturbation_func(unprotected_data = model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              privatized_responses = privatized_model_response
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              # current_response = matrix(current_response, nrow=length(current_response),
              #                           ncol = 1, byrow = TRUE)
              
              privatized_dataframe = model_data[,organized_vars$predictors]
              
              privatized_dataframe[,colnames(privatized_responses)[response_index]] = 
                current_response
              
              # Get the sample from privatized df
              drawing = sample(1:nrow(privatized_dataframe), size = n[samp_size_index], replace = FALSE)
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              x_dat_holdout = cbind(1,as.matrix(model_data)[-drawing,organized_vars$predictors])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after gapd
              x_dat = as.matrix(cbind(1,
                                      sample_private_data[,organized_vars$predictors]))
              
              # print(dim(current_response))
              
              # Neural Net Regression on data after gapd
              gapd_nnet = nnet(x = x_dat, y = current_response[drawing], size = hidden_layer_count,
                               maxit = 200, linout = TRUE)
              
              
              gapd_nnet_reg_pred = predict(gapd_nnet,x_dat_holdout)
              
              
              gapd_nnet_reg_mse = calc_mse(orig_data_response[-drawing],
                                        gapd_nnet_reg_pred)
              
              # Neural Net on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              # sample_orig_data_response = matrix(sample_orig_data_response, 
              #                                    nrow = length(sample_orig_data_response),
              #                                    ncol = 1,
              #                                    byrow = TRUE)
              
              no_gapd_nnet = nnet(x = x_dat, y = sample_orig_data_response, 
                                  size = hidden_layer_count,
                                  maxit = 200, linout = TRUE)
              
              
              
              nnet_reg_unprivatized_pred = predict(no_gapd_nnet, x_dat_holdout)
              
              nnet_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         nnet_reg_unprivatized_pred)
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "sample_size", "y", "mse", "gadp")
              
              
              # construct the dataframe of results
              nnet_reg_df = data.frame(dat.seed,iteration, n[samp_size_index], 
                                         response_index, gapd_nnet_reg_mse, 1)
              
              # Adding results for LASSO without GAPD
              nnet_reg_df[2,] = c(dat.seed,iteration, n[samp_size_index],
                                    response_index, nnet_reg_mse_orig_data, 0)
              
              colnames(nnet_reg_df) = data_frame_names
              
              
              return(nnet_reg_df)
            }
  
  stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  # Add name of the estimator
  results$estimator = "NN"
  results$perturb_value = theta_value
  return(results)
}




## SVM Radial ---------------------------------
# initial code for evaluating the estimator
# Focused on the Linear Regression
estimator_evaluation_svm_radial_reg = function(model_data, # the dataset
                                         m = 1000, # Number of Monte Carlo simulations
                                         n = c(500, 800), # sample sizes to be used
                                         cores = 2,
                                         response_prefix = "y",
                                         acquire_variable_names = get_variables,
                                         sdc_form = sdc_formula_generation,
                                         perturbation_func = gadp_func,
                                         theta_value = 0.4) # Number of CPU threads/ logical processors to use
{
  
  # Getting variable names:
  all_variable_names = colnames(model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  
  # print(organized_vars)
  # sdc_data_formula = sdc_form(target_variables = organized_vars$response, 
  #                             predictor_variables = organized_vars$predictors)
  
  
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  sim_cluster = makeCluster(cores)
  registerDoSNOW(sim_cluster)
  
  results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
                    .packages = c("e1071","sdcMicro"),
                    .inorder = FALSE) %:%
    foreach( iteration = 1:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",.inorder = FALSE) %dopar% {
              
              # setting the seed
              dat.seed = iteration + n[samp_size_index]
              set.seed(dat.seed)
              
              # Performing SDC GADP method
              privatized_model_response = 
                perturbation_func(unprotected_data = model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              privatized_responses = privatized_model_response
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              # current_response = matrix(current_response, nrow=length(current_response),
              #                           ncol = 1, byrow = TRUE)
              
              privatized_dataframe = model_data[,organized_vars$predictors]
              
              privatized_dataframe[,colnames(privatized_responses)[response_index]] = 
                current_response
              
              # Get the sample from privatized df
              drawing = sample(1:nrow(privatized_dataframe), size = n[samp_size_index], replace = FALSE)
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              x_dat_holdout = cbind(1,as.matrix(model_data)[-drawing,organized_vars$predictors])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after gapd
              x_dat = as.matrix(cbind(1,
                                      sample_private_data[,organized_vars$predictors]))
              
              # print(dim(current_response))
              
              # SVM Regression on data after gapd
              gapd_svm_tune = tune.svm(x = x_dat, y = current_response[drawing],
                                       gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                       degree = 3:4,
                                   kernel = c("radial"))
              
              gapd_svm = gapd_svm_tune$best.model
              
              gapd_svm_reg_pred = predict(gapd_svm,x_dat_holdout)
              
              
              gapd_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                           gapd_svm_reg_pred)
              
              # LASSO on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              # sample_orig_data_response = matrix(sample_orig_data_response, 
              #                                    nrow = length(sample_orig_data_response),
              #                                    ncol = 1,
              #                                    byrow = TRUE)
              
              no_gapd_svm_tune = tune.svm(x = x_dat, 
                                          y = sample_orig_data_response,
                                          gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                          degree = 3:4,
                                       kernel = c("radial"))
              
              no_gapd_svm = no_gapd_svm_tune$best.model
              
              svm_reg_unprivatized_pred = predict(no_gapd_svm, x_dat_holdout)
              
              svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         svm_reg_unprivatized_pred)
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "sample_size", "y", "mse", "gadp")
              
              
              # construct the dataframe of results
              svm_reg_df = data.frame(dat.seed,iteration, n[samp_size_index], 
                                       response_index, gapd_svm_reg_mse, 1)
              
              # Adding results for LASSO without GAPD
              svm_reg_df[2,] = c(dat.seed,iteration, n[samp_size_index],
                                  response_index, svm_reg_mse_orig_data, 0)
              
              colnames(svm_reg_df) = data_frame_names
              
              
              return(svm_reg_df)
            }
  
  stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  # Add name of the estimator
  results$estimator = "SVM_Radial"
  results$perturb_value = theta_value
  return(results)
}


## SVM Linear ---------------------------------
# initial code for evaluating the estimator
# Focused on the Linear Regression
estimator_evaluation_svm_linear_reg = function(model_data, # the dataset
                                               m = 1000, # Number of Monte Carlo simulations
                                               n = c(500, 800), # sample sizes to be used
                                               cores = 2,
                                               response_prefix = "y",
                                               acquire_variable_names = get_variables,
                                               sdc_form = sdc_formula_generation,
                                               perturbation_func = gadp_func,
                                               theta_value = 0.4) # Number of CPU threads/ logical processors to use
{
  
  # Getting variable names:
  all_variable_names = colnames(model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  
  # print(organized_vars)
  # sdc_data_formula = sdc_form(target_variables = organized_vars$response, 
  #                             predictor_variables = organized_vars$predictors)
  
  
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  sim_cluster = makeCluster(cores)
  registerDoSNOW(sim_cluster)
  
  results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
                    .packages = c("e1071","sdcMicro"),
                    .inorder = FALSE) %:%
    foreach( iteration = 1:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",.inorder = FALSE) %dopar% {
              
              # setting the seed
              dat.seed = iteration + n[samp_size_index]
              set.seed(dat.seed)
              
              # Performing SDC GADP method
              privatized_model_response = 
                perturbation_func(unprotected_data = model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              privatized_responses = privatized_model_response
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              # current_response = matrix(current_response, nrow=length(current_response),
              #                           ncol = 1, byrow = TRUE)
              
              privatized_dataframe = model_data[,organized_vars$predictors]
              
              privatized_dataframe[,colnames(privatized_responses)[response_index]] = 
                current_response
              
              # Get the sample from privatized df
              drawing = sample(1:nrow(privatized_dataframe), size = n[samp_size_index], replace = FALSE)
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              x_dat_holdout = cbind(1,as.matrix(model_data)[-drawing,organized_vars$predictors])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after gapd
              x_dat = as.matrix(cbind(1,
                                      sample_private_data[,organized_vars$predictors]))
              
              # print(dim(current_response))
              
              # Linear Regression on data after gapd
              gapd_svm_tune = tune.svm(x = x_dat, y = current_response[drawing],
                                       gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                       degree = 3:4,
                                       kernel = c("linear"))
              
              gapd_svm = gapd_svm_tune$best.model
              
              gapd_svm_reg_pred = predict(gapd_svm,x_dat_holdout)
              
              
              gapd_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                          gapd_svm_reg_pred)
              
              # LASSO on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              # sample_orig_data_response = matrix(sample_orig_data_response, 
              #                                    nrow = length(sample_orig_data_response),
              #                                    ncol = 1,
              #                                    byrow = TRUE)
              
              no_gapd_svm_tune = tune.svm(x = x_dat, 
                                          y = sample_orig_data_response,
                                          gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                          degree = 3:4,
                                          kernel = c("linear"))
              
              no_gapd_svm = no_gapd_svm_tune$best.model
              
              svm_reg_unprivatized_pred = predict(no_gapd_svm, x_dat_holdout)
              
              svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         svm_reg_unprivatized_pred)
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "sample_size", "y", "mse", "gadp")
              
              
              # construct the dataframe of results
              svm_reg_df = data.frame(dat.seed,iteration, n[samp_size_index], 
                                      response_index, gapd_svm_reg_mse, 1)
              
              # Adding results for LASSO without GAPD
              svm_reg_df[2,] = c(dat.seed,iteration, n[samp_size_index],
                                 response_index, svm_reg_mse_orig_data, 0)
              
              colnames(svm_reg_df) = data_frame_names
              
              
              return(svm_reg_df)
            }
  
  stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  # Add name of the estimator
  results$estimator = "SVM_Linear"
  results$perturb_value = theta_value
  return(results)
}


## SVM Poly ---------------------------------
# initial code for evaluating the estimator
# Focused on the SVM Regression
estimator_evaluation_svm_poly_reg = function(model_data, # the dataset
                                               m = 1000, # Number of Monte Carlo simulations
                                               n = c(500, 800), # sample sizes to be used
                                               cores = 2,
                                               response_prefix = "y",
                                               acquire_variable_names = get_variables,
                                               sdc_form = sdc_formula_generation,
                                               perturbation_func = gadp_func,
                                               theta_value = 0.4) # Number of CPU threads/ logical processors to use
{
  
  # Getting variable names:
  all_variable_names = colnames(model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  
  # print(organized_vars)
  # sdc_data_formula = sdc_form(target_variables = organized_vars$response, 
  #                             predictor_variables = organized_vars$predictors)
  
  
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  sim_cluster = makeCluster(cores)
  registerDoSNOW(sim_cluster)
  
  results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
                    .packages = c("e1071","sdcMicro"),
                    .inorder = FALSE) %:%
    foreach( iteration = 1:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",.inorder = FALSE) %dopar% {
              
              # setting the seed
              dat.seed = iteration + n[samp_size_index]
              set.seed(dat.seed)
              
              # Performing SDC GADP method
              privatized_model_response = 
                perturbation_func(unprotected_data = model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              privatized_responses = privatized_model_response
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              # current_response = matrix(current_response, nrow=length(current_response),
              #                           ncol = 1, byrow = TRUE)
              
              privatized_dataframe = model_data[,organized_vars$predictors]
              
              privatized_dataframe[,colnames(privatized_responses)[response_index]] = 
                current_response
              
              # Get the sample from privatized df
              drawing = sample(1:nrow(privatized_dataframe), size = n[samp_size_index], replace = FALSE)
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              x_dat_holdout = cbind(1,as.matrix(model_data)[-drawing,organized_vars$predictors])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after gapd
              x_dat = as.matrix(cbind(1,
                                      sample_private_data[,organized_vars$predictors]))
              
              # print(dim(current_response))
              
              # Linear Regression on data after gapd
              gapd_svm_tune = tune.svm(x = x_dat, y = current_response[drawing],
                                       gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                       degree = 3:4,
                                       kernel = c("polynomial"))
              
              gapd_svm = gapd_svm_tune$best.model
              
              gapd_svm_reg_pred = predict(gapd_svm,x_dat_holdout)
              
              
              gapd_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                          gapd_svm_reg_pred)
              
              # LASSO on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              # sample_orig_data_response = matrix(sample_orig_data_response, 
              #                                    nrow = length(sample_orig_data_response),
              #                                    ncol = 1,
              #                                    byrow = TRUE)
              
              no_gapd_svm_tune = tune.svm(x = x_dat, 
                                          y = sample_orig_data_response,
                                          gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                          degree = 3:4,
                                          kernel = c("polynomial"))
              
              no_gapd_svm = no_gapd_svm_tune$best.model
              
              svm_reg_unprivatized_pred = predict(no_gapd_svm, x_dat_holdout)
              
              svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         svm_reg_unprivatized_pred)
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "sample_size", "y", "mse", "gadp")
              
              
              # construct the dataframe of results
              svm_reg_df = data.frame(dat.seed,iteration, n[samp_size_index], 
                                      response_index, gapd_svm_reg_mse, 1)
              
              # Adding results for LASSO without GAPD
              svm_reg_df[2,] = c(dat.seed,iteration, n[samp_size_index],
                                 response_index, svm_reg_mse_orig_data, 0)
              
              colnames(svm_reg_df) = data_frame_names
              
              
              return(svm_reg_df)
            }
  
  stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  # Add name of the estimator
  results$estimator = "SVM_Poly"
  results$perturb_value = theta_value
  return(results)
}






## Random forest -----------------------
# Initial code for evaluating the estimator
# Focused on the Random Forest
estimator_evaluation_ranger = function(model_data, # the dataset
                                      m = 1000, # Number of Monte Carlo simulations
                                      n = c(500, 800), # sample sizes to be used
                                      cores = 2,
                                      response_prefix = "y",
                                      acquire_variable_names = get_variables,
                                      sdc_form = sdc_formula_generation,
                                      perturbation_func = gadp_func,
                                      theta_value = 0.4) # Number of CPU threads/ logical processors to use
{
  
  # Getting variable names:
  all_variable_names = colnames(model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  
  # print(organized_vars)
  # sdc_data_formula = sdc_form(target_variables = organized_vars$response, 
  #                             predictor_variables = organized_vars$predictors)
  
  
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  sim_cluster = makeCluster(cores)
  registerDoSNOW(sim_cluster)
  
  results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
                    .packages = c("ranger", "sdcMicro"),
                    .inorder = FALSE) %:%
    foreach( iteration = 1:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",.inorder = FALSE) %dopar% {
              
              # setting the seed
              dat.seed = iteration + n[samp_size_index]
              set.seed(dat.seed)
              
              # Performing SDC GADP method
              privatized_model_response = 
                perturbation_func(unprotected_data = model_data, 
                          target_indices = target_variables_indices, 
                          theta = theta_value, dat.seed = dat.seed)
              
              privatized_responses = privatized_model_response
              
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,organized_vars$predictors]
              
              privatized_dataframe[,colnames(privatized_responses)[response_index]] = 
                current_response
              
              # Get the sample from privatized df
              drawing = sample(1:nrow(privatized_dataframe), size = n[samp_size_index], replace = FALSE)
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              num_x_vars = length(organized_vars$predictors)
              # Random forest on data after gapd
              ranger_rf_regress_model = ranger( 
                x = as.matrix(sample_private_data[,organized_vars$predictors]),
                y = current_response[drawing], 
                num.trees = 500,
                mtry = ceiling(1 / 3 * num_x_vars))
              
              ranger_rf_pred = predict(ranger_rf_regress_model, 
                                   privatized_dataframe[-drawing,organized_vars$predictors])$predictions
              
              ranger_rf_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                   ranger_rf_pred)
              
              # LASSO on data before gapd
              
              
              ranger_rf_model_orig_data = ranger( 
                x = as.matrix(sample_unprotected_data[,organized_vars$predictors]),
                y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]), 
                num.trees = 500,
                mtry = ceiling(1 / 3 * num_x_vars))
                
              
              
              ranger_rf_pred_orig_data = predict(ranger_rf_model_orig_data,
                                             model_data[-drawing,organized_vars$predictors])$predictions
              
              ranger_rf_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         ranger_rf_pred_orig_data)
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "sample_size", "y", "mse", "gadp")
              
              
              # construct the dataframe of results
              ranger_rf_df = data.frame(dat.seed,iteration, n[samp_size_index], 
                                    response_index, ranger_rf_mse, 1)
              
              # Adding results for LASSO without GAPD
              ranger_rf_df[2,] = c(dat.seed,iteration, n[samp_size_index],
                               response_index, ranger_rf_mse_orig_data, 0)
              
              colnames(ranger_rf_df) = data_frame_names
              
              
              return(ranger_rf_df)
            }
  
  stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  # Add name of the estimator
  results$estimator = "RF"
  results$perturb_value = theta_value
  return(results)
}

## XGBoost Tree -------------------------------------------------------------

estimator_evaluation_xgbt = function(model_data, # the dataset
                                      m = 1000, # Number of Monte Carlo simulations
                                      n = c(500, 800), # sample sizes to be used
                                      cores = 2, # Number of CPU threads/ logical processors to use
                                      response_prefix = "y",
                                      acquire_variable_names = get_variables,
                                      sdc_form = sdc_formula_generation,
                                      perturbation_func = gadp_func,
                                      theta_value = 0.4)  # Value to control the level of security via peturbation
{
  
  # Getting variable names:
  all_variable_names = colnames(model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  
  # sdc_data_formula = sdc_form(target_variables = organized_vars$responses, 
  #          predictor_variables = organized_vars$predictors)
  
  
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  sim_cluster = makeCluster(cores)
  registerDoSNOW(sim_cluster)
  
  results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
                    .packages = c("xgboost", "sdcMicro"),
                    .inorder = FALSE) %:%
    foreach( iteration = 1:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$responses), 
            .combine = "rbind",.inorder = FALSE) %dopar% {
              
              # setting the seed
              dat.seed = iteration + n[samp_size_index]
              set.seed(dat.seed)
              
              # Performing SDC GADP method
              privatized_model_response = 
                perturbation_func(unprotected_data = model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              privatized_responses = privatized_model_response
              
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,organized_vars$predictors]
              
              privatized_dataframe[,colnames(privatized_responses)[response_index]] = 
                current_response
              
              # Get the sample from privatized df
              drawing = sample(1:nrow(privatized_dataframe), size = n[samp_size_index], replace = FALSE)
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # XGBT on data after gapd
              xgbt_regress_model = xgboost( 
                data = as.matrix(sample_private_data[,organized_vars$predictors]),
                label = current_response[drawing], 
                nrounds = 40,
                nthread = 1)
              
              xgbt_pred = predict(xgbt_regress_model, 
                                   as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]))
              
              xgbt_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                   xgbt_pred)
              
              # XGBT on data before gapd
              
              
              xgbt_regress_model_orig_data =
                xgboost( data = as.matrix(sample_unprotected_data[,organized_vars$predictors]),
                           label = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]),
                           nrounds = 40,
                         nthread = 1)
              
              
              xgbt_pred_orig_data = predict(xgbt_regress_model_orig_data,
                                             as.matrix(model_data[-drawing,organized_vars$predictors]))
              
              xgbt_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         xgbt_pred_orig_data)
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "sample_size", "y", "mse", "gadp")
              
              
              # construct the dataframe of results
              xgbt_df = data.frame(dat.seed,iteration, n[samp_size_index], 
                                    response_index, xgbt_mse, 1)
              
              # Adding results for LASSO without GAPD
              xgbt_df[2,] = c(dat.seed,iteration, n[samp_size_index],
                               response_index, xgbt_mse_orig_data, 0)
              
              colnames(xgbt_df) = data_frame_names
              
              
              return(xgbt_df)
            }
  
  stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  # Add name of the estimator
  results$estimator = "XGBTree"
  results$perturb_value = theta_value
  return(results)
}

## Stacked Ensemble Static Dataset ---------
estimator_evaluation_stacked_ensemb = function(model_data, # the dataset
                                               m_start = 1, # Primarily used for debugging determines which iteration to start from
                                           m = 1000, # Number of Monte Carlo simulations
                                           n = c(500, 800), # sample sizes to be used
                                           cores = 2,
                                           response_prefix = "y",
                                           acquire_variable_names = get_variables,
                                           sdc_form = sdc_formula_generation,
                                           perturbation_func = gadp_func,
                                           theta_value = 0.4) # Number of CPU threads/ logical processors to use
{
  
  # Getting variable names:
  all_variable_names = colnames(model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  
  # print(organized_vars)
  # sdc_data_formula = sdc_form(target_variables = organized_vars$response, 
  #                             predictor_variables = organized_vars$predictors)
  
  
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  sim_cluster = makeCluster(cores)
  registerDoSNOW(sim_cluster)
  
  results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
                    .packages = c("glmnet", "ranger", 
                                  "nnet", "e1071", "xgboost"),
                    .inorder = FALSE) %:%
    foreach( iteration = m_start:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",.inorder = FALSE) %dopar% {
              
              # setting the seed
              dat.seed = iteration + n[samp_size_index]
              set.seed(dat.seed)
              
              # Performing SDC GADP method
              privatized_model_response = 
                perturbation_func(unprotected_data = model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              privatized_responses = privatized_model_response
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,organized_vars$predictors]
              
              privatized_dataframe[,colnames(privatized_responses)[response_index]] = 
                current_response
              
              # Get the sample from privatized df
              drawing = sample(1:nrow(privatized_dataframe), size = n[samp_size_index], replace = FALSE)
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              x_dat_holdout = cbind(1,as.matrix(model_data)[-drawing,organized_vars$predictors])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after gapd
              x_dat = as.matrix(cbind(1,
                                      sample_private_data[,organized_vars$predictors]))
              
              ### GAPD Used----
              
              # Linear Regression on data after gapd
              betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(current_response[drawing])
              
              
              linear_reg_pred = as.vector(x_dat_holdout %*% betas_linear)
              
              # Get the predictions for the training data
              
              training_lin_reg_pred = as.vector(x_dat %*% betas_linear)
              
              
              # Get the mse for the training data
              
              
              training_lin_reg_mse = calc_mse(as.matrix(current_response[drawing]),
                                              training_lin_reg_pred)
              
              # Get the mse for the test data
              
              
              linear_reg_mse = calc_mse(orig_data_response[-drawing],
                                        linear_reg_pred)
              
              
              # LASSO on data after gapd
              lasso_regress_model = cv.glmnet( 
                x = as.matrix(sample_private_data[,organized_vars$predictors]),
                y = current_response[drawing])
              
              lasso_pred = predict(lasso_regress_model, 
                                   as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]),
                                   s = "lambda.min")
              
              # get predictions on the training data
              training_lasso_pred = predict(lasso_regress_model, 
                                            as.matrix(privatized_dataframe[drawing,organized_vars$predictors]),
                                            s = "lambda.min")
              
              # Get the MSE for the training data
              training_lasso_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_lasso_pred)
              
              # # MSE for the testing data
              
              lasso_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                   lasso_pred)
              
              # Neural Net Regression on data after gapd
              gapd_nnet = nnet(x = x_dat[,organized_vars$predictors], y = current_response[drawing], 
                               size = 3,
                               maxit = 200, linout = TRUE)
              
              
              gapd_nnet_reg_pred = predict(gapd_nnet,x_dat_holdout[,organized_vars$predictors])
              
              # training data predictions
              training_gapd_nnet_reg_pred = predict(gapd_nnet,x_dat)
              
              # training data MSE
              training_gapd_nnet_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                       training_gapd_nnet_reg_pred)
              
              
              # testing data MSE
              gapd_nnet_reg_mse = calc_mse(orig_data_response[-drawing],
                                           gapd_nnet_reg_pred)
              
              # SVM Radial Regression on data after gapd
              gapd_radial_svm_tune = tune.svm(x = x_dat[,organized_vars$predictors], y = current_response[drawing],
                                       gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                       degree = 3:4,
                                       kernel = c("radial"))
              
              gapd_radial_svm = gapd_radial_svm_tune$best.model
              
              gapd_radial_svm_reg_pred = predict(gapd_radial_svm,x_dat_holdout[,organized_vars$predictors])
              
              # Get the training predictions of the radial svm
              training_gapd_radial_svm_reg_pred =
                predict(gapd_radial_svm,x_dat[,organized_vars$predictors])
              
              # Get the training mse for radial svm
              training_gapd_radial_svm_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_radial_svm_reg_pred)
              
              # Testing data MSE
              gapd_radial_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                          gapd_radial_svm_reg_pred)
              
              
              
              # SVM Linear Regression on data after gapd
              gapd_linear_svm_tune = tune.svm(x = x_dat[,organized_vars$predictors], y = current_response[drawing],
                                              gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                              degree = 3:4,
                                              kernel = c("linear"))
              
              gapd_linear_svm = gapd_linear_svm_tune$best.model
              
              # Predictions for the training data for Linear SVM
              training_gapd_linear_svm_reg_pred =
                predict(gapd_linear_svm,x_dat_holdout[,organized_vars$predictors])
              
              # Training MSE for Linear SVM
              training_gapd_linear_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_linear_svm_reg_pred)
              
              # Predictions for the test dataset
              gapd_linear_svm_reg_pred = predict(gapd_linear_svm,x_dat_holdout[,organized_vars$predictors])
              
              # Testing data MSE
              gapd_linear_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                 gapd_linear_svm_reg_pred)
              
              # SVM Polynomial Regression on data after gapd
              gapd_poly_svm_tune = tune.svm(x = x_dat[,organized_vars$predictors], y = current_response[drawing],
                                              gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                              degree = 3:4,
                                              kernel = c("polynomial"))
              
              gapd_poly_svm = gapd_poly_svm_tune$best.model
              
              # Training data polynomial SVM predictions
              training_gapd_poly_svm_reg_pred = predict(gapd_poly_svm,x_dat[,organized_vars$predictors])
              
              # Training data MSE polynomial SVM
              training_gapd_poly_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_poly_svm_reg_pred)
              
              # Predictions for testing dataset
              gapd_poly_svm_reg_pred = predict(gapd_poly_svm,x_dat_holdout[,organized_vars$predictors])
              
              # MSE for testing dataset for polynomial kernel SVM
              gapd_poly_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                 gapd_poly_svm_reg_pred)
              
              # Random forest on data after gapd
              num_x_vars = length(organized_vars$predictors)
              
              ranger_rf_regress_model = ranger( 
                x = as.matrix(sample_private_data[,organized_vars$predictors]),
                y = current_response[drawing], 
                num.trees = 500,
                num.threads = 1,
                mtry = ceiling(1 / 3 * num_x_vars))
              
              # Training dataset predicions for the random forest
              training_ranger_rf_pred = 
                predict(ranger_rf_regress_model, 
                        privatized_dataframe[drawing,organized_vars$predictors])$predictions
              
              # training dataset MSE for the radom forest
              training_ranger_rf_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_ranger_rf_pred)
              
              
              # Testing dataset prediction for random forest
              ranger_rf_pred = predict(ranger_rf_regress_model, 
                                       privatized_dataframe[-drawing,organized_vars$predictors])$predictions
              
              # Testing dataset MSE for random forest
              ranger_rf_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                       ranger_rf_pred)
              
              # XGBT on data after gapd
              xgbt_regress_model = xgboost( 
                data = as.matrix(sample_private_data[,organized_vars$predictors]),
                label = current_response[drawing], 
                nrounds = 40,
                nthread = 1)
              
              # Training data predictions for the XGB tree
              training_xgbt_pred = 
                predict(xgbt_regress_model,
                        as.matrix(privatized_dataframe[drawing,organized_vars$predictors]))
              
              # training data MSE for the XGB tree
              training_xgbt_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_xgbt_pred)
              
              # Testing data predicitons for the XGB tree
              xgbt_pred = predict(xgbt_regress_model, 
                                  as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]))
              
              # Testing data MSE for the XGB tree
              xgbt_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                  xgbt_pred)
              
              ##
              ## Before GAPD (On original data) ------
              
              # Linear Regression on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              
              # Beta coefficients for the original data
              orig_data_betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(sample_orig_data_response)
              
              # Predictions on the training data no GADP
              training_lin_reg_unprivat_pred = as.vector(x_dat %*%orig_data_betas_linear)
              
              
              # MSE for training data for linear regression, no GADP
              training_linear_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_lin_reg_unprivat_pred)
              
              # Testing data MSE for linear regression no ADP
              linear_reg_unprivatized_pred = as.vector(x_dat_holdout %*% orig_data_betas_linear)
              
              linear_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_reg_unprivatized_pred)
              
              # LASSO on data before gapd (original data)
              
              
              lasso_regress_model_orig_data =
                cv.glmnet( x = as.matrix(sample_unprotected_data[,organized_vars$predictors]),
                           y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # Training data predictions for LASSO
              
              training_lasso_pred_orig_data = 
                predict(lasso_regress_model_orig_data,
                        as.matrix(model_data[drawing,organized_vars$predictors]), 
                        s = "lambda.min")
              
              # Training data MSE for LASSO
              training_lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_lasso_pred_orig_data)
              # Predictions for the test dataset for the LASSO model
              lasso_pred_orig_data = predict(lasso_regress_model_orig_data,
                                             as.matrix(model_data[-drawing,organized_vars$predictors]), 
                                             s = "lambda.min")
              # MSE for the LASSO model for the test data
              lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         lasso_pred_orig_data)
              
              
              # Neural Net before GAPD 
              
              no_gapd_nnet = nnet(x = x_dat[,organized_vars$predictors], y = sample_orig_data_response, 
                                  size = 3,
                                  maxit = 200, linout = TRUE)
              
              # The predictions of the training set for neural network
              training_nnet_reg_unprivatized_pred = predict(no_gapd_nnet, x_dat[,organized_vars$predictors])
              
              # The MSE for the training set for the neural network
              training_nnet_reg_unprivatized_mse = 
                calc_mse(orig_data_response[drawing],
                         training_nnet_reg_unprivatized_pred)
              
              # Predicitons on testing set for neural net
              nnet_reg_unprivatized_pred = predict(no_gapd_nnet, x_dat_holdout[,organized_vars$predictors])
              
              # MSE for testing set for neural network
              nnet_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         nnet_reg_unprivatized_pred)
              
              # SVM Radial before GAPD
              
              
              no_gapd_radial_svm_tune = tune.svm(x = x_dat[,organized_vars$predictors], 
                                          y = sample_orig_data_response,
                                          gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                          degree = 3:4,
                                          kernel = c("radial"))
              
              # Get the best model for Radial SVM
              no_gapd_radial_svm = no_gapd_radial_svm_tune$best.model
              
              # Predictions on the training set for radial SVM
              training_radial_svm_reg_unprivatized_pred = 
                predict(no_gapd_radial_svm, x_dat[,organized_vars$predictors])
              # The MSE for the training set for radial SVM
              training_radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_radial_svm_reg_unprivatized_pred)
              
              # Get the predictions of Radial SVM for the testing set
              radial_svm_reg_unprivatized_pred = 
                predict(no_gapd_radial_svm, x_dat_holdout[,organized_vars$predictors])
              # Get the MSE of the Radial SVM for the testing set
              radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         radial_svm_reg_unprivatized_pred)
              
              # Linear SVM before GAPD
              no_gapd_linear_svm_tune = tune.svm(x = x_dat[,organized_vars$predictors], 
                                                 y = sample_orig_data_response,
                                                 gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                                 degree = 3:4,
                                                 kernel = c("linear"))
              # The Linear SVM best tuned model
              no_gapd_linear_svm = no_gapd_linear_svm_tune$best.model
              
              # The predictions on the training set, Linear SVM
              training_linear_svm_reg_unprivatized_pred = 
                predict(no_gapd_linear_svm, x_dat[,organized_vars$predictors])
              # the MSE on the training set, Linear SVM
              training_linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_linear_svm_reg_unprivatized_pred)
              # The predictions for the testing set Linear SVM
              linear_svm_reg_unprivatized_pred = 
                predict(no_gapd_linear_svm, x_dat_holdout[,organized_vars$predictors])
              
              # The MSE for the testing set Linear SVM
              linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_svm_reg_unprivatized_pred)
              
              
              # Polynomial SVM before GAPD
              
              no_gapd_poly_svm_tune = tune.svm(x = x_dat[,organized_vars$predictors], 
                                                 y = sample_orig_data_response,
                                                 gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                                 degree = 3:4,
                                                 kernel = c("polynomial"))
              # Get the best tuned polynomial SVM
              no_gapd_poly_svm = no_gapd_poly_svm_tune$best.model
              
              # The training set predicitons for the polynomial SVM
              training_poly_svm_reg_unprivatized_pred = 
                predict(no_gapd_poly_svm, x_dat[,organized_vars$predictors])
              # the training set MSE for the polynomial SVM
              training_poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_poly_svm_reg_unprivatized_pred)
              
              # The testing set predicitons using the polynomial SVM
              poly_svm_reg_unprivatized_pred = 
                predict(no_gapd_poly_svm, x_dat_holdout[,organized_vars$predictors])
              # The testing set MSE for the polynomial SVM
              poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         poly_svm_reg_unprivatized_pred)
              
              # Random forest on data before gapd
              
              ranger_rf_model_orig_data = ranger( 
                x = as.matrix(sample_unprotected_data[,organized_vars$predictors]),
                y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]), 
                num.trees = 500,
                num.threads = 1,
                mtry = ceiling(1 / 3 * num_x_vars))
              
              # The training set predictions from random forest
              training_ranger_rf_pred_orig_data = 
                predict(ranger_rf_model_orig_data,
                        model_data[drawing,organized_vars$predictors])$predictions
              # The training set MSE for the random forest
              training_ranger_rf_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_ranger_rf_pred_orig_data)
              
              # Testing set predictions for random forest
              ranger_rf_pred_orig_data = predict(ranger_rf_model_orig_data,
                                                 model_data[-drawing,organized_vars$predictors])$predictions
              # Testing set MSE for random forest
              ranger_rf_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         ranger_rf_pred_orig_data)
              
              # XGBT on data before gapd
              xgbt_regress_model_orig_data =
                xgboost( data = as.matrix(sample_unprotected_data[,organized_vars$predictors]),
                         label = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]),
                         nrounds = 40,
                         nthread = 1)
              
              # The training set predictions for XGB tree
              training_xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[drawing,organized_vars$predictors]))
              # Training set MSE for XGB tree for 
              training_xgbt_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_xgbt_pred_orig_data)
              
              # The testing set XGB tree predictions
              xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[-drawing,organized_vars$predictors]))
              # The testing set MSE for XGB tree
              xgbt_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         xgbt_pred_orig_data)
              
              
              ###
              ### Stacked Ensemble Equal Weighting ----
              # GAPD
              avg_stacked_ensemble_predictions_gapd = (linear_reg_pred  + 
                lasso_pred  +  gapd_nnet_reg_pred  + 
                gapd_radial_svm_reg_pred  + 
                gapd_linear_svm_reg_pred  + 
                gapd_poly_svm_reg_pred  +
                ranger_rf_pred  + xgbt_pred) / 8
              
              avg_stacked_ensemble_gapd_mse =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         avg_stacked_ensemble_predictions_gapd)
              # No GAPD
              avg_stacked_ensemble_pred_no_gapd = 
                (linear_reg_unprivatized_pred  + 
                lasso_pred_orig_data  + 
                nnet_reg_unprivatized_pred + 
                radial_svm_reg_unprivatized_pred  + 
                linear_svm_reg_unprivatized_pred  + 
                poly_svm_reg_unprivatized_pred  +
                ranger_rf_pred_orig_data  + xgbt_pred_orig_data) / 8
              
              avg_stacked_ensemble_no_gapd_mse =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         avg_stacked_ensemble_pred_no_gapd)
              
              
              ### Stacked Ensemble Equal Weighting by Proportion of variance Explained ----------------
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_gapd = 
                c(training_lin_reg_mse, training_lasso_mse, 
                training_gapd_nnet_reg_mse, training_gapd_radial_svm_reg_mse,
                training_gapd_linear_svm_reg_mse, 
                training_gapd_poly_svm_reg_mse, training_ranger_rf_mse,
                training_xgbt_mse)
              
              # The models with the lowest MSE get the highest weight
              prop_var_exp_gapd_weighting = 
                (training_set_mses_gapd / sum(training_set_mses_gapd))^(-1)
              prop_var_exp_gapd_weighting = prop_var_exp_gapd_weighting / 
                sum(prop_var_exp_gapd_weighting)
              
              
              
              # Get the predictions for the stacked ensemble for the prop
              # of variance explained
              prop_var_stacked_ens_gapd_pred = 
                linear_reg_pred * prop_var_exp_gapd_weighting[1] +
                lasso_pred * prop_var_exp_gapd_weighting[2] + 
                gapd_nnet_reg_pred * prop_var_exp_gapd_weighting[3] + 
                gapd_radial_svm_reg_pred * prop_var_exp_gapd_weighting[4] + 
                gapd_linear_svm_reg_pred * prop_var_exp_gapd_weighting[5] + 
                gapd_poly_svm_reg_pred * prop_var_exp_gapd_weighting[6] +
                ranger_rf_pred * prop_var_exp_gapd_weighting[7] + 
                xgbt_pred * prop_var_exp_gapd_weighting[8]
              
              
              # Get the MSE
              prop_var_stacked_ens_gapd_mse = 
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         prop_var_stacked_ens_gapd_pred)
              
              
              # Get the Stacked Ensemble (Prop. of Var. explained) with No GADP 
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_no_gapd = 
                c(training_linear_reg_mse_orig_data, 
                  training_lasso_mse_orig_data, 
                  training_nnet_reg_unprivatized_mse, 
                  training_radial_svm_reg_mse_orig_data,
                  training_linear_svm_reg_mse_orig_data, 
                  training_poly_svm_reg_mse_orig_data, 
                  training_ranger_rf_mse_orig_data,
                  training_xgbt_mse_orig_data)
              
              # weighting for the SE
              prop_var_exp_no_gapd_weighting = 
                (training_set_mses_no_gapd / sum(training_set_mses_no_gapd))^(-1)
              prop_var_exp_no_gapd_weighting = prop_var_exp_no_gapd_weighting /
                sum(prop_var_exp_no_gapd_weighting)
              
              
              # get the predictions
              prop_var_stacked_ens_no_gapd_pred = 
                linear_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[1] +
                lasso_pred_orig_data * prop_var_exp_no_gapd_weighting[2] + 
                nnet_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[3] + 
                radial_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[4] + 
                linear_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[5] + 
                poly_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[6] +
                ranger_rf_pred_orig_data * prop_var_exp_no_gapd_weighting[7] + 
                xgbt_pred_orig_data * prop_var_exp_no_gapd_weighting[8]
              
              
              # Get the MSE
              prop_var_stacked_ens_no_gapd_mse = 
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         prop_var_stacked_ens_no_gapd_pred)
              
              
              # Compile MSEs into a vector
              gapd_mse_vector = 
                c(linear_reg_mse, lasso_mse, gapd_nnet_reg_mse,
                  gapd_radial_svm_reg_mse, gapd_linear_svm_reg_mse,
                  gapd_poly_svm_reg_mse, ranger_rf_mse,
                  xgbt_mse, avg_stacked_ensemble_gapd_mse, 
                  prop_var_stacked_ens_gapd_mse)
              
              no_gapd_mse_vector =
                c(linear_reg_mse_orig_data, lasso_mse_orig_data,
                  nnet_reg_mse_orig_data, radial_svm_reg_mse_orig_data,
                  linear_svm_reg_mse_orig_data, poly_svm_reg_mse_orig_data,
                  ranger_rf_mse_orig_data, xgbt_mse_orig_data,
                  avg_stacked_ensemble_no_gapd_mse, 
                  prop_var_stacked_ens_no_gapd_mse)
              
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "sample_size", "y", 
                                   "LR", "LASSO", "NNET", "Radial_SVM",
                                   "Linear_SVM", "Poly_SVM", "RF", "XGBTree",
                                   "Stacked_equal", "Stacked_prop","gadp", "theta")
              
              
              # construct the dataframe of results
              estimator_reg_df = data.frame(rbind(c(dat.seed,iteration, n[samp_size_index], 
                                         response_index, gapd_mse_vector, 1, theta_value)))
              
              # Adding results for LASSO without GAPD
              estimator_reg_df[2,] = rbind(c(dat.seed, iteration, 
                                             n[samp_size_index],
                                    response_index, no_gapd_mse_vector, 0, theta_value))
              
              colnames(estimator_reg_df) = data_frame_names
              
              
              return(estimator_reg_df)
            }
  
  stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  # Add name of the estimators
  
  return(results)
}

## Training Dataset Iteratively generated ----------------------------------------------------
### GADP -----------------------------------------------
estimator_eval_stacked_ensemb_iterative = function( # the dataset
                                               m_start = 1, # Primarily used for debugging determines which iteration to start from
                                               m = 1000, # Number of Monte Carlo simulations
                                               n = c(500, 1000, 2000), # sample sizes to be used for training set
                                               cores = 2, # Number of CPU threads/ logical processors to use
                                               response_prefix = "y", # prefix for response variables
                                               acquire_variable_names = get_variables, #Function for getting the variables
                                               perturbation_func = gadp_func, # CGADP or GADP function
                                               theta_value = 0.4, # privacy parameter value
                                               model_generating_function, # function to generate the model
                                               testing_data_seed = 30000, # Seed for the testing set
                                               model_predictor_quant = 10, # Quantity of the predictors for the model
                                               model_rho = 0.2, # correlation of predictors in the model
                                               model_noise = 0.2, # Noise to be used in the model
                                               testing_n = 500) # Size of the testing set
{
  # generate the testing data
  testing_model_data = model_generating_function(N = testing_n, rho = model_rho, 
                                                 max_pred = model_predictor_quant, 
                                                 dat.seed = testing_data_seed,
                                         noise = model_noise)
  
  # Getting variable names:
  all_variable_names = colnames(testing_model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  # Variables serving as the predictors
  predictor_vars = organized_vars$predictors
  
  
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  # Create the cluster
  sim_cluster = makeCluster(cores)
  # Register the cluster
  registerDoSNOW(sim_cluster)
  
  # Store the results in a matrix
  results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
                    .packages = c("glmnet", "ranger", 
                                  "nnet", "e1071", "xgboost", "dplyr"),
                    .inorder = FALSE) %:%
    foreach( iteration = m_start:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",.inorder = FALSE) %dopar% {
              
              # setting the seed
              dat.seed = iteration + n[samp_size_index]
              set.seed(dat.seed)
              
              # Training dataset
              # Generated with the same parameters sxcluding the seed
              training_model_data = 
                model_generating_function(N = n[samp_size_index], 
                                          rho = model_rho, 
                                        max_pred = model_predictor_quant, 
                                        dat.seed = dat.seed,
                                        noise = model_noise)
              
              
              # Performing SDC GADP method
              # Testing data
              privatized_responses_testing = 
                perturbation_func(unprotected_data = testing_model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              # Training data
              privatized_responses_training = 
                perturbation_func(unprotected_data = training_model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              # Combining the training and the test datasets
              privatized_responses = 
                rbind(privatized_responses_training,
                           privatized_responses_testing)
              
              model_data = rbind(training_model_data, 
                                    testing_model_data)
              
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,predictor_vars]
              
              privatized_dataframe[,colnames(privatized_responses)[response_index]] = 
                current_response
              
              # Get the training set form the df, first 1:n is training, n+1 
              # onwards is the testing data set. 
              drawing = 1:n[samp_size_index]
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              x_dat_holdout = cbind(rep(1, nrow(model_data[-drawing,])),
                                    as.matrix(model_data)[-drawing,predictor_vars])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after gapd
              x_dat = as.matrix(cbind(rep(1, nrow(sample_private_data[,predictor_vars])),
                                      sample_private_data[,predictor_vars]))
              
              ### GAPD Used----
              
              # Linear Regression on data after gapd
              betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(current_response[drawing])
              
              
              linear_reg_pred = as.vector(x_dat_holdout %*% betas_linear)
              
              # Get the predictions for the training data
              
              training_lin_reg_pred = as.vector(x_dat %*% betas_linear)
              
              
              # Get the mse for the training data
              
              
              training_lin_reg_mse = calc_mse(as.matrix(current_response[drawing]),
                                              training_lin_reg_pred)
              
              # Get the mse for the test data
              
              
              linear_reg_mse = calc_mse(orig_data_response[-drawing],
                                        linear_reg_pred)
              
              
              # LASSO on data after gapd
              lasso_regress_model = cv.glmnet( 
                x = as.matrix(sample_private_data[,predictor_vars]),
                y = current_response[drawing])
              
              lasso_pred = predict(lasso_regress_model, 
                                   as.matrix(privatized_dataframe[-drawing,predictor_vars]),
                                   s = "lambda.min")
              
              # get predictions on the training data
              training_lasso_pred = predict(lasso_regress_model, 
                                            as.matrix(privatized_dataframe[drawing,predictor_vars]),
                                            s = "lambda.min")
              
              # Get the MSE for the training data
              training_lasso_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_lasso_pred)
              
              # # MSE for the testing data
              
              lasso_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                   lasso_pred)
              
              ### Neural Net Regression on data after gapd -------------
              
              
              gapd_nnet = nnet(x = x_dat[,predictor_vars], y = current_response[drawing], 
                               size = 3,
                               maxit = 200, linout = TRUE)
              
              
              gapd_nnet_reg_pred = predict(gapd_nnet,x_dat_holdout[,predictor_vars])
              
              # training data predictions
              training_gapd_nnet_reg_pred = predict(gapd_nnet,x_dat[,predictor_vars])
              
              # training data MSE
              training_gapd_nnet_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_nnet_reg_pred)
              
              
              # testing data MSE
              gapd_nnet_reg_mse = calc_mse(orig_data_response[-drawing],
                                           gapd_nnet_reg_pred)
              
              # SVM Radial Regression on data after gapd
              gapd_radial_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
                                              gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                              degree = 3:4,
                                              kernel = c("radial"))
              
              gapd_radial_svm = gapd_radial_svm_tune$best.model
              
              gapd_radial_svm_reg_pred = predict(gapd_radial_svm,x_dat_holdout[,predictor_vars])
              
              # Get the training predictions of the radial svm
              training_gapd_radial_svm_reg_pred =
                predict(gapd_radial_svm,x_dat[,predictor_vars])
              
              # Get the training mse for radial svm
              training_gapd_radial_svm_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_radial_svm_reg_pred)
              
              # Testing data MSE
              gapd_radial_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                 gapd_radial_svm_reg_pred)
              
              
              
              # SVM Linear Regression on data after gapd
              gapd_linear_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
                                              gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                              degree = 3:4,
                                              kernel = c("linear"))
              
              gapd_linear_svm = gapd_linear_svm_tune$best.model
              
              # Predictions for the training data for Linear SVM
              training_gapd_linear_svm_reg_pred =
                predict(gapd_linear_svm,x_dat_holdout[,predictor_vars])
              
              # Training MSE for Linear SVM
              training_gapd_linear_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_linear_svm_reg_pred)
              
              # Predictions for the test dataset
              gapd_linear_svm_reg_pred = predict(gapd_linear_svm,x_dat_holdout[,predictor_vars])
              
              # Testing data MSE
              gapd_linear_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                 gapd_linear_svm_reg_pred)
              
              # SVM Polynomial Regression on data after gapd
              gapd_poly_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
                                            gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                            degree = 3:4,
                                            kernel = c("polynomial"))
              
              gapd_poly_svm = gapd_poly_svm_tune$best.model
              
              # Training data polynomial SVM predictions
              training_gapd_poly_svm_reg_pred = predict(gapd_poly_svm,x_dat[,predictor_vars])
              
              # Training data MSE polynomial SVM
              training_gapd_poly_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_poly_svm_reg_pred)
              
              # Predictions for testing dataset
              gapd_poly_svm_reg_pred = predict(gapd_poly_svm,x_dat_holdout[,predictor_vars])
              
              # MSE for testing dataset for polynomial kernel SVM
              gapd_poly_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                               gapd_poly_svm_reg_pred)
              
              # Random forest on data after gapd
              num_x_vars = length(predictor_vars)
              
              ranger_rf_regress_model = ranger( 
                x = as.matrix(sample_private_data[,predictor_vars]),
                y = current_response[drawing], 
                num.trees = 500,
                num.threads = 1,
                mtry = ceiling(1 / 3 * num_x_vars))
              
              # Training dataset predicions for the random forest
              training_ranger_rf_pred = 
                predict(ranger_rf_regress_model, 
                        privatized_dataframe[drawing,predictor_vars])$predictions
              
              # training dataset MSE for the radom forest
              training_ranger_rf_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_ranger_rf_pred)
              
              
              # Testing dataset prediction for random forest
              ranger_rf_pred = predict(ranger_rf_regress_model, 
                                       privatized_dataframe[-drawing,predictor_vars])$predictions
              
              # Testing dataset MSE for random forest
              ranger_rf_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                       ranger_rf_pred)
              
              # XGBT on data after gapd
              xgbt_regress_model = xgboost( 
                data = as.matrix(sample_private_data[,predictor_vars]),
                label = current_response[drawing], 
                nrounds = 40,
                nthread = 1)
              
              # Training data predictions for the XGB tree
              training_xgbt_pred = 
                predict(xgbt_regress_model,
                        as.matrix(privatized_dataframe[drawing,predictor_vars]))
              
              # training data MSE for the XGB tree
              training_xgbt_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_xgbt_pred)
              
              # Testing data predicitons for the XGB tree
              xgbt_pred = predict(xgbt_regress_model, 
                                  as.matrix(privatized_dataframe[-drawing,predictor_vars]))
              
              # Testing data MSE for the XGB tree
              xgbt_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                  xgbt_pred)
              
              ##
              ## Before GAPD (On original data) ------
              print(" running checkpoint 1")
              # Linear Regression on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              
              # Beta coefficients for the original data
              orig_data_betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(sample_orig_data_response)
              
              # Predictions on the training data no GADP
              training_lin_reg_unprivat_pred = as.vector(x_dat %*%orig_data_betas_linear)
              
              
              # MSE for training data for linear regression, no GADP
              training_linear_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_lin_reg_unprivat_pred)
              
              # Testing data MSE for linear regression no ADP
              linear_reg_unprivatized_pred = as.vector(x_dat_holdout %*% orig_data_betas_linear)
              
              linear_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_reg_unprivatized_pred)
              
              # LASSO on data before gapd (original data)
              
              
              lasso_regress_model_orig_data =
                cv.glmnet( x = as.matrix(sample_unprotected_data[,predictor_vars]),
                           y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # Training data predictions for LASSO
              
              training_lasso_pred_orig_data = 
                predict(lasso_regress_model_orig_data,
                        as.matrix(model_data[drawing,predictor_vars]), 
                        s = "lambda.min")
              
              # Training data MSE for LASSO
              training_lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_lasso_pred_orig_data)
              # Predictions for the test dataset for the LASSO model
              lasso_pred_orig_data = predict(lasso_regress_model_orig_data,
                                             as.matrix(model_data[-drawing,predictor_vars]), 
                                             s = "lambda.min")
              # MSE for the LASSO model for the test data
              lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         lasso_pred_orig_data)
              
              
              # Neural Net before GAPD 
              
              no_gapd_nnet = nnet(x = x_dat[,predictor_vars], y = sample_orig_data_response, 
                                  size = 3,
                                  maxit = 200, linout = TRUE)
              
              # The predictions of the training set for neural network
              training_nnet_reg_unprivatized_pred = predict(no_gapd_nnet, x_dat[,predictor_vars])
              
              # The MSE for the training set for the neural network
              training_nnet_reg_unprivatized_mse = 
                calc_mse(orig_data_response[drawing],
                         training_nnet_reg_unprivatized_pred)
              
              # Predicitons on testing set for neural net
              nnet_reg_unprivatized_pred = predict(no_gapd_nnet, x_dat_holdout[,predictor_vars])
              
              # MSE for testing set for neural network
              nnet_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         nnet_reg_unprivatized_pred)
              
              # SVM Radial before GAPD
              
              
              no_gapd_radial_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
                                                 y = sample_orig_data_response,
                                                 gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                                 degree = 3:4,
                                                 kernel = c("radial"))
              
              # Get the best model for Radial SVM
              no_gapd_radial_svm = no_gapd_radial_svm_tune$best.model
              
              # Predictions on the training set for radial SVM
              training_radial_svm_reg_unprivatized_pred = 
                predict(no_gapd_radial_svm, x_dat[,predictor_vars])
              # The MSE for the training set for radial SVM
              training_radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_radial_svm_reg_unprivatized_pred)
              
              # Get the predictions of Radial SVM for the testing set
              radial_svm_reg_unprivatized_pred = 
                predict(no_gapd_radial_svm, x_dat_holdout[,predictor_vars])
              # Get the MSE of the Radial SVM for the testing set
              radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         radial_svm_reg_unprivatized_pred)
              
              # Linear SVM before GAPD
              no_gapd_linear_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
                                                 y = sample_orig_data_response,
                                                 gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                                 degree = 3:4,
                                                 kernel = c("linear"))
              # The Linear SVM best tuned model
              no_gapd_linear_svm = no_gapd_linear_svm_tune$best.model
              
              # The predictions on the training set, Linear SVM
              training_linear_svm_reg_unprivatized_pred = 
                predict(no_gapd_linear_svm, x_dat[,predictor_vars])
              # the MSE on the training set, Linear SVM
              training_linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_linear_svm_reg_unprivatized_pred)
              # The predictions for the testing set Linear SVM
              linear_svm_reg_unprivatized_pred = 
                predict(no_gapd_linear_svm, x_dat_holdout[,predictor_vars])
              
              # The MSE for the testing set Linear SVM
              linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_svm_reg_unprivatized_pred)
              
              
              # Polynomial SVM before GAPD
              
              no_gapd_poly_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
                                               y = sample_orig_data_response,
                                               gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                               degree = 3:4,
                                               kernel = c("polynomial"))
              # Get the best tuned polynomial SVM
              no_gapd_poly_svm = no_gapd_poly_svm_tune$best.model
              
              # The training set predicitons for the polynomial SVM
              training_poly_svm_reg_unprivatized_pred = 
                predict(no_gapd_poly_svm, x_dat[,predictor_vars])
              # the training set MSE for the polynomial SVM
              training_poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_poly_svm_reg_unprivatized_pred)
              
              # The testing set predicitons using the polynomial SVM
              poly_svm_reg_unprivatized_pred = 
                predict(no_gapd_poly_svm, x_dat_holdout[,predictor_vars])
              # The testing set MSE for the polynomial SVM
              poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         poly_svm_reg_unprivatized_pred)
              
              # Random forest on data before gapd
              
              ranger_rf_model_orig_data = ranger( 
                x = as.matrix(sample_unprotected_data[,predictor_vars]),
                y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]), 
                num.trees = 500,
                num.threads = 1,
                mtry = ceiling(1 / 3 * num_x_vars))
              
              # The training set predictions from random forest
              training_ranger_rf_pred_orig_data = 
                predict(ranger_rf_model_orig_data,
                        model_data[drawing,predictor_vars])$predictions
              # The training set MSE for the random forest
              training_ranger_rf_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_ranger_rf_pred_orig_data)
              
              # Testing set predictions for random forest
              ranger_rf_pred_orig_data = predict(ranger_rf_model_orig_data,
                                                 model_data[-drawing,predictor_vars])$predictions
              # Testing set MSE for random forest
              ranger_rf_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         ranger_rf_pred_orig_data)
              
              # XGBT on data before gapd
              xgbt_regress_model_orig_data =
                xgboost( data = as.matrix(sample_unprotected_data[,predictor_vars]),
                         label = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]),
                         nrounds = 40,
                         nthread = 1)
              
              # The training set predictions for XGB tree
              training_xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[drawing,predictor_vars]))
              # Training set MSE for XGB tree for 
              training_xgbt_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_xgbt_pred_orig_data)
              
              # The testing set XGB tree predictions
              xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[-drawing,predictor_vars]))
              # The testing set MSE for XGB tree
              xgbt_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         xgbt_pred_orig_data)
              
              
              ###
              ### Stacked Ensemble Equal Weighting ----
              # GAPD
              avg_stacked_ensemble_predictions_gapd = (linear_reg_pred  + 
                                                         lasso_pred  +  gapd_nnet_reg_pred  + 
                                                         gapd_radial_svm_reg_pred  + 
                                                         gapd_linear_svm_reg_pred  + 
                                                         gapd_poly_svm_reg_pred  +
                                                         ranger_rf_pred  + xgbt_pred) / 8
              
              avg_stacked_ensemble_gapd_mse =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         avg_stacked_ensemble_predictions_gapd)
              # No GAPD
              avg_stacked_ensemble_pred_no_gapd = 
                (linear_reg_unprivatized_pred  + 
                   lasso_pred_orig_data  + 
                   nnet_reg_unprivatized_pred + 
                   radial_svm_reg_unprivatized_pred  + 
                   linear_svm_reg_unprivatized_pred  + 
                   poly_svm_reg_unprivatized_pred  +
                   ranger_rf_pred_orig_data  + xgbt_pred_orig_data) / 8
              
              avg_stacked_ensemble_no_gapd_mse =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         avg_stacked_ensemble_pred_no_gapd)
              
              
              ### Stacked Ensemble Equal Weighting by Proportion of variance Explained ----------------
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_gapd = 
                c(training_lin_reg_mse, training_lasso_mse, 
                  training_gapd_nnet_reg_mse, training_gapd_radial_svm_reg_mse,
                  training_gapd_linear_svm_reg_mse, 
                  training_gapd_poly_svm_reg_mse, training_ranger_rf_mse,
                  training_xgbt_mse)
              
              # The models with the lowest MSE get the highest weight
              prop_var_exp_gapd_weighting = 
                (training_set_mses_gapd / sum(training_set_mses_gapd))^(-1)
              prop_var_exp_gapd_weighting = prop_var_exp_gapd_weighting / 
                sum(prop_var_exp_gapd_weighting)
              
              
              
              # Get the predictions for the stacked ensemble for the prop
              # of variance explained
              prop_var_stacked_ens_gapd_pred = 
                linear_reg_pred * prop_var_exp_gapd_weighting[1] +
                lasso_pred * prop_var_exp_gapd_weighting[2] + 
                gapd_nnet_reg_pred * prop_var_exp_gapd_weighting[3] + 
                gapd_radial_svm_reg_pred * prop_var_exp_gapd_weighting[4] + 
                gapd_linear_svm_reg_pred * prop_var_exp_gapd_weighting[5] + 
                gapd_poly_svm_reg_pred * prop_var_exp_gapd_weighting[6] +
                ranger_rf_pred * prop_var_exp_gapd_weighting[7] + 
                xgbt_pred * prop_var_exp_gapd_weighting[8]
              
              
              # Get the MSE
              prop_var_stacked_ens_gapd_mse = 
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         prop_var_stacked_ens_gapd_pred)
              
              
              # Get the Stacked Ensemble (Prop. of Var. explained) with No GADP 
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_no_gapd = 
                c(training_linear_reg_mse_orig_data, 
                  training_lasso_mse_orig_data, 
                  training_nnet_reg_unprivatized_mse, 
                  training_radial_svm_reg_mse_orig_data,
                  training_linear_svm_reg_mse_orig_data, 
                  training_poly_svm_reg_mse_orig_data, 
                  training_ranger_rf_mse_orig_data,
                  training_xgbt_mse_orig_data)
              
              # weighting for the SE
              prop_var_exp_no_gapd_weighting = 
                (training_set_mses_no_gapd / sum(training_set_mses_no_gapd))^(-1)
              prop_var_exp_no_gapd_weighting = prop_var_exp_no_gapd_weighting /
                sum(prop_var_exp_no_gapd_weighting)
              
              
              # get the predictions
              prop_var_stacked_ens_no_gapd_pred = 
                linear_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[1] +
                lasso_pred_orig_data * prop_var_exp_no_gapd_weighting[2] + 
                nnet_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[3] + 
                radial_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[4] + 
                linear_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[5] + 
                poly_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[6] +
                ranger_rf_pred_orig_data * prop_var_exp_no_gapd_weighting[7] + 
                xgbt_pred_orig_data * prop_var_exp_no_gapd_weighting[8]
              
              
              # Get the MSE
              prop_var_stacked_ens_no_gapd_mse = 
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         prop_var_stacked_ens_no_gapd_pred)
              
              
              # Compile MSEs into a vector
              gapd_mse_vector = 
                c(linear_reg_mse, lasso_mse, gapd_nnet_reg_mse,
                  gapd_radial_svm_reg_mse, gapd_linear_svm_reg_mse,
                  gapd_poly_svm_reg_mse, ranger_rf_mse,
                  xgbt_mse, avg_stacked_ensemble_gapd_mse, 
                  prop_var_stacked_ens_gapd_mse)
              
              no_gapd_mse_vector =
                c(linear_reg_mse_orig_data, lasso_mse_orig_data,
                  nnet_reg_mse_orig_data, radial_svm_reg_mse_orig_data,
                  linear_svm_reg_mse_orig_data, poly_svm_reg_mse_orig_data,
                  ranger_rf_mse_orig_data, xgbt_mse_orig_data,
                  avg_stacked_ensemble_no_gapd_mse, 
                  prop_var_stacked_ens_no_gapd_mse)
              
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "training_size", 
                                   "testing_size",
                                   "y", 
                                   "LR", "LASSO", "NNET", "Radial_SVM",
                                   "Linear_SVM", "Poly_SVM", "RF", "XGBTree",
                                   "Stacked_equal", "Stacked_prop","gadp", "theta")
              
              
              # construct the dataframe of results
              estimator_reg_df = data.frame(rbind(c(dat.seed,iteration, n[samp_size_index], testing_n,
                                                    response_index, gapd_mse_vector, 1, theta_value)))
              
              # Adding results for LASSO without GAPD
              estimator_reg_df[2,] = rbind(c(dat.seed, iteration, 
                                             n[samp_size_index], testing_n,
                                             response_index, no_gapd_mse_vector, 0, theta_value))
              
              
              colnames(estimator_reg_df) = data_frame_names
              
              
              return(estimator_reg_df)
            }
  
  stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  # Add name of the estimators
  
  results_list = list("output" = results,
                      "testing_set" = testing_model_data)
  
  return(results_list)
}


# Update on 02-24-2024
## Training Dataset Iteratively generated Version 2 ----------------------------------------------------
### GADP -----------------------------------------------
estimator_eval_stacked_ensemb_iterative_ver_2 = function( # the dataset
  m_start = 1, # Primarily used for debugging determines which iteration to start from
  m = 1000, # Number of Monte Carlo simulations
  n = c(500, 1000, 2000), # sample sizes to be used for training set
  cores = 2, # Number of CPU threads/ logical processors to use
  response_prefix = "y", # prefix for response variables
  acquire_variable_names = get_variables, #Function for getting the variables
  perturbation_func = gadp_func, # CGADP or GADP function
  theta_value = 0.4, # privacy parameter value
  model_generating_function, # function to generate the model
  testing_data_seed = 30000, # Seed for the testing set
  model_predictor_quant = 10, # Quantity of the predictors for the model
  model_rho = 0.2, # correlation of predictors in the model
  model_noise = 0.2, # Noise to be used in the model
  testing_n = 500) # Size of the testing set
{
  # generate the testing data
  testing_model_data = model_generating_function(N = testing_n, rho = model_rho, 
                                                 max_pred = model_predictor_quant, 
                                                 dat.seed = testing_data_seed,
                                                 noise = model_noise)
  
  # Getting variable names:
  all_variable_names = colnames(testing_model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  # Variables serving as the predictors
  predictor_vars = organized_vars$predictors
  
  
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  # Create the cluster
  sim_cluster = makeCluster(cores)
  # Register the cluster
  registerDoSNOW(sim_cluster)
  
  # Store the results in a matrix
  results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
                    .packages = c("glmnet", "ranger", 
                                  "nnet", "e1071", "xgboost", "dplyr"),
                    .inorder = FALSE) %:%
    foreach( iteration = m_start:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",.inorder = FALSE) %dopar% {
              
              # setting the seed
              dat.seed = iteration + n[samp_size_index]
              set.seed(dat.seed)
              
              # Training dataset
              # Generated with the same parameters sxcluding the seed
              training_model_data = 
                model_generating_function(N = n[samp_size_index], 
                                          rho = model_rho, 
                                          max_pred = model_predictor_quant, 
                                          dat.seed = dat.seed,
                                          noise = model_noise)
              
              
              # Performing SDC GADP method
              # Testing data
              privatized_responses_testing = 
                perturbation_func(unprotected_data = testing_model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              # Training data
              privatized_responses_training = 
                perturbation_func(unprotected_data = training_model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              # Combining the training and the test datasets
              privatized_responses = 
                rbind(privatized_responses_training,
                      privatized_responses_testing)
              
              model_data = rbind(training_model_data, 
                                 testing_model_data)
              
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,predictor_vars]
              
              privatized_dataframe[,colnames(privatized_responses)[response_index]] = 
                current_response
              
              # Get the training set form the df, first 1:n is training, n+1 
              # onwards is the testing data set. 
              drawing = 1:n[samp_size_index]
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              x_dat_holdout = cbind(rep(1, nrow(model_data[-drawing,])),
                                    as.matrix(model_data)[-drawing,predictor_vars])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after gapd
              x_dat = as.matrix(cbind(rep(1, nrow(sample_private_data[,predictor_vars])),
                                      sample_private_data[,predictor_vars]))
              
              ### GAPD Used----
              
              # Linear Regression on data after gapd
              betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(current_response[drawing])
              
              
              linear_reg_pred = as.vector(x_dat_holdout %*% betas_linear)
              
              # Get the predictions for the training data
              
              training_lin_reg_pred = as.vector(x_dat %*% betas_linear)
              
              
              # Get the mse for the training data
              
              
              training_lin_reg_mse = calc_mse(as.matrix(current_response[drawing]),
                                              training_lin_reg_pred)
              
              # Get the mse for the test data
              
              
              linear_reg_mse = calc_mse(orig_data_response[-drawing],
                                        linear_reg_pred)
              
              
              # LASSO on data after gapd
              lasso_regress_model = cv.glmnet( 
                x = as.matrix(sample_private_data[,predictor_vars]),
                y = current_response[drawing])
              
              lasso_pred = predict(lasso_regress_model, 
                                   as.matrix(privatized_dataframe[-drawing,predictor_vars]),
                                   s = "lambda.min")
              
              # get predictions on the training data
              training_lasso_pred = predict(lasso_regress_model, 
                                            as.matrix(privatized_dataframe[drawing,predictor_vars]),
                                            s = "lambda.min")
              
              # Get the MSE for the training data
              training_lasso_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_lasso_pred)
              
              # # MSE for the testing data
              
              lasso_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                   lasso_pred)
              
              ### Neural Net Regression on data after gapd -------------
              
              
              gapd_nnet = nnet(x = x_dat[,predictor_vars], y = current_response[drawing], 
                               size = 3,
                               maxit = 200, linout = TRUE)
              
              
              gapd_nnet_reg_pred = predict(gapd_nnet,x_dat_holdout[,predictor_vars])
              
              # training data predictions
              training_gapd_nnet_reg_pred = predict(gapd_nnet,x_dat[,predictor_vars])
              
              # training data MSE
              training_gapd_nnet_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_nnet_reg_pred)
              
              
              # testing data MSE
              gapd_nnet_reg_mse = calc_mse(orig_data_response[-drawing],
                                           gapd_nnet_reg_pred)
              
              # SVM Radial Regression on data after gapd
              gapd_radial_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
                                              gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                              degree = 3:4,
                                              kernel = c("radial"))
              
              gapd_radial_svm = gapd_radial_svm_tune$best.model
              
              gapd_radial_svm_reg_pred = predict(gapd_radial_svm,x_dat_holdout[,predictor_vars])
              
              # Get the training predictions of the radial svm
              training_gapd_radial_svm_reg_pred =
                predict(gapd_radial_svm,x_dat[,predictor_vars])
              
              # Get the training mse for radial svm
              training_gapd_radial_svm_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_radial_svm_reg_pred)
              
              # Testing data MSE
              gapd_radial_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                 gapd_radial_svm_reg_pred)
              
              
              
              # SVM Linear Regression on data after gapd
              gapd_linear_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
                                              gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                              degree = 3:4,
                                              kernel = c("linear"))
              
              gapd_linear_svm = gapd_linear_svm_tune$best.model
              
              # Predictions for the training data for Linear SVM
              training_gapd_linear_svm_reg_pred =
                predict(gapd_linear_svm,x_dat_holdout[,predictor_vars])
              
              # Training MSE for Linear SVM
              training_gapd_linear_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_linear_svm_reg_pred)
              
              # Predictions for the test dataset
              gapd_linear_svm_reg_pred = predict(gapd_linear_svm,x_dat_holdout[,predictor_vars])
              
              # Testing data MSE
              gapd_linear_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                 gapd_linear_svm_reg_pred)
              
              # SVM Polynomial Regression on data after gapd
              gapd_poly_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
                                            gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                            degree = 3:4,
                                            kernel = c("polynomial"))
              
              gapd_poly_svm = gapd_poly_svm_tune$best.model
              
              # Training data polynomial SVM predictions
              training_gapd_poly_svm_reg_pred = predict(gapd_poly_svm,x_dat[,predictor_vars])
              
              # Training data MSE polynomial SVM
              training_gapd_poly_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_poly_svm_reg_pred)
              
              # Predictions for testing dataset
              gapd_poly_svm_reg_pred = predict(gapd_poly_svm,x_dat_holdout[,predictor_vars])
              
              # MSE for testing dataset for polynomial kernel SVM
              gapd_poly_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                               gapd_poly_svm_reg_pred)
              
              # Random forest on data after gapd
              num_x_vars = length(predictor_vars)
              
              ranger_rf_regress_model = ranger( 
                x = as.matrix(sample_private_data[,predictor_vars]),
                y = current_response[drawing], 
                num.trees = 500,
                num.threads = 1,
                mtry = ceiling(1 / 3 * num_x_vars))
              
              # Training dataset predicions for the random forest
              training_ranger_rf_pred = 
                predict(ranger_rf_regress_model, 
                        privatized_dataframe[drawing,predictor_vars])$predictions
              
              # training dataset MSE for the radom forest
              training_ranger_rf_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_ranger_rf_pred)
              
              
              # Testing dataset prediction for random forest
              ranger_rf_pred = predict(ranger_rf_regress_model, 
                                       privatized_dataframe[-drawing,predictor_vars])$predictions
              
              # Testing dataset MSE for random forest
              ranger_rf_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                       ranger_rf_pred)
              
              # XGBT on data after gapd
              xgbt_regress_model = xgboost( 
                data = as.matrix(sample_private_data[,predictor_vars]),
                label = current_response[drawing], 
                nrounds = 40,
                nthread = 1)
              
              # Training data predictions for the XGB tree
              training_xgbt_pred = 
                predict(xgbt_regress_model,
                        as.matrix(privatized_dataframe[drawing,predictor_vars]))
              
              # training data MSE for the XGB tree
              training_xgbt_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_xgbt_pred)
              
              # Testing data predicitons for the XGB tree
              xgbt_pred = predict(xgbt_regress_model, 
                                  as.matrix(privatized_dataframe[-drawing,predictor_vars]))
              
              # Testing data MSE for the XGB tree
              xgbt_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                  xgbt_pred)
              
              ##
              ## Before GAPD (On original data) ------
              print(" running checkpoint 1")
              # Linear Regression on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              
              # Beta coefficients for the original data
              orig_data_betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(sample_orig_data_response)
              
              # Predictions on the training data no GADP
              training_lin_reg_unprivat_pred = as.vector(x_dat %*%orig_data_betas_linear)
              
              
              # MSE for training data for linear regression, no GADP
              training_linear_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_lin_reg_unprivat_pred)
              
              # Testing data MSE for linear regression no ADP
              linear_reg_unprivatized_pred = as.vector(x_dat_holdout %*% orig_data_betas_linear)
              
              linear_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_reg_unprivatized_pred)
              
              # LASSO on data before gapd (original data)
              
              
              lasso_regress_model_orig_data =
                cv.glmnet( x = as.matrix(sample_unprotected_data[,predictor_vars]),
                           y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # Training data predictions for LASSO
              
              training_lasso_pred_orig_data = 
                predict(lasso_regress_model_orig_data,
                        as.matrix(model_data[drawing,predictor_vars]), 
                        s = "lambda.min")
              
              # Training data MSE for LASSO
              training_lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_lasso_pred_orig_data)
              # Predictions for the test dataset for the LASSO model
              lasso_pred_orig_data = predict(lasso_regress_model_orig_data,
                                             as.matrix(model_data[-drawing,predictor_vars]), 
                                             s = "lambda.min")
              # MSE for the LASSO model for the test data
              lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         lasso_pred_orig_data)
              
              
              # Neural Net before GAPD 
              
              no_gapd_nnet = nnet(x = x_dat[,predictor_vars], y = sample_orig_data_response, 
                                  size = 3,
                                  maxit = 200, linout = TRUE)
              
              # The predictions of the training set for neural network
              training_nnet_reg_unprivatized_pred = predict(no_gapd_nnet, x_dat[,predictor_vars])
              
              # The MSE for the training set for the neural network
              training_nnet_reg_unprivatized_mse = 
                calc_mse(orig_data_response[drawing],
                         training_nnet_reg_unprivatized_pred)
              
              # Predicitons on testing set for neural net
              nnet_reg_unprivatized_pred = predict(no_gapd_nnet, x_dat_holdout[,predictor_vars])
              
              # MSE for testing set for neural network
              nnet_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         nnet_reg_unprivatized_pred)
              
              # SVM Radial before GAPD
              
              
              no_gapd_radial_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
                                                 y = sample_orig_data_response,
                                                 gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                                 degree = 3:4,
                                                 kernel = c("radial"))
              
              # Get the best model for Radial SVM
              no_gapd_radial_svm = no_gapd_radial_svm_tune$best.model
              
              # Predictions on the training set for radial SVM
              training_radial_svm_reg_unprivatized_pred = 
                predict(no_gapd_radial_svm, x_dat[,predictor_vars])
              # The MSE for the training set for radial SVM
              training_radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_radial_svm_reg_unprivatized_pred)
              
              # Get the predictions of Radial SVM for the testing set
              radial_svm_reg_unprivatized_pred = 
                predict(no_gapd_radial_svm, x_dat_holdout[,predictor_vars])
              # Get the MSE of the Radial SVM for the testing set
              radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         radial_svm_reg_unprivatized_pred)
              
              # Linear SVM before GAPD
              no_gapd_linear_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
                                                 y = sample_orig_data_response,
                                                 gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                                 degree = 3:4,
                                                 kernel = c("linear"))
              # The Linear SVM best tuned model
              no_gapd_linear_svm = no_gapd_linear_svm_tune$best.model
              
              # The predictions on the training set, Linear SVM
              training_linear_svm_reg_unprivatized_pred = 
                predict(no_gapd_linear_svm, x_dat[,predictor_vars])
              # the MSE on the training set, Linear SVM
              training_linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_linear_svm_reg_unprivatized_pred)
              # The predictions for the testing set Linear SVM
              linear_svm_reg_unprivatized_pred = 
                predict(no_gapd_linear_svm, x_dat_holdout[,predictor_vars])
              
              # The MSE for the testing set Linear SVM
              linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_svm_reg_unprivatized_pred)
              
              
              # Polynomial SVM before GAPD
              
              no_gapd_poly_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
                                               y = sample_orig_data_response,
                                               gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                               degree = 3:4,
                                               kernel = c("polynomial"))
              # Get the best tuned polynomial SVM
              no_gapd_poly_svm = no_gapd_poly_svm_tune$best.model
              
              # The training set predicitons for the polynomial SVM
              training_poly_svm_reg_unprivatized_pred = 
                predict(no_gapd_poly_svm, x_dat[,predictor_vars])
              # the training set MSE for the polynomial SVM
              training_poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_poly_svm_reg_unprivatized_pred)
              
              # The testing set predicitons using the polynomial SVM
              poly_svm_reg_unprivatized_pred = 
                predict(no_gapd_poly_svm, x_dat_holdout[,predictor_vars])
              # The testing set MSE for the polynomial SVM
              poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         poly_svm_reg_unprivatized_pred)
              
              # Random forest on data before gapd
              
              ranger_rf_model_orig_data = ranger( 
                x = as.matrix(sample_unprotected_data[,predictor_vars]),
                y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]), 
                num.trees = 500,
                num.threads = 1,
                mtry = ceiling(1 / 3 * num_x_vars))
              
              # The training set predictions from random forest
              training_ranger_rf_pred_orig_data = 
                predict(ranger_rf_model_orig_data,
                        model_data[drawing,predictor_vars])$predictions
              # The training set MSE for the random forest
              training_ranger_rf_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_ranger_rf_pred_orig_data)
              
              # Testing set predictions for random forest
              ranger_rf_pred_orig_data = predict(ranger_rf_model_orig_data,
                                                 model_data[-drawing,predictor_vars])$predictions
              # Testing set MSE for random forest
              ranger_rf_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         ranger_rf_pred_orig_data)
              
              # XGBT on data before gapd
              xgbt_regress_model_orig_data =
                xgboost( data = as.matrix(sample_unprotected_data[,predictor_vars]),
                         label = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]),
                         nrounds = 40,
                         nthread = 1)
              
              # The training set predictions for XGB tree
              training_xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[drawing,predictor_vars]))
              # Training set MSE for XGB tree for 
              training_xgbt_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_xgbt_pred_orig_data)
              
              # The testing set XGB tree predictions
              xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[-drawing,predictor_vars]))
              # The testing set MSE for XGB tree
              xgbt_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         xgbt_pred_orig_data)
              
              
              ###
              ### Stacked Ensemble Equal Weighting ----
              # GAPD
              avg_stacked_ensemble_predictions_gapd = (linear_reg_pred  + 
                                                         lasso_pred  +  gapd_nnet_reg_pred  + 
                                                         gapd_radial_svm_reg_pred  + 
                                                         gapd_linear_svm_reg_pred  + 
                                                         gapd_poly_svm_reg_pred  +
                                                         ranger_rf_pred  + xgbt_pred) / 8
              
              avg_stacked_ensemble_gapd_mse =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         avg_stacked_ensemble_predictions_gapd)
              # No GAPD
              avg_stacked_ensemble_pred_no_gapd = 
                (linear_reg_unprivatized_pred  + 
                   lasso_pred_orig_data  + 
                   nnet_reg_unprivatized_pred + 
                   radial_svm_reg_unprivatized_pred  + 
                   linear_svm_reg_unprivatized_pred  + 
                   poly_svm_reg_unprivatized_pred  +
                   ranger_rf_pred_orig_data  + xgbt_pred_orig_data) / 8
              
              avg_stacked_ensemble_no_gapd_mse =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         avg_stacked_ensemble_pred_no_gapd)
              
              
              ### Stacked Ensemble Equal Weighting by Proportion of variance Explained ----------------
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_gapd = 
                c(training_lin_reg_mse, training_lasso_mse, 
                  training_gapd_nnet_reg_mse, training_gapd_radial_svm_reg_mse,
                  training_gapd_linear_svm_reg_mse, 
                  training_gapd_poly_svm_reg_mse, training_ranger_rf_mse,
                  training_xgbt_mse)
              
              # The models with the lowest MSE get the highest weight
              prop_var_exp_gapd_weighting = 
                (training_set_mses_gapd / sum(training_set_mses_gapd))^(-1)
              prop_var_exp_gapd_weighting = prop_var_exp_gapd_weighting / 
                sum(prop_var_exp_gapd_weighting)
              
              
              
              # Get the predictions for the stacked ensemble for the prop
              # of variance explained
              prop_var_stacked_ens_gapd_pred = 
                linear_reg_pred * prop_var_exp_gapd_weighting[1] +
                lasso_pred * prop_var_exp_gapd_weighting[2] + 
                gapd_nnet_reg_pred * prop_var_exp_gapd_weighting[3] + 
                gapd_radial_svm_reg_pred * prop_var_exp_gapd_weighting[4] + 
                gapd_linear_svm_reg_pred * prop_var_exp_gapd_weighting[5] + 
                gapd_poly_svm_reg_pred * prop_var_exp_gapd_weighting[6] +
                ranger_rf_pred * prop_var_exp_gapd_weighting[7] + 
                xgbt_pred * prop_var_exp_gapd_weighting[8]
              
              
              # Get the MSE
              prop_var_stacked_ens_gapd_mse = 
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         prop_var_stacked_ens_gapd_pred)
              
              
              # Get the Stacked Ensemble (Prop. of Var. explained) with No GADP 
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_no_gapd = 
                c(training_linear_reg_mse_orig_data, 
                  training_lasso_mse_orig_data, 
                  training_nnet_reg_unprivatized_mse, 
                  training_radial_svm_reg_mse_orig_data,
                  training_linear_svm_reg_mse_orig_data, 
                  training_poly_svm_reg_mse_orig_data, 
                  training_ranger_rf_mse_orig_data,
                  training_xgbt_mse_orig_data)
              
              # weighting for the SE
              prop_var_exp_no_gapd_weighting = 
                (training_set_mses_no_gapd / sum(training_set_mses_no_gapd))^(-1)
              prop_var_exp_no_gapd_weighting = prop_var_exp_no_gapd_weighting /
                sum(prop_var_exp_no_gapd_weighting)
              
              
              # get the predictions
              prop_var_stacked_ens_no_gapd_pred = 
                linear_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[1] +
                lasso_pred_orig_data * prop_var_exp_no_gapd_weighting[2] + 
                nnet_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[3] + 
                radial_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[4] + 
                linear_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[5] + 
                poly_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[6] +
                ranger_rf_pred_orig_data * prop_var_exp_no_gapd_weighting[7] + 
                xgbt_pred_orig_data * prop_var_exp_no_gapd_weighting[8]
              
              
              # Get the MSE
              prop_var_stacked_ens_no_gapd_mse = 
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         prop_var_stacked_ens_no_gapd_pred)
              
              
              # Compile MSEs into a vector
              gapd_mse_vector = 
                c(linear_reg_mse, lasso_mse, gapd_nnet_reg_mse,
                  gapd_radial_svm_reg_mse, gapd_linear_svm_reg_mse,
                  gapd_poly_svm_reg_mse, ranger_rf_mse,
                  xgbt_mse, avg_stacked_ensemble_gapd_mse, 
                  prop_var_stacked_ens_gapd_mse)
              
              no_gapd_mse_vector =
                c(linear_reg_mse_orig_data, lasso_mse_orig_data,
                  nnet_reg_mse_orig_data, radial_svm_reg_mse_orig_data,
                  linear_svm_reg_mse_orig_data, poly_svm_reg_mse_orig_data,
                  ranger_rf_mse_orig_data, xgbt_mse_orig_data,
                  avg_stacked_ensemble_no_gapd_mse, 
                  prop_var_stacked_ens_no_gapd_mse)
              
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "training_size", 
                                   "testing_size",
                                   "y", 
                                   "LR", "LASSO", "NNET", "Radial_SVM",
                                   "Linear_SVM", "Poly_SVM", "RF", "XGBTree",
                                   "Stacked_equal", "Stacked_prop","gadp", "theta", "rho")
              
              
              # construct the dataframe of results
              estimator_reg_df = data.frame(rbind(c(dat.seed,iteration, n[samp_size_index], testing_n,
                                                    response_index, gapd_mse_vector, 1, theta_value, model_rho)))
              
              # Adding results for predictive models without GAPD
              estimator_reg_df[2,] = rbind(c(dat.seed, iteration, 
                                             n[samp_size_index], testing_n,
                                             response_index, no_gapd_mse_vector, 0, theta_value, model_rho))
              
              
              colnames(estimator_reg_df) = data_frame_names
              
              
              return(estimator_reg_df)
            }
  
  stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  # Add name of the estimators
  
  results_list = list("output" = results,
                      "testing_set" = testing_model_data)
  
  return(results_list)
}

## Test and Modify GADP ----
# estimator_eval_stacked_ensemb_iterative_ver_2_mod = function( # the dataset
#   m_start = 1, # Primarily used for debugging determines which iteration to start from
#   m = 1000, # Number of Monte Carlo simulations
#   n = c(500, 1000, 2000), # sample sizes to be used for training set
#   cores = 2, # Number of CPU threads/ logical processors to use
#   response_prefix = "y", # prefix for response variables
#   acquire_variable_names = get_variables, #Function for getting the variables
#   perturbation_func = gadp_func, # CGADP or GADP function
#   theta_value = 0.4, # privacy parameter value
#   model_generating_function, # function to generate the model
#   testing_data_seed = 30000, # Seed for the testing set
#   model_predictor_quant = 10, # Quantity of the predictors for the model
#   model_rho = 0.2, # correlation of predictors in the model
#   model_noise = 0.2, # Noise to be used in the model
#   testing_n = 500) # Size of the testing set
# {
#   # generate the testing data
#   testing_model_data = model_generating_function(N = testing_n, rho = model_rho, 
#                                                  max_pred = model_predictor_quant, 
#                                                  dat.seed = testing_data_seed,
#                                                  noise = model_noise)
#   
#   # Getting variable names:
#   all_variable_names = colnames(testing_model_data)
#   organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
#                                           variable_names = all_variable_names)
#   
#   
#   # Get target variable indices
#   target_variables_indices = which(all_variable_names %in% organized_vars$responses)
#   # Variables serving as the predictors
#   predictor_vars = organized_vars$predictors
#   
#   
#   calc_mse = function(y_actual, y_hat)
#   {
#     mse = mean((y_actual - y_hat)^2)
#     return(mse)
#   }
#   
#   # Create the cluster
#   sim_cluster = makeCluster(cores)
#   # Register the cluster
#   registerDoSNOW(sim_cluster)
#   
#   # Store the results in a matrix
#   results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
#                     .packages = c("glmnet", "ranger", 
#                                   "nnet", "e1071", "xgboost", "dplyr"),
#                     .inorder = FALSE) %:%
#     foreach( iteration = m_start:m, 
#              .combine = "rbind",.inorder = FALSE) %:%
#     foreach(response_index = 1:length(organized_vars$response), 
#             .combine = "rbind",.inorder = FALSE) %dopar% {
#               
#               # setting the seed
#               dat.seed = iteration + n[samp_size_index]
#               set.seed(dat.seed)
#               
#               # Training dataset
#               # Generated with the same parameters sxcluding the seed
#               training_model_data = 
#                 model_generating_function(N = n[samp_size_index], 
#                                           rho = model_rho, 
#                                           max_pred = model_predictor_quant, 
#                                           dat.seed = dat.seed,
#                                           noise = model_noise)
#               
#               
#               # Performing SDC GADP method
#               # Testing data
#               privatized_responses_testing = 
#                 perturbation_func(unprotected_data = testing_model_data, 
#                                   target_indices = target_variables_indices, 
#                                   theta = theta_value, dat.seed = dat.seed)
#               
#               # Training data
#               privatized_responses_training = 
#                 perturbation_func(unprotected_data = training_model_data, 
#                                   target_indices = target_variables_indices, 
#                                   theta = theta_value, dat.seed = dat.seed)
#               
#               # Combining the training and the test datasets
#               privatized_responses = 
#                 rbind(privatized_responses_training,
#                       privatized_responses_testing)
#               
#               model_data = rbind(training_model_data, 
#                                  testing_model_data)
#               
#               
#               # Current response from the privatized data, 
#               # as indicated by the index for the response matrix
#               current_response = privatized_responses[,response_index]
#               
#               privatized_dataframe = model_data[,predictor_vars]
#               
#               privatized_dataframe[,colnames(privatized_responses)[response_index]] = 
#                 current_response
#               
#               # Get the training set form the df, first 1:n is training, n+1 
#               # onwards is the testing data set. 
#               drawing = 1:n[samp_size_index]
#               
#               # Sample data after gapd process
#               sample_private_data = privatized_dataframe[drawing,]
#               
#               # Sample data without gapd process
#               sample_unprotected_data = model_data[drawing,]
#               
#               # x predictors for test set
#               x_dat_holdout = cbind(rep(1, nrow(model_data[-drawing,])),
#                                     as.matrix(model_data)[-drawing,predictor_vars])
#               
#               # for the population
#               orig_data_response =
#                 as.matrix(model_data)[,
#                                       organized_vars$responses[response_index]]
#               
#               # x data does not change after gapd
#               x_dat = as.matrix(cbind(rep(1, nrow(sample_private_data[,predictor_vars])),
#                                       sample_private_data[,predictor_vars]))
#               
#               ### GAPD Used----
#               
#               # Linear Regression on data after gapd
#               betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
#                 as.matrix(current_response[drawing])
#               
#               
#               linear_reg_pred = as.vector(x_dat_holdout %*% betas_linear)
#               
#               # Get the predictions for the training data
#               
#               training_lin_reg_pred = as.vector(x_dat %*% betas_linear)
#               
#               
#               # Get the mse for the training data
#               
#               
#               training_lin_reg_mse = calc_mse(as.matrix(current_response[drawing]),
#                                               training_lin_reg_pred)
#               
#               # Get the mse for the test data
#               
#               
#               linear_reg_mse = calc_mse(orig_data_response[-drawing],
#                                         linear_reg_pred)
#               
#               
#               # LASSO on data after gapd
#               lasso_regress_model = cv.glmnet( 
#                 x = as.matrix(sample_private_data[,predictor_vars]),
#                 y = current_response[drawing])
#               
#               lasso_pred = predict(lasso_regress_model, 
#                                    as.matrix(privatized_dataframe[-drawing,predictor_vars]),
#                                    s = "lambda.min")
#               
#               # get predictions on the training data
#               training_lasso_pred = predict(lasso_regress_model, 
#                                             as.matrix(privatized_dataframe[drawing,predictor_vars]),
#                                             s = "lambda.min")
#               
#               # Get the MSE for the training data
#               training_lasso_mse = 
#                 calc_mse(as.matrix(current_response[drawing]),
#                          training_lasso_pred)
#               
#               # # MSE for the testing data
#               
#               lasso_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
#                                    lasso_pred)
#               
#               ### Neural Net Regression on data after gapd -------------
#               
#               
#               gapd_nnet = nnet(x = x_dat[,predictor_vars], y = current_response[drawing], 
#                                size = 3,
#                                maxit = 200, linout = TRUE)
#               
#               
#               gapd_nnet_reg_pred = predict(gapd_nnet,x_dat_holdout[,predictor_vars])
#               
#               # training data predictions
#               training_gapd_nnet_reg_pred = predict(gapd_nnet,x_dat[,predictor_vars])
#               
#               # training data MSE
#               training_gapd_nnet_reg_mse = 
#                 calc_mse(as.matrix(current_response[drawing]),
#                          training_gapd_nnet_reg_pred)
#               
#               
#               # testing data MSE
#               gapd_nnet_reg_mse = calc_mse(orig_data_response[-drawing],
#                                            gapd_nnet_reg_pred)
#               
#               # SVM Radial Regression on data after gapd
#               gapd_radial_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
#                                               gamma = seq(0.001,0.1,length.out = 2), 
#                                               cost = seq(0.1,1,length.out = 2),
#                                               kernel = c("radial"))
#               
#               gapd_radial_svm = gapd_radial_svm_tune$best.model
#               
#               gapd_radial_svm_reg_pred = predict(gapd_radial_svm,x_dat_holdout[,predictor_vars])
#               
#               # Get the training predictions of the radial svm
#               training_gapd_radial_svm_reg_pred =
#                 predict(gapd_radial_svm,x_dat[,predictor_vars])
#               
#               # Get the training mse for radial svm
#               training_gapd_radial_svm_reg_mse = 
#                 calc_mse(as.matrix(current_response[drawing]),
#                          training_gapd_radial_svm_reg_pred)
#               
#               # Testing data MSE
#               gapd_radial_svm_reg_mse = calc_mse(orig_data_response[-drawing],
#                                                  gapd_radial_svm_reg_pred)
#               
#               
#               
#               # SVM Linear Regression on data after gapd
#               gapd_linear_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
#                                               
#                                               cost = seq(0.1,1,length.out = 2),
#                                               kernel = c("linear"))
#               
#               gapd_linear_svm = gapd_linear_svm_tune$best.model
#               
#               # Predictions for the training data for Linear SVM
#               training_gapd_linear_svm_reg_pred =
#                 predict(gapd_linear_svm,x_dat_holdout[,predictor_vars])
#               
#               # Training MSE for Linear SVM
#               training_gapd_linear_svm_reg_mse =
#                 calc_mse(as.matrix(current_response[drawing]),
#                          training_gapd_linear_svm_reg_pred)
#               
#               # Predictions for the test dataset
#               gapd_linear_svm_reg_pred = predict(gapd_linear_svm,x_dat_holdout[,predictor_vars])
#               
#               # Testing data MSE
#               gapd_linear_svm_reg_mse = calc_mse(orig_data_response[-drawing],
#                                                  gapd_linear_svm_reg_pred)
#               
#               # SVM Polynomial Regression on data after gapd
#               gapd_poly_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
#                                             gamma = seq(0.001,0.1,length.out = 2), 
#                                             cost = seq(0.1,1,length.out = 2),
#                                             degree = seq(1,3,length.out=2),
#                                             kernel = c("polynomial"))
#               
#               gapd_poly_svm = gapd_poly_svm_tune$best.model
#               
#               # Training data polynomial SVM predictions
#               training_gapd_poly_svm_reg_pred = predict(gapd_poly_svm,x_dat[,predictor_vars])
#               
#               # Training data MSE polynomial SVM
#               training_gapd_poly_svm_reg_mse =
#                 calc_mse(as.matrix(current_response[drawing]),
#                          training_gapd_poly_svm_reg_pred)
#               
#               # Predictions for testing dataset
#               gapd_poly_svm_reg_pred = predict(gapd_poly_svm,x_dat_holdout[,predictor_vars])
#               
#               # MSE for testing dataset for polynomial kernel SVM
#               gapd_poly_svm_reg_mse = calc_mse(orig_data_response[-drawing],
#                                                gapd_poly_svm_reg_pred)
#               
#               # Random forest on data after gapd
#               num_x_vars = length(predictor_vars)
#               
#               ranger_rf_regress_model = ranger( 
#                 x = as.matrix(sample_private_data[,predictor_vars]),
#                 y = current_response[drawing], 
#                 num.trees = 500,
#                 num.threads = 1,
#                 mtry = ceiling(1 / 3 * num_x_vars))
#               
#               # Training dataset predicions for the random forest
#               training_ranger_rf_pred = 
#                 predict(ranger_rf_regress_model, 
#                         privatized_dataframe[drawing,predictor_vars])$predictions
#               
#               # training dataset MSE for the radom forest
#               training_ranger_rf_mse =
#                 calc_mse(as.matrix(current_response[drawing]),
#                          training_ranger_rf_pred)
#               
#               
#               # Testing dataset prediction for random forest
#               ranger_rf_pred = predict(ranger_rf_regress_model, 
#                                        privatized_dataframe[-drawing,predictor_vars])$predictions
#               
#               # Testing dataset MSE for random forest
#               ranger_rf_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
#                                        ranger_rf_pred)
#               
#               # XGBT on data after gapd
#               xgbt_regress_model = xgboost( 
#                 data = as.matrix(sample_private_data[,predictor_vars]),
#                 label = current_response[drawing], 
#                 nrounds = 40,
#                 nthread = 1)
#               
#               # Training data predictions for the XGB tree
#               training_xgbt_pred = 
#                 predict(xgbt_regress_model,
#                         as.matrix(privatized_dataframe[drawing,predictor_vars]))
#               
#               # training data MSE for the XGB tree
#               training_xgbt_mse = 
#                 calc_mse(as.matrix(current_response[drawing]),
#                          training_xgbt_pred)
#               
#               # Testing data predicitons for the XGB tree
#               xgbt_pred = predict(xgbt_regress_model, 
#                                   as.matrix(privatized_dataframe[-drawing,predictor_vars]))
#               
#               # Testing data MSE for the XGB tree
#               xgbt_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
#                                   xgbt_pred)
#               
#               ##
#               ## Before GAPD (On original data) ------
#               print(" running checkpoint 1")
#               # Linear Regression on data before gapd (unprotected original data)
#               
#               sample_orig_data_response = 
#                 as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
#               
#               
#               # Beta coefficients for the original data
#               orig_data_betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
#                 as.matrix(sample_orig_data_response)
#               
#               # Predictions on the training data no GADP
#               training_lin_reg_unprivat_pred = as.vector(x_dat %*%orig_data_betas_linear)
#               
#               
#               # MSE for training data for linear regression, no GADP
#               training_linear_reg_mse_orig_data =
#                 calc_mse(orig_data_response[drawing],
#                          training_lin_reg_unprivat_pred)
#               
#               # Testing data MSE for linear regression no ADP
#               linear_reg_unprivatized_pred = as.vector(x_dat_holdout %*% orig_data_betas_linear)
#               
#               linear_reg_mse_orig_data =
#                 calc_mse(orig_data_response[-drawing],
#                          linear_reg_unprivatized_pred)
#               
#               # LASSO on data before gapd (original data)
#               
#               
#               lasso_regress_model_orig_data =
#                 cv.glmnet( x = as.matrix(sample_unprotected_data[,predictor_vars]),
#                            y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
#               
#               # Training data predictions for LASSO
#               
#               training_lasso_pred_orig_data = 
#                 predict(lasso_regress_model_orig_data,
#                         as.matrix(model_data[drawing,predictor_vars]), 
#                         s = "lambda.min")
#               
#               # Training data MSE for LASSO
#               training_lasso_mse_orig_data =
#                 calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
#                          training_lasso_pred_orig_data)
#               # Predictions for the test dataset for the LASSO model
#               lasso_pred_orig_data = predict(lasso_regress_model_orig_data,
#                                              as.matrix(model_data[-drawing,predictor_vars]), 
#                                              s = "lambda.min")
#               # MSE for the LASSO model for the test data
#               lasso_mse_orig_data =
#                 calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
#                          lasso_pred_orig_data)
#               
#               
#               # Neural Net before GAPD 
#               
#               no_gapd_nnet = nnet(x = x_dat[,predictor_vars], y = sample_orig_data_response, 
#                                   size = 3,
#                                   maxit = 200, linout = TRUE)
#               
#               # The predictions of the training set for neural network
#               training_nnet_reg_unprivatized_pred = predict(no_gapd_nnet, x_dat[,predictor_vars])
#               
#               # The MSE for the training set for the neural network
#               training_nnet_reg_unprivatized_mse = 
#                 calc_mse(orig_data_response[drawing],
#                          training_nnet_reg_unprivatized_pred)
#               
#               # Predicitons on testing set for neural net
#               nnet_reg_unprivatized_pred = predict(no_gapd_nnet, x_dat_holdout[,predictor_vars])
#               
#               # MSE for testing set for neural network
#               nnet_reg_mse_orig_data =
#                 calc_mse(orig_data_response[-drawing],
#                          nnet_reg_unprivatized_pred)
#               
#               # SVM Radial before GAPD
#               
#               
#               no_gapd_radial_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
#                                                  y = sample_orig_data_response,
#                                                  gamma = seq(0.001,0.1,length.out = 2), 
#                                                  cost = seq(0.1,1,length.out = 2),
#                                                 
#                                                  kernel = c("radial"))
#               
#               # Get the best model for Radial SVM
#               no_gapd_radial_svm = no_gapd_radial_svm_tune$best.model
#               
#               # Predictions on the training set for radial SVM
#               training_radial_svm_reg_unprivatized_pred = 
#                 predict(no_gapd_radial_svm, x_dat[,predictor_vars])
#               # The MSE for the training set for radial SVM
#               training_radial_svm_reg_mse_orig_data =
#                 calc_mse(orig_data_response[drawing],
#                          training_radial_svm_reg_unprivatized_pred)
#               
#               # Get the predictions of Radial SVM for the testing set
#               radial_svm_reg_unprivatized_pred = 
#                 predict(no_gapd_radial_svm, x_dat_holdout[,predictor_vars])
#               # Get the MSE of the Radial SVM for the testing set
#               radial_svm_reg_mse_orig_data =
#                 calc_mse(orig_data_response[-drawing],
#                          radial_svm_reg_unprivatized_pred)
#               
#               # Linear SVM before GAPD
#               no_gapd_linear_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
#                                                  y = sample_orig_data_response, 
#                                                  cost = seq(0.1,1,length.out = 2),
#                                                  kernel = c("linear"))
#               # The Linear SVM best tuned model
#               no_gapd_linear_svm = no_gapd_linear_svm_tune$best.model
#               
#               # The predictions on the training set, Linear SVM
#               training_linear_svm_reg_unprivatized_pred = 
#                 predict(no_gapd_linear_svm, x_dat[,predictor_vars])
#               # the MSE on the training set, Linear SVM
#               training_linear_svm_reg_mse_orig_data =
#                 calc_mse(orig_data_response[drawing],
#                          training_linear_svm_reg_unprivatized_pred)
#               # The predictions for the testing set Linear SVM
#               linear_svm_reg_unprivatized_pred = 
#                 predict(no_gapd_linear_svm, x_dat_holdout[,predictor_vars])
#               
#               # The MSE for the testing set Linear SVM
#               linear_svm_reg_mse_orig_data =
#                 calc_mse(orig_data_response[-drawing],
#                          linear_svm_reg_unprivatized_pred)
#               
#               
#               # Polynomial SVM before GAPD
#               
#               no_gapd_poly_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
#                                                y = sample_orig_data_response,
#                                                gamma = seq(0.001,0.1,length.out = 2), 
#                                                cost = seq(0.1,1,length.out = 2),
#                                                degree = seq(1,3,length.out=2),
#                                                kernel = c("polynomial"))
#               # Get the best tuned polynomial SVM
#               no_gapd_poly_svm = no_gapd_poly_svm_tune$best.model
#               
#               # The training set predicitons for the polynomial SVM
#               training_poly_svm_reg_unprivatized_pred = 
#                 predict(no_gapd_poly_svm, x_dat[,predictor_vars])
#               # the training set MSE for the polynomial SVM
#               training_poly_svm_reg_mse_orig_data =
#                 calc_mse(orig_data_response[drawing],
#                          training_poly_svm_reg_unprivatized_pred)
#               
#               # The testing set predicitons using the polynomial SVM
#               poly_svm_reg_unprivatized_pred = 
#                 predict(no_gapd_poly_svm, x_dat_holdout[,predictor_vars])
#               # The testing set MSE for the polynomial SVM
#               poly_svm_reg_mse_orig_data =
#                 calc_mse(orig_data_response[-drawing],
#                          poly_svm_reg_unprivatized_pred)
#               
#               # Random forest on data before gapd
#               
#               ranger_rf_model_orig_data = ranger( 
#                 x = as.matrix(sample_unprotected_data[,predictor_vars]),
#                 y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]), 
#                 num.trees = 500,
#                 num.threads = 1,
#                 mtry = ceiling(1 / 3 * num_x_vars))
#               
#               # The training set predictions from random forest
#               training_ranger_rf_pred_orig_data = 
#                 predict(ranger_rf_model_orig_data,
#                         model_data[drawing,predictor_vars])$predictions
#               # The training set MSE for the random forest
#               training_ranger_rf_mse_orig_data =
#                 calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
#                          training_ranger_rf_pred_orig_data)
#               
#               # Testing set predictions for random forest
#               ranger_rf_pred_orig_data = predict(ranger_rf_model_orig_data,
#                                                  model_data[-drawing,predictor_vars])$predictions
#               # Testing set MSE for random forest
#               ranger_rf_mse_orig_data =
#                 calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
#                          ranger_rf_pred_orig_data)
#               
#               # XGBT on data before gapd
#               xgbt_regress_model_orig_data =
#                 xgboost( data = as.matrix(sample_unprotected_data[,predictor_vars]),
#                          label = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]),
#                          nrounds = 40,
#                          nthread = 1)
#               
#               # The training set predictions for XGB tree
#               training_xgbt_pred_orig_data = 
#                 predict(xgbt_regress_model_orig_data,
#                         as.matrix(model_data[drawing,predictor_vars]))
#               # Training set MSE for XGB tree for 
#               training_xgbt_mse_orig_data =
#                 calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
#                          training_xgbt_pred_orig_data)
#               
#               # The testing set XGB tree predictions
#               xgbt_pred_orig_data = 
#                 predict(xgbt_regress_model_orig_data,
#                         as.matrix(model_data[-drawing,predictor_vars]))
#               # The testing set MSE for XGB tree
#               xgbt_mse_orig_data =
#                 calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
#                          xgbt_pred_orig_data)
#               
#               
#               ###
#               ### Stacked Ensemble Equal Weighting ----
#               # GAPD
#               avg_stacked_ensemble_predictions_gapd = (linear_reg_pred  + 
#                                                          lasso_pred  +  gapd_nnet_reg_pred  + 
#                                                          gapd_radial_svm_reg_pred  + 
#                                                          gapd_linear_svm_reg_pred  + 
#                                                          gapd_poly_svm_reg_pred  +
#                                                          ranger_rf_pred  + xgbt_pred) / 8
#               
#               avg_stacked_ensemble_gapd_mse =
#                 calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
#                          avg_stacked_ensemble_predictions_gapd)
#               # No GAPD
#               avg_stacked_ensemble_pred_no_gapd = 
#                 (linear_reg_unprivatized_pred  + 
#                    lasso_pred_orig_data  + 
#                    nnet_reg_unprivatized_pred + 
#                    radial_svm_reg_unprivatized_pred  + 
#                    linear_svm_reg_unprivatized_pred  + 
#                    poly_svm_reg_unprivatized_pred  +
#                    ranger_rf_pred_orig_data  + xgbt_pred_orig_data) / 8
#               
#               avg_stacked_ensemble_no_gapd_mse =
#                 calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
#                          avg_stacked_ensemble_pred_no_gapd)
#               
#               
#               ### Stacked Ensemble Equal Weighting by Proportion of variance Explained ----------------
#               
#               # in this order:
#               # linear regression, LASSO, neural net, radial SVM, linear SVM,
#               # poly SVM, RF, XGB Tree
#               training_set_mses_gapd = 
#                 c(training_lin_reg_mse, training_lasso_mse, 
#                   training_gapd_nnet_reg_mse, training_gapd_radial_svm_reg_mse,
#                   training_gapd_linear_svm_reg_mse, 
#                   training_gapd_poly_svm_reg_mse, training_ranger_rf_mse,
#                   training_xgbt_mse)
#               
#               # The models with the lowest MSE get the highest weight
#               prop_var_exp_gapd_weighting = 
#                 (training_set_mses_gapd / sum(training_set_mses_gapd))^(-1)
#               prop_var_exp_gapd_weighting = prop_var_exp_gapd_weighting / 
#                 sum(prop_var_exp_gapd_weighting)
#               
#               
#               
#               # Get the predictions for the stacked ensemble for the prop
#               # of variance explained
#               prop_var_stacked_ens_gapd_pred = 
#                 linear_reg_pred * prop_var_exp_gapd_weighting[1] +
#                 lasso_pred * prop_var_exp_gapd_weighting[2] + 
#                 gapd_nnet_reg_pred * prop_var_exp_gapd_weighting[3] + 
#                 gapd_radial_svm_reg_pred * prop_var_exp_gapd_weighting[4] + 
#                 gapd_linear_svm_reg_pred * prop_var_exp_gapd_weighting[5] + 
#                 gapd_poly_svm_reg_pred * prop_var_exp_gapd_weighting[6] +
#                 ranger_rf_pred * prop_var_exp_gapd_weighting[7] + 
#                 xgbt_pred * prop_var_exp_gapd_weighting[8]
#               
#               
#               # Get the MSE
#               prop_var_stacked_ens_gapd_mse = 
#                 calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
#                          prop_var_stacked_ens_gapd_pred)
#               
#               
#               # Get the Stacked Ensemble (Prop. of Var. explained) with No GADP 
#               
#               # in this order:
#               # linear regression, LASSO, neural net, radial SVM, linear SVM,
#               # poly SVM, RF, XGB Tree
#               training_set_mses_no_gapd = 
#                 c(training_linear_reg_mse_orig_data, 
#                   training_lasso_mse_orig_data, 
#                   training_nnet_reg_unprivatized_mse, 
#                   training_radial_svm_reg_mse_orig_data,
#                   training_linear_svm_reg_mse_orig_data, 
#                   training_poly_svm_reg_mse_orig_data, 
#                   training_ranger_rf_mse_orig_data,
#                   training_xgbt_mse_orig_data)
#               
#               # weighting for the SE
#               prop_var_exp_no_gapd_weighting = 
#                 (training_set_mses_no_gapd / sum(training_set_mses_no_gapd))^(-1)
#               prop_var_exp_no_gapd_weighting = prop_var_exp_no_gapd_weighting /
#                 sum(prop_var_exp_no_gapd_weighting)
#               
#               
#               # get the predictions
#               prop_var_stacked_ens_no_gapd_pred = 
#                 linear_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[1] +
#                 lasso_pred_orig_data * prop_var_exp_no_gapd_weighting[2] + 
#                 nnet_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[3] + 
#                 radial_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[4] + 
#                 linear_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[5] + 
#                 poly_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[6] +
#                 ranger_rf_pred_orig_data * prop_var_exp_no_gapd_weighting[7] + 
#                 xgbt_pred_orig_data * prop_var_exp_no_gapd_weighting[8]
#               
#               
#               # Get the MSE
#               prop_var_stacked_ens_no_gapd_mse = 
#                 calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
#                          prop_var_stacked_ens_no_gapd_pred)
#               
#               
#               # Compile MSEs into a vector
#               gapd_mse_vector = 
#                 c(linear_reg_mse, lasso_mse, gapd_nnet_reg_mse,
#                   gapd_radial_svm_reg_mse, gapd_linear_svm_reg_mse,
#                   gapd_poly_svm_reg_mse, ranger_rf_mse,
#                   xgbt_mse, avg_stacked_ensemble_gapd_mse, 
#                   prop_var_stacked_ens_gapd_mse)
#               
#               no_gapd_mse_vector =
#                 c(linear_reg_mse_orig_data, lasso_mse_orig_data,
#                   nnet_reg_mse_orig_data, radial_svm_reg_mse_orig_data,
#                   linear_svm_reg_mse_orig_data, poly_svm_reg_mse_orig_data,
#                   ranger_rf_mse_orig_data, xgbt_mse_orig_data,
#                   avg_stacked_ensemble_no_gapd_mse, 
#                   prop_var_stacked_ens_no_gapd_mse)
#               
#               
#               # Dataframe names
#               data_frame_names = c("seed", "iteration", "training_size", 
#                                    "testing_size",
#                                    "y", 
#                                    "LR", "LASSO", "NNET", "Radial_SVM",
#                                    "Linear_SVM", "Poly_SVM", "RF", "XGBTree",
#                                    "Stacked_equal", "Stacked_prop","gadp", "theta", "rho")
#               
#               
#               # construct the dataframe of results
#               estimator_reg_df = data.frame(rbind(c(dat.seed,iteration, n[samp_size_index], testing_n,
#                                                     response_index, gapd_mse_vector, 1, theta_value, model_rho)))
#               
#               # Adding results for predictive models without GAPD
#               estimator_reg_df[2,] = rbind(c(dat.seed, iteration, 
#                                              n[samp_size_index], testing_n,
#                                              response_index, no_gapd_mse_vector, 0, theta_value, model_rho))
#               
#               
#               colnames(estimator_reg_df) = data_frame_names
#               
#               
#               return(estimator_reg_df)
#             }
#   
#   stopCluster(sim_cluster)
#   
#   results = data.frame(results)
#   
#   # Add name of the estimators
#   
#   results_list = list("output" = results,
#                       "testing_set" = testing_model_data)
#   
#   return(results_list)
# }






# Stacked Ensemble CGADP --------
## Stacked Ensemble Static Dataset ---------
estimator_evaluation_stacked_ensemb_cgadp = function(model_data, # the dataset
                                               m = 1000, # Number of Monte Carlo simulations
                                               m_start = 1, # Start of iterations, used for debugging if different than 1
                                               n = c(500, 800), # sample sizes to be used
                                               cores = 2,
                                               response_prefix = "y",
                                               acquire_variable_names = get_variables,
                                               sdc_form = sdc_formula_generation,
                                               perturbation_func = NA,
                                               theta_value = 0.4) # Number of CPU threads/ logical processors to use
{
  
  # Getting variable names:
  all_variable_names = colnames(model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  
  # print(organized_vars)
  # sdc_data_formula = sdc_form(target_variables = organized_vars$response, 
  #                             predictor_variables = organized_vars$predictors)
  
  
  calc_mse = function(y_actual, y_hat)
  {
    print("Actual y")
    print(length(y_actual))
    print("Predicted y")
    print(length(y_hat))
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  sim_cluster = makeCluster(cores)
  registerDoSNOW(sim_cluster)
  
  results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
                    .packages = c( "Matrix", "glmnet", "ranger", 
                                  "nnet", "e1071", "xgboost"),
                    .inorder = FALSE) %:%
    foreach( iteration = m_start:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",
            .inorder = FALSE) %dopar% {
              
              # setting the seed
              dat.seed = iteration + n[samp_size_index]
              set.seed(dat.seed)
              
              
              # Performing SDC GADP method
              privatized_dataframe = 
                perturbation_func(unprotected_data = model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              
              print("cgadp successful")
              
              privatized_responses = privatized_dataframe[, target_variables_indices]
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,organized_vars$predictors]
              
              # Get the sample from privatized df
              drawing = sample(1:nrow(privatized_dataframe), size = n[samp_size_index], replace = FALSE)
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              # Adding intercept term for linear regression
              x_dat_holdout = cbind(1,as.matrix(privatized_dataframe)[-drawing,organized_vars$predictors])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after cgapd, but preparing for case when both x and y are randomized
              x_dat = as.matrix(cbind(1,
                                      sample_private_data[,organized_vars$predictors]))
              
              
              # Get the predictors for the training set for unprotected data
              x_dat_unprotected = 
                as.matrix(cbind(1, sample_unprotected_data[,organized_vars$predictors]))
              
              # Get the predictors for the testing set for unprotected data
              x_dat_holdout_unprotected = 
                as.matrix(cbind(1, model_data[-drawing,organized_vars$predictors]))
              
              
              
              ### CGAPD Used----
              print("CGADP Data transformation...")
              
              # Linear Regression on data after cgapd
              betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(current_response[drawing])
              
              
              linear_reg_pred = as.vector(x_dat_holdout %*% betas_linear)
              
              # Get the predictions for the training data
              
              training_lin_reg_pred = as.vector(x_dat %*% betas_linear)
              
              
              # Get the mse for the training data
              
              
              training_lin_reg_mse = calc_mse(as.matrix(current_response[drawing]),
                                              training_lin_reg_pred)
              
              # Get the mse for the test data
              
              
              linear_reg_mse = calc_mse(orig_data_response[-drawing],
                                        linear_reg_pred)
              
              
              # LASSO on data after cgapd
              
              lasso_regress_model = cv.glmnet( 
                x = as.matrix(sample_private_data[,organized_vars$predictors]),
                y = current_response[drawing])
              
              
              lasso_pred = predict(lasso_regress_model, 
                                   as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]),
                                   s = "lambda.min")
              
              # get predictions on the training data
              training_lasso_pred = predict(lasso_regress_model, 
                                            as.matrix(privatized_dataframe[drawing,organized_vars$predictors]),
                                            s = "lambda.min")
              
              # Get the MSE for the training data
              training_lasso_mse = 
                calc_mse(orig_data_response[drawing],
                         training_lasso_pred)
              
              # # MSE for the testing data
              
              lasso_mse = calc_mse(orig_data_response[-drawing],
                                   lasso_pred)
              
              
              # Neural Net Regression on data after cgapd
              cgapd_nnet = nnet(x = x_dat[,-1], y = current_response[drawing], 
                               size = 3,
                               maxit = 200, linout = TRUE)
              
              
              cgapd_nnet_reg_pred = predict(cgapd_nnet,x_dat_holdout[,-1])
              
              # training data predictions
              training_cgapd_nnet_reg_pred = predict(cgapd_nnet,x_dat[,-1])
              
              # training data MSE
              training_cgapd_nnet_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_nnet_reg_pred)
              
              
              # testing data MSE
              cgapd_nnet_reg_mse = calc_mse(orig_data_response[-drawing],
                                           cgapd_nnet_reg_pred)
              
              #print("Neural network run for CGADP finished.")
              
              ## SVM Runs for CGADP -------
              
              
              # SVM Radial Regression on data after cgapd
              cgapd_radial_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                              gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                              degree = 3:4,
                                              kernel = c("radial"))
              
              cgapd_radial_svm = cgapd_radial_svm_tune$best.model
              
              cgapd_radial_svm_reg_pred = predict(cgapd_radial_svm,x_dat_holdout[,-1])
              
              # Get the training predictions of the radial svm
              training_cgapd_radial_svm_reg_pred =
                predict(cgapd_radial_svm, x_dat[,-1])
              
              # Get the training mse for radial svm
              training_cgapd_radial_svm_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_radial_svm_reg_pred)
              
              # Testing data MSE
              cgapd_radial_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                 cgapd_radial_svm_reg_pred)
              
              
              
              # SVM Linear Regression on data after cgapd
              cgapd_linear_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                              gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                              degree = 3:4,
                                              kernel = c("linear"))
              
              cgapd_linear_svm = cgapd_linear_svm_tune$best.model
              
              # Predictions for the training data for Linear SVM
              training_cgapd_linear_svm_reg_pred =
                predict(cgapd_linear_svm,x_dat_holdout[,-1])
              
              # Training MSE for Linear SVM
              training_cgapd_linear_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_linear_svm_reg_pred)
              
              # Predictions for the test dataset
              cgapd_linear_svm_reg_pred = predict(cgapd_linear_svm,x_dat_holdout[,-1])
              
              # Testing data MSE
              cgapd_linear_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                 cgapd_linear_svm_reg_pred)
              
              # SVM Polynomial Regression on data after gapd
              cgapd_poly_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                            gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                            degree = 3:4,
                                            kernel = c("polynomial"))
              
              cgapd_poly_svm = cgapd_poly_svm_tune$best.model
              
              # Training data polynomial SVM predictions
              training_cgapd_poly_svm_reg_pred = predict(cgapd_poly_svm,x_dat[,-1])
              
              # Training data MSE polynomial SVM
              training_cgapd_poly_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_poly_svm_reg_pred)
              
              # Predictions for testing dataset
              cgapd_poly_svm_reg_pred = predict(cgapd_poly_svm,x_dat_holdout[,-1])
              
              # MSE for testing dataset for polynomial kernel SVM
              cgapd_poly_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                               cgapd_poly_svm_reg_pred)
              
              ## Random Forest Run ---------------------------
              
              # Random forest on data after cgapd
              num_x_vars = length(organized_vars$predictors)
              
              ranger_rf_regress_model = ranger( 
                x = as.matrix(sample_private_data[,organized_vars$predictors]),
                y = current_response[drawing], 
                num.trees = 500,
                num.threads = 1,
                mtry = ceiling(1 / 3 * num_x_vars))
              
              # Training dataset predicions for the random forest
              training_ranger_rf_pred = 
                predict(ranger_rf_regress_model, 
                        privatized_dataframe[drawing,organized_vars$predictors])$predictions
              
              # training dataset MSE for the radom forest
              training_ranger_rf_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_ranger_rf_pred)
              
              
              # Testing dataset prediction for random forest
              ranger_rf_pred = predict(ranger_rf_regress_model, 
                                       privatized_dataframe[-drawing,organized_vars$predictors])$predictions
              
              # Testing dataset MSE for random forest
              ranger_rf_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                       ranger_rf_pred)
              
              # XGBT on data after cgapd
              xgbt_regress_model = xgboost( 
                data = as.matrix(sample_private_data[,organized_vars$predictors]),
                label = current_response[drawing], 
                nrounds = 40,
                nthread = 1)
              
              # Training data predictions for the XGB tree
              training_xgbt_pred = 
                predict(xgbt_regress_model,
                        as.matrix(privatized_dataframe[drawing,organized_vars$predictors]))
              
              # training data MSE for the XGB tree
              training_xgbt_mse = 
                calc_mse(as.matrix(current_response[drawing]) ,
                         training_xgbt_pred)
              
              # Testing data predicitons for the XGB tree
              xgbt_pred = predict(xgbt_regress_model, 
                                  as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]))
              
              # Testing data MSE for the XGB tree
              xgbt_mse = calc_mse(orig_data_response[-drawing],
                                  xgbt_pred)
              
              ##
              ## Before CGAPD (On original data) ------
              print("Using original data without CGADP")
              
              
              # Linear Regression on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              
              # Beta coefficients for the original data
              orig_data_betas_linear = solve(t(x_dat_unprotected) %*% 
                                               x_dat_unprotected) %*% t(x_dat_unprotected) %*% 
                as.matrix(sample_orig_data_response)
              
              # Predictions on the training data no GADP
              training_lin_reg_unprivat_pred = 
                as.vector(x_dat_unprotected %*% orig_data_betas_linear)
              
              
              # MSE for training data for linear regression, no GADP
              training_linear_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_lin_reg_unprivat_pred)
              
              # Testing data MSE for linear regression no ADP
              linear_reg_unprivatized_pred = as.vector(x_dat_holdout_unprotected %*% 
                                                         orig_data_betas_linear)
              
              linear_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_reg_unprivatized_pred)
              
              # LASSO on data before gapd (original data)
              
              
              lasso_regress_model_orig_data =
                cv.glmnet( x = as.matrix(model_data[drawing,organized_vars$predictors]), # Get rid of intercept term
                           y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # Training data predictions for LASSO
              
              training_lasso_pred_orig_data = 
                predict(lasso_regress_model_orig_data,
                        as.matrix(model_data[drawing,organized_vars$predictors]), 
                        s = "lambda.min")
              
              # Training data MSE for LASSO
              training_lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_lasso_pred_orig_data)
              
              # Predictions for the test dataset for the LASSO model
              lasso_pred_orig_data = predict(lasso_regress_model_orig_data,
                                             as.matrix(model_data[-drawing,organized_vars$predictors]), 
                                             s = "lambda.min")
              # MSE for the LASSO model for the test data
              lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         lasso_pred_orig_data)
              
              
              # Neural Net before GAPD 
              
              no_cgapd_nnet = nnet(x = x_dat_unprotected[,-1], 
                                   y = sample_orig_data_response, 
                                  size = 3,
                                  maxit = 200, linout = TRUE)
              
              # The predictions of the training set for neural network
              training_nnet_reg_unprivatized_pred = predict(no_cgapd_nnet, 
                                                            x_dat_unprotected[,-1])
              
              # The MSE for the training set for the neural network
              training_nnet_reg_unprivatized_mse = 
                calc_mse(orig_data_response[drawing],
                         training_nnet_reg_unprivatized_pred)
              
              # Predicitons on testing set for neural net
              nnet_reg_unprivatized_pred = predict(no_cgapd_nnet, 
                                                   x_dat_holdout_unprotected[,-1])
              
              # MSE for testing set for neural network
              nnet_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         nnet_reg_unprivatized_pred)
              
              # SVM Radial before GAPD
              
              
              no_cgapd_radial_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                 y = sample_orig_data_response,
                                                 gamma = seq(0.001,0.1,length.out = 2), 
                                                 cost = seq(1,3,length.out = 2),
                                                 degree = 3:4,
                                                 kernel = c("radial"))
              
              # Get the best model for Radial SVM
              no_cgapd_radial_svm = no_cgapd_radial_svm_tune$best.model
              
              # Predictions on the training set for radial SVM
              training_radial_svm_reg_unprivatized_pred = 
                predict(no_cgapd_radial_svm, x_dat_unprotected[,-1])
              
              # The MSE for the training set for radial SVM
              training_radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_radial_svm_reg_unprivatized_pred)
              
              # Get the predictions of Radial SVM for the testing set
              radial_svm_reg_unprivatized_pred = 
                predict(no_cgapd_radial_svm, x_dat_holdout_unprotected[,-1])
              
              # Get the MSE of the Radial SVM for the testing set
              radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         radial_svm_reg_unprivatized_pred)
              
              # Linear SVM before GAPD
              no_cgapd_linear_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                 y = sample_orig_data_response,
                                                 gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                                 degree = 3:4,
                                                 kernel = c("linear"))
              # The Linear SVM best tuned model
              no_cgapd_linear_svm = no_cgapd_linear_svm_tune$best.model
              
              # The predictions on the training set, Linear SVM
              training_linear_svm_reg_unprivatized_pred = 
                predict(no_cgapd_linear_svm, x_dat_unprotected[,-1])
              
              # the MSE on the training set, Linear SVM
              training_linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_linear_svm_reg_unprivatized_pred)
              
              # The predictions for the testing set Linear SVM
              linear_svm_reg_unprivatized_pred = 
                predict(no_cgapd_linear_svm, x_dat_holdout_unprotected[,-1])
              
              # The MSE for the testing set Linear SVM
              linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_svm_reg_unprivatized_pred)
              
              
              # Polynomial SVM before GAPD
              
              no_cgapd_poly_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                               y = sample_orig_data_response,
                                               gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                               degree = 3:4,
                                               kernel = c("polynomial"))
              # Get the best tuned polynomial SVM
              no_cgapd_poly_svm = no_cgapd_poly_svm_tune$best.model
              
              # The training set predicitons for the polynomial SVM
              training_poly_svm_reg_unprivatized_pred = 
                predict(no_cgapd_poly_svm, x_dat_unprotected[,-1])
              
              # the training set MSE for the polynomial SVM
              training_poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_poly_svm_reg_unprivatized_pred)
              
              # The testing set predicitons using the polynomial SVM
              poly_svm_reg_unprivatized_pred = 
                predict(no_cgapd_poly_svm, x_dat_holdout_unprotected[,-1])
              
              # The testing set MSE for the polynomial SVM
              poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         poly_svm_reg_unprivatized_pred)
              
              # Random forest on data before cgapd
              
              ranger_rf_model_orig_data = ranger( 
                x = model_data[drawing,organized_vars$predictors],
                y = orig_data_response[drawing], 
                num.trees = 500,
                num.threads = 1,
                mtry = ceiling(1 / 3 * num_x_vars))
              
              # The training set predictions from random forest
              training_ranger_rf_pred_orig_data = 
                predict(ranger_rf_model_orig_data,
                        model_data[drawing,organized_vars$predictors])$predictions
              
              # The training set MSE for the random forest
              training_ranger_rf_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_ranger_rf_pred_orig_data)
              
              # Testing set predictions for random forest
              ranger_rf_pred_orig_data = predict(ranger_rf_model_orig_data,
                                                 model_data[-drawing,organized_vars$predictors])$predictions
              # Testing set MSE for random forest
              ranger_rf_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         ranger_rf_pred_orig_data)
              
              # XGBT on data before cgapd
              xgbt_regress_model_orig_data =
                xgboost(data = as.matrix(sample_unprotected_data[,organized_vars$predictors]),
                         label = orig_data_response[drawing],
                         nrounds = 40,
                         nthread = 1)
              
              # The training set predictions for XGB tree
              training_xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(sample_unprotected_data[,organized_vars$predictors]))
              # Training set MSE for XGB tree for 
              training_xgbt_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_xgbt_pred_orig_data)
              
              # The testing set XGB tree predictions
              xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[-drawing,organized_vars$predictors]))
              
              # The testing set MSE for XGB tree
              xgbt_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         xgbt_pred_orig_data)
              
              
             #print("Constructing the stacked ensemble")
              ###
              ### Stacked Ensemble Equal Weighting ----
              # CGAPD
              avg_stacked_ensemble_predictions_cgapd = (linear_reg_pred  + 
                lasso_pred  +  cgapd_nnet_reg_pred + 
                cgapd_radial_svm_reg_pred  + 
                cgapd_linear_svm_reg_pred  + 
                cgapd_poly_svm_reg_pred  +
                ranger_rf_pred  + xgbt_pred) / 8
              
              avg_stacked_ensemble_cgapd_mse =
                calc_mse(orig_data_response[-drawing],
                         avg_stacked_ensemble_predictions_cgapd)
              # No GAPD
              avg_stacked_ensemble_pred_no_cgapd = 
                (linear_reg_unprivatized_pred  + 
                lasso_pred_orig_data  + 
                nnet_reg_unprivatized_pred  + 
                radial_svm_reg_unprivatized_pred  + 
                linear_svm_reg_unprivatized_pred  + 
                poly_svm_reg_unprivatized_pred  +
                ranger_rf_pred_orig_data  + xgbt_pred_orig_data) / 8
              
              avg_stacked_ensemble_no_cgapd_mse =
                calc_mse(orig_data_response[-drawing],
                         avg_stacked_ensemble_pred_no_cgapd)
              
              
              ### Stacked Ensemble Equal Weighting by Proportion of variance Explained ----------------
              # under CGADP
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_cgapd = 
                c(training_lin_reg_mse, training_lasso_mse, 
                  training_cgapd_nnet_reg_mse, training_cgapd_radial_svm_reg_mse,
                  training_cgapd_linear_svm_reg_mse, 
                  training_cgapd_poly_svm_reg_mse, training_ranger_rf_mse,
                  training_xgbt_mse)
              
              # The models with the lowest MSE get the highest weight
              prop_var_exp_cgapd_weighting = 
                (training_set_mses_cgapd / sum(training_set_mses_cgapd))^(-1)
              
              prop_var_exp_cgapd_weighting = prop_var_exp_cgapd_weighting / 
                sum(prop_var_exp_cgapd_weighting)
              
              
              
              # Get the predictions for the stacked ensemble for the prop
              # of variance explained
              prop_var_stacked_ens_cgapd_pred = 
                linear_reg_pred * prop_var_exp_cgapd_weighting[1] +
                lasso_pred * prop_var_exp_cgapd_weighting[2] + 
                cgapd_nnet_reg_pred * prop_var_exp_cgapd_weighting[3] + 
                cgapd_radial_svm_reg_pred * prop_var_exp_cgapd_weighting[4] + 
                cgapd_linear_svm_reg_pred * prop_var_exp_cgapd_weighting[5] + 
                cgapd_poly_svm_reg_pred * prop_var_exp_cgapd_weighting[6] +
                ranger_rf_pred * prop_var_exp_cgapd_weighting[7] + 
                xgbt_pred * prop_var_exp_cgapd_weighting[8]
              
              
              
              # Get the MSE
              prop_var_stacked_ens_cgapd_mse = 
                calc_mse(orig_data_response[-drawing],
                         prop_var_stacked_ens_cgapd_pred)
              
              
              # Get the Stacked Ensemble (Prop. of Var. explained) with No GADP 
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_no_cgapd = 
                c(training_linear_reg_mse_orig_data, 
                  training_lasso_mse_orig_data, 
                  training_nnet_reg_unprivatized_mse, 
                  training_radial_svm_reg_mse_orig_data,
                  training_linear_svm_reg_mse_orig_data, 
                  training_poly_svm_reg_mse_orig_data, 
                  training_ranger_rf_mse_orig_data,
                  training_xgbt_mse_orig_data)
              
              # weighting for the SE
              prop_var_exp_no_cgapd_weighting = 
                (training_set_mses_no_cgapd / sum(training_set_mses_no_cgapd))^(-1)
              
              prop_var_exp_no_cgapd_weighting = prop_var_exp_no_cgapd_weighting /
                sum(prop_var_exp_no_cgapd_weighting)
              
              
              # get the predictions
              prop_var_stacked_ens_no_cgapd_pred = 
                linear_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[1] +
                lasso_pred_orig_data * prop_var_exp_no_cgapd_weighting[2] + 
                nnet_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[3] + 
                radial_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[4] + 
                linear_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[5] + 
                poly_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[6] +
                ranger_rf_pred_orig_data * prop_var_exp_no_cgapd_weighting[7] + 
                xgbt_pred_orig_data * prop_var_exp_no_cgapd_weighting[8]
              
              # Get the MSE
              prop_var_stacked_ens_no_cgapd_mse = 
                calc_mse(orig_data_response[-drawing],
                         prop_var_stacked_ens_no_cgapd_pred)
              
              
              # Compile MSEs into a vector
              cgapd_mse_vector = 
                c(linear_reg_mse, lasso_mse, cgapd_nnet_reg_mse,
                  cgapd_radial_svm_reg_mse, cgapd_linear_svm_reg_mse,
                  cgapd_poly_svm_reg_mse, ranger_rf_mse,
                  xgbt_mse, avg_stacked_ensemble_cgapd_mse, 
                  prop_var_stacked_ens_cgapd_mse)
              
              no_cgapd_mse_vector =
                c(linear_reg_mse_orig_data, lasso_mse_orig_data,
                  nnet_reg_mse_orig_data, radial_svm_reg_mse_orig_data,
                  linear_svm_reg_mse_orig_data, poly_svm_reg_mse_orig_data,
                  ranger_rf_mse_orig_data, xgbt_mse_orig_data,
                  avg_stacked_ensemble_no_cgapd_mse, 
                  prop_var_stacked_ens_no_cgapd_mse)
              
              #print("Constructing the final results")
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "sample_size", "y", 
                                   "LR", "LASSO", "NNET", "Radial_SVM",
                                   "Linear_SVM", "Poly_SVM", "RF", "XGBTree",
                                   "Stacked_equal", "Stacked_prop","cgadp", "theta")
              
              
              # construct the dataframe of results
              # With CGADP
             
              estimator_reg_df = matrix( rep(0, length(data_frame_names) * 2), nrow = 2, ncol = length(data_frame_names), byrow = TRUE)
              estimator_reg_df[1,] = c(dat.seed,iteration, n[samp_size_index], 
                                                    response_index, cgapd_mse_vector, 1, theta_value)
              
              # Adding results for LASSO without CGAPD
              estimator_reg_df[2,] = c(dat.seed, iteration, 
                                             n[samp_size_index],
                                             response_index, no_cgapd_mse_vector, 0, theta_value)
              
              
              colnames(estimator_reg_df) = data_frame_names
              
              return(estimator_reg_df)
            }
  
  stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  # Add name of the estimators
  
  return(results)
}
# Training Dataset Iteratively Generated --------------
## CGADP --------------------
estimator_evaluation_stacked_ensemb_cgadp_iter = function(# the dataset
                                                    m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                    m = 1000, # Number of Monte Carlo simulations
                                                    n = c(500, 1000, 2000), # sample sizes to be used for training set
                                                    cores = 2, # Number of CPU threads/ logical processors to use
                                                    response_prefix = "y", # prefix for response variables
                                                    acquire_variable_names = get_variables, #Function for getting the variables
                                                    perturbation_func, # CGADP or GADP function
                                                    theta_value = 0.4, # privacy parameter value
                                                    model_generating_function, # function to generate the model
                                                    testing_data_seed = 30000, # Seed for the testing set
                                                    model_predictor_quant = 10, # Quantity of the predictors for the model
                                                    model_rho = 0.2, # correlation of predictors in the model
                                                    model_noise = 0.2, # Noise to be used in the model
                                                    testing_n = 500) # Size of the testing set
{
  
  # generate the testing data
  testing_model_data = model_generating_function(N = testing_n, rho = model_rho, 
                                                 max_pred = model_predictor_quant, 
                                                 dat.seed = testing_data_seed,
                                                 noise = model_noise)
  
  # Getting variable names:
  all_variable_names = colnames(testing_model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  
  # sdc_data_formula = sdc_form(target_variables = organized_vars$response, 
  #                             predictor_variables = organized_vars$predictors)
  
  # Calculate MSE function
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  sim_cluster = makeCluster(cores)
  registerDoSNOW(sim_cluster)
  
  results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
                    .packages = c( "Matrix",  "ranger", 
                                   "nnet", "e1071", "xgboost", "methods", "glmnet"),
                    .inorder = FALSE) %:%
    foreach( iteration = m_start:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",
            .inorder = FALSE) %dopar% {
              
              # setting the seed
              dat.seed = iteration + n[samp_size_index]
              set.seed(dat.seed)
              
              # Training dataset
              # Generated with the same parameters sxcluding the seed
              training_model_data = 
                model_generating_function(N = n[samp_size_index], 
                                          rho = model_rho, 
                                          max_pred = model_predictor_quant, 
                                          dat.seed = dat.seed,
                                          noise = model_noise)
              
              model_data = rbind(training_model_data, 
                                 testing_model_data)
              
              # Performing SDC GADP method
              privatized_dataframe = 
                perturbation_func(unprotected_data = model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              
              privatized_responses = privatized_dataframe[, target_variables_indices]
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,organized_vars$predictors]
              
              # Get the sample from privatized df
              drawing = 1:n[samp_size_index]
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              # Adding intercept term for linear regression
              x_dat_holdout = cbind(1,as.matrix(privatized_dataframe)[-drawing,organized_vars$predictors])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after cgapd, but preparing for case when both x and y are randomized
              x_dat = as.matrix(cbind(1,
                                      sample_private_data[,organized_vars$predictors]))
              
              
              # Get the predictors for the training set for unprotected data
              x_dat_unprotected = 
                as.matrix(cbind(1, sample_unprotected_data[,organized_vars$predictors]))
              
              # Get the predictors for the testing set for unprotected data
              x_dat_holdout_unprotected = 
                as.matrix(cbind(1, model_data[-drawing,organized_vars$predictors]))
              
              
              
              ### CGAPD Used----
              
              # Linear Regression on data after cgapd
              betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(current_response[drawing])
              
              
              linear_reg_pred = as.vector(x_dat_holdout %*% betas_linear)
              
              # Get the predictions for the training data
              
              training_lin_reg_pred = as.vector(x_dat %*% betas_linear)
              
              
              # Get the mse for the training data
              
              
              training_lin_reg_mse = calc_mse(as.matrix(current_response[drawing]),
                                              training_lin_reg_pred)
              
              # Get the mse for the test data
              
              
              linear_reg_mse = calc_mse(orig_data_response[-drawing],
                                        linear_reg_pred)
              
              # LASSO on data after cgapd ---------------------
  
              # Build the LASSO regression model
              # using cross validation
              lasso_regress_model = cv.glmnet( 
                x = as.matrix(sample_private_data[,organized_vars$predictors]),
                y = current_response[drawing])
              
              
              lasso_pred = predict(lasso_regress_model, 
                                   as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]),
                                   s = "lambda.min")
              
              # get predictions on the training data
              training_lasso_pred = predict(lasso_regress_model, 
                                            as.matrix(privatized_dataframe[drawing,organized_vars$predictors]),
                                            s = "lambda.min")
              
              # Get the MSE for the training data
              training_lasso_mse = 
                calc_mse(orig_data_response[drawing],
                         training_lasso_pred)
              
              # # MSE for the testing data
              
              lasso_mse = calc_mse(orig_data_response[-drawing],
                                   lasso_pred)
              
              # Neural Net Regression on data after cgapd
              cgapd_nnet = nnet(x = x_dat[,-1], y = current_response[drawing], 
                                size = 3,
                                maxit = 200, linout = TRUE)
              
              
              cgapd_nnet_reg_pred = predict(cgapd_nnet,x_dat_holdout[,-1])
              
              # training data predictions
              training_cgapd_nnet_reg_pred = predict(cgapd_nnet,x_dat[,-1])
              
              # training data MSE
              training_cgapd_nnet_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_nnet_reg_pred)
              
              
              # testing data MSE
              cgapd_nnet_reg_mse = calc_mse(orig_data_response[-drawing],
                                            cgapd_nnet_reg_pred)
              
              ## SVM Runs for CGADP -------
              
              # SVM Radial Regression on data after cgapd
              cgapd_radial_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                               gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                               degree = 3:4,
                                               kernel = c("radial"))
              
              cgapd_radial_svm = cgapd_radial_svm_tune$best.model
              
              cgapd_radial_svm_reg_pred = predict(cgapd_radial_svm,x_dat_holdout[,-1])
              
              # Get the training predictions of the radial svm
              training_cgapd_radial_svm_reg_pred =
                predict(cgapd_radial_svm, x_dat[,-1])
              
              # Get the training mse for radial svm
              training_cgapd_radial_svm_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_radial_svm_reg_pred)
              
              # Testing data MSE
              cgapd_radial_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                  cgapd_radial_svm_reg_pred)
              
              
              
              # SVM Linear Regression on data after cgapd
              cgapd_linear_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                               gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                               degree = 3:4,
                                               kernel = c("linear"))
              
              cgapd_linear_svm = cgapd_linear_svm_tune$best.model
              
              # Predictions for the training data for Linear SVM
              training_cgapd_linear_svm_reg_pred =
                predict(cgapd_linear_svm,x_dat_holdout[,-1])
              
              # Training MSE for Linear SVM
              training_cgapd_linear_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_linear_svm_reg_pred)
              
              # Predictions for the test dataset
              cgapd_linear_svm_reg_pred = predict(cgapd_linear_svm,x_dat_holdout[,-1])
              
              # Testing data MSE
              cgapd_linear_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                  cgapd_linear_svm_reg_pred)
              
              # SVM Polynomial Regression on data after gapd
              cgapd_poly_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                             gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                             degree = 3:4,
                                             kernel = c("polynomial"))
              
              cgapd_poly_svm = cgapd_poly_svm_tune$best.model
              
              # Training data polynomial SVM predictions
              training_cgapd_poly_svm_reg_pred = predict(cgapd_poly_svm,x_dat[,-1])
              
              # Training data MSE polynomial SVM
              training_cgapd_poly_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_poly_svm_reg_pred)
              
              # Predictions for testing dataset
              cgapd_poly_svm_reg_pred = predict(cgapd_poly_svm,x_dat_holdout[,-1])
              
              # MSE for testing dataset for polynomial kernel SVM
              cgapd_poly_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                cgapd_poly_svm_reg_pred)
              
              ## Random Forest Run ---------------------------
              
              # Random forest on data after cgapd
              num_x_vars = length(organized_vars$predictors)
              
              ranger_rf_regress_model = ranger( 
                x = as.matrix(sample_private_data[,organized_vars$predictors]),
                y = current_response[drawing], 
                num.trees = 500,
                num.threads = 1,
                mtry = ceiling(1 / 3 * num_x_vars))
              
              # Training dataset predicions for the random forest
              training_ranger_rf_pred = 
                predict(ranger_rf_regress_model, 
                        privatized_dataframe[drawing,organized_vars$predictors])$predictions
              
              # training dataset MSE for the radom forest
              training_ranger_rf_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_ranger_rf_pred)
              
              
              # Testing dataset prediction for random forest
              ranger_rf_pred = predict(ranger_rf_regress_model, 
                                       privatized_dataframe[-drawing,organized_vars$predictors])$predictions
              
              # Testing dataset MSE for random forest
              ranger_rf_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                       ranger_rf_pred)
              
              # XGBT on data after cgapd
              xgbt_regress_model = xgboost( 
                data = as.matrix(sample_private_data[,organized_vars$predictors]),
                label = current_response[drawing], 
                nrounds = 40,
                nthread = 1)
              
              # Training data predictions for the XGB tree
              training_xgbt_pred = 
                predict(xgbt_regress_model,
                        as.matrix(privatized_dataframe[drawing,organized_vars$predictors]))
              
              # training data MSE for the XGB tree
              training_xgbt_mse = 
                calc_mse(as.matrix(current_response[drawing]) ,
                         training_xgbt_pred)
              
              # Testing data predicitons for the XGB tree
              xgbt_pred = predict(xgbt_regress_model, 
                                  as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]))
              
              # Testing data MSE for the XGB tree
              xgbt_mse = calc_mse(orig_data_response[-drawing],
                                  xgbt_pred)
              
              ##
              ## Before CGAPD (On original data) -----------------------------------
              # Linear Regression on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              
              # Beta coefficients for the original data
              orig_data_betas_linear = solve(t(x_dat_unprotected) %*% 
                                               x_dat_unprotected) %*% t(x_dat_unprotected) %*% 
                as.matrix(sample_orig_data_response)
              
              # Predictions on the training data no GADP
              training_lin_reg_unprivat_pred = 
                as.vector(x_dat_unprotected %*% orig_data_betas_linear)
              
              
              # MSE for training data for linear regression, no GADP
              training_linear_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_lin_reg_unprivat_pred)
              
              # Testing data MSE for linear regression no ADP
              linear_reg_unprivatized_pred = as.vector(x_dat_holdout_unprotected %*% 
                                                         orig_data_betas_linear)
              
              linear_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_reg_unprivatized_pred)
              
              # LASSO on data before gapd (original data)
              
              
              lasso_regress_model_orig_data =
                cv.glmnet( x = as.matrix(model_data[drawing,organized_vars$predictors]), # Get rid of intercept term
                           y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # Training data predictions for LASSO
              
              training_lasso_pred_orig_data = 
                predict(lasso_regress_model_orig_data,
                        as.matrix(model_data[drawing,organized_vars$predictors]), 
                        s = "lambda.min")
              
              # Training data MSE for LASSO
              training_lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_lasso_pred_orig_data)
              
              # Predictions for the test dataset for the LASSO model
              lasso_pred_orig_data = predict(lasso_regress_model_orig_data,
                                             as.matrix(model_data[-drawing,organized_vars$predictors]), 
                                             s = "lambda.min")
              # MSE for the LASSO model for the test data
              lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         lasso_pred_orig_data)
              
              
              # Neural Net before GAPD 
              
              no_cgapd_nnet = nnet(x = x_dat_unprotected[,-1], 
                                   y = sample_orig_data_response, 
                                   size = 3,
                                   maxit = 200, linout = TRUE)
              
              # The predictions of the training set for neural network
              training_nnet_reg_unprivatized_pred = predict(no_cgapd_nnet, 
                                                            x_dat_unprotected[,-1])
              
              # The MSE for the training set for the neural network
              training_nnet_reg_unprivatized_mse = 
                calc_mse(orig_data_response[drawing],
                         training_nnet_reg_unprivatized_pred)
              
              # Predicitons on testing set for neural net
              nnet_reg_unprivatized_pred = predict(no_cgapd_nnet, 
                                                   x_dat_holdout_unprotected[,-1])
              
              # MSE for testing set for neural network
              nnet_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         nnet_reg_unprivatized_pred)
              
              # SVM Radial before GAPD
              no_cgapd_radial_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                  y = sample_orig_data_response,
                                                  gamma = seq(0.001,0.1,length.out = 2), 
                                                  cost = seq(1,3,length.out = 2),
                                                  degree = 3:4,
                                                  kernel = c("radial"))
              
              # Get the best model for Radial SVM
              no_cgapd_radial_svm = no_cgapd_radial_svm_tune$best.model
              
              # Predictions on the training set for radial SVM
              training_radial_svm_reg_unprivatized_pred = 
                predict(no_cgapd_radial_svm, x_dat_unprotected[,-1])
              
              # The MSE for the training set for radial SVM
              training_radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_radial_svm_reg_unprivatized_pred)
              
              # Get the predictions of Radial SVM for the testing set
              radial_svm_reg_unprivatized_pred = 
                predict(no_cgapd_radial_svm, x_dat_holdout_unprotected[,-1])
              
              # Get the MSE of the Radial SVM for the testing set
              radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         radial_svm_reg_unprivatized_pred)
              
              # Linear SVM before GAPD
              no_cgapd_linear_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                  y = sample_orig_data_response,
                                                  gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                                  degree = 3:4,
                                                  kernel = c("linear"))
              # The Linear SVM best tuned model
              no_cgapd_linear_svm = no_cgapd_linear_svm_tune$best.model
              
              # The predictions on the training set, Linear SVM
              training_linear_svm_reg_unprivatized_pred = 
                predict(no_cgapd_linear_svm, x_dat_unprotected[,-1])
              
              # the MSE on the training set, Linear SVM
              training_linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_linear_svm_reg_unprivatized_pred)
              
              # The predictions for the testing set Linear SVM
              linear_svm_reg_unprivatized_pred = 
                predict(no_cgapd_linear_svm, x_dat_holdout_unprotected[,-1])
              
              # The MSE for the testing set Linear SVM
              linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_svm_reg_unprivatized_pred)
              
              
              # Polynomial SVM before GAPD
              no_cgapd_poly_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                y = sample_orig_data_response,
                                                gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                                degree = 3:4,
                                                kernel = c("polynomial"))
              # Get the best tuned polynomial SVM
              no_cgapd_poly_svm = no_cgapd_poly_svm_tune$best.model
              
              # The training set predicitons for the polynomial SVM
              training_poly_svm_reg_unprivatized_pred = 
                predict(no_cgapd_poly_svm, x_dat_unprotected[,-1])
              
              # the training set MSE for the polynomial SVM
              training_poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_poly_svm_reg_unprivatized_pred)
              
              # The testing set predicitons using the polynomial SVM
              poly_svm_reg_unprivatized_pred = 
                predict(no_cgapd_poly_svm, x_dat_holdout_unprotected[,-1])
              
              # The testing set MSE for the polynomial SVM
              poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         poly_svm_reg_unprivatized_pred)
              
              # Random forest on data before cgapd
              ranger_rf_model_orig_data = ranger( 
                x = model_data[drawing,organized_vars$predictors],
                y = orig_data_response[drawing], 
                num.trees = 500,
                num.threads = 1,
                mtry = ceiling(1 / 3 * num_x_vars))
              
              # The training set predictions from random forest
              training_ranger_rf_pred_orig_data = 
                predict(ranger_rf_model_orig_data,
                        model_data[drawing,organized_vars$predictors])$predictions
              
              # The training set MSE for the random forest
              training_ranger_rf_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_ranger_rf_pred_orig_data)
              
              # Testing set predictions for random forest
              ranger_rf_pred_orig_data = predict(ranger_rf_model_orig_data,
                                                 model_data[-drawing,organized_vars$predictors])$predictions
              # Testing set MSE for random forest
              ranger_rf_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         ranger_rf_pred_orig_data)
              
              # XGBT on data before cgapd
              xgbt_regress_model_orig_data =
                xgboost(data = as.matrix(sample_unprotected_data[,organized_vars$predictors]),
                        label = orig_data_response[drawing],
                        nrounds = 40,
                        nthread = 1)
              
              # The training set predictions for XGB tree
              training_xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(sample_unprotected_data[,organized_vars$predictors]))
              # Training set MSE for XGB tree for 
              training_xgbt_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_xgbt_pred_orig_data)
              
              # The testing set XGB tree predictions
              xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[-drawing,organized_vars$predictors]))
              
              # The testing set MSE for XGB tree
              xgbt_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         xgbt_pred_orig_data)
              
              
              #print("Constructing the stacked ensemble")
              ###
              ### Stacked Ensemble Equal Weighting ----
              # CGAPD
              avg_stacked_ensemble_predictions_cgapd = (linear_reg_pred  + 
                                                          lasso_pred  +  cgapd_nnet_reg_pred + 
                                                          cgapd_radial_svm_reg_pred  + 
                                                          cgapd_linear_svm_reg_pred  + 
                                                          cgapd_poly_svm_reg_pred  +
                                                          ranger_rf_pred  + xgbt_pred) / 8
              
              avg_stacked_ensemble_cgapd_mse =
                calc_mse(orig_data_response[-drawing],
                         avg_stacked_ensemble_predictions_cgapd)
              # No GAPD
              avg_stacked_ensemble_pred_no_cgapd = 
                (linear_reg_unprivatized_pred  + 
                   lasso_pred_orig_data  + 
                   nnet_reg_unprivatized_pred  + 
                   radial_svm_reg_unprivatized_pred  + 
                   linear_svm_reg_unprivatized_pred  + 
                   poly_svm_reg_unprivatized_pred  +
                   ranger_rf_pred_orig_data  + xgbt_pred_orig_data) / 8
              
              avg_stacked_ensemble_no_cgapd_mse =
                calc_mse(orig_data_response[-drawing],
                         avg_stacked_ensemble_pred_no_cgapd)
              
              
              ### Stacked Ensemble Equal Weighting by Proportion of variance Explained ----------------
              # under CGADP
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_cgapd = 
                c(training_lin_reg_mse, training_lasso_mse, 
                  training_cgapd_nnet_reg_mse, training_cgapd_radial_svm_reg_mse,
                  training_cgapd_linear_svm_reg_mse, 
                  training_cgapd_poly_svm_reg_mse, training_ranger_rf_mse,
                  training_xgbt_mse)
              
              # The models with the lowest MSE get the highest weight
              prop_var_exp_cgapd_weighting = 
                (training_set_mses_cgapd / sum(training_set_mses_cgapd))^(-1)
              
              prop_var_exp_cgapd_weighting = prop_var_exp_cgapd_weighting / 
                sum(prop_var_exp_cgapd_weighting)
              
              
              
              # Get the predictions for the stacked ensemble for the prop
              # of variance explained
              prop_var_stacked_ens_cgapd_pred = 
                linear_reg_pred * prop_var_exp_cgapd_weighting[1] +
                lasso_pred * prop_var_exp_cgapd_weighting[2] + 
                cgapd_nnet_reg_pred * prop_var_exp_cgapd_weighting[3] + 
                cgapd_radial_svm_reg_pred * prop_var_exp_cgapd_weighting[4] + 
                cgapd_linear_svm_reg_pred * prop_var_exp_cgapd_weighting[5] + 
                cgapd_poly_svm_reg_pred * prop_var_exp_cgapd_weighting[6] +
                ranger_rf_pred * prop_var_exp_cgapd_weighting[7] + 
                xgbt_pred * prop_var_exp_cgapd_weighting[8]
              
              
              
              # Get the MSE
              prop_var_stacked_ens_cgapd_mse = 
                calc_mse(orig_data_response[-drawing],
                         prop_var_stacked_ens_cgapd_pred)
              
              
              # Get the Stacked Ensemble (Prop. of Var. explained) with No GADP 
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_no_cgapd = 
                c(training_linear_reg_mse_orig_data, 
                  training_lasso_mse_orig_data, 
                  training_nnet_reg_unprivatized_mse, 
                  training_radial_svm_reg_mse_orig_data,
                  training_linear_svm_reg_mse_orig_data, 
                  training_poly_svm_reg_mse_orig_data, 
                  training_ranger_rf_mse_orig_data,
                  training_xgbt_mse_orig_data)
              
              # weighting for the SE
              prop_var_exp_no_cgapd_weighting = 
                (training_set_mses_no_cgapd / sum(training_set_mses_no_cgapd))^(-1)
              
              prop_var_exp_no_cgapd_weighting = prop_var_exp_no_cgapd_weighting /
                sum(prop_var_exp_no_cgapd_weighting)
              
              
              # get the predictions
              prop_var_stacked_ens_no_cgapd_pred = 
                linear_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[1] +
                lasso_pred_orig_data * prop_var_exp_no_cgapd_weighting[2] + 
                nnet_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[3] + 
                radial_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[4] + 
                linear_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[5] + 
                poly_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[6] +
                ranger_rf_pred_orig_data * prop_var_exp_no_cgapd_weighting[7] + 
                xgbt_pred_orig_data * prop_var_exp_no_cgapd_weighting[8]
              
              # Get the MSE
              prop_var_stacked_ens_no_cgapd_mse = 
                calc_mse(orig_data_response[-drawing],
                         prop_var_stacked_ens_no_cgapd_pred)
              
              
              # Compile MSEs into a vector
              cgapd_mse_vector = 
                c(linear_reg_mse, lasso_mse, cgapd_nnet_reg_mse,
                  cgapd_radial_svm_reg_mse, cgapd_linear_svm_reg_mse,
                  cgapd_poly_svm_reg_mse, ranger_rf_mse,
                  xgbt_mse, avg_stacked_ensemble_cgapd_mse, 
                  prop_var_stacked_ens_cgapd_mse)
              
              no_cgapd_mse_vector =
                c(linear_reg_mse_orig_data, lasso_mse_orig_data,
                  nnet_reg_mse_orig_data, radial_svm_reg_mse_orig_data,
                  linear_svm_reg_mse_orig_data, poly_svm_reg_mse_orig_data,
                  ranger_rf_mse_orig_data, xgbt_mse_orig_data,
                  avg_stacked_ensemble_no_cgapd_mse, 
                  prop_var_stacked_ens_no_cgapd_mse)
              
              #print("Constructing the final results")
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "sample_size", "y", 
                                   "LR", "LASSO", "NNET", "Radial_SVM",
                                   "Linear_SVM", "Poly_SVM", "RF", "XGBTree",
                                   "Stacked_equal", "Stacked_prop","cgadp", "theta")
              
              
              # construct the dataframe of results
              # With CGADP
              
              estimator_reg_df = matrix( rep(0, length(data_frame_names) * 2), nrow = 2, ncol = length(data_frame_names), byrow = TRUE)
              estimator_reg_df[1,] = c(dat.seed,iteration, n[samp_size_index], 
                                       response_index, cgapd_mse_vector, 1, theta_value)
              
              # Adding results for LASSO without CGAPD
              estimator_reg_df[2,] = c(dat.seed, iteration, 
                                       n[samp_size_index],
                                       response_index, no_cgapd_mse_vector, 0, theta_value)
              
              
              colnames(estimator_reg_df) = data_frame_names
              
              return(estimator_reg_df)
            }
  #### Stopping Cluster -------------
  stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  # Add name of the estimators
  
  return(results)
}

#CGADP  Version 2 ------

estimator_evaluation_stacked_ensemb_cgadp_iter_ver_2 = function(# the dataset
        m_start = 1, # Primarily used for debugging determines which iteration to start from
        m = 1000, # Number of Monte Carlo simulations
        n = c(500, 1000, 2000), # sample sizes to be used for training set
        cores = 2, # Number of CPU threads/ logical processors to use
        response_prefix = "y", # prefix for response variables
        acquire_variable_names = get_variables, #Function for getting the variables
        perturbation_func, # CGADP or GADP function
        theta_value = 0.4, # privacy parameter value
        model_generating_function, # function to generate the model
        testing_data_seed = 30000, # Seed for the testing set
        model_predictor_quant = 10, # Quantity of the predictors for the model
        model_rho = 0.2, # correlation of predictors in the model
        model_noise = 0.2, # Noise to be used in the model
        testing_n = 500) # Size of the testing set
{
  
  # generate the testing data
  testing_model_data = model_generating_function(N = testing_n, rho = model_rho, 
                                                 max_pred = model_predictor_quant, 
                                                 dat.seed = testing_data_seed,
                                                 noise = model_noise)
  
  # Getting variable names:
  all_variable_names = colnames(testing_model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  
  # sdc_data_formula = sdc_form(target_variables = organized_vars$response, 
  #                             predictor_variables = organized_vars$predictors)
  
  # Calculate MSE function
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  sim_cluster = makeCluster(cores)
  registerDoSNOW(sim_cluster)
  
  results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
                    .packages = c( "Matrix",  "ranger", 
                                   "nnet", "e1071", "xgboost", "methods", "glmnet"),
                    .inorder = FALSE) %:%
    foreach( iteration = m_start:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",
            .inorder = FALSE) %dopar% {
              
              # setting the seed
              dat.seed = iteration + n[samp_size_index]
              set.seed(dat.seed)
              
              # Training dataset
              # Generated with the same parameters sxcluding the seed
              training_model_data = 
                model_generating_function(N = n[samp_size_index], 
                                          rho = model_rho, 
                                          max_pred = model_predictor_quant, 
                                          dat.seed = dat.seed,
                                          noise = model_noise)
              
              model_data = rbind(training_model_data, 
                                 testing_model_data)
              
              # Performing SDC GADP method
              privatized_dataframe = 
                perturbation_func(unprotected_data = model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              
              privatized_responses = privatized_dataframe[, target_variables_indices]
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,organized_vars$predictors]
              
              # Get the sample from privatized df
              drawing = 1:n[samp_size_index]
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              # Adding intercept term for linear regression
              x_dat_holdout = cbind(1,as.matrix(privatized_dataframe)[-drawing,organized_vars$predictors])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after cgapd, but preparing for case when both x and y are randomized
              x_dat = as.matrix(cbind(1,
                                      sample_private_data[,organized_vars$predictors]))
              
              
              # Get the predictors for the training set for unprotected data
              x_dat_unprotected = 
                as.matrix(cbind(1, sample_unprotected_data[,organized_vars$predictors]))
              
              # Get the predictors for the testing set for unprotected data
              x_dat_holdout_unprotected = 
                as.matrix(cbind(1, model_data[-drawing,organized_vars$predictors]))
              
              
              
              ### CGAPD Used----
              
              # Linear Regression on data after cgapd
              betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(current_response[drawing])
              
              
              linear_reg_pred = as.vector(x_dat_holdout %*% betas_linear)
              
              # Get the predictions for the training data
              
              training_lin_reg_pred = as.vector(x_dat %*% betas_linear)
              
              
              # Get the mse for the training data
              
              
              training_lin_reg_mse = calc_mse(as.matrix(current_response[drawing]),
                                              training_lin_reg_pred)
              
              # Get the mse for the test data
              
              
              linear_reg_mse = calc_mse(orig_data_response[-drawing],
                                        linear_reg_pred)
              
              # LASSO on data after cgapd ---------------------
              
              # Build the LASSO regression model
              # using cross validation
              lasso_regress_model = cv.glmnet( 
                x = as.matrix(sample_private_data[,organized_vars$predictors]),
                y = current_response[drawing])
              
              
              lasso_pred = predict(lasso_regress_model, 
                                   as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]),
                                   s = "lambda.min")
              
              # get predictions on the training data
              training_lasso_pred = predict(lasso_regress_model, 
                                            as.matrix(privatized_dataframe[drawing,organized_vars$predictors]),
                                            s = "lambda.min")
              
              # Get the MSE for the training data
              training_lasso_mse = 
                calc_mse(orig_data_response[drawing],
                         training_lasso_pred)
              
              # # MSE for the testing data
              
              lasso_mse = calc_mse(orig_data_response[-drawing],
                                   lasso_pred)
              
              # Neural Net Regression on data after cgapd
              cgapd_nnet = nnet(x = x_dat[,-1], y = current_response[drawing], 
                                size = 3,
                                maxit = 200, linout = TRUE)
              
              
              cgapd_nnet_reg_pred = predict(cgapd_nnet,x_dat_holdout[,-1])
              
              # training data predictions
              training_cgapd_nnet_reg_pred = predict(cgapd_nnet,x_dat[,-1])
              
              # training data MSE
              training_cgapd_nnet_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_nnet_reg_pred)
              
              
              # testing data MSE
              cgapd_nnet_reg_mse = calc_mse(orig_data_response[-drawing],
                                            cgapd_nnet_reg_pred)
              
              ## SVM Runs for CGADP -------
              
              # SVM Radial Regression on data after cgapd
              cgapd_radial_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                               gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                               degree = 3:4,
                                               kernel = c("radial"))
              
              cgapd_radial_svm = cgapd_radial_svm_tune$best.model
              
              cgapd_radial_svm_reg_pred = predict(cgapd_radial_svm,x_dat_holdout[,-1])
              
              # Get the training predictions of the radial svm
              training_cgapd_radial_svm_reg_pred =
                predict(cgapd_radial_svm, x_dat[,-1])
              
              # Get the training mse for radial svm
              training_cgapd_radial_svm_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_radial_svm_reg_pred)
              
              # Testing data MSE
              cgapd_radial_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                  cgapd_radial_svm_reg_pred)
              
              
              
              # SVM Linear Regression on data after cgapd
              cgapd_linear_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                               gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                               degree = 3:4,
                                               kernel = c("linear"))
              
              cgapd_linear_svm = cgapd_linear_svm_tune$best.model
              
              # Predictions for the training data for Linear SVM
              training_cgapd_linear_svm_reg_pred =
                predict(cgapd_linear_svm,x_dat_holdout[,-1])
              
              # Training MSE for Linear SVM
              training_cgapd_linear_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_linear_svm_reg_pred)
              
              # Predictions for the test dataset
              cgapd_linear_svm_reg_pred = predict(cgapd_linear_svm,x_dat_holdout[,-1])
              
              # Testing data MSE
              cgapd_linear_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                  cgapd_linear_svm_reg_pred)
              
              # SVM Polynomial Regression on data after gapd
              cgapd_poly_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                             gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                             degree = 3:4,
                                             kernel = c("polynomial"))
              
              cgapd_poly_svm = cgapd_poly_svm_tune$best.model
              
              # Training data polynomial SVM predictions
              training_cgapd_poly_svm_reg_pred = predict(cgapd_poly_svm,x_dat[,-1])
              
              # Training data MSE polynomial SVM
              training_cgapd_poly_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_poly_svm_reg_pred)
              
              # Predictions for testing dataset
              cgapd_poly_svm_reg_pred = predict(cgapd_poly_svm,x_dat_holdout[,-1])
              
              # MSE for testing dataset for polynomial kernel SVM
              cgapd_poly_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                cgapd_poly_svm_reg_pred)
              
              ## Random Forest Run ---------------------------
              
              # Random forest on data after cgapd
              num_x_vars = length(organized_vars$predictors)
              
              ranger_rf_regress_model = ranger( 
                x = as.matrix(sample_private_data[,organized_vars$predictors]),
                y = current_response[drawing], 
                num.trees = 500,
                num.threads = 1,
                mtry = ceiling(1 / 3 * num_x_vars))
              
              # Training dataset predicions for the random forest
              training_ranger_rf_pred = 
                predict(ranger_rf_regress_model, 
                        privatized_dataframe[drawing,organized_vars$predictors])$predictions
              
              # training dataset MSE for the radom forest
              training_ranger_rf_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_ranger_rf_pred)
              
              
              # Testing dataset prediction for random forest
              ranger_rf_pred = predict(ranger_rf_regress_model, 
                                       privatized_dataframe[-drawing,organized_vars$predictors])$predictions
              
              # Testing dataset MSE for random forest
              ranger_rf_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                       ranger_rf_pred)
              
              # XGBT on data after cgapd
              xgbt_regress_model = xgboost( 
                data = as.matrix(sample_private_data[,organized_vars$predictors]),
                label = current_response[drawing], 
                nrounds = 40,
                nthread = 1)
              
              # Training data predictions for the XGB tree
              training_xgbt_pred = 
                predict(xgbt_regress_model,
                        as.matrix(privatized_dataframe[drawing,organized_vars$predictors]))
              
              # training data MSE for the XGB tree
              training_xgbt_mse = 
                calc_mse(as.matrix(current_response[drawing]) ,
                         training_xgbt_pred)
              
              # Testing data predicitons for the XGB tree
              xgbt_pred = predict(xgbt_regress_model, 
                                  as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]))
              
              # Testing data MSE for the XGB tree
              xgbt_mse = calc_mse(orig_data_response[-drawing],
                                  xgbt_pred)
              
              ##
              ## Before CGAPD (On original data) -----------------------------------
              # Linear Regression on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              
              # Beta coefficients for the original data
              orig_data_betas_linear = solve(t(x_dat_unprotected) %*% 
                                               x_dat_unprotected) %*% t(x_dat_unprotected) %*% 
                as.matrix(sample_orig_data_response)
              
              # Predictions on the training data no GADP
              training_lin_reg_unprivat_pred = 
                as.vector(x_dat_unprotected %*% orig_data_betas_linear)
              
              
              # MSE for training data for linear regression, no GADP
              training_linear_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_lin_reg_unprivat_pred)
              
              # Testing data MSE for linear regression no ADP
              linear_reg_unprivatized_pred = as.vector(x_dat_holdout_unprotected %*% 
                                                         orig_data_betas_linear)
              
              linear_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_reg_unprivatized_pred)
              
              # LASSO on data before gapd (original data)
              
              
              lasso_regress_model_orig_data =
                cv.glmnet( x = as.matrix(model_data[drawing,organized_vars$predictors]), # Get rid of intercept term
                           y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # Training data predictions for LASSO
              
              training_lasso_pred_orig_data = 
                predict(lasso_regress_model_orig_data,
                        as.matrix(model_data[drawing,organized_vars$predictors]), 
                        s = "lambda.min")
              
              # Training data MSE for LASSO
              training_lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_lasso_pred_orig_data)
              
              # Predictions for the test dataset for the LASSO model
              lasso_pred_orig_data = predict(lasso_regress_model_orig_data,
                                             as.matrix(model_data[-drawing,organized_vars$predictors]), 
                                             s = "lambda.min")
              # MSE for the LASSO model for the test data
              lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         lasso_pred_orig_data)
              
              
              # Neural Net before GAPD 
              
              no_cgapd_nnet = nnet(x = x_dat_unprotected[,-1], 
                                   y = sample_orig_data_response, 
                                   size = 3,
                                   maxit = 200, linout = TRUE)
              
              # The predictions of the training set for neural network
              training_nnet_reg_unprivatized_pred = predict(no_cgapd_nnet, 
                                                            x_dat_unprotected[,-1])
              
              # The MSE for the training set for the neural network
              training_nnet_reg_unprivatized_mse = 
                calc_mse(orig_data_response[drawing],
                         training_nnet_reg_unprivatized_pred)
              
              # Predicitons on testing set for neural net
              nnet_reg_unprivatized_pred = predict(no_cgapd_nnet, 
                                                   x_dat_holdout_unprotected[,-1])
              
              # MSE for testing set for neural network
              nnet_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         nnet_reg_unprivatized_pred)
              
              # SVM Radial before GAPD
              no_cgapd_radial_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                  y = sample_orig_data_response,
                                                  gamma = seq(0.001,0.1,length.out = 2), 
                                                  cost = seq(1,3,length.out = 2),
                                                  degree = 3:4,
                                                  kernel = c("radial"))
              
              # Get the best model for Radial SVM
              no_cgapd_radial_svm = no_cgapd_radial_svm_tune$best.model
              
              # Predictions on the training set for radial SVM
              training_radial_svm_reg_unprivatized_pred = 
                predict(no_cgapd_radial_svm, x_dat_unprotected[,-1])
              
              # The MSE for the training set for radial SVM
              training_radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_radial_svm_reg_unprivatized_pred)
              
              # Get the predictions of Radial SVM for the testing set
              radial_svm_reg_unprivatized_pred = 
                predict(no_cgapd_radial_svm, x_dat_holdout_unprotected[,-1])
              
              # Get the MSE of the Radial SVM for the testing set
              radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         radial_svm_reg_unprivatized_pred)
              
              # Linear SVM before GAPD
              no_cgapd_linear_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                  y = sample_orig_data_response,
                                                  gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                                  degree = 3:4,
                                                  kernel = c("linear"))
              # The Linear SVM best tuned model
              no_cgapd_linear_svm = no_cgapd_linear_svm_tune$best.model
              
              # The predictions on the training set, Linear SVM
              training_linear_svm_reg_unprivatized_pred = 
                predict(no_cgapd_linear_svm, x_dat_unprotected[,-1])
              
              # the MSE on the training set, Linear SVM
              training_linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_linear_svm_reg_unprivatized_pred)
              
              # The predictions for the testing set Linear SVM
              linear_svm_reg_unprivatized_pred = 
                predict(no_cgapd_linear_svm, x_dat_holdout_unprotected[,-1])
              
              # The MSE for the testing set Linear SVM
              linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_svm_reg_unprivatized_pred)
              
              
              # Polynomial SVM before GAPD
              no_cgapd_poly_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                y = sample_orig_data_response,
                                                gamma = seq(0.001,0.1,length.out = 2), cost = seq(1,3,length.out = 2),
                                                degree = 3:4,
                                                kernel = c("polynomial"))
              # Get the best tuned polynomial SVM
              no_cgapd_poly_svm = no_cgapd_poly_svm_tune$best.model
              
              # The training set predicitons for the polynomial SVM
              training_poly_svm_reg_unprivatized_pred = 
                predict(no_cgapd_poly_svm, x_dat_unprotected[,-1])
              
              # the training set MSE for the polynomial SVM
              training_poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_poly_svm_reg_unprivatized_pred)
              
              # The testing set predicitons using the polynomial SVM
              poly_svm_reg_unprivatized_pred = 
                predict(no_cgapd_poly_svm, x_dat_holdout_unprotected[,-1])
              
              # The testing set MSE for the polynomial SVM
              poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         poly_svm_reg_unprivatized_pred)
              
              # Random forest on data before cgapd
              ranger_rf_model_orig_data = ranger( 
                x = model_data[drawing,organized_vars$predictors],
                y = orig_data_response[drawing], 
                num.trees = 500,
                num.threads = 1,
                mtry = ceiling(1 / 3 * num_x_vars))
              
              # The training set predictions from random forest
              training_ranger_rf_pred_orig_data = 
                predict(ranger_rf_model_orig_data,
                        model_data[drawing,organized_vars$predictors])$predictions
              
              # The training set MSE for the random forest
              training_ranger_rf_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_ranger_rf_pred_orig_data)
              
              # Testing set predictions for random forest
              ranger_rf_pred_orig_data = predict(ranger_rf_model_orig_data,
                                                 model_data[-drawing,organized_vars$predictors])$predictions
              # Testing set MSE for random forest
              ranger_rf_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         ranger_rf_pred_orig_data)
              
              # XGBT on data before cgapd
              xgbt_regress_model_orig_data =
                xgboost(data = as.matrix(sample_unprotected_data[,organized_vars$predictors]),
                        label = orig_data_response[drawing],
                        nrounds = 40,
                        nthread = 1)
              
              # The training set predictions for XGB tree
              training_xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(sample_unprotected_data[,organized_vars$predictors]))
              # Training set MSE for XGB tree for 
              training_xgbt_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_xgbt_pred_orig_data)
              
              # The testing set XGB tree predictions
              xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[-drawing,organized_vars$predictors]))
              
              # The testing set MSE for XGB tree
              xgbt_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         xgbt_pred_orig_data)
              
              
              #print("Constructing the stacked ensemble")
              ###
              ### Stacked Ensemble Equal Weighting ----
              # CGAPD
              avg_stacked_ensemble_predictions_cgapd = (linear_reg_pred  + 
                                                          lasso_pred  +  cgapd_nnet_reg_pred + 
                                                          cgapd_radial_svm_reg_pred  + 
                                                          cgapd_linear_svm_reg_pred  + 
                                                          cgapd_poly_svm_reg_pred  +
                                                          ranger_rf_pred  + xgbt_pred) / 8
              
              avg_stacked_ensemble_cgapd_mse =
                calc_mse(orig_data_response[-drawing],
                         avg_stacked_ensemble_predictions_cgapd)
              # No GAPD
              avg_stacked_ensemble_pred_no_cgapd = 
                (linear_reg_unprivatized_pred  + 
                   lasso_pred_orig_data  + 
                   nnet_reg_unprivatized_pred  + 
                   radial_svm_reg_unprivatized_pred  + 
                   linear_svm_reg_unprivatized_pred  + 
                   poly_svm_reg_unprivatized_pred  +
                   ranger_rf_pred_orig_data  + xgbt_pred_orig_data) / 8
              
              avg_stacked_ensemble_no_cgapd_mse =
                calc_mse(orig_data_response[-drawing],
                         avg_stacked_ensemble_pred_no_cgapd)
              
              
              ### Stacked Ensemble Equal Weighting by Proportion of variance Explained ----------------
              # under CGADP
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_cgapd = 
                c(training_lin_reg_mse, training_lasso_mse, 
                  training_cgapd_nnet_reg_mse, training_cgapd_radial_svm_reg_mse,
                  training_cgapd_linear_svm_reg_mse, 
                  training_cgapd_poly_svm_reg_mse, training_ranger_rf_mse,
                  training_xgbt_mse)
              
              # The models with the lowest MSE get the highest weight
              prop_var_exp_cgapd_weighting = 
                (training_set_mses_cgapd / sum(training_set_mses_cgapd))^(-1)
              
              prop_var_exp_cgapd_weighting = prop_var_exp_cgapd_weighting / 
                sum(prop_var_exp_cgapd_weighting)
              
              
              
              # Get the predictions for the stacked ensemble for the prop
              # of variance explained
              prop_var_stacked_ens_cgapd_pred = 
                linear_reg_pred * prop_var_exp_cgapd_weighting[1] +
                lasso_pred * prop_var_exp_cgapd_weighting[2] + 
                cgapd_nnet_reg_pred * prop_var_exp_cgapd_weighting[3] + 
                cgapd_radial_svm_reg_pred * prop_var_exp_cgapd_weighting[4] + 
                cgapd_linear_svm_reg_pred * prop_var_exp_cgapd_weighting[5] + 
                cgapd_poly_svm_reg_pred * prop_var_exp_cgapd_weighting[6] +
                ranger_rf_pred * prop_var_exp_cgapd_weighting[7] + 
                xgbt_pred * prop_var_exp_cgapd_weighting[8]
              
              
              
              # Get the MSE
              prop_var_stacked_ens_cgapd_mse = 
                calc_mse(orig_data_response[-drawing],
                         prop_var_stacked_ens_cgapd_pred)
              
              
              # Get the Stacked Ensemble (Prop. of Var. explained) with No GADP 
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_no_cgapd = 
                c(training_linear_reg_mse_orig_data, 
                  training_lasso_mse_orig_data, 
                  training_nnet_reg_unprivatized_mse, 
                  training_radial_svm_reg_mse_orig_data,
                  training_linear_svm_reg_mse_orig_data, 
                  training_poly_svm_reg_mse_orig_data, 
                  training_ranger_rf_mse_orig_data,
                  training_xgbt_mse_orig_data)
              
              # weighting for the SE
              prop_var_exp_no_cgapd_weighting = 
                (training_set_mses_no_cgapd / sum(training_set_mses_no_cgapd))^(-1)
              
              prop_var_exp_no_cgapd_weighting = prop_var_exp_no_cgapd_weighting /
                sum(prop_var_exp_no_cgapd_weighting)
              
              
              # get the predictions
              prop_var_stacked_ens_no_cgapd_pred = 
                linear_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[1] +
                lasso_pred_orig_data * prop_var_exp_no_cgapd_weighting[2] + 
                nnet_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[3] + 
                radial_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[4] + 
                linear_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[5] + 
                poly_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[6] +
                ranger_rf_pred_orig_data * prop_var_exp_no_cgapd_weighting[7] + 
                xgbt_pred_orig_data * prop_var_exp_no_cgapd_weighting[8]
              
              # Get the MSE
              prop_var_stacked_ens_no_cgapd_mse = 
                calc_mse(orig_data_response[-drawing],
                         prop_var_stacked_ens_no_cgapd_pred)
              
              
              # Compile MSEs into a vector
              cgapd_mse_vector = 
                c(linear_reg_mse, lasso_mse, cgapd_nnet_reg_mse,
                  cgapd_radial_svm_reg_mse, cgapd_linear_svm_reg_mse,
                  cgapd_poly_svm_reg_mse, ranger_rf_mse,
                  xgbt_mse, avg_stacked_ensemble_cgapd_mse, 
                  prop_var_stacked_ens_cgapd_mse)
              
              no_cgapd_mse_vector =
                c(linear_reg_mse_orig_data, lasso_mse_orig_data,
                  nnet_reg_mse_orig_data, radial_svm_reg_mse_orig_data,
                  linear_svm_reg_mse_orig_data, poly_svm_reg_mse_orig_data,
                  ranger_rf_mse_orig_data, xgbt_mse_orig_data,
                  avg_stacked_ensemble_no_cgapd_mse, 
                  prop_var_stacked_ens_no_cgapd_mse)
              
              #print("Constructing the final results")
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "training_size","testing_size", "y", 
                                   "LR", "LASSO", "NNET", "Radial_SVM",
                                   "Linear_SVM", "Poly_SVM", "RF", "XGBTree",
                                   "Stacked_equal", "Stacked_prop","cgadp", "theta")
              
              
              # construct the dataframe of results
              # With CGADP
              
              estimator_reg_df = matrix( rep(0, length(data_frame_names) * 2), nrow = 2, ncol = length(data_frame_names), byrow = TRUE)
              estimator_reg_df[1,] = c(dat.seed,iteration, n[samp_size_index], testing_n, 
                                       response_index, cgapd_mse_vector, 1, theta_value)
              
              # Adding results for LASSO without CGAPD
              estimator_reg_df[2,] = c(dat.seed, iteration, 
                                       n[samp_size_index], testing_n,
                                       response_index, no_cgapd_mse_vector, 0, theta_value)
              
              
              colnames(estimator_reg_df) = data_frame_names
              
              return(estimator_reg_df)
            }
  #### Stopping Cluster -------------
  stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  # Add name of the estimators
  
  return(results)
}

# More Hyperparameter Tuning --------------------

# Code for Performing Machine Learning Evaluations With GADP
## GADP ------------------------

estimator_eval_stacked_ensemb_iterative_cv = function( # the dataset
  m_start = 1, # Primarily used for debugging determines which iteration to start from
  m = 1000, # Number of Monte Carlo simulations
  samp_n = c(500, 1000, 2000), # sample sizes to be used for training set
  cores = 2, # Number of CPU threads/ logical processors to use
  response_prefix = "y", # prefix for response variables
  acquire_variable_names = get_variables, #Function for getting the variables
  perturbation_func = gadp_func, # CGADP or GADP function
  theta_value = 0.4, # privacy parameter value
  model_generating_function, # function to generate the model
  testing_data_seed = 30000, # Seed for the testing set
  model_predictor_quant = 10, # Quantity of the predictors for the model
  model_rho = 0.2, # correlation of predictors in the model
  model_noise = 0.2, # Noise to be used in the model
  testing_n = 500) # Size of the testing set
{
  # Cross-Validation Funcitons -----------
  # XGBoosted Trees
  xgbt_cv = function(x_pred, y_resp, 
                     parameter_grid = list("nrounds" = c(10,20, 40),
                                           "max_depth" = c(2,4),
                                           "eta" = c(0.2, 0.4, 0.6),
                                           "gamma" = c(0, 0.5, 1),
                                           "colsample_bytree" = c(0.2, 0.5, 1),
                                           "min_child_weight" = c(0.5, 1),
                                           "subsample" = c(0.4, 0.5, 0.6)),
                     cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = expand.grid(parameter_grid)
    
    # Construct the XGBoosted trees
    xgbt_control= caret::trainControl(method = "cv", number = cv_folds)
    # Train XGBoosted trees
    xgbt_obj = caret::train(x = x_pred, y = as.vector(y_resp),
                            method = "xgbTree", tuneGrid = param_grid,
                            trControl = xgbt_control,
                            nthread = 1,
                            verbose=0,
                            print_every_n =0)
    
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_xgbt_obj = xgbt_obj$finalModel
    return(final_xgbt_obj)
  }
  
  # Random forest
  ranger_cv = function(x_pred, y_resp, num_trees = 500,
                       frac_pred = c(1/3, 1/4), 
                       min_node_size = c(5,10),
                       cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = as.matrix(expand_grid(num_trees, frac_pred, min_node_size))
    # Set up folds to be used in cross-validation
    fold_ids = split(1:nrow(x_pred), 1:cv_folds)
    # Vector of average mse for each combination of hyperparameters
    avg_mse_vals = numeric(nrow(param_grid))
    # Loop through each set of hyperparameters
    for(param_index in 1:nrow(param_grid))
    {
      # Vector to store mses per each test fold
      cv_mse_vec = numeric(cv_folds)
      # Cycle through each potential test fold
      for(fold_index in 1:cv_folds)
      {
        # Folds that will be used for training
        cv_sample = unsplit(fold_ids[-fold_index], rep(1:(cv_folds), cv_folds-1))
        # Fold that will be used for test
        cv_test = unsplit(fold_ids[-fold_index], 1:(cv_folds))
        # COnstruct the random forest
        ranger_rf_obj = ranger(x = x_pred[cv_sample,],
                               y = y_resp[cv_sample],
                               mtry = ceiling(ncol(x_pred) * param_grid[param_index, "frac_pred"]),
                               min.node.size = param_grid[param_index, "min_node_size"],
                               num.trees = param_grid[param_index, "num_trees"],
                               num.threads = 1)
        # Get the predictions
        ranger_predictions = predict(ranger_rf_obj, x_pred[cv_sample,])$predictions
        # Calculate the MSE
        mse_inst = 1 / length(cv_sample) * sum((y_resp[cv_sample] - ranger_predictions)^2)
        # Store the MSEs across each fold
        cv_mse_vec[fold_index] = mse_inst
      }
      # Store the average of the MSEs across eeach fold
      avg_mse_vals[param_index] = mean(cv_mse_vec) 
    }
    # Taking the minimum MSE or the first occurrence of the minimum MSE
    min_mse_position = which.min(avg_mse_vals)
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_ranger_obj = ranger(x = x_pred, y = y_resp,
                              mtry = ceiling(ncol(x_pred) * param_grid[min_mse_position, "frac_pred"]),
                              min.node.size = param_grid[min_mse_position, "min_node_size"],
                              num.trees = param_grid[min_mse_position, "num_trees"],
                              num.threads = 1)
    return(final_ranger_obj)
  }
  
  # Neural Network
  nnet_cv = function(x_pred, y_resp, 
                     parameter_grid = list("size" = c(3, 5, 10, 20),
                                           "decay" = c(0.001, 0.01, 0.1) ),
                     cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = expand.grid(parameter_grid)
    
    # Construct the XGBoosted trees
    nnet_control= caret::trainControl(method = "cv", number = cv_folds)
    # Train XGBoosted trees
    nnet_obj = caret::train(x = x_pred, y = as.vector(y_resp),
                            method = "nnet", tuneGrid = param_grid,
                            trControl = nnet_control,
                            maxit = 500,
                            linout=FALSE,
                            trace=FALSE)
    
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_nnet_obj = nnet_obj$finalModel
    return(final_nnet_obj)
  }
  
  # generate the testing data
  testing_model_data = model_generating_function(N = testing_n, rho = model_rho, 
                                                 max_pred = model_predictor_quant, 
                                                 dat.seed = testing_data_seed,
                                                 noise_1 = model_noise)
  
  # Getting variable names:
  all_variable_names = colnames(testing_model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  # Variables serving as the predictors
  predictor_vars = organized_vars$predictors
  
  
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  # Create the cluster (doParallel, doSNOW)
  # sim_cluster = makeCluster(cores)
  # Register the cluster
  # registerDoFuture()
  
  # Use futures with callr to bypass R connections limits
  # plan(multisession, workers = cores)
  plan(callr, workers = cores)
  
  # don't use %<-% as it will throw an error.
  # Store the results in a matrix
  results <- foreach(samp_size_index = 1:length(samp_n), .combine = "rbind", 
                     .options.future=list(packages = c("caret","glmnet", "ranger", 
                                                       "nnet", "e1071", "xgboost", "dplyr",
                                                       "tidyr", "caret")),
                     .inorder = FALSE) %:%
    foreach( iteration = m_start:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",.inorder = FALSE) %dofuture% {
              
              # setting the seed
              dat.seed = iteration + samp_n[samp_size_index]
              set.seed(dat.seed)
              
              # Training dataset
              # Generated with the same parameters sxcluding the seed
              training_model_data = 
                model_generating_function(N = samp_n[samp_size_index], 
                                          rho = model_rho, 
                                          max_pred = model_predictor_quant, 
                                          dat.seed = dat.seed,
                                          noise_1 = model_noise)
              
              
              # Performing SDC GADP method
              # Testing data
              privatized_responses_testing = 
                perturbation_func(unprotected_data = testing_model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              # Training data
              privatized_responses_training = 
                perturbation_func(unprotected_data = training_model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              # Combining the training and the test datasets
              privatized_responses = 
                rbind(privatized_responses_training,
                      privatized_responses_testing)
              
              model_data = rbind(training_model_data, 
                                 testing_model_data)
              
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,predictor_vars]
              
              privatized_dataframe[,colnames(privatized_responses)[response_index]] = 
                current_response
              
              # Get the training set form the df, first 1:n is training, n+1 
              # onwards is the testing data set. 
              drawing = 1:samp_n[samp_size_index]
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              x_dat_holdout = cbind(rep(1, nrow(model_data[-drawing,])),
                                    as.matrix(model_data)[-drawing,predictor_vars])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after gapd
              x_dat = as.matrix(cbind(rep(1, nrow(sample_private_data[,predictor_vars])),
                                      sample_private_data[,predictor_vars]))
              
              ### GAPD Used----
              
              # Linear Regression on data after gapd
              betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(current_response[drawing])
              
              
              linear_reg_pred = as.vector(x_dat_holdout %*% betas_linear)
              
              # Get the predictions for the training data
              
              training_lin_reg_pred = as.vector(x_dat %*% betas_linear)
              
              
              # Get the mse for the training data
              
              
              training_lin_reg_mse = calc_mse(as.matrix(current_response[drawing]),
                                              training_lin_reg_pred)
              
              # Get the mse for the test data
              
              
              linear_reg_mse = calc_mse(orig_data_response[-drawing],
                                        linear_reg_pred)
              
              
              # LASSO on data after gapd
              lasso_regress_model = cv.glmnet( 
                x = as.matrix(sample_private_data[,predictor_vars]),
                y = current_response[drawing])
              
              lasso_pred = predict(lasso_regress_model, 
                                   as.matrix(privatized_dataframe[-drawing,predictor_vars]),
                                   s = "lambda.min")
              
              # get predictions on the training data
              training_lasso_pred = predict(lasso_regress_model, 
                                            as.matrix(privatized_dataframe[drawing,predictor_vars]),
                                            s = "lambda.min")
              
              # Get the MSE for the training data
              training_lasso_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_lasso_pred)
              
              # # MSE for the testing data
              
              lasso_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                   lasso_pred)
              
              ### Neural Net Regression on data after gapd -------------
              
              
              gapd_nnet = nnet_cv(x_pred = x_dat[,predictor_vars], y_resp = current_response[drawing])
              
              
              gapd_nnet_reg_pred = predict(gapd_nnet,x_dat_holdout[,predictor_vars])
              
              # training data predictions
              training_gapd_nnet_reg_pred = predict(gapd_nnet,x_dat[,predictor_vars])
              
              # training data MSE
              training_gapd_nnet_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_nnet_reg_pred)
              
              
              # testing data MSE
              gapd_nnet_reg_mse = calc_mse(orig_data_response[-drawing],
                                           gapd_nnet_reg_pred)
              
              # SVM Radial Regression on data after gapd
              gapd_radial_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
                                              gamma = seq(0.001,0.1,length.out = 2), 
                                              cost = seq(0.1,1,length.out = 2),
                                              kernel = c("radial"))
              
              gapd_radial_svm = gapd_radial_svm_tune$best.model
              
              gapd_radial_svm_reg_pred = predict(gapd_radial_svm,x_dat_holdout[,predictor_vars])
              
              # Get the training predictions of the radial svm
              training_gapd_radial_svm_reg_pred =
                predict(gapd_radial_svm,x_dat[,predictor_vars])
              
              # Get the training mse for radial svm
              training_gapd_radial_svm_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_radial_svm_reg_pred)
              
              # Testing data MSE
              gapd_radial_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                 gapd_radial_svm_reg_pred)
              
              
              
              # SVM Linear Regression on data after gapd
              gapd_linear_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
                                              
                                              cost = seq(0.1,1,length.out = 2),
                                              kernel = c("linear"))
              
              gapd_linear_svm = gapd_linear_svm_tune$best.model
              
              # Predictions for the training data for Linear SVM
              training_gapd_linear_svm_reg_pred =
                predict(gapd_linear_svm,x_dat_holdout[,predictor_vars])
              
              # Training MSE for Linear SVM
              training_gapd_linear_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_linear_svm_reg_pred)
              
              # Predictions for the test dataset
              gapd_linear_svm_reg_pred = predict(gapd_linear_svm,x_dat_holdout[,predictor_vars])
              
              # Testing data MSE
              gapd_linear_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                 gapd_linear_svm_reg_pred)
              
              # SVM Polynomial Regression on data after gapd
              gapd_poly_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
                                            gamma = seq(0.001,0.1,length.out = 2), 
                                            cost = seq(0.1,1,length.out = 2),
                                            degree = seq(1,2,length.out=2),
                                            kernel = c("polynomial"),
                                            cross = 10)
              
              gapd_poly_svm = gapd_poly_svm_tune$best.model
              
              # Training data polynomial SVM predictions
              training_gapd_poly_svm_reg_pred = predict(gapd_poly_svm,x_dat[,predictor_vars])
              
              # Training data MSE polynomial SVM
              training_gapd_poly_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_poly_svm_reg_pred)
              
              # Predictions for testing dataset
              gapd_poly_svm_reg_pred = predict(gapd_poly_svm,x_dat_holdout[,predictor_vars])
              
              # MSE for testing dataset for polynomial kernel SVM
              gapd_poly_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                               gapd_poly_svm_reg_pred)
              
              # Random forest on data after gapd
              num_x_vars = length(predictor_vars)
              
              ranger_rf_regress_model = ranger_cv( 
                x_pred = as.matrix(sample_private_data[,predictor_vars]),
                y_resp = current_response[drawing])
              
              # Training dataset predicions for the random forest
              training_ranger_rf_pred = 
                predict(ranger_rf_regress_model, 
                        privatized_dataframe[drawing,predictor_vars])$predictions
              
              # training dataset MSE for the radom forest
              training_ranger_rf_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_ranger_rf_pred)
              
              
              # Testing dataset prediction for random forest
              ranger_rf_pred = predict(ranger_rf_regress_model, 
                                       privatized_dataframe[-drawing,predictor_vars])$predictions
              
              # Testing dataset MSE for random forest
              ranger_rf_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                       ranger_rf_pred)
              
              # XGBT on data after gapd
              xgbt_regress_model = xgbt_cv( 
                x_pred = as.matrix(sample_private_data[,predictor_vars]),
                y_resp = current_response[drawing])
              
              # Training data predictions for the XGB tree
              training_xgbt_pred = 
                predict(xgbt_regress_model,
                        as.matrix(privatized_dataframe[drawing,predictor_vars]))
              
              # training data MSE for the XGB tree
              training_xgbt_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_xgbt_pred)
              
              # Testing data predicitons for the XGB tree
              xgbt_pred = predict(xgbt_regress_model, 
                                  as.matrix(privatized_dataframe[-drawing,predictor_vars]))
              
              # Testing data MSE for the XGB tree
              xgbt_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                  xgbt_pred)
              
              ##
              ## Before GAPD (On original data) ------
              print(" running checkpoint 1")
              # Linear Regression on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              
              # Beta coefficients for the original data
              orig_data_betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(sample_orig_data_response)
              
              # Predictions on the training data no GADP
              training_lin_reg_unprivat_pred = as.vector(x_dat %*%orig_data_betas_linear)
              
              
              # MSE for training data for linear regression, no GADP
              training_linear_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_lin_reg_unprivat_pred)
              
              # Testing data MSE for linear regression no ADP
              linear_reg_unprivatized_pred = as.vector(x_dat_holdout %*% orig_data_betas_linear)
              
              linear_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_reg_unprivatized_pred)
              
              # LASSO on data before gapd (original data)
              
              
              lasso_regress_model_orig_data =
                cv.glmnet( x = as.matrix(sample_unprotected_data[,predictor_vars]),
                           y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # Training data predictions for LASSO
              
              training_lasso_pred_orig_data = 
                predict(lasso_regress_model_orig_data,
                        as.matrix(model_data[drawing,predictor_vars]), 
                        s = "lambda.min")
              
              # Training data MSE for LASSO
              training_lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_lasso_pred_orig_data)
              # Predictions for the test dataset for the LASSO model
              lasso_pred_orig_data = predict(lasso_regress_model_orig_data,
                                             as.matrix(model_data[-drawing,predictor_vars]), 
                                             s = "lambda.min")
              # MSE for the LASSO model for the test data
              lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         lasso_pred_orig_data)
              
              
              # Neural Net before GAPD 
              
              no_gapd_nnet = nnet_cv(x_pred = x_dat[,predictor_vars], 
                                     y_resp = sample_orig_data_response)
              
              # The predictions of the training set for neural network
              training_nnet_reg_unprivatized_pred = predict(no_gapd_nnet, x_dat[,predictor_vars])
              
              # The MSE for the training set for the neural network
              training_nnet_reg_unprivatized_mse = 
                calc_mse(orig_data_response[drawing],
                         training_nnet_reg_unprivatized_pred)
              
              # Predicitons on testing set for neural net
              nnet_reg_unprivatized_pred = predict(no_gapd_nnet, x_dat_holdout[,predictor_vars])
              
              # MSE for testing set for neural network
              nnet_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         nnet_reg_unprivatized_pred)
              
              # SVM Radial before GAPD
              
              
              no_gapd_radial_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
                                                 y = sample_orig_data_response,
                                                 gamma = seq(0.001,0.1,length.out = 2), 
                                                 cost = seq(0.1,1,length.out = 2),
                                                 
                                                 kernel = c("radial"))
              
              # Get the best model for Radial SVM
              no_gapd_radial_svm = no_gapd_radial_svm_tune$best.model
              
              # Predictions on the training set for radial SVM
              training_radial_svm_reg_unprivatized_pred = 
                predict(no_gapd_radial_svm, x_dat[,predictor_vars])
              # The MSE for the training set for radial SVM
              training_radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_radial_svm_reg_unprivatized_pred)
              
              # Get the predictions of Radial SVM for the testing set
              radial_svm_reg_unprivatized_pred = 
                predict(no_gapd_radial_svm, x_dat_holdout[,predictor_vars])
              # Get the MSE of the Radial SVM for the testing set
              radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         radial_svm_reg_unprivatized_pred)
              
              # Linear SVM before GAPD
              no_gapd_linear_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
                                                 y = sample_orig_data_response, 
                                                 cost = seq(0.1,1,length.out = 2),
                                                 kernel = c("linear"))
              # The Linear SVM best tuned model
              no_gapd_linear_svm = no_gapd_linear_svm_tune$best.model
              
              # The predictions on the training set, Linear SVM
              training_linear_svm_reg_unprivatized_pred = 
                predict(no_gapd_linear_svm, x_dat[,predictor_vars])
              # the MSE on the training set, Linear SVM
              training_linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_linear_svm_reg_unprivatized_pred)
              # The predictions for the testing set Linear SVM
              linear_svm_reg_unprivatized_pred = 
                predict(no_gapd_linear_svm, x_dat_holdout[,predictor_vars])
              
              # The MSE for the testing set Linear SVM
              linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_svm_reg_unprivatized_pred)
              
              
              # Polynomial SVM before GAPD
              
              no_gapd_poly_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
                                               y = sample_orig_data_response,
                                               gamma = seq(0.001,0.1,length.out = 2), 
                                               cost = seq(0.1,1,length.out = 2),
                                               degree = seq(1,2,length.out=2),
                                               kernel = c("polynomial"),
                                               cross = 10)
              # Get the best tuned polynomial SVM
              no_gapd_poly_svm = no_gapd_poly_svm_tune$best.model
              
              # The training set predicitons for the polynomial SVM
              training_poly_svm_reg_unprivatized_pred = 
                predict(no_gapd_poly_svm, x_dat[,predictor_vars])
              # the training set MSE for the polynomial SVM
              training_poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_poly_svm_reg_unprivatized_pred)
              
              # The testing set predicitons using the polynomial SVM
              poly_svm_reg_unprivatized_pred = 
                predict(no_gapd_poly_svm, x_dat_holdout[,predictor_vars])
              # The testing set MSE for the polynomial SVM
              poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         poly_svm_reg_unprivatized_pred)
              
              # Random forest on data before gapd
              
              ranger_rf_model_orig_data = ranger_cv( 
                x_pred = as.matrix(sample_unprotected_data[,predictor_vars]),
                y_resp = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # The training set predictions from random forest
              training_ranger_rf_pred_orig_data = 
                predict(ranger_rf_model_orig_data,
                        model_data[drawing,predictor_vars])$predictions
              # The training set MSE for the random forest
              training_ranger_rf_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_ranger_rf_pred_orig_data)
              
              # Testing set predictions for random forest
              ranger_rf_pred_orig_data = predict(ranger_rf_model_orig_data,
                                                 model_data[-drawing,predictor_vars])$predictions
              # Testing set MSE for random forest
              ranger_rf_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         ranger_rf_pred_orig_data)
              
              # XGBT on data before gapd
              xgbt_regress_model_orig_data =
                xgbt_cv( x_pred = as.matrix(sample_unprotected_data[,predictor_vars]),
                         y_resp = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # The training set predictions for XGB tree
              training_xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[drawing,predictor_vars]))
              # Training set MSE for XGB tree for 
              training_xgbt_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_xgbt_pred_orig_data)
              
              # The testing set XGB tree predictions
              xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[-drawing,predictor_vars]))
              # The testing set MSE for XGB tree
              xgbt_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         xgbt_pred_orig_data)
              
              
              ###
              ### Stacked Ensemble Equal Weighting ----
              # GAPD
              avg_stacked_ensemble_predictions_gapd = (linear_reg_pred  + 
                                                         lasso_pred  +  gapd_nnet_reg_pred  + 
                                                         gapd_radial_svm_reg_pred  + 
                                                         gapd_linear_svm_reg_pred  + 
                                                         gapd_poly_svm_reg_pred  +
                                                         ranger_rf_pred  + xgbt_pred) / 8
              
              avg_stacked_ensemble_gapd_mse =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         avg_stacked_ensemble_predictions_gapd)
              # No GAPD
              avg_stacked_ensemble_pred_no_gapd = 
                (linear_reg_unprivatized_pred  + 
                   lasso_pred_orig_data  + 
                   nnet_reg_unprivatized_pred + 
                   radial_svm_reg_unprivatized_pred  + 
                   linear_svm_reg_unprivatized_pred  + 
                   poly_svm_reg_unprivatized_pred  +
                   ranger_rf_pred_orig_data  + xgbt_pred_orig_data) / 8
              
              avg_stacked_ensemble_no_gapd_mse =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         avg_stacked_ensemble_pred_no_gapd)
              
              
              ### Stacked Ensemble Equal Weighting by Proportion of variance Explained ----------------
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_gapd = 
                c(training_lin_reg_mse, training_lasso_mse, 
                  training_gapd_nnet_reg_mse, training_gapd_radial_svm_reg_mse,
                  training_gapd_linear_svm_reg_mse, 
                  training_gapd_poly_svm_reg_mse, training_ranger_rf_mse,
                  training_xgbt_mse)
              
              # The models with the lowest MSE get the highest weight
              prop_var_exp_gapd_weighting = 
                (training_set_mses_gapd / sum(training_set_mses_gapd))^(-1)
              prop_var_exp_gapd_weighting = prop_var_exp_gapd_weighting / 
                sum(prop_var_exp_gapd_weighting)
              
              
              
              # Get the predictions for the stacked ensemble for the prop
              # of variance explained
              prop_var_stacked_ens_gapd_pred = 
                linear_reg_pred * prop_var_exp_gapd_weighting[1] +
                lasso_pred * prop_var_exp_gapd_weighting[2] + 
                gapd_nnet_reg_pred * prop_var_exp_gapd_weighting[3] + 
                gapd_radial_svm_reg_pred * prop_var_exp_gapd_weighting[4] + 
                gapd_linear_svm_reg_pred * prop_var_exp_gapd_weighting[5] + 
                gapd_poly_svm_reg_pred * prop_var_exp_gapd_weighting[6] +
                ranger_rf_pred * prop_var_exp_gapd_weighting[7] + 
                xgbt_pred * prop_var_exp_gapd_weighting[8]
              
              
              # Get the MSE
              prop_var_stacked_ens_gapd_mse = 
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         prop_var_stacked_ens_gapd_pred)
              
              
              # Get the Stacked Ensemble (Prop. of Var. explained) with No GADP 
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_no_gapd = 
                c(training_linear_reg_mse_orig_data, 
                  training_lasso_mse_orig_data, 
                  training_nnet_reg_unprivatized_mse, 
                  training_radial_svm_reg_mse_orig_data,
                  training_linear_svm_reg_mse_orig_data, 
                  training_poly_svm_reg_mse_orig_data, 
                  training_ranger_rf_mse_orig_data,
                  training_xgbt_mse_orig_data)
              
              # weighting for the SE
              prop_var_exp_no_gapd_weighting = 
                (training_set_mses_no_gapd / sum(training_set_mses_no_gapd))^(-1)
              prop_var_exp_no_gapd_weighting = prop_var_exp_no_gapd_weighting /
                sum(prop_var_exp_no_gapd_weighting)
              
              
              # get the predictions
              prop_var_stacked_ens_no_gapd_pred = 
                linear_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[1] +
                lasso_pred_orig_data * prop_var_exp_no_gapd_weighting[2] + 
                nnet_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[3] + 
                radial_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[4] + 
                linear_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[5] + 
                poly_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[6] +
                ranger_rf_pred_orig_data * prop_var_exp_no_gapd_weighting[7] + 
                xgbt_pred_orig_data * prop_var_exp_no_gapd_weighting[8]
              
              
              # Get the MSE
              prop_var_stacked_ens_no_gapd_mse = 
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         prop_var_stacked_ens_no_gapd_pred)
              
              
              # Compile MSEs into a vector
              gapd_mse_vector = 
                c(linear_reg_mse, lasso_mse, gapd_nnet_reg_mse,
                  gapd_radial_svm_reg_mse, gapd_linear_svm_reg_mse,
                  gapd_poly_svm_reg_mse, ranger_rf_mse,
                  xgbt_mse, avg_stacked_ensemble_gapd_mse, 
                  prop_var_stacked_ens_gapd_mse)
              
              no_gapd_mse_vector =
                c(linear_reg_mse_orig_data, lasso_mse_orig_data,
                  nnet_reg_mse_orig_data, radial_svm_reg_mse_orig_data,
                  linear_svm_reg_mse_orig_data, poly_svm_reg_mse_orig_data,
                  ranger_rf_mse_orig_data, xgbt_mse_orig_data,
                  avg_stacked_ensemble_no_gapd_mse, 
                  prop_var_stacked_ens_no_gapd_mse)
              
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "training_size", 
                                   "testing_size",
                                   "y", 
                                   "LR", "LASSO", "NNET", "Radial_SVM",
                                   "Linear_SVM", "Poly_SVM", "RF", "XGBTree",
                                   "Stacked_equal", "Stacked_prop","gadp", "theta", "rho")
              
              
              # construct the dataframe of results
              estimator_reg_df = data.frame(rbind(c(dat.seed,iteration, samp_n[samp_size_index], testing_n,
                                                    response_index, gapd_mse_vector, 1, theta_value, model_rho)))
              
              # Adding results for predictive models without GAPD
              estimator_reg_df[2,] = rbind(c(dat.seed, iteration, 
                                             samp_n[samp_size_index], testing_n,
                                             response_index, no_gapd_mse_vector, 0, theta_value, model_rho))
              
              
              colnames(estimator_reg_df) = data_frame_names
              
              
              return(estimator_reg_df)
            }
  
  # stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  return(results)
}

## Stagnant Data For Real Data Application ------
# Train on static dataset loaded into the function, versus iteratively
# generated dataset
estimator_eval_stacked_ensemb_cv_stag = function( # the dataset
  m_start = 1, # Primarily used for debugging determines which iteration to start from
  m = 1000, # Number of Monte Carlo simulations
  samp_n = c(500, 1000, 2000), # sample sizes to be used for training set
  cores = 2, # Number of CPU threads/ logical processors to use
  response_prefix = "y", # prefix for response variables
  acquire_variable_names = get_variables, #Function for getting the variables
  perturbation_func = gadp_func, # CGADP or GADP function
  theta_value = 0.4, # privacy parameter value
  model_data) # Data to be loaded into the model
{
  # Cross-Validation Funcitons -----------
  # XGBoosted Trees
  xgbt_cv = function(x_pred, y_resp, 
                     parameter_grid = list("nrounds" = c(10,20, 40),
                                           "max_depth" = c(2,4),
                                           "eta" = c(0.2, 0.4, 0.6),
                                           "gamma" = c(0, 0.5, 1),
                                           "colsample_bytree" = c(0.2, 0.5, 1),
                                           "min_child_weight" = c(0.5, 1),
                                           "subsample" = c(0.4, 0.5, 0.6)),
                     cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = expand.grid(parameter_grid)
    
    # Construct the XGBoosted trees
    xgbt_control= caret::trainControl(method = "cv", number = cv_folds)
    # Train XGBoosted trees
    xgbt_obj = caret::train(x = x_pred, y = as.vector(y_resp),
                            method = "xgbTree", tuneGrid = param_grid,
                            trControl = xgbt_control,
                            nthread = 1,
                            verbose=0,
                            print_every_n =0)
    
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_xgbt_obj = xgbt_obj$finalModel
    return(final_xgbt_obj)
  }
  
  # Random forest
  ranger_cv = function(x_pred, y_resp, num_trees = 500,
                       frac_pred = c(1/3, 1/4), 
                       min_node_size = c(5,10),
                       cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = as.matrix(expand_grid(num_trees, frac_pred, min_node_size))
    # Set up folds to be used in cross-validation
    fold_ids = split(1:nrow(x_pred), 1:cv_folds)
    # Vector of average mse for each combination of hyperparameters
    avg_mse_vals = numeric(nrow(param_grid))
    # Loop through each set of hyperparameters
    for(param_index in 1:nrow(param_grid))
    {
      # Vector to store mses per each test fold
      cv_mse_vec = numeric(cv_folds)
      # Cycle through each potential test fold
      for(fold_index in 1:cv_folds)
      {
        # Folds that will be used for training
        cv_sample = unsplit(fold_ids[-fold_index], rep(1:(cv_folds), cv_folds-1))
        # Fold that will be used for test
        cv_test = unsplit(fold_ids[-fold_index], 1:(cv_folds))
        # COnstruct the random forest
        ranger_rf_obj = ranger(x = x_pred[cv_sample,],
                               y = y_resp[cv_sample],
                               mtry = ceiling(ncol(x_pred) * param_grid[param_index, "frac_pred"]),
                               min.node.size = param_grid[param_index, "min_node_size"],
                               num.trees = param_grid[param_index, "num_trees"],
                               num.threads = 1)
        # Get the predictions
        ranger_predictions = predict(ranger_rf_obj, x_pred[cv_sample,])$predictions
        # Calculate the MSE
        mse_inst = 1 / length(cv_sample) * sum((y_resp[cv_sample] - ranger_predictions)^2)
        # Store the MSEs across each fold
        cv_mse_vec[fold_index] = mse_inst
      }
      # Store the average of the MSEs across eeach fold
      avg_mse_vals[param_index] = mean(cv_mse_vec) 
    }
    # Taking the minimum MSE or the first occurrence of the minimum MSE
    min_mse_position = which.min(avg_mse_vals)
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_ranger_obj = ranger(x = x_pred, y = y_resp,
                              mtry = ceiling(ncol(x_pred) * param_grid[min_mse_position, "frac_pred"]),
                              min.node.size = param_grid[min_mse_position, "min_node_size"],
                              num.trees = param_grid[min_mse_position, "num_trees"],
                              num.threads = 1)
    return(final_ranger_obj)
  }
  
  # Neural Network
  nnet_cv = function(x_pred, y_resp, 
                     parameter_grid = list("size" = c(3, 5, 10, 20),
                                           "decay" = c(0.001, 0.01, 0.1) ),
                     cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = expand.grid(parameter_grid)
    
    # Construct the XGBoosted trees
    nnet_control= caret::trainControl(method = "cv", number = cv_folds)
    # Train XGBoosted trees
    nnet_obj = caret::train(x = x_pred, y = as.vector(y_resp),
                            method = "nnet", tuneGrid = param_grid,
                            trControl = nnet_control,
                            maxit = 500,
                            linout=FALSE,
                            trace=FALSE)
    
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_nnet_obj = nnet_obj$finalModel
    return(final_nnet_obj)
  }
  
  # generate the testing data
  
  
  # Getting variable names:
  all_variable_names = colnames(model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  # Variables serving as the predictors
  predictor_vars = organized_vars$predictors
  
  
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  # Create the cluster (doParallel, doSNOW)
  # sim_cluster = makeCluster(cores)
  # Register the cluster
  # registerDoFuture()
  
  # Use futures with callr to bypass R connections limits
  # plan(multisession, workers = cores)
  plan(callr, workers = cores)
  
  # don't use %<-% as it will throw an error.
  # Store the results in a matrix
  results <- foreach(samp_size_index = 1:length(samp_n), .combine = "rbind", 
                     .options.future=list(packages = c("caret","glmnet", "ranger", 
                                                       "nnet", "e1071", "xgboost", "dplyr",
                                                       "tidyr", "caret")),
                     .inorder = FALSE) %:%
    foreach( iteration = m_start:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",.inorder = FALSE) %dofuture% {
              
              # setting the seed
              dat.seed = iteration + samp_n[samp_size_index]
              set.seed(dat.seed)
              
              # Performing SDC GADP method
              privatized_responses = 
                perturbation_func(unprotected_data = model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,organized_vars$predictors]
              
              privatized_dataframe[,colnames(privatized_responses)[response_index]] = 
                current_response
              
              # Get the sample from the data
              drawing = sample(1:nrow(model_data), size = samp_n[samp_size_index], replace = FALSE)
              
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,predictor_vars]
              
              privatized_dataframe[,colnames(privatized_responses)[response_index]] = 
                current_response
              
              
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              x_dat_holdout = cbind(rep(1, nrow(model_data[-drawing,])),
                                    as.matrix(model_data)[-drawing,predictor_vars])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after gapd
              x_dat = as.matrix(cbind(intercept = rep(1, nrow(sample_private_data[,predictor_vars])),
                                      sample_private_data[,predictor_vars]))

              
              ### GAPD Used----
              
              # Linear Regression on data after gapd
              betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(current_response[drawing])
              
              
              linear_reg_pred = as.vector(x_dat_holdout %*% betas_linear)
              
              # Get the predictions for the training data
              
              training_lin_reg_pred = as.vector(x_dat %*% betas_linear)
              
              
              # Get the mse for the training data
              
              
              training_lin_reg_mse = calc_mse(as.matrix(current_response[drawing]),
                                              training_lin_reg_pred)
              
              # Get the mse for the test data
              
              
              linear_reg_mse = calc_mse(orig_data_response[-drawing],
                                        linear_reg_pred)
              
              
              # LASSO on data after gapd
              lasso_regress_model = cv.glmnet( 
                x = as.matrix(sample_private_data[,predictor_vars]),
                y = current_response[drawing])
              
              lasso_pred = predict(lasso_regress_model, 
                                   as.matrix(privatized_dataframe[-drawing,predictor_vars]),
                                   s = "lambda.min")
              
              # get predictions on the training data
              training_lasso_pred = predict(lasso_regress_model, 
                                            as.matrix(privatized_dataframe[drawing,predictor_vars]),
                                            s = "lambda.min")
              
              # Get the MSE for the training data
              training_lasso_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_lasso_pred)
              
              # # MSE for the testing data
              
              lasso_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                   lasso_pred)
              
              ### Neural Net Regression on data after gapd -------------
              
              
              gapd_nnet = nnet_cv(x_pred = x_dat[,predictor_vars], y_resp = current_response[drawing])
              
              
              gapd_nnet_reg_pred = predict(gapd_nnet,x_dat_holdout[,predictor_vars])
              
              # training data predictions
              training_gapd_nnet_reg_pred = predict(gapd_nnet,x_dat[,predictor_vars])
              
              # training data MSE
              training_gapd_nnet_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_nnet_reg_pred)
              
              
              # testing data MSE
              gapd_nnet_reg_mse = calc_mse(orig_data_response[-drawing],
                                           gapd_nnet_reg_pred)
              
              # SVM Radial Regression on data after gapd
              gapd_radial_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
                                              gamma = seq(0.001,0.1,length.out = 2), 
                                              cost = seq(0.1,1,length.out = 2),
                                              kernel = c("radial"))
              
              gapd_radial_svm = gapd_radial_svm_tune$best.model
              
              gapd_radial_svm_reg_pred = predict(gapd_radial_svm,x_dat_holdout[,predictor_vars])
              
              # Get the training predictions of the radial svm
              training_gapd_radial_svm_reg_pred =
                predict(gapd_radial_svm,x_dat[,predictor_vars])
              
              # Get the training mse for radial svm
              training_gapd_radial_svm_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_radial_svm_reg_pred)
              
              # Testing data MSE
              gapd_radial_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                 gapd_radial_svm_reg_pred)
              
              
              
              # SVM Linear Regression on data after gapd
              gapd_linear_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
                                              
                                              cost = seq(0.1,1,length.out = 2),
                                              kernel = c("linear"))
              
              gapd_linear_svm = gapd_linear_svm_tune$best.model
              
              # Predictions for the training data for Linear SVM
              training_gapd_linear_svm_reg_pred =
                predict(gapd_linear_svm,x_dat_holdout[,predictor_vars])
              
              # Training MSE for Linear SVM
              training_gapd_linear_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_linear_svm_reg_pred)
              
              # Predictions for the test dataset
              gapd_linear_svm_reg_pred = predict(gapd_linear_svm,x_dat_holdout[,predictor_vars])
              
              # Testing data MSE
              gapd_linear_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                 gapd_linear_svm_reg_pred)
              
              # SVM Polynomial Regression on data after gapd
              gapd_poly_svm_tune = tune.svm(x = x_dat[,predictor_vars], y = current_response[drawing],
                                            gamma = seq(0.001,0.1,length.out = 2), 
                                            cost = seq(0.1,1,length.out = 2),
                                            degree = seq(1,2,length.out=2),
                                            kernel = c("polynomial"),
                                            cross = 10)
              
              gapd_poly_svm = gapd_poly_svm_tune$best.model
              
              # Training data polynomial SVM predictions
              training_gapd_poly_svm_reg_pred = predict(gapd_poly_svm,x_dat[,predictor_vars])
              
              # Training data MSE polynomial SVM
              training_gapd_poly_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_gapd_poly_svm_reg_pred)
              
              # Predictions for testing dataset
              gapd_poly_svm_reg_pred = predict(gapd_poly_svm,x_dat_holdout[,predictor_vars])
              
              # MSE for testing dataset for polynomial kernel SVM
              gapd_poly_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                               gapd_poly_svm_reg_pred)
              
              # Random forest on data after gapd
              num_x_vars = length(predictor_vars)
              
              ranger_rf_regress_model = ranger_cv( 
                x_pred = as.matrix(sample_private_data[,predictor_vars]),
                y_resp = current_response[drawing])
              
              # Training dataset predicions for the random forest
              training_ranger_rf_pred = 
                predict(ranger_rf_regress_model, 
                        privatized_dataframe[drawing,predictor_vars])$predictions
              
              # training dataset MSE for the radom forest
              training_ranger_rf_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_ranger_rf_pred)
              
              
              # Testing dataset prediction for random forest
              ranger_rf_pred = predict(ranger_rf_regress_model, 
                                       privatized_dataframe[-drawing,predictor_vars])$predictions
              
              # Testing dataset MSE for random forest
              ranger_rf_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                       ranger_rf_pred)
              
              # XGBT on data after gapd
              xgbt_regress_model = xgbt_cv( 
                x_pred = as.matrix(sample_private_data[,predictor_vars]),
                y_resp = current_response[drawing])
              
              # Training data predictions for the XGB tree
              training_xgbt_pred = 
                predict(xgbt_regress_model,
                        as.matrix(privatized_dataframe[drawing,predictor_vars]))
              
              # training data MSE for the XGB tree
              training_xgbt_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_xgbt_pred)
              
              # Testing data predicitons for the XGB tree
              xgbt_pred = predict(xgbt_regress_model, 
                                  as.matrix(privatized_dataframe[-drawing,predictor_vars]))
              
              # Testing data MSE for the XGB tree
              xgbt_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                  xgbt_pred)
              
              ##
              ## Before GAPD (On original data) ------
              print(" running checkpoint 1")
              # Linear Regression on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              
              # Beta coefficients for the original data
              orig_data_betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(sample_orig_data_response)
              
              # Predictions on the training data no GADP
              training_lin_reg_unprivat_pred = as.vector(x_dat %*%orig_data_betas_linear)
              
              
              # MSE for training data for linear regression, no GADP
              training_linear_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_lin_reg_unprivat_pred)
              
              # Testing data MSE for linear regression no ADP
              linear_reg_unprivatized_pred = as.vector(x_dat_holdout %*% orig_data_betas_linear)
              
              linear_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_reg_unprivatized_pred)
              
              # LASSO on data before gapd (original data)
              
              
              lasso_regress_model_orig_data =
                cv.glmnet( x = as.matrix(sample_unprotected_data[,predictor_vars]),
                           y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # Training data predictions for LASSO
              
              training_lasso_pred_orig_data = 
                predict(lasso_regress_model_orig_data,
                        as.matrix(model_data[drawing,predictor_vars]), 
                        s = "lambda.min")
              
              # Training data MSE for LASSO
              training_lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_lasso_pred_orig_data)
              # Predictions for the test dataset for the LASSO model
              lasso_pred_orig_data = predict(lasso_regress_model_orig_data,
                                             as.matrix(model_data[-drawing,predictor_vars]), 
                                             s = "lambda.min")
              # MSE for the LASSO model for the test data
              lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         lasso_pred_orig_data)
              
              
              # Neural Net before GAPD 
              
              no_gapd_nnet = nnet_cv(x_pred = x_dat[,predictor_vars], 
                                     y_resp = sample_orig_data_response)
              
              # The predictions of the training set for neural network
              training_nnet_reg_unprivatized_pred = predict(no_gapd_nnet, x_dat[,predictor_vars])
              
              # The MSE for the training set for the neural network
              training_nnet_reg_unprivatized_mse = 
                calc_mse(orig_data_response[drawing],
                         training_nnet_reg_unprivatized_pred)
              
              # Predicitons on testing set for neural net
              nnet_reg_unprivatized_pred = predict(no_gapd_nnet, x_dat_holdout[,predictor_vars])
              
              # MSE for testing set for neural network
              nnet_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         nnet_reg_unprivatized_pred)
              
              # SVM Radial before GAPD
              
              
              no_gapd_radial_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
                                                 y = sample_orig_data_response,
                                                 gamma = seq(0.001,0.1,length.out = 2), 
                                                 cost = seq(0.1,1,length.out = 2),
                                                 
                                                 kernel = c("radial"))
              
              # Get the best model for Radial SVM
              no_gapd_radial_svm = no_gapd_radial_svm_tune$best.model
              
              # Predictions on the training set for radial SVM
              training_radial_svm_reg_unprivatized_pred = 
                predict(no_gapd_radial_svm, x_dat[,predictor_vars])
              # The MSE for the training set for radial SVM
              training_radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_radial_svm_reg_unprivatized_pred)
              
              # Get the predictions of Radial SVM for the testing set
              radial_svm_reg_unprivatized_pred = 
                predict(no_gapd_radial_svm, x_dat_holdout[,predictor_vars])
              # Get the MSE of the Radial SVM for the testing set
              radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         radial_svm_reg_unprivatized_pred)
              
              # Linear SVM before GAPD
              no_gapd_linear_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
                                                 y = sample_orig_data_response, 
                                                 cost = seq(0.1,1,length.out = 2),
                                                 kernel = c("linear"))
              # The Linear SVM best tuned model
              no_gapd_linear_svm = no_gapd_linear_svm_tune$best.model
              
              # The predictions on the training set, Linear SVM
              training_linear_svm_reg_unprivatized_pred = 
                predict(no_gapd_linear_svm, x_dat[,predictor_vars])
              # the MSE on the training set, Linear SVM
              training_linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_linear_svm_reg_unprivatized_pred)
              # The predictions for the testing set Linear SVM
              linear_svm_reg_unprivatized_pred = 
                predict(no_gapd_linear_svm, x_dat_holdout[,predictor_vars])
              
              # The MSE for the testing set Linear SVM
              linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_svm_reg_unprivatized_pred)
              
              
              # Polynomial SVM before GAPD
              
              no_gapd_poly_svm_tune = tune.svm(x = x_dat[,predictor_vars], 
                                               y = sample_orig_data_response,
                                               gamma = seq(0.001,0.1,length.out = 2), 
                                               cost = seq(0.1,1,length.out = 2),
                                               degree = seq(1,2,length.out=2),
                                               kernel = c("polynomial"),
                                               cross = 10)
              # Get the best tuned polynomial SVM
              no_gapd_poly_svm = no_gapd_poly_svm_tune$best.model
              
              # The training set predicitons for the polynomial SVM
              training_poly_svm_reg_unprivatized_pred = 
                predict(no_gapd_poly_svm, x_dat[,predictor_vars])
              # the training set MSE for the polynomial SVM
              training_poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_poly_svm_reg_unprivatized_pred)
              
              # The testing set predicitons using the polynomial SVM
              poly_svm_reg_unprivatized_pred = 
                predict(no_gapd_poly_svm, x_dat_holdout[,predictor_vars])
              # The testing set MSE for the polynomial SVM
              poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         poly_svm_reg_unprivatized_pred)
              
              # Random forest on data before gapd
              
              ranger_rf_model_orig_data = ranger_cv( 
                x_pred = as.matrix(sample_unprotected_data[,predictor_vars]),
                y_resp = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # The training set predictions from random forest
              training_ranger_rf_pred_orig_data = 
                predict(ranger_rf_model_orig_data,
                        model_data[drawing,predictor_vars])$predictions
              # The training set MSE for the random forest
              training_ranger_rf_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_ranger_rf_pred_orig_data)
              
              # Testing set predictions for random forest
              ranger_rf_pred_orig_data = predict(ranger_rf_model_orig_data,
                                                 model_data[-drawing,predictor_vars])$predictions
              # Testing set MSE for random forest
              ranger_rf_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         ranger_rf_pred_orig_data)
              
              # XGBT on data before gapd
              xgbt_regress_model_orig_data =
                xgbt_cv( x_pred = as.matrix(sample_unprotected_data[,predictor_vars]),
                         y_resp = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # The training set predictions for XGB tree
              training_xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[drawing,predictor_vars]))
              # Training set MSE for XGB tree for 
              training_xgbt_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_xgbt_pred_orig_data)
              
              # The testing set XGB tree predictions
              xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[-drawing,predictor_vars]))
              # The testing set MSE for XGB tree
              xgbt_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         xgbt_pred_orig_data)
              
              
              ###
              ### Stacked Ensemble Equal Weighting ----
              # GAPD
              avg_stacked_ensemble_predictions_gapd = (linear_reg_pred  + 
                                                         lasso_pred  +  gapd_nnet_reg_pred  + 
                                                         gapd_radial_svm_reg_pred  + 
                                                         gapd_linear_svm_reg_pred  + 
                                                         gapd_poly_svm_reg_pred  +
                                                         ranger_rf_pred  + xgbt_pred) / 8
              
              avg_stacked_ensemble_gapd_mse =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         avg_stacked_ensemble_predictions_gapd)
              # No GAPD
              avg_stacked_ensemble_pred_no_gapd = 
                (linear_reg_unprivatized_pred  + 
                   lasso_pred_orig_data  + 
                   nnet_reg_unprivatized_pred + 
                   radial_svm_reg_unprivatized_pred  + 
                   linear_svm_reg_unprivatized_pred  + 
                   poly_svm_reg_unprivatized_pred  +
                   ranger_rf_pred_orig_data  + xgbt_pred_orig_data) / 8
              
              avg_stacked_ensemble_no_gapd_mse =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         avg_stacked_ensemble_pred_no_gapd)
              
              
              ### Stacked Ensemble Equal Weighting by Proportion of variance Explained ----------------
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_gapd = 
                c(training_lin_reg_mse, training_lasso_mse, 
                  training_gapd_nnet_reg_mse, training_gapd_radial_svm_reg_mse,
                  training_gapd_linear_svm_reg_mse, 
                  training_gapd_poly_svm_reg_mse, training_ranger_rf_mse,
                  training_xgbt_mse)
              
              # The models with the lowest MSE get the highest weight
              prop_var_exp_gapd_weighting = 
                (training_set_mses_gapd / sum(training_set_mses_gapd))^(-1)
              prop_var_exp_gapd_weighting = prop_var_exp_gapd_weighting / 
                sum(prop_var_exp_gapd_weighting)
              
              
              
              # Get the predictions for the stacked ensemble for the prop
              # of variance explained
              prop_var_stacked_ens_gapd_pred = 
                linear_reg_pred * prop_var_exp_gapd_weighting[1] +
                lasso_pred * prop_var_exp_gapd_weighting[2] + 
                gapd_nnet_reg_pred * prop_var_exp_gapd_weighting[3] + 
                gapd_radial_svm_reg_pred * prop_var_exp_gapd_weighting[4] + 
                gapd_linear_svm_reg_pred * prop_var_exp_gapd_weighting[5] + 
                gapd_poly_svm_reg_pred * prop_var_exp_gapd_weighting[6] +
                ranger_rf_pred * prop_var_exp_gapd_weighting[7] + 
                xgbt_pred * prop_var_exp_gapd_weighting[8]
              
              
              # Get the MSE
              prop_var_stacked_ens_gapd_mse = 
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         prop_var_stacked_ens_gapd_pred)
              
              
              # Get the Stacked Ensemble (Prop. of Var. explained) with No GADP 
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_no_gapd = 
                c(training_linear_reg_mse_orig_data, 
                  training_lasso_mse_orig_data, 
                  training_nnet_reg_unprivatized_mse, 
                  training_radial_svm_reg_mse_orig_data,
                  training_linear_svm_reg_mse_orig_data, 
                  training_poly_svm_reg_mse_orig_data, 
                  training_ranger_rf_mse_orig_data,
                  training_xgbt_mse_orig_data)
              
              # weighting for the SE
              prop_var_exp_no_gapd_weighting = 
                (training_set_mses_no_gapd / sum(training_set_mses_no_gapd))^(-1)
              prop_var_exp_no_gapd_weighting = prop_var_exp_no_gapd_weighting /
                sum(prop_var_exp_no_gapd_weighting)
              
              
              # get the predictions
              prop_var_stacked_ens_no_gapd_pred = 
                linear_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[1] +
                lasso_pred_orig_data * prop_var_exp_no_gapd_weighting[2] + 
                nnet_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[3] + 
                radial_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[4] + 
                linear_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[5] + 
                poly_svm_reg_unprivatized_pred * prop_var_exp_no_gapd_weighting[6] +
                ranger_rf_pred_orig_data * prop_var_exp_no_gapd_weighting[7] + 
                xgbt_pred_orig_data * prop_var_exp_no_gapd_weighting[8]
              
              
              # Get the MSE
              prop_var_stacked_ens_no_gapd_mse = 
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         prop_var_stacked_ens_no_gapd_pred)
              
              
              # Compile MSEs into a vector
              gapd_mse_vector = 
                c(linear_reg_mse, lasso_mse, gapd_nnet_reg_mse,
                  gapd_radial_svm_reg_mse, gapd_linear_svm_reg_mse,
                  gapd_poly_svm_reg_mse, ranger_rf_mse,
                  xgbt_mse, avg_stacked_ensemble_gapd_mse, 
                  prop_var_stacked_ens_gapd_mse)
              
              no_gapd_mse_vector =
                c(linear_reg_mse_orig_data, lasso_mse_orig_data,
                  nnet_reg_mse_orig_data, radial_svm_reg_mse_orig_data,
                  linear_svm_reg_mse_orig_data, poly_svm_reg_mse_orig_data,
                  ranger_rf_mse_orig_data, xgbt_mse_orig_data,
                  avg_stacked_ensemble_no_gapd_mse, 
                  prop_var_stacked_ens_no_gapd_mse)
              
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "training_size", 
                                   
                                   "y", 
                                   "LR", "LASSO", "NNET", "Radial_SVM",
                                   "Linear_SVM", "Poly_SVM", "RF", "XGBTree",
                                   "Stacked_equal", "Stacked_prop","gadp", "theta")
              
              
              # construct the dataframe of results
              estimator_reg_df = data.frame(rbind(c(dat.seed,iteration, samp_n[samp_size_index],
                                                    response_index, gapd_mse_vector, 1, theta_value)))
              
              # Adding results for predictive models without GAPD
              estimator_reg_df[2,] = rbind(c(dat.seed, iteration, 
                                             samp_n[samp_size_index],
                                             response_index, no_gapd_mse_vector, 0, theta_value))
              
              
              colnames(estimator_reg_df) = data_frame_names
              
              
              return(estimator_reg_df)
            }
  
  # stopCluster(sim_cluster)
  
  results = data.frame(results)
  
  return(results)
}


# More Hyperparameter tuning with CGADP

## CGADP (Model 1 and 2, to expensive for Poly SVM) --------------------

estimator_evaluation_stacked_ensemb_cgadp_cv = function(# the dataset
  m_start = 1, # Primarily used for debugging determines which iteration to start from
  m = 1000, # Number of Monte Carlo simulations
  samp_n = c(500, 1000, 2000), # sample sizes to be used for training set
  cores = 2, # Number of CPU threads/ logical processors to use
  response_prefix = "y", # prefix for response variables
  acquire_variable_names = get_variables, #Function for getting the variables
  perturbation_func, # CGADP or GADP function
  theta_value = 0.4, # privacy parameter value
  model_generating_function, # function to generate the model
  testing_data_seed = 30000, # Seed for the testing set
  model_predictor_quant = 10, # Quantity of the predictors for the model
  model_rho = 0.2, # correlation of predictors in the model
  model_noise = 0.2, # Noise to be used in the model
  testing_n = 500) # Size of the testing set
{
  
  # Cross-Validation Funcitons -----------
  # XGBoosted Trees
  xgbt_cv = function(x_pred, y_resp, 
                     parameter_grid = list("nrounds" = c(10,20, 40),
                                           "max_depth" = c(2,4),
                                           "eta" = c(0.2, 0.4, 0.6),
                                           "gamma" = c(0, 0.5, 1),
                                           "colsample_bytree" = c(0.2, 0.5, 1),
                                           "min_child_weight" = c(0.5, 1),
                                           "subsample" = c(0.4, 0.5, 0.6)),
                     cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = expand.grid(parameter_grid)
    
    # Construct the XGBoosted trees
    xgbt_control= caret::trainControl(method = "cv", number = cv_folds)
    # Train XGBoosted trees
    xgbt_obj = caret::train(x = x_pred, y = as.vector(y_resp),
                            method = "xgbTree", tuneGrid = param_grid,
                            trControl = xgbt_control,
                            nthread = 1,
                            verbose=0,
                            print_every_n =0)
    
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_xgbt_obj = xgbt_obj$finalModel
    return(final_xgbt_obj)
  }
  
  # Random forest
  ranger_cv = function(x_pred, y_resp, num_trees = 500,
                       frac_pred = c(1/3, 1/4), 
                       min_node_size = c(5,10),
                       cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = as.matrix(expand_grid(num_trees, frac_pred, min_node_size))
    # Set up folds to be used in cross-validation
    fold_ids = split(1:nrow(x_pred), 1:cv_folds)
    # Vector of average mse for each combination of hyperparameters
    avg_mse_vals = numeric(nrow(param_grid))
    # Loop through each set of hyperparameters
    for(param_index in 1:nrow(param_grid))
    {
      # Vector to store mses per each test fold
      cv_mse_vec = numeric(cv_folds)
      # Cycle through each potential test fold
      for(fold_index in 1:cv_folds)
      {
        # Folds that will be used for training
        cv_sample = unsplit(fold_ids[-fold_index], rep(1:(cv_folds), cv_folds-1))
        # Fold that will be used for test
        cv_test = unsplit(fold_ids[-fold_index], 1:(cv_folds))
        # COnstruct the random forest
        ranger_rf_obj = ranger(x = x_pred[cv_sample,],
                               y = y_resp[cv_sample],
                               mtry = ceiling(ncol(x_pred) * param_grid[param_index, "frac_pred"]),
                               min.node.size = param_grid[param_index, "min_node_size"],
                               num.trees = param_grid[param_index, "num_trees"],
                               num.threads = 1)
        # Get the predictions
        ranger_predictions = predict(ranger_rf_obj, x_pred[cv_sample,])$predictions
        # Calculate the MSE
        mse_inst = 1 / length(cv_sample) * sum((y_resp[cv_sample] - ranger_predictions)^2)
        # Store the MSEs across each fold
        cv_mse_vec[fold_index] = mse_inst
      }
      # Store the average of the MSEs across eeach fold
      avg_mse_vals[param_index] = mean(cv_mse_vec) 
    }
    # Taking the minimum MSE or the first occurrence of the minimum MSE
    min_mse_position = which.min(avg_mse_vals)
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_ranger_obj = ranger(x = x_pred, y = y_resp,
                              mtry = ceiling(ncol(x_pred) * param_grid[min_mse_position, "frac_pred"]),
                              min.node.size = param_grid[min_mse_position, "min_node_size"],
                              num.trees = param_grid[min_mse_position, "num_trees"],
                              num.threads = 1)
    return(final_ranger_obj)
  }
  
  # Neural Network
  nnet_cv = function(x_pred, y_resp, 
                     parameter_grid = list("size" = c(3, 5, 10, 20),
                                           "decay" = c(0.001, 0.01, 0.1) ),
                     cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = expand.grid(parameter_grid)
    
    # Construct the XGBoosted trees
    nnet_control= caret::trainControl(method = "cv", number = cv_folds)
    # Train XGBoosted trees
    nnet_obj = caret::train(x = x_pred, y = as.vector(y_resp),
                            method = "nnet", tuneGrid = param_grid,
                            trControl = nnet_control,
                            maxit = 500,
                            linout=FALSE,
                            trace=FALSE)
    
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_nnet_obj = nnet_obj$finalModel
    return(final_nnet_obj)
  }
  
  # generate the testing data
  testing_model_data = model_generating_function(N = testing_n, rho = model_rho, 
                                                 max_pred = model_predictor_quant, 
                                                 dat.seed = testing_data_seed,
                                                 noise_1 = model_noise)
  
  # Getting variable names:
  all_variable_names = colnames(testing_model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  
  # sdc_data_formula = sdc_form(target_variables = organized_vars$response, 
  #                             predictor_variables = organized_vars$predictors)
  
  # Calculate MSE function
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  # Use futures with callr to bypass R connections limits
  # plan(multisession, workers = core_alloc)
  plan(callr, workers = cores)
  
  results = foreach(samp_size_index = 1:length(samp_n), .combine = "rbind", 
                    .options.future= list(packages = c( "Matrix",  "ranger", 
                                   "nnet", "e1071", "xgboost", "methods", "glmnet",
                                   "tidyr", "caret")),
                    .inorder = FALSE) %:%
    foreach( iteration = m_start:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",
            .inorder = FALSE) %dofuture% {
              
              # setting the seed
              dat.seed = iteration + samp_n[samp_size_index]
              set.seed(dat.seed)
              
              # Training dataset
              # Generated with the same parameters sxcluding the seed
              training_model_data = 
                model_generating_function(N = samp_n[samp_size_index], 
                                          rho = model_rho, 
                                          max_pred = model_predictor_quant, 
                                          dat.seed = dat.seed,
                                          noise_1 = model_noise)
              
              model_data = rbind(training_model_data, 
                                 testing_model_data)
              
              # Performing SDC GADP method
              privatized_dataframe = 
                perturbation_func(unprotected_data = model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              
              privatized_responses = privatized_dataframe[, target_variables_indices]
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,organized_vars$predictors]
              
              # Get the sample from privatized df
              drawing = 1:samp_n[samp_size_index]
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              # Adding intercept term for linear regression
              x_dat_holdout = cbind(1,as.matrix(privatized_dataframe)[-drawing,organized_vars$predictors])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after cgapd, but preparing for case when both x and y are randomized
              x_dat = as.matrix(cbind(1,
                                      sample_private_data[,organized_vars$predictors]))
              
              
              # Get the predictors for the training set for unprotected data
              x_dat_unprotected = 
                as.matrix(cbind(1, sample_unprotected_data[,organized_vars$predictors]))
              
              # Get the predictors for the testing set for unprotected data
              x_dat_holdout_unprotected = 
                as.matrix(cbind(1, model_data[-drawing,organized_vars$predictors]))
              
              
              
              ### CGAPD Used----
              
              # Linear Regression on data after cgapd
              betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(current_response[drawing])
              
              
              linear_reg_pred = as.vector(x_dat_holdout %*% betas_linear)
              
              # Get the predictions for the training data
              
              training_lin_reg_pred = as.vector(x_dat %*% betas_linear)
              
              
              # Get the mse for the training data
              
              
              training_lin_reg_mse = calc_mse(as.matrix(current_response[drawing]),
                                              training_lin_reg_pred)
              
              # Get the mse for the test data
              
              
              linear_reg_mse = calc_mse(orig_data_response[-drawing],
                                        linear_reg_pred)
              
              # LASSO on data after cgapd ---------------------
              
              # Build the LASSO regression model
              # using cross validation
              lasso_regress_model = cv.glmnet( 
                x = as.matrix(sample_private_data[,organized_vars$predictors]),
                y = current_response[drawing])
              
              
              lasso_pred = predict(lasso_regress_model, 
                                   as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]),
                                   s = "lambda.min")
              
              # get predictions on the training data
              training_lasso_pred = predict(lasso_regress_model, 
                                            as.matrix(privatized_dataframe[drawing,organized_vars$predictors]),
                                            s = "lambda.min")
              
              # Get the MSE for the training data
              training_lasso_mse = 
                calc_mse(orig_data_response[drawing],
                         training_lasso_pred)
              
              # # MSE for the testing data
              
              lasso_mse = calc_mse(orig_data_response[-drawing],
                                   lasso_pred)
              
              # Neural Net Regression on data after cgapd
              cgapd_nnet = nnet_cv(x_pred = x_dat[,-1], 
                                   y_resp = current_response[drawing])
              
              
              cgapd_nnet_reg_pred = predict(cgapd_nnet,x_dat_holdout[,-1])
              
              # training data predictions
              training_cgapd_nnet_reg_pred = predict(cgapd_nnet,x_dat[,-1])
              
              # training data MSE
              training_cgapd_nnet_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_nnet_reg_pred)
              
              
              # testing data MSE
              cgapd_nnet_reg_mse = calc_mse(orig_data_response[-drawing],
                                            cgapd_nnet_reg_pred)
              
              ## SVM Runs for CGADP -------
              
              # SVM Radial Regression on data after cgapd
              cgapd_radial_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                               gamma = seq(0.001,0.1,length.out = 2), 
                                               cost = seq(0.1,1,length.out = 2),
                                               
                                               kernel = c("radial"))
              
              cgapd_radial_svm = cgapd_radial_svm_tune$best.model
              
              cgapd_radial_svm_reg_pred = predict(cgapd_radial_svm,x_dat_holdout[,-1])
              
              # Get the training predictions of the radial svm
              training_cgapd_radial_svm_reg_pred =
                predict(cgapd_radial_svm, x_dat[,-1])
              
              # Get the training mse for radial svm
              training_cgapd_radial_svm_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_radial_svm_reg_pred)
              
              # Testing data MSE
              cgapd_radial_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                  cgapd_radial_svm_reg_pred)
              
              
              
              # SVM Linear Regression on data after cgapd
              cgapd_linear_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                               
                                               cost = seq(0.1,1,length.out = 2),
                                               
                                               kernel = c("linear"))
              
              cgapd_linear_svm = cgapd_linear_svm_tune$best.model
              
              # Predictions for the training data for Linear SVM
              training_cgapd_linear_svm_reg_pred =
                predict(cgapd_linear_svm,x_dat_holdout[,-1])
              
              # Training MSE for Linear SVM
              training_cgapd_linear_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_linear_svm_reg_pred)
              
              # Predictions for the test dataset
              cgapd_linear_svm_reg_pred = predict(cgapd_linear_svm,x_dat_holdout[,-1])
              
              # Testing data MSE
              cgapd_linear_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                  cgapd_linear_svm_reg_pred)
              
              # SVM Polynomial Regression on data after gapd
              cgapd_poly_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                             gamma = seq(0.001,0.1,length.out = 2), 
                                             cost = seq(0.1,1,length.out = 2),
                                             degree = seq(1,3, length.out=2),
                                             kernel = c("polynomial"),
                                             cross = 10)
              
              cgapd_poly_svm = cgapd_poly_svm_tune$best.model
              
              # Training data polynomial SVM predictions
              training_cgapd_poly_svm_reg_pred = predict(cgapd_poly_svm,x_dat[,-1])
              
              # Training data MSE polynomial SVM
              training_cgapd_poly_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_poly_svm_reg_pred)
              
              # Predictions for testing dataset
              cgapd_poly_svm_reg_pred = predict(cgapd_poly_svm,x_dat_holdout[,-1])
              
              # MSE for testing dataset for polynomial kernel SVM
              cgapd_poly_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                cgapd_poly_svm_reg_pred)
              
              ## Random Forest Run ---------------------------
              
              # Random forest on data after cgapd
              num_x_vars = length(organized_vars$predictors)
              
              ranger_rf_regress_model = ranger_cv( 
                x_pred = as.matrix(sample_private_data[,organized_vars$predictors]),
                y_resp = current_response[drawing])
              
              # Training dataset predicions for the random forest
              training_ranger_rf_pred = 
                predict(ranger_rf_regress_model, 
                        privatized_dataframe[drawing,organized_vars$predictors])$predictions
              
              # training dataset MSE for the radom forest
              training_ranger_rf_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_ranger_rf_pred)
              
              
              # Testing dataset prediction for random forest
              ranger_rf_pred = predict(ranger_rf_regress_model, 
                                       privatized_dataframe[-drawing,organized_vars$predictors])$predictions
              
              # Testing dataset MSE for random forest
              ranger_rf_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                       ranger_rf_pred)
              
              # XGBT on data after cgapd
              xgbt_regress_model = xgbt_cv( 
                x_pred = as.matrix(sample_private_data[,organized_vars$predictors]),
                y_resp = current_response[drawing])
              
              # Training data predictions for the XGB tree
              training_xgbt_pred = 
                predict(xgbt_regress_model,
                        as.matrix(privatized_dataframe[drawing,organized_vars$predictors]))
              
              # training data MSE for the XGB tree
              training_xgbt_mse = 
                calc_mse(as.matrix(current_response[drawing]) ,
                         training_xgbt_pred)
              
              # Testing data predicitons for the XGB tree
              xgbt_pred = predict(xgbt_regress_model, 
                                  as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]))
              
              # Testing data MSE for the XGB tree
              xgbt_mse = calc_mse(orig_data_response[-drawing],
                                  xgbt_pred)
              
              ##
              ## Before CGAPD (On original data) -----------------------------------
              # Linear Regression on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              
              # Beta coefficients for the original data
              orig_data_betas_linear = solve(t(x_dat_unprotected) %*% 
                                               x_dat_unprotected) %*% t(x_dat_unprotected) %*% 
                as.matrix(sample_orig_data_response)
              
              # Predictions on the training data no GADP
              training_lin_reg_unprivat_pred = 
                as.vector(x_dat_unprotected %*% orig_data_betas_linear)
              
              
              # MSE for training data for linear regression, no GADP
              training_linear_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_lin_reg_unprivat_pred)
              
              # Testing data MSE for linear regression no ADP
              linear_reg_unprivatized_pred = as.vector(x_dat_holdout_unprotected %*% 
                                                         orig_data_betas_linear)
              
              linear_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_reg_unprivatized_pred)
              
              # LASSO on data before gapd (original data)
              
              
              lasso_regress_model_orig_data =
                cv.glmnet( x = as.matrix(model_data[drawing,organized_vars$predictors]), # Get rid of intercept term
                           y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # Training data predictions for LASSO
              
              training_lasso_pred_orig_data = 
                predict(lasso_regress_model_orig_data,
                        as.matrix(model_data[drawing,organized_vars$predictors]), 
                        s = "lambda.min")
              
              # Training data MSE for LASSO
              training_lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_lasso_pred_orig_data)
              
              # Predictions for the test dataset for the LASSO model
              lasso_pred_orig_data = predict(lasso_regress_model_orig_data,
                                             as.matrix(model_data[-drawing,organized_vars$predictors]), 
                                             s = "lambda.min")
              # MSE for the LASSO model for the test data
              lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         lasso_pred_orig_data)
              
              
              # Neural Net before GAPD 
              
              no_cgapd_nnet = nnet_cv(x_pred = x_dat_unprotected[,-1], 
                                      y_resp = sample_orig_data_response)
              
              # The predictions of the training set for neural network
              training_nnet_reg_unprivatized_pred = predict(no_cgapd_nnet, 
                                                            x_dat_unprotected[,-1])
              
              # The MSE for the training set for the neural network
              training_nnet_reg_unprivatized_mse = 
                calc_mse(orig_data_response[drawing],
                         training_nnet_reg_unprivatized_pred)
              
              # Predicitons on testing set for neural net
              nnet_reg_unprivatized_pred = predict(no_cgapd_nnet, 
                                                   x_dat_holdout_unprotected[,-1])
              
              # MSE for testing set for neural network
              nnet_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         nnet_reg_unprivatized_pred)
              
              # SVM Radial before GAPD
              no_cgapd_radial_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                  y = sample_orig_data_response,
                                                  gamma = seq(0.001,0.1,length.out = 2), 
                                                  cost = seq(0.1,1,length.out = 2),
                                                  
                                                  kernel = c("radial"))
              
              # Get the best model for Radial SVM
              no_cgapd_radial_svm = no_cgapd_radial_svm_tune$best.model
              
              # Predictions on the training set for radial SVM
              training_radial_svm_reg_unprivatized_pred = 
                predict(no_cgapd_radial_svm, x_dat_unprotected[,-1])
              
              # The MSE for the training set for radial SVM
              training_radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_radial_svm_reg_unprivatized_pred)
              
              # Get the predictions of Radial SVM for the testing set
              radial_svm_reg_unprivatized_pred = 
                predict(no_cgapd_radial_svm, x_dat_holdout_unprotected[,-1])
              
              # Get the MSE of the Radial SVM for the testing set
              radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         radial_svm_reg_unprivatized_pred)
              
              # Linear SVM before GAPD
              no_cgapd_linear_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                  y = sample_orig_data_response,
                                                  cost = seq(0.1,1,length.out = 2),
                                                  
                                                  kernel = c("linear"))
              # The Linear SVM best tuned model
              no_cgapd_linear_svm = no_cgapd_linear_svm_tune$best.model
              
              # The predictions on the training set, Linear SVM
              training_linear_svm_reg_unprivatized_pred = 
                predict(no_cgapd_linear_svm, x_dat_unprotected[,-1])
              
              # the MSE on the training set, Linear SVM
              training_linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_linear_svm_reg_unprivatized_pred)
              
              # The predictions for the testing set Linear SVM
              linear_svm_reg_unprivatized_pred = 
                predict(no_cgapd_linear_svm, x_dat_holdout_unprotected[,-1])
              
              # The MSE for the testing set Linear SVM
              linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_svm_reg_unprivatized_pred)
              
              
              # Polynomial SVM before GAPD
              no_cgapd_poly_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                y = sample_orig_data_response,
                                                gamma = seq(0.001,0.1,length.out = 2), 
                                                cost = seq(0.1,1,length.out = 2),
                                                degree = seq(1,3, length.out=2),
                                                kernel = c("polynomial"),
                                                cross = 10)
              # Get the best tuned polynomial SVM
              no_cgapd_poly_svm = no_cgapd_poly_svm_tune$best.model
              
              # The training set predicitons for the polynomial SVM
              training_poly_svm_reg_unprivatized_pred = 
                predict(no_cgapd_poly_svm, x_dat_unprotected[,-1])
              
              # the training set MSE for the polynomial SVM
              training_poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_poly_svm_reg_unprivatized_pred)
              
              # The testing set predicitons using the polynomial SVM
              poly_svm_reg_unprivatized_pred = 
                predict(no_cgapd_poly_svm, x_dat_holdout_unprotected[,-1])
              
              # The testing set MSE for the polynomial SVM
              poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         poly_svm_reg_unprivatized_pred)
              
              # Random forest on data before cgapd
              ranger_rf_model_orig_data = ranger_cv( 
                x_pred = model_data[drawing,organized_vars$predictors],
                y_resp = orig_data_response[drawing])
              
              # The training set predictions from random forest
              training_ranger_rf_pred_orig_data = 
                predict(ranger_rf_model_orig_data,
                        model_data[drawing,organized_vars$predictors])$predictions
              
              # The training set MSE for the random forest
              training_ranger_rf_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_ranger_rf_pred_orig_data)
              
              # Testing set predictions for random forest
              ranger_rf_pred_orig_data = predict(ranger_rf_model_orig_data,
                                                 model_data[-drawing,organized_vars$predictors])$predictions
              # Testing set MSE for random forest
              ranger_rf_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         ranger_rf_pred_orig_data)
              
              # XGBT on data before cgapd
              xgbt_regress_model_orig_data =
                xgbt_cv(x_pred = as.matrix(sample_unprotected_data[,organized_vars$predictors]),
                        y_resp = orig_data_response[drawing])
              
              # The training set predictions for XGB tree
              training_xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(sample_unprotected_data[,organized_vars$predictors]))
              # Training set MSE for XGB tree for 
              training_xgbt_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_xgbt_pred_orig_data)
              
              # The testing set XGB tree predictions
              xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[-drawing,organized_vars$predictors]))
              
              # The testing set MSE for XGB tree
              xgbt_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         xgbt_pred_orig_data)
              
              
              #print("Constructing the stacked ensemble")
              ###
              ### Stacked Ensemble Equal Weighting ----
              # CGAPD
              avg_stacked_ensemble_predictions_cgapd = (linear_reg_pred  + 
                                                          lasso_pred  +  cgapd_nnet_reg_pred + 
                                                          cgapd_radial_svm_reg_pred  + 
                                                          cgapd_linear_svm_reg_pred  + 
                                                          cgapd_poly_svm_reg_pred  +
                                                          ranger_rf_pred  + xgbt_pred) / 8
              
              avg_stacked_ensemble_cgapd_mse =
                calc_mse(orig_data_response[-drawing],
                         avg_stacked_ensemble_predictions_cgapd)
              # No GAPD
              avg_stacked_ensemble_pred_no_cgapd = 
                (linear_reg_unprivatized_pred  + 
                   lasso_pred_orig_data  + 
                   nnet_reg_unprivatized_pred  + 
                   radial_svm_reg_unprivatized_pred  + 
                   linear_svm_reg_unprivatized_pred  + 
                   poly_svm_reg_unprivatized_pred  +
                   ranger_rf_pred_orig_data  + xgbt_pred_orig_data) / 8
              
              avg_stacked_ensemble_no_cgapd_mse =
                calc_mse(orig_data_response[-drawing],
                         avg_stacked_ensemble_pred_no_cgapd)
              
              
              ### Stacked Ensemble Equal Weighting by Proportion of variance Explained ----------------
              # under CGADP
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_cgapd = 
                c(training_lin_reg_mse, training_lasso_mse, 
                  training_cgapd_nnet_reg_mse, training_cgapd_radial_svm_reg_mse,
                  training_cgapd_linear_svm_reg_mse, 
                  training_cgapd_poly_svm_reg_mse, training_ranger_rf_mse,
                  training_xgbt_mse)
              
              # The models with the lowest MSE get the highest weight
              prop_var_exp_cgapd_weighting = 
                (training_set_mses_cgapd / sum(training_set_mses_cgapd))^(-1)
              
              prop_var_exp_cgapd_weighting = prop_var_exp_cgapd_weighting / 
                sum(prop_var_exp_cgapd_weighting)
              
              
              
              # Get the predictions for the stacked ensemble for the prop
              # of variance explained
              prop_var_stacked_ens_cgapd_pred = 
                linear_reg_pred * prop_var_exp_cgapd_weighting[1] +
                lasso_pred * prop_var_exp_cgapd_weighting[2] + 
                cgapd_nnet_reg_pred * prop_var_exp_cgapd_weighting[3] + 
                cgapd_radial_svm_reg_pred * prop_var_exp_cgapd_weighting[4] + 
                cgapd_linear_svm_reg_pred * prop_var_exp_cgapd_weighting[5] + 
                cgapd_poly_svm_reg_pred * prop_var_exp_cgapd_weighting[6] +
                ranger_rf_pred * prop_var_exp_cgapd_weighting[7] + 
                xgbt_pred * prop_var_exp_cgapd_weighting[8]
              
              
              
              # Get the MSE
              prop_var_stacked_ens_cgapd_mse = 
                calc_mse(orig_data_response[-drawing],
                         prop_var_stacked_ens_cgapd_pred)
              
              
              # Get the Stacked Ensemble (Prop. of Var. explained) with No GADP 
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_no_cgapd = 
                c(training_linear_reg_mse_orig_data, 
                  training_lasso_mse_orig_data, 
                  training_nnet_reg_unprivatized_mse, 
                  training_radial_svm_reg_mse_orig_data,
                  training_linear_svm_reg_mse_orig_data, 
                  training_poly_svm_reg_mse_orig_data, 
                  training_ranger_rf_mse_orig_data,
                  training_xgbt_mse_orig_data)
              
              # weighting for the SE
              prop_var_exp_no_cgapd_weighting = 
                (training_set_mses_no_cgapd / sum(training_set_mses_no_cgapd))^(-1)
              
              prop_var_exp_no_cgapd_weighting = prop_var_exp_no_cgapd_weighting /
                sum(prop_var_exp_no_cgapd_weighting)
              
              
              # get the predictions
              prop_var_stacked_ens_no_cgapd_pred = 
                linear_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[1] +
                lasso_pred_orig_data * prop_var_exp_no_cgapd_weighting[2] + 
                nnet_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[3] + 
                radial_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[4] + 
                linear_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[5] + 
                poly_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[6] +
                ranger_rf_pred_orig_data * prop_var_exp_no_cgapd_weighting[7] + 
                xgbt_pred_orig_data * prop_var_exp_no_cgapd_weighting[8]
              
              # Get the MSE
              prop_var_stacked_ens_no_cgapd_mse = 
                calc_mse(orig_data_response[-drawing],
                         prop_var_stacked_ens_no_cgapd_pred)
              
              
              # Compile MSEs into a vector
              cgapd_mse_vector = 
                c(linear_reg_mse, lasso_mse, cgapd_nnet_reg_mse,
                  cgapd_radial_svm_reg_mse, cgapd_linear_svm_reg_mse,
                  cgapd_poly_svm_reg_mse, ranger_rf_mse,
                  xgbt_mse, avg_stacked_ensemble_cgapd_mse, 
                  prop_var_stacked_ens_cgapd_mse)
              
              no_cgapd_mse_vector =
                c(linear_reg_mse_orig_data, lasso_mse_orig_data,
                  nnet_reg_mse_orig_data, radial_svm_reg_mse_orig_data,
                  linear_svm_reg_mse_orig_data, poly_svm_reg_mse_orig_data,
                  ranger_rf_mse_orig_data, xgbt_mse_orig_data,
                  avg_stacked_ensemble_no_cgapd_mse, 
                  prop_var_stacked_ens_no_cgapd_mse)
              
              #print("Constructing the final results")
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "training_size","testing_size", "y", 
                                   "LR", "LASSO", "NNET", "Radial_SVM",
                                   "Linear_SVM", "Poly_SVM", "RF", "XGBTree",
                                   "Stacked_equal", "Stacked_prop","cgadp", "theta")
              
              
              # construct the dataframe of results
              # With CGADP
              
              estimator_reg_df = matrix( rep(0, length(data_frame_names) * 2), nrow = 2, ncol = length(data_frame_names), byrow = TRUE)
              estimator_reg_df[1,] = c(dat.seed,iteration, samp_n[samp_size_index], testing_n, 
                                       response_index, cgapd_mse_vector, 1, theta_value)
              
              # Adding results for LASSO without CGAPD
              estimator_reg_df[2,] = c(dat.seed, iteration, 
                                       samp_n[samp_size_index], testing_n,
                                       response_index, no_cgapd_mse_vector, 0, theta_value)
              
              
              colnames(estimator_reg_df) = data_frame_names
              
              return(estimator_reg_df)
            }
  #### Stopping Cluster -------------
  
  results = data.frame(results)
  
  # Add name of the estimators
  
  return(results)
}

# CGADP (Low Poly, less expensive for poly_SVM) --------
est_evaluation_stacked_ensemb_cgadp_cv_low_poly = function(# the dataset
  m_start = 1, # Primarily used for debugging determines which iteration to start from
  m = 1000, # Number of Monte Carlo simulations
  samp_n = c(500, 1000, 2000), # sample sizes to be used for training set
  cores = 2, # Number of CPU threads/ logical processors to use
  response_prefix = "y", # prefix for response variables
  acquire_variable_names = get_variables, #Function for getting the variables
  perturbation_func, # CGADP or GADP function
  theta_value = 0.4, # privacy parameter value
  model_generating_function, # function to generate the model
  testing_data_seed = 30000, # Seed for the testing set
  model_predictor_quant = 10, # Quantity of the predictors for the model
  model_rho = 0.2, # correlation of predictors in the model
  model_noise = 0.2, # Noise to be used in the model
  testing_n = 500) # Size of the testing set
{
  
  # Cross-Validation Funcitons -----------
  # XGBoosted Trees
  xgbt_cv = function(x_pred, y_resp, 
                     parameter_grid = list("nrounds" = c(10,20, 40),
                                           "max_depth" = c(2,4),
                                           "eta" = c(0.2, 0.4, 0.6),
                                           "gamma" = c(0, 0.5, 1),
                                           "colsample_bytree" = c(0.2, 0.5, 1),
                                           "min_child_weight" = c(0.5, 1),
                                           "subsample" = c(0.4, 0.5, 0.6)),
                     cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = expand.grid(parameter_grid)
    
    # Construct the XGBoosted trees
    xgbt_control= caret::trainControl(method = "cv", number = cv_folds)
    # Train XGBoosted trees
    xgbt_obj = caret::train(x = x_pred, y = as.vector(y_resp),
                            method = "xgbTree", tuneGrid = param_grid,
                            trControl = xgbt_control,
                            nthread = 1,
                            verbose=0,
                            print_every_n =0)
    
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_xgbt_obj = xgbt_obj$finalModel
    return(final_xgbt_obj)
  }
  
  # Random forest
  ranger_cv = function(x_pred, y_resp, num_trees = 500,
                       frac_pred = c(1/3, 1/4), 
                       min_node_size = c(5,10),
                       cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = as.matrix(expand_grid(num_trees, frac_pred, min_node_size))
    # Set up folds to be used in cross-validation
    fold_ids = split(1:nrow(x_pred), 1:cv_folds)
    # Vector of average mse for each combination of hyperparameters
    avg_mse_vals = numeric(nrow(param_grid))
    # Loop through each set of hyperparameters
    for(param_index in 1:nrow(param_grid))
    {
      # Vector to store mses per each test fold
      cv_mse_vec = numeric(cv_folds)
      # Cycle through each potential test fold
      for(fold_index in 1:cv_folds)
      {
        # Folds that will be used for training
        cv_sample = unsplit(fold_ids[-fold_index], rep(1:(cv_folds), cv_folds-1))
        # Fold that will be used for test
        cv_test = unsplit(fold_ids[-fold_index], 1:(cv_folds))
        # COnstruct the random forest
        ranger_rf_obj = ranger(x = x_pred[cv_sample,],
                               y = y_resp[cv_sample],
                               mtry = ceiling(ncol(x_pred) * param_grid[param_index, "frac_pred"]),
                               min.node.size = param_grid[param_index, "min_node_size"],
                               num.trees = param_grid[param_index, "num_trees"],
                               num.threads = 1)
        # Get the predictions
        ranger_predictions = predict(ranger_rf_obj, x_pred[cv_sample,])$predictions
        # Calculate the MSE
        mse_inst = 1 / length(cv_sample) * sum((y_resp[cv_sample] - ranger_predictions)^2)
        # Store the MSEs across each fold
        cv_mse_vec[fold_index] = mse_inst
      }
      # Store the average of the MSEs across eeach fold
      avg_mse_vals[param_index] = mean(cv_mse_vec) 
    }
    # Taking the minimum MSE or the first occurrence of the minimum MSE
    min_mse_position = which.min(avg_mse_vals)
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_ranger_obj = ranger(x = x_pred, y = y_resp,
                              mtry = ceiling(ncol(x_pred) * param_grid[min_mse_position, "frac_pred"]),
                              min.node.size = param_grid[min_mse_position, "min_node_size"],
                              num.trees = param_grid[min_mse_position, "num_trees"],
                              num.threads = 1)
    return(final_ranger_obj)
  }
  
  # Neural Network
  nnet_cv = function(x_pred, y_resp, 
                     parameter_grid = list("size" = c(3, 5, 10, 20),
                                           "decay" = c(0.001, 0.01, 0.1) ),
                     cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = expand.grid(parameter_grid)
    
    # Construct the XGBoosted trees
    nnet_control= caret::trainControl(method = "cv", number = cv_folds)
    # Train XGBoosted trees
    nnet_obj = caret::train(x = x_pred, y = as.vector(y_resp),
                            method = "nnet", tuneGrid = param_grid,
                            trControl = nnet_control,
                            maxit = 500,
                            linout=FALSE,
                            trace=FALSE)
    
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_nnet_obj = nnet_obj$finalModel
    return(final_nnet_obj)
  }
  
  # generate the testing data
  testing_model_data = model_generating_function(N = testing_n, rho = model_rho, 
                                                 max_pred = model_predictor_quant, 
                                                 dat.seed = testing_data_seed,
                                                 noise_1 = model_noise)
  
  # Getting variable names:
  all_variable_names = colnames(testing_model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  
  # sdc_data_formula = sdc_form(target_variables = organized_vars$response, 
  #                             predictor_variables = organized_vars$predictors)
  
  # Calculate MSE function
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  # Use futures with callr to bypass R connections limits
  # plan(multisession, workers = core_alloc)
  plan(callr, workers = cores)
  
  results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
                    .options.future= list(packages = c( "Matrix",  "ranger", 
                                                         "nnet", "e1071", "xgboost", "methods", "glmnet",
                                                         "tidyr", "caret")),
                    .inorder = FALSE) %:%
    foreach( iteration = m_start:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",
            .inorder = FALSE) %dofuture% {
              
              # setting the seed
              dat.seed = iteration + samp_n[samp_size_index]
              set.seed(dat.seed)
              
              # Training dataset
              # Generated with the same parameters sxcluding the seed
              training_model_data = 
                model_generating_function(N = samp_n[samp_size_index], 
                                          rho = model_rho, 
                                          max_pred = model_predictor_quant, 
                                          dat.seed = dat.seed,
                                          noise_1 = model_noise)
              
              model_data = rbind(training_model_data, 
                                 testing_model_data)
              
              # Performing SDC GADP method
              privatized_dataframe = 
                perturbation_func(unprotected_data = model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              
              privatized_responses = privatized_dataframe[, target_variables_indices]
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,organized_vars$predictors]
              
              # Get the sample from privatized df
              drawing = 1:n[samp_size_index]
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              # Adding intercept term for linear regression
              x_dat_holdout = cbind(1,as.matrix(privatized_dataframe)[-drawing,organized_vars$predictors])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after cgapd, but preparing for case when both x and y are randomized
              x_dat = as.matrix(cbind(1,
                                      sample_private_data[,organized_vars$predictors]))
              
              
              # Get the predictors for the training set for unprotected data
              x_dat_unprotected = 
                as.matrix(cbind(1, sample_unprotected_data[,organized_vars$predictors]))
              
              # Get the predictors for the testing set for unprotected data
              x_dat_holdout_unprotected = 
                as.matrix(cbind(1, model_data[-drawing,organized_vars$predictors]))
              
              
              
              ### CGAPD Used----
              
              # Linear Regression on data after cgapd
              betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(current_response[drawing])
              
              
              linear_reg_pred = as.vector(x_dat_holdout %*% betas_linear)
              
              # Get the predictions for the training data
              
              training_lin_reg_pred = as.vector(x_dat %*% betas_linear)
              
              
              # Get the mse for the training data
              
              
              training_lin_reg_mse = calc_mse(as.matrix(current_response[drawing]),
                                              training_lin_reg_pred)
              
              # Get the mse for the test data
              
              
              linear_reg_mse = calc_mse(orig_data_response[-drawing],
                                        linear_reg_pred)
              
              # LASSO on data after cgapd ---------------------
              
              # Build the LASSO regression model
              # using cross validation
              lasso_regress_model = cv.glmnet( 
                x = as.matrix(sample_private_data[,organized_vars$predictors]),
                y = current_response[drawing])
              
              
              lasso_pred = predict(lasso_regress_model, 
                                   as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]),
                                   s = "lambda.min")
              
              # get predictions on the training data
              training_lasso_pred = predict(lasso_regress_model, 
                                            as.matrix(privatized_dataframe[drawing,organized_vars$predictors]),
                                            s = "lambda.min")
              
              # Get the MSE for the training data
              training_lasso_mse = 
                calc_mse(orig_data_response[drawing],
                         training_lasso_pred)
              
              # # MSE for the testing data
              
              lasso_mse = calc_mse(orig_data_response[-drawing],
                                   lasso_pred)
              
              # Neural Net Regression on data after cgapd
              cgapd_nnet = nnet_cv(x_pred = x_dat[,-1], 
                                   y_resp = current_response[drawing])
              
              
              cgapd_nnet_reg_pred = predict(cgapd_nnet,x_dat_holdout[,-1])
              
              # training data predictions
              training_cgapd_nnet_reg_pred = predict(cgapd_nnet,x_dat[,-1])
              
              # training data MSE
              training_cgapd_nnet_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_nnet_reg_pred)
              
              
              # testing data MSE
              cgapd_nnet_reg_mse = calc_mse(orig_data_response[-drawing],
                                            cgapd_nnet_reg_pred)
              
              ## SVM Runs for CGADP -------
              
              # SVM Radial Regression on data after cgapd
              cgapd_radial_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                               gamma = seq(0.001,0.1,length.out = 2), 
                                               cost = seq(0.1,1,length.out = 2),
                                               
                                               kernel = c("radial"))
              
              cgapd_radial_svm = cgapd_radial_svm_tune$best.model
              
              cgapd_radial_svm_reg_pred = predict(cgapd_radial_svm,x_dat_holdout[,-1])
              
              # Get the training predictions of the radial svm
              training_cgapd_radial_svm_reg_pred =
                predict(cgapd_radial_svm, x_dat[,-1])
              
              # Get the training mse for radial svm
              training_cgapd_radial_svm_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_radial_svm_reg_pred)
              
              # Testing data MSE
              cgapd_radial_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                  cgapd_radial_svm_reg_pred)
              
              
              
              # SVM Linear Regression on data after cgapd
              cgapd_linear_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                               
                                               cost = seq(0.1,1,length.out = 2),
                                               
                                               kernel = c("linear"))
              
              cgapd_linear_svm = cgapd_linear_svm_tune$best.model
              
              # Predictions for the training data for Linear SVM
              training_cgapd_linear_svm_reg_pred =
                predict(cgapd_linear_svm,x_dat_holdout[,-1])
              
              # Training MSE for Linear SVM
              training_cgapd_linear_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_linear_svm_reg_pred)
              
              # Predictions for the test dataset
              cgapd_linear_svm_reg_pred = predict(cgapd_linear_svm,x_dat_holdout[,-1])
              
              # Testing data MSE
              cgapd_linear_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                  cgapd_linear_svm_reg_pred)
              
              # SVM Polynomial Regression on data after gapd
              cgapd_poly_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                             gamma = seq(0.001,0.1,length.out = 2), 
                                             cost = seq(0.1,1,length.out = 2),
                                             degree = seq(1,2, length.out=2),
                                             kernel = c("polynomial"),
                                             cross = 10)
              
              cgapd_poly_svm = cgapd_poly_svm_tune$best.model
              
              # Training data polynomial SVM predictions
              training_cgapd_poly_svm_reg_pred = predict(cgapd_poly_svm,x_dat[,-1])
              
              # Training data MSE polynomial SVM
              training_cgapd_poly_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_poly_svm_reg_pred)
              
              # Predictions for testing dataset
              cgapd_poly_svm_reg_pred = predict(cgapd_poly_svm,x_dat_holdout[,-1])
              
              # MSE for testing dataset for polynomial kernel SVM
              cgapd_poly_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                cgapd_poly_svm_reg_pred)
              
              ## Random Forest Run ---------------------------
              
              # Random forest on data after cgapd
              num_x_vars = length(organized_vars$predictors)
              
              ranger_rf_regress_model = ranger_cv( 
                x_pred = as.matrix(sample_private_data[,organized_vars$predictors]),
                y_resp = current_response[drawing])
              
              # Training dataset predicions for the random forest
              training_ranger_rf_pred = 
                predict(ranger_rf_regress_model, 
                        privatized_dataframe[drawing,organized_vars$predictors])$predictions
              
              # training dataset MSE for the radom forest
              training_ranger_rf_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_ranger_rf_pred)
              
              
              # Testing dataset prediction for random forest
              ranger_rf_pred = predict(ranger_rf_regress_model, 
                                       privatized_dataframe[-drawing,organized_vars$predictors])$predictions
              
              # Testing dataset MSE for random forest
              ranger_rf_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                       ranger_rf_pred)
              
              # XGBT on data after cgapd
              xgbt_regress_model = xgbt_cv( 
                x_pred = as.matrix(sample_private_data[,organized_vars$predictors]),
                y_resp = current_response[drawing])
              
              # Training data predictions for the XGB tree
              training_xgbt_pred = 
                predict(xgbt_regress_model,
                        as.matrix(privatized_dataframe[drawing,organized_vars$predictors]))
              
              # training data MSE for the XGB tree
              training_xgbt_mse = 
                calc_mse(as.matrix(current_response[drawing]) ,
                         training_xgbt_pred)
              
              # Testing data predicitons for the XGB tree
              xgbt_pred = predict(xgbt_regress_model, 
                                  as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]))
              
              # Testing data MSE for the XGB tree
              xgbt_mse = calc_mse(orig_data_response[-drawing],
                                  xgbt_pred)
              
              ##
              ## Before CGAPD (On original data) -----------------------------------
              # Linear Regression on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              
              # Beta coefficients for the original data
              orig_data_betas_linear = solve(t(x_dat_unprotected) %*% 
                                               x_dat_unprotected) %*% t(x_dat_unprotected) %*% 
                as.matrix(sample_orig_data_response)
              
              # Predictions on the training data no GADP
              training_lin_reg_unprivat_pred = 
                as.vector(x_dat_unprotected %*% orig_data_betas_linear)
              
              
              # MSE for training data for linear regression, no GADP
              training_linear_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_lin_reg_unprivat_pred)
              
              # Testing data MSE for linear regression no ADP
              linear_reg_unprivatized_pred = as.vector(x_dat_holdout_unprotected %*% 
                                                         orig_data_betas_linear)
              
              linear_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_reg_unprivatized_pred)
              
              # LASSO on data before gapd (original data)
              
              
              lasso_regress_model_orig_data =
                cv.glmnet( x = as.matrix(model_data[drawing,organized_vars$predictors]), # Get rid of intercept term
                           y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # Training data predictions for LASSO
              
              training_lasso_pred_orig_data = 
                predict(lasso_regress_model_orig_data,
                        as.matrix(model_data[drawing,organized_vars$predictors]), 
                        s = "lambda.min")
              
              # Training data MSE for LASSO
              training_lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_lasso_pred_orig_data)
              
              # Predictions for the test dataset for the LASSO model
              lasso_pred_orig_data = predict(lasso_regress_model_orig_data,
                                             as.matrix(model_data[-drawing,organized_vars$predictors]), 
                                             s = "lambda.min")
              # MSE for the LASSO model for the test data
              lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         lasso_pred_orig_data)
              
              
              # Neural Net before GAPD 
              
              no_cgapd_nnet = nnet_cv(x_pred = x_dat_unprotected[,-1], 
                                      y_resp = sample_orig_data_response)
              
              # The predictions of the training set for neural network
              training_nnet_reg_unprivatized_pred = predict(no_cgapd_nnet, 
                                                            x_dat_unprotected[,-1])
              
              # The MSE for the training set for the neural network
              training_nnet_reg_unprivatized_mse = 
                calc_mse(orig_data_response[drawing],
                         training_nnet_reg_unprivatized_pred)
              
              # Predicitons on testing set for neural net
              nnet_reg_unprivatized_pred = predict(no_cgapd_nnet, 
                                                   x_dat_holdout_unprotected[,-1])
              
              # MSE for testing set for neural network
              nnet_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         nnet_reg_unprivatized_pred)
              
              # SVM Radial before GAPD
              no_cgapd_radial_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                  y = sample_orig_data_response,
                                                  gamma = seq(0.001,0.1,length.out = 2), 
                                                  cost = seq(0.1,1,length.out = 2),
                                                  
                                                  kernel = c("radial"))
              
              # Get the best model for Radial SVM
              no_cgapd_radial_svm = no_cgapd_radial_svm_tune$best.model
              
              # Predictions on the training set for radial SVM
              training_radial_svm_reg_unprivatized_pred = 
                predict(no_cgapd_radial_svm, x_dat_unprotected[,-1])
              
              # The MSE for the training set for radial SVM
              training_radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_radial_svm_reg_unprivatized_pred)
              
              # Get the predictions of Radial SVM for the testing set
              radial_svm_reg_unprivatized_pred = 
                predict(no_cgapd_radial_svm, x_dat_holdout_unprotected[,-1])
              
              # Get the MSE of the Radial SVM for the testing set
              radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         radial_svm_reg_unprivatized_pred)
              
              # Linear SVM before GAPD
              no_cgapd_linear_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                  y = sample_orig_data_response,
                                                  cost = seq(0.1,1,length.out = 2),
                                                  
                                                  kernel = c("linear"))
              # The Linear SVM best tuned model
              no_cgapd_linear_svm = no_cgapd_linear_svm_tune$best.model
              
              # The predictions on the training set, Linear SVM
              training_linear_svm_reg_unprivatized_pred = 
                predict(no_cgapd_linear_svm, x_dat_unprotected[,-1])
              
              # the MSE on the training set, Linear SVM
              training_linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_linear_svm_reg_unprivatized_pred)
              
              # The predictions for the testing set Linear SVM
              linear_svm_reg_unprivatized_pred = 
                predict(no_cgapd_linear_svm, x_dat_holdout_unprotected[,-1])
              
              # The MSE for the testing set Linear SVM
              linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_svm_reg_unprivatized_pred)
              
              
              # Polynomial SVM before GAPD
              no_cgapd_poly_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                y = sample_orig_data_response,
                                                gamma = seq(0.001,0.1,length.out = 2), 
                                                cost = seq(0.1,1,length.out = 2),
                                                degree = seq(1,2, length.out=2),
                                                kernel = c("polynomial"),
                                                cross = 10)
              # Get the best tuned polynomial SVM
              no_cgapd_poly_svm = no_cgapd_poly_svm_tune$best.model
              
              # The training set predicitons for the polynomial SVM
              training_poly_svm_reg_unprivatized_pred = 
                predict(no_cgapd_poly_svm, x_dat_unprotected[,-1])
              
              # the training set MSE for the polynomial SVM
              training_poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_poly_svm_reg_unprivatized_pred)
              
              # The testing set predicitons using the polynomial SVM
              poly_svm_reg_unprivatized_pred = 
                predict(no_cgapd_poly_svm, x_dat_holdout_unprotected[,-1])
              
              # The testing set MSE for the polynomial SVM
              poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         poly_svm_reg_unprivatized_pred)
              
              # Random forest on data before cgapd
              ranger_rf_model_orig_data = ranger_cv( 
                x_pred = model_data[drawing,organized_vars$predictors],
                y_resp = orig_data_response[drawing])
              
              # The training set predictions from random forest
              training_ranger_rf_pred_orig_data = 
                predict(ranger_rf_model_orig_data,
                        model_data[drawing,organized_vars$predictors])$predictions
              
              # The training set MSE for the random forest
              training_ranger_rf_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_ranger_rf_pred_orig_data)
              
              # Testing set predictions for random forest
              ranger_rf_pred_orig_data = predict(ranger_rf_model_orig_data,
                                                 model_data[-drawing,organized_vars$predictors])$predictions
              # Testing set MSE for random forest
              ranger_rf_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         ranger_rf_pred_orig_data)
              
              # XGBT on data before cgapd
              xgbt_regress_model_orig_data =
                xgbt_cv(x_pred = as.matrix(sample_unprotected_data[,organized_vars$predictors]),
                        y_resp = orig_data_response[drawing])
              
              # The training set predictions for XGB tree
              training_xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(sample_unprotected_data[,organized_vars$predictors]))
              # Training set MSE for XGB tree for 
              training_xgbt_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_xgbt_pred_orig_data)
              
              # The testing set XGB tree predictions
              xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[-drawing,organized_vars$predictors]))
              
              # The testing set MSE for XGB tree
              xgbt_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         xgbt_pred_orig_data)
              
              
              #print("Constructing the stacked ensemble")
              ###
              ### Stacked Ensemble Equal Weighting ----
              # CGAPD
              avg_stacked_ensemble_predictions_cgapd = (linear_reg_pred  + 
                                                          lasso_pred  +  cgapd_nnet_reg_pred + 
                                                          cgapd_radial_svm_reg_pred  + 
                                                          cgapd_linear_svm_reg_pred  + 
                                                          cgapd_poly_svm_reg_pred  +
                                                          ranger_rf_pred  + xgbt_pred) / 8
              
              avg_stacked_ensemble_cgapd_mse =
                calc_mse(orig_data_response[-drawing],
                         avg_stacked_ensemble_predictions_cgapd)
              # No GAPD
              avg_stacked_ensemble_pred_no_cgapd = 
                (linear_reg_unprivatized_pred  + 
                   lasso_pred_orig_data  + 
                   nnet_reg_unprivatized_pred  + 
                   radial_svm_reg_unprivatized_pred  + 
                   linear_svm_reg_unprivatized_pred  + 
                   poly_svm_reg_unprivatized_pred  +
                   ranger_rf_pred_orig_data  + xgbt_pred_orig_data) / 8
              
              avg_stacked_ensemble_no_cgapd_mse =
                calc_mse(orig_data_response[-drawing],
                         avg_stacked_ensemble_pred_no_cgapd)
              
              
              ### Stacked Ensemble Equal Weighting by Proportion of variance Explained ----------------
              # under CGADP
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_cgapd = 
                c(training_lin_reg_mse, training_lasso_mse, 
                  training_cgapd_nnet_reg_mse, training_cgapd_radial_svm_reg_mse,
                  training_cgapd_linear_svm_reg_mse, 
                  training_cgapd_poly_svm_reg_mse, training_ranger_rf_mse,
                  training_xgbt_mse)
              
              # The models with the lowest MSE get the highest weight
              prop_var_exp_cgapd_weighting = 
                (training_set_mses_cgapd / sum(training_set_mses_cgapd))^(-1)
              
              prop_var_exp_cgapd_weighting = prop_var_exp_cgapd_weighting / 
                sum(prop_var_exp_cgapd_weighting)
              
              
              
              # Get the predictions for the stacked ensemble for the prop
              # of variance explained
              prop_var_stacked_ens_cgapd_pred = 
                linear_reg_pred * prop_var_exp_cgapd_weighting[1] +
                lasso_pred * prop_var_exp_cgapd_weighting[2] + 
                cgapd_nnet_reg_pred * prop_var_exp_cgapd_weighting[3] + 
                cgapd_radial_svm_reg_pred * prop_var_exp_cgapd_weighting[4] + 
                cgapd_linear_svm_reg_pred * prop_var_exp_cgapd_weighting[5] + 
                cgapd_poly_svm_reg_pred * prop_var_exp_cgapd_weighting[6] +
                ranger_rf_pred * prop_var_exp_cgapd_weighting[7] + 
                xgbt_pred * prop_var_exp_cgapd_weighting[8]
              
              
              
              # Get the MSE
              prop_var_stacked_ens_cgapd_mse = 
                calc_mse(orig_data_response[-drawing],
                         prop_var_stacked_ens_cgapd_pred)
              
              
              # Get the Stacked Ensemble (Prop. of Var. explained) with No GADP 
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_no_cgapd = 
                c(training_linear_reg_mse_orig_data, 
                  training_lasso_mse_orig_data, 
                  training_nnet_reg_unprivatized_mse, 
                  training_radial_svm_reg_mse_orig_data,
                  training_linear_svm_reg_mse_orig_data, 
                  training_poly_svm_reg_mse_orig_data, 
                  training_ranger_rf_mse_orig_data,
                  training_xgbt_mse_orig_data)
              
              # weighting for the SE
              prop_var_exp_no_cgapd_weighting = 
                (training_set_mses_no_cgapd / sum(training_set_mses_no_cgapd))^(-1)
              
              prop_var_exp_no_cgapd_weighting = prop_var_exp_no_cgapd_weighting /
                sum(prop_var_exp_no_cgapd_weighting)
              
              
              # get the predictions
              prop_var_stacked_ens_no_cgapd_pred = 
                linear_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[1] +
                lasso_pred_orig_data * prop_var_exp_no_cgapd_weighting[2] + 
                nnet_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[3] + 
                radial_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[4] + 
                linear_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[5] + 
                poly_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[6] +
                ranger_rf_pred_orig_data * prop_var_exp_no_cgapd_weighting[7] + 
                xgbt_pred_orig_data * prop_var_exp_no_cgapd_weighting[8]
              
              # Get the MSE
              prop_var_stacked_ens_no_cgapd_mse = 
                calc_mse(orig_data_response[-drawing],
                         prop_var_stacked_ens_no_cgapd_pred)
              
              
              # Compile MSEs into a vector
              cgapd_mse_vector = 
                c(linear_reg_mse, lasso_mse, cgapd_nnet_reg_mse,
                  cgapd_radial_svm_reg_mse, cgapd_linear_svm_reg_mse,
                  cgapd_poly_svm_reg_mse, ranger_rf_mse,
                  xgbt_mse, avg_stacked_ensemble_cgapd_mse, 
                  prop_var_stacked_ens_cgapd_mse)
              
              no_cgapd_mse_vector =
                c(linear_reg_mse_orig_data, lasso_mse_orig_data,
                  nnet_reg_mse_orig_data, radial_svm_reg_mse_orig_data,
                  linear_svm_reg_mse_orig_data, poly_svm_reg_mse_orig_data,
                  ranger_rf_mse_orig_data, xgbt_mse_orig_data,
                  avg_stacked_ensemble_no_cgapd_mse, 
                  prop_var_stacked_ens_no_cgapd_mse)
              
              #print("Constructing the final results")
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "training_size","testing_size", "y", 
                                   "LR", "LASSO", "NNET", "Radial_SVM",
                                   "Linear_SVM", "Poly_SVM", "RF", "XGBTree",
                                   "Stacked_equal", "Stacked_prop","cgadp", "theta")
              
              
              # construct the dataframe of results
              # With CGADP
              
              estimator_reg_df = matrix( rep(0, length(data_frame_names) * 2), nrow = 2, ncol = length(data_frame_names), byrow = TRUE)
              estimator_reg_df[1,] = c(dat.seed,iteration, samp_n[samp_size_index], testing_n, 
                                       response_index, cgapd_mse_vector, 1, theta_value)
              
              # Adding results for LASSO without CGAPD
              estimator_reg_df[2,] = c(dat.seed, iteration, 
                                       samp_n[samp_size_index], testing_n,
                                       response_index, no_cgapd_mse_vector, 0, theta_value)
              
              
              colnames(estimator_reg_df) = data_frame_names
              
              return(estimator_reg_df)
            }
  #### Stopping Cluster -------------
  
  results = data.frame(results)
  
  # Add name of the estimators
  
  return(results)
}




## Stagnant Data -----------------------------
# Used for the CE real data application


estimator_evaluation_stacked_ensemb_cgadp_cv_stag = function(# the dataset
  m_start = 1, # Primarily used for debugging determines which iteration to start from
  m = 1000, # Number of Monte Carlo simulations
  samp_n = c(500, 1000, 2000), # sample sizes to be used for training set
  cores = 2, # Number of CPU threads/ logical processors to use
  response_prefix = "y", # prefix for response variables
  acquire_variable_names = get_variables, #Function for getting the variables
  perturbation_func, # CGADP or GADP function
  theta_value = 0.4, # privacy parameter value
  model_data # Data for the simulations
  ) 
{
  
  # Cross-Validation Funcitons -----------
  # XGBoosted Trees
  xgbt_cv = function(x_pred, y_resp, 
                     parameter_grid = list("nrounds" = c(10,20, 40),
                                           "max_depth" = c(2,4),
                                           "eta" = c(0.2, 0.4, 0.6),
                                           "gamma" = c(0, 0.5, 1),
                                           "colsample_bytree" = c(0.2, 0.5, 1),
                                           "min_child_weight" = c(0.5, 1),
                                           "subsample" = c(0.4, 0.5, 0.6)),
                     cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = expand.grid(parameter_grid)
    
    # Construct the XGBoosted trees
    xgbt_control= caret::trainControl(method = "cv", number = cv_folds)
    # Train XGBoosted trees
    xgbt_obj = caret::train(x = x_pred, y = as.vector(y_resp),
                            method = "xgbTree", tuneGrid = param_grid,
                            trControl = xgbt_control,
                            nthread = 1,
                            verbose=0,
                            print_every_n =0)
    
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_xgbt_obj = xgbt_obj$finalModel
    return(final_xgbt_obj)
  }
  
  # Random forest
  ranger_cv = function(x_pred, y_resp, num_trees = 500,
                       frac_pred = c(1/3, 1/4), 
                       min_node_size = c(5,10),
                       cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = as.matrix(expand_grid(num_trees, frac_pred, min_node_size))
    # Set up folds to be used in cross-validation
    fold_ids = split(1:nrow(x_pred), 1:cv_folds)
    # Vector of average mse for each combination of hyperparameters
    avg_mse_vals = numeric(nrow(param_grid))
    # Loop through each set of hyperparameters
    for(param_index in 1:nrow(param_grid))
    {
      # Vector to store mses per each test fold
      cv_mse_vec = numeric(cv_folds)
      # Cycle through each potential test fold
      for(fold_index in 1:cv_folds)
      {
        # Folds that will be used for training
        cv_sample = unsplit(fold_ids[-fold_index], rep(1:(cv_folds), cv_folds-1))
        # Fold that will be used for test
        cv_test = unsplit(fold_ids[-fold_index], 1:(cv_folds))
        # COnstruct the random forest
        ranger_rf_obj = ranger(x = x_pred[cv_sample,],
                               y = y_resp[cv_sample],
                               mtry = ceiling(ncol(x_pred) * param_grid[param_index, "frac_pred"]),
                               min.node.size = param_grid[param_index, "min_node_size"],
                               num.trees = param_grid[param_index, "num_trees"],
                               num.threads = 1)
        # Get the predictions
        ranger_predictions = predict(ranger_rf_obj, x_pred[cv_sample,])$predictions
        # Calculate the MSE
        mse_inst = 1 / length(cv_sample) * sum((y_resp[cv_sample] - ranger_predictions)^2)
        # Store the MSEs across each fold
        cv_mse_vec[fold_index] = mse_inst
      }
      # Store the average of the MSEs across eeach fold
      avg_mse_vals[param_index] = mean(cv_mse_vec) 
    }
    # Taking the minimum MSE or the first occurrence of the minimum MSE
    min_mse_position = which.min(avg_mse_vals)
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_ranger_obj = ranger(x = x_pred, y = y_resp,
                              mtry = ceiling(ncol(x_pred) * param_grid[min_mse_position, "frac_pred"]),
                              min.node.size = param_grid[min_mse_position, "min_node_size"],
                              num.trees = param_grid[min_mse_position, "num_trees"],
                              num.threads = 1)
    return(final_ranger_obj)
  }
  
  # Neural Network
  nnet_cv = function(x_pred, y_resp, 
                     parameter_grid = list("size" = c(3, 5, 10, 20),
                                           "decay" = c(0.001, 0.01, 0.1) ),
                     cv_folds = 5)
  {
    # Set up the hyperparameter combinations
    param_grid = expand.grid(parameter_grid)
    
    # Construct the XGBoosted trees
    nnet_control= caret::trainControl(method = "cv", number = cv_folds)
    # Train XGBoosted trees
    nnet_obj = caret::train(x = x_pred, y = as.vector(y_resp),
                            method = "nnet", tuneGrid = param_grid,
                            trControl = nnet_control,
                            maxit = 500,
                            linout=FALSE,
                            trace=FALSE)
    
    # Construct the final random forest based on hyperparameters with lowest MSE
    final_nnet_obj = nnet_obj$finalModel
    return(final_nnet_obj)
  }
  
  
  
  # Getting variable names:
  all_variable_names = colnames(model_data)
  organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
                                          variable_names = all_variable_names)
  
  
  # Get target variable indices
  target_variables_indices = which(all_variable_names %in% organized_vars$responses)
  
  # sdc_data_formula = sdc_form(target_variables = organized_vars$response, 
  #                             predictor_variables = organized_vars$predictors)
  
  # Calculate MSE function
  calc_mse = function(y_actual, y_hat)
  {
    mse = mean((y_actual - y_hat)^2)
    return(mse)
  }
  
  plan(callr, workers = cores)
  
  results = foreach(samp_size_index = 1:length(samp_n), .combine = "rbind", 
                    .options.future=list(packages = c( "Matrix",  "ranger", 
                                   "nnet", "e1071", "xgboost", "methods", "glmnet",
                                   "tidyr", "caret")),
                    .inorder = FALSE) %:%
    foreach( iteration = m_start:m, 
             .combine = "rbind",.inorder = FALSE) %:%
    foreach(response_index = 1:length(organized_vars$response), 
            .combine = "rbind",
            .inorder = FALSE) %dofuture% {
              
              
              # setting the seed
              dat.seed = iteration + samp_n[samp_size_index]
              set.seed(dat.seed)
              
              
              # Performing SDC GADP method
              privatized_dataframe = 
                perturbation_func(unprotected_data = model_data, 
                                  target_indices = target_variables_indices, 
                                  theta = theta_value, dat.seed = dat.seed)
              
              privatized_responses = privatized_dataframe[, target_variables_indices]
              
              # Current response from the privatized data, 
              # as indicated by the index for the response matrix
              current_response = privatized_responses[,response_index]
              
              privatized_dataframe = model_data[,organized_vars$predictors]
              
              # Get the sample from privatized df
              drawing = sample(1:nrow(privatized_dataframe), size = samp_n[samp_size_index], replace = FALSE)
              
              # Sample data after gapd process
              sample_private_data = privatized_dataframe[drawing,]
              
              # Sample data without gapd process
              sample_unprotected_data = model_data[drawing,]
              
              # x predictors for test set
              # Adding intercept term for linear regression
              x_dat_holdout = cbind(1,as.matrix(privatized_dataframe)[-drawing,organized_vars$predictors])
              
              # for the population
              orig_data_response =
                as.matrix(model_data)[,
                                      organized_vars$responses[response_index]]
              
              # x data does not change after cgapd, but preparing for case when both x and y are randomized
              x_dat = as.matrix(cbind(1,
                                      sample_private_data[,organized_vars$predictors]))
              
              
              # Get the predictors for the training set for unprotected data
              x_dat_unprotected = 
                as.matrix(cbind(1, sample_unprotected_data[,organized_vars$predictors]))
              
              # Get the predictors for the testing set for unprotected data
              x_dat_holdout_unprotected = 
                as.matrix(cbind(1, model_data[-drawing,organized_vars$predictors]))
              
              
              
              ### CGAPD Used----
              
              # Linear Regression on data after cgapd
              betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
                as.matrix(current_response[drawing])
              
              
              linear_reg_pred = as.vector(x_dat_holdout %*% betas_linear)
              
              # Get the predictions for the training data
              
              training_lin_reg_pred = as.vector(x_dat %*% betas_linear)
              
              
              # Get the mse for the training data
              
              
              training_lin_reg_mse = calc_mse(as.matrix(current_response[drawing]),
                                              training_lin_reg_pred)
              
              # Get the mse for the test data
              
              
              linear_reg_mse = calc_mse(orig_data_response[-drawing],
                                        linear_reg_pred)
              
              # LASSO on data after cgapd ---------------------
              
              # Build the LASSO regression model
              # using cross validation
              lasso_regress_model = cv.glmnet( 
                x = as.matrix(sample_private_data[,organized_vars$predictors]),
                y = current_response[drawing])
              
              
              lasso_pred = predict(lasso_regress_model, 
                                   as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]),
                                   s = "lambda.min")
              
              # get predictions on the training data
              training_lasso_pred = predict(lasso_regress_model, 
                                            as.matrix(privatized_dataframe[drawing,organized_vars$predictors]),
                                            s = "lambda.min")
              
              # Get the MSE for the training data
              training_lasso_mse = 
                calc_mse(orig_data_response[drawing],
                         training_lasso_pred)
              
              # # MSE for the testing data
              
              lasso_mse = calc_mse(orig_data_response[-drawing],
                                   lasso_pred)
              
              # Neural Net Regression on data after cgapd
              cgapd_nnet = nnet_cv(x_pred = x_dat[,-1], 
                                   y_resp = current_response[drawing])
              
              
              cgapd_nnet_reg_pred = predict(cgapd_nnet,x_dat_holdout[,-1])
              
              # training data predictions
              training_cgapd_nnet_reg_pred = predict(cgapd_nnet,x_dat[,-1])
              
              # training data MSE
              training_cgapd_nnet_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_nnet_reg_pred)
              
              
              # testing data MSE
              cgapd_nnet_reg_mse = calc_mse(orig_data_response[-drawing],
                                            cgapd_nnet_reg_pred)
              
              ## SVM Runs for CGADP -------
              
              # SVM Radial Regression on data after cgapd
              cgapd_radial_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                               gamma = seq(0.001,0.1,length.out = 2), 
                                               cost = seq(0.1,1,length.out = 2),
                                               
                                               kernel = c("radial"))
              
              cgapd_radial_svm = cgapd_radial_svm_tune$best.model
              
              cgapd_radial_svm_reg_pred = predict(cgapd_radial_svm,x_dat_holdout[,-1])
              
              # Get the training predictions of the radial svm
              training_cgapd_radial_svm_reg_pred =
                predict(cgapd_radial_svm, x_dat[,-1])
              
              # Get the training mse for radial svm
              training_cgapd_radial_svm_reg_mse = 
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_radial_svm_reg_pred)
              
              # Testing data MSE
              cgapd_radial_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                  cgapd_radial_svm_reg_pred)
              
              
              
              # SVM Linear Regression on data after cgapd
              cgapd_linear_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                               
                                               cost = seq(0.1,1,length.out = 2),
                                               
                                               kernel = c("linear"))
              
              cgapd_linear_svm = cgapd_linear_svm_tune$best.model
              
              # Predictions for the training data for Linear SVM
              training_cgapd_linear_svm_reg_pred =
                predict(cgapd_linear_svm,x_dat_holdout[,-1])
              
              # Training MSE for Linear SVM
              training_cgapd_linear_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_linear_svm_reg_pred)
              
              # Predictions for the test dataset
              cgapd_linear_svm_reg_pred = predict(cgapd_linear_svm,x_dat_holdout[,-1])
              
              # Testing data MSE
              cgapd_linear_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                  cgapd_linear_svm_reg_pred)
              
              # SVM Polynomial Regression on data after gapd
              cgapd_poly_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
                                             gamma = seq(0.001,0.1,length.out = 2), 
                                             cost = seq(0.1,1,length.out = 2),
                                             degree = seq(1,2, length.out=2),
                                             kernel = c("polynomial"),
                                             cross = 10)
              
              cgapd_poly_svm = cgapd_poly_svm_tune$best.model
              
              # Training data polynomial SVM predictions
              training_cgapd_poly_svm_reg_pred = predict(cgapd_poly_svm,x_dat[,-1])
              
              # Training data MSE polynomial SVM
              training_cgapd_poly_svm_reg_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_cgapd_poly_svm_reg_pred)
              
              # Predictions for testing dataset
              cgapd_poly_svm_reg_pred = predict(cgapd_poly_svm,x_dat_holdout[,-1])
              
              # MSE for testing dataset for polynomial kernel SVM
              cgapd_poly_svm_reg_mse = calc_mse(orig_data_response[-drawing],
                                                cgapd_poly_svm_reg_pred)
              
              ## Random Forest Run ---------------------------
              
              # Random forest on data after cgapd
              num_x_vars = length(organized_vars$predictors)
              
              ranger_rf_regress_model = ranger_cv( 
                x_pred = as.matrix(sample_private_data[,organized_vars$predictors]),
                y_resp = current_response[drawing])
              
              # Training dataset predicions for the random forest
              training_ranger_rf_pred = 
                predict(ranger_rf_regress_model, 
                        privatized_dataframe[drawing,organized_vars$predictors])$predictions
              
              # training dataset MSE for the radom forest
              training_ranger_rf_mse =
                calc_mse(as.matrix(current_response[drawing]),
                         training_ranger_rf_pred)
              
              
              # Testing dataset prediction for random forest
              ranger_rf_pred = predict(ranger_rf_regress_model, 
                                       privatized_dataframe[-drawing,organized_vars$predictors])$predictions
              
              # Testing dataset MSE for random forest
              ranger_rf_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                                       ranger_rf_pred)
              
              # XGBT on data after cgapd
              xgbt_regress_model = xgbt_cv( 
                x_pred = as.matrix(sample_private_data[,organized_vars$predictors]),
                y_resp = current_response[drawing])
              
              # Training data predictions for the XGB tree
              training_xgbt_pred = 
                predict(xgbt_regress_model,
                        as.matrix(privatized_dataframe[drawing,organized_vars$predictors]))
              
              # training data MSE for the XGB tree
              training_xgbt_mse = 
                calc_mse(as.matrix(current_response[drawing]) ,
                         training_xgbt_pred)
              
              # Testing data predicitons for the XGB tree
              xgbt_pred = predict(xgbt_regress_model, 
                                  as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]))
              
              # Testing data MSE for the XGB tree
              xgbt_mse = calc_mse(orig_data_response[-drawing],
                                  xgbt_pred)
              
              ##
              ## Before CGAPD (On original data) -----------------------------------
              # Linear Regression on data before gapd (unprotected original data)
              
              sample_orig_data_response = 
                as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
              
              
              # Beta coefficients for the original data
              orig_data_betas_linear = solve(t(x_dat_unprotected) %*% 
                                               x_dat_unprotected) %*% t(x_dat_unprotected) %*% 
                as.matrix(sample_orig_data_response)
              
              # Predictions on the training data no GADP
              training_lin_reg_unprivat_pred = 
                as.vector(x_dat_unprotected %*% orig_data_betas_linear)
              
              
              # MSE for training data for linear regression, no GADP
              training_linear_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_lin_reg_unprivat_pred)
              
              # Testing data MSE for linear regression no ADP
              linear_reg_unprivatized_pred = as.vector(x_dat_holdout_unprotected %*% 
                                                         orig_data_betas_linear)
              
              linear_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_reg_unprivatized_pred)
              
              # LASSO on data before gapd (original data)
              
              
              lasso_regress_model_orig_data =
                cv.glmnet( x = as.matrix(model_data[drawing,organized_vars$predictors]), # Get rid of intercept term
                           y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
              
              # Training data predictions for LASSO
              
              training_lasso_pred_orig_data = 
                predict(lasso_regress_model_orig_data,
                        as.matrix(model_data[drawing,organized_vars$predictors]), 
                        s = "lambda.min")
              
              # Training data MSE for LASSO
              training_lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
                         training_lasso_pred_orig_data)
              
              # Predictions for the test dataset for the LASSO model
              lasso_pred_orig_data = predict(lasso_regress_model_orig_data,
                                             as.matrix(model_data[-drawing,organized_vars$predictors]), 
                                             s = "lambda.min")
              # MSE for the LASSO model for the test data
              lasso_mse_orig_data =
                calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
                         lasso_pred_orig_data)
              
              
              # Neural Net before GAPD 
              
              no_cgapd_nnet = nnet_cv(x_pred = x_dat_unprotected[,-1], 
                                      y_resp = sample_orig_data_response)
              
              # The predictions of the training set for neural network
              training_nnet_reg_unprivatized_pred = predict(no_cgapd_nnet, 
                                                            x_dat_unprotected[,-1])
              
              # The MSE for the training set for the neural network
              training_nnet_reg_unprivatized_mse = 
                calc_mse(orig_data_response[drawing],
                         training_nnet_reg_unprivatized_pred)
              
              # Predicitons on testing set for neural net
              nnet_reg_unprivatized_pred = predict(no_cgapd_nnet, 
                                                   x_dat_holdout_unprotected[,-1])
              
              # MSE for testing set for neural network
              nnet_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         nnet_reg_unprivatized_pred)
              
              # SVM Radial before GAPD
              no_cgapd_radial_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                  y = sample_orig_data_response,
                                                  gamma = seq(0.001,0.1,length.out = 2), 
                                                  cost = seq(0.1,1,length.out = 2),
                                                  
                                                  kernel = c("radial"))
              
              # Get the best model for Radial SVM
              no_cgapd_radial_svm = no_cgapd_radial_svm_tune$best.model
              
              # Predictions on the training set for radial SVM
              training_radial_svm_reg_unprivatized_pred = 
                predict(no_cgapd_radial_svm, x_dat_unprotected[,-1])
              
              # The MSE for the training set for radial SVM
              training_radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_radial_svm_reg_unprivatized_pred)
              
              # Get the predictions of Radial SVM for the testing set
              radial_svm_reg_unprivatized_pred = 
                predict(no_cgapd_radial_svm, x_dat_holdout_unprotected[,-1])
              
              # Get the MSE of the Radial SVM for the testing set
              radial_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         radial_svm_reg_unprivatized_pred)
              
              # Linear SVM before GAPD
              no_cgapd_linear_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                  y = sample_orig_data_response,
                                                  cost = seq(0.1,1,length.out = 2),
                                                  
                                                  kernel = c("linear"))
              # The Linear SVM best tuned model
              no_cgapd_linear_svm = no_cgapd_linear_svm_tune$best.model
              
              # The predictions on the training set, Linear SVM
              training_linear_svm_reg_unprivatized_pred = 
                predict(no_cgapd_linear_svm, x_dat_unprotected[,-1])
              
              # the MSE on the training set, Linear SVM
              training_linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_linear_svm_reg_unprivatized_pred)
              
              # The predictions for the testing set Linear SVM
              linear_svm_reg_unprivatized_pred = 
                predict(no_cgapd_linear_svm, x_dat_holdout_unprotected[,-1])
              
              # The MSE for the testing set Linear SVM
              linear_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         linear_svm_reg_unprivatized_pred)
              
              
              # Polynomial SVM before GAPD
              no_cgapd_poly_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
                                                y = sample_orig_data_response,
                                                gamma = seq(0.001,0.1,length.out = 2), 
                                                cost = seq(0.1,1,length.out = 2),
                                                degree = seq(1,2, length.out=2),
                                                kernel = c("polynomial"),
                                                cross = 10)
              # Get the best tuned polynomial SVM
              no_cgapd_poly_svm = no_cgapd_poly_svm_tune$best.model
              
              # The training set predicitons for the polynomial SVM
              training_poly_svm_reg_unprivatized_pred = 
                predict(no_cgapd_poly_svm, x_dat_unprotected[,-1])
              
              # the training set MSE for the polynomial SVM
              training_poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_poly_svm_reg_unprivatized_pred)
              
              # The testing set predicitons using the polynomial SVM
              poly_svm_reg_unprivatized_pred = 
                predict(no_cgapd_poly_svm, x_dat_holdout_unprotected[,-1])
              
              # The testing set MSE for the polynomial SVM
              poly_svm_reg_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         poly_svm_reg_unprivatized_pred)
              
              # Random forest on data before cgapd
              ranger_rf_model_orig_data = ranger_cv( 
                x_pred = model_data[drawing,organized_vars$predictors],
                y_resp = orig_data_response[drawing])
              
              # The training set predictions from random forest
              training_ranger_rf_pred_orig_data = 
                predict(ranger_rf_model_orig_data,
                        model_data[drawing,organized_vars$predictors])$predictions
              
              # The training set MSE for the random forest
              training_ranger_rf_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_ranger_rf_pred_orig_data)
              
              # Testing set predictions for random forest
              ranger_rf_pred_orig_data = predict(ranger_rf_model_orig_data,
                                                 model_data[-drawing,organized_vars$predictors])$predictions
              # Testing set MSE for random forest
              ranger_rf_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         ranger_rf_pred_orig_data)
              
              # XGBT on data before cgapd
              xgbt_regress_model_orig_data =
                xgbt_cv(x_pred = as.matrix(sample_unprotected_data[,organized_vars$predictors]),
                        y_resp = orig_data_response[drawing])
              
              # The training set predictions for XGB tree
              training_xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(sample_unprotected_data[,organized_vars$predictors]))
              # Training set MSE for XGB tree for 
              training_xgbt_mse_orig_data =
                calc_mse(orig_data_response[drawing],
                         training_xgbt_pred_orig_data)
              
              # The testing set XGB tree predictions
              xgbt_pred_orig_data = 
                predict(xgbt_regress_model_orig_data,
                        as.matrix(model_data[-drawing,organized_vars$predictors]))
              
              # The testing set MSE for XGB tree
              xgbt_mse_orig_data =
                calc_mse(orig_data_response[-drawing],
                         xgbt_pred_orig_data)
              
              
              #print("Constructing the stacked ensemble")
              ###
              ### Stacked Ensemble Equal Weighting ----
              # CGAPD
              avg_stacked_ensemble_predictions_cgapd = (linear_reg_pred  + 
                                                          lasso_pred  +  cgapd_nnet_reg_pred + 
                                                          cgapd_radial_svm_reg_pred  + 
                                                          cgapd_linear_svm_reg_pred  + 
                                                          cgapd_poly_svm_reg_pred  +
                                                          ranger_rf_pred  + xgbt_pred) / 8
              
              avg_stacked_ensemble_cgapd_mse =
                calc_mse(orig_data_response[-drawing],
                         avg_stacked_ensemble_predictions_cgapd)
              # No GAPD
              avg_stacked_ensemble_pred_no_cgapd = 
                (linear_reg_unprivatized_pred  + 
                   lasso_pred_orig_data  + 
                   nnet_reg_unprivatized_pred  + 
                   radial_svm_reg_unprivatized_pred  + 
                   linear_svm_reg_unprivatized_pred  + 
                   poly_svm_reg_unprivatized_pred  +
                   ranger_rf_pred_orig_data  + xgbt_pred_orig_data) / 8
              
              avg_stacked_ensemble_no_cgapd_mse =
                calc_mse(orig_data_response[-drawing],
                         avg_stacked_ensemble_pred_no_cgapd)
              
              
              ### Stacked Ensemble Equal Weighting by Proportion of variance Explained ----------------
              # under CGADP
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_cgapd = 
                c(training_lin_reg_mse, training_lasso_mse, 
                  training_cgapd_nnet_reg_mse, training_cgapd_radial_svm_reg_mse,
                  training_cgapd_linear_svm_reg_mse, 
                  training_cgapd_poly_svm_reg_mse, training_ranger_rf_mse,
                  training_xgbt_mse)
              
              # The models with the lowest MSE get the highest weight
              prop_var_exp_cgapd_weighting = 
                (training_set_mses_cgapd / sum(training_set_mses_cgapd))^(-1)
              
              prop_var_exp_cgapd_weighting = prop_var_exp_cgapd_weighting / 
                sum(prop_var_exp_cgapd_weighting)
              
              
              
              # Get the predictions for the stacked ensemble for the prop
              # of variance explained
              prop_var_stacked_ens_cgapd_pred = 
                linear_reg_pred * prop_var_exp_cgapd_weighting[1] +
                lasso_pred * prop_var_exp_cgapd_weighting[2] + 
                cgapd_nnet_reg_pred * prop_var_exp_cgapd_weighting[3] + 
                cgapd_radial_svm_reg_pred * prop_var_exp_cgapd_weighting[4] + 
                cgapd_linear_svm_reg_pred * prop_var_exp_cgapd_weighting[5] + 
                cgapd_poly_svm_reg_pred * prop_var_exp_cgapd_weighting[6] +
                ranger_rf_pred * prop_var_exp_cgapd_weighting[7] + 
                xgbt_pred * prop_var_exp_cgapd_weighting[8]
              
              
              
              # Get the MSE
              prop_var_stacked_ens_cgapd_mse = 
                calc_mse(orig_data_response[-drawing],
                         prop_var_stacked_ens_cgapd_pred)
              
              
              # Get the Stacked Ensemble (Prop. of Var. explained) with No GADP 
              
              # in this order:
              # linear regression, LASSO, neural net, radial SVM, linear SVM,
              # poly SVM, RF, XGB Tree
              training_set_mses_no_cgapd = 
                c(training_linear_reg_mse_orig_data, 
                  training_lasso_mse_orig_data, 
                  training_nnet_reg_unprivatized_mse, 
                  training_radial_svm_reg_mse_orig_data,
                  training_linear_svm_reg_mse_orig_data, 
                  training_poly_svm_reg_mse_orig_data, 
                  training_ranger_rf_mse_orig_data,
                  training_xgbt_mse_orig_data)
              
              # weighting for the SE
              prop_var_exp_no_cgapd_weighting = 
                (training_set_mses_no_cgapd / sum(training_set_mses_no_cgapd))^(-1)
              
              prop_var_exp_no_cgapd_weighting = prop_var_exp_no_cgapd_weighting /
                sum(prop_var_exp_no_cgapd_weighting)
              
              
              # get the predictions
              prop_var_stacked_ens_no_cgapd_pred = 
                linear_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[1] +
                lasso_pred_orig_data * prop_var_exp_no_cgapd_weighting[2] + 
                nnet_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[3] + 
                radial_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[4] + 
                linear_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[5] + 
                poly_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[6] +
                ranger_rf_pred_orig_data * prop_var_exp_no_cgapd_weighting[7] + 
                xgbt_pred_orig_data * prop_var_exp_no_cgapd_weighting[8]
              
              # Get the MSE
              prop_var_stacked_ens_no_cgapd_mse = 
                calc_mse(orig_data_response[-drawing],
                         prop_var_stacked_ens_no_cgapd_pred)
              
              
              # Compile MSEs into a vector
              cgapd_mse_vector = 
                c(linear_reg_mse, lasso_mse, cgapd_nnet_reg_mse,
                  cgapd_radial_svm_reg_mse, cgapd_linear_svm_reg_mse,
                  cgapd_poly_svm_reg_mse, ranger_rf_mse,
                  xgbt_mse, avg_stacked_ensemble_cgapd_mse, 
                  prop_var_stacked_ens_cgapd_mse)
              
              no_cgapd_mse_vector =
                c(linear_reg_mse_orig_data, lasso_mse_orig_data,
                  nnet_reg_mse_orig_data, radial_svm_reg_mse_orig_data,
                  linear_svm_reg_mse_orig_data, poly_svm_reg_mse_orig_data,
                  ranger_rf_mse_orig_data, xgbt_mse_orig_data,
                  avg_stacked_ensemble_no_cgapd_mse, 
                  prop_var_stacked_ens_no_cgapd_mse)
              
              #print("Constructing the final results")
              
              # Dataframe names
              data_frame_names = c("seed", "iteration", "training_size", "y", 
                                   "LR", "LASSO", "NNET", "Radial_SVM",
                                   "Linear_SVM", "Poly_SVM", "RF", "XGBTree",
                                   "Stacked_equal", "Stacked_prop","cgadp", "theta")
              
              
              # construct the dataframe of results
              # With CGADP
              
              estimator_reg_df = matrix( rep(0, length(data_frame_names) * 2), nrow = 2, ncol = length(data_frame_names), byrow = TRUE)
              estimator_reg_df[1,] = c(dat.seed,iteration, samp_n[samp_size_index],  
                                       response_index, cgapd_mse_vector, 1, theta_value)
              
              # Adding results for LASSO without CGAPD
              estimator_reg_df[2,] = c(dat.seed, iteration, 
                                       samp_n[samp_size_index], 
                                       response_index, no_cgapd_mse_vector, 0, theta_value)
              
              
              colnames(estimator_reg_df) = data_frame_names
              
              return(estimator_reg_df)
            }
  
  
  results = data.frame(results)
  
  # Add name of the estimators
  
  return(results)
}


## Test and Modify CGADP -----
# estimator_evaluation_stacked_ensemb_cgadp_iter_ver_2_mod = function(# the dataset
#   m_start = 1, # Primarily used for debugging determines which iteration to start from
#   m = 1000, # Number of Monte Carlo simulations
#   n = c(500, 1000, 2000), # sample sizes to be used for training set
#   cores = 2, # Number of CPU threads/ logical processors to use
#   response_prefix = "y", # prefix for response variables
#   acquire_variable_names = get_variables, #Function for getting the variables
#   perturbation_func, # CGADP or GADP function
#   theta_value = 0.4, # privacy parameter value
#   model_generating_function, # function to generate the model
#   testing_data_seed = 30000, # Seed for the testing set
#   model_predictor_quant = 10, # Quantity of the predictors for the model
#   model_rho = 0.2, # correlation of predictors in the model
#   model_noise = 0.2, # Noise to be used in the model
#   testing_n = 500) # Size of the testing set
# {
#   
#   # generate the testing data
#   testing_model_data = model_generating_function(N = testing_n, rho = model_rho, 
#                                                  max_pred = model_predictor_quant, 
#                                                  dat.seed = testing_data_seed,
#                                                  noise = model_noise)
#   
#   # Getting variable names:
#   all_variable_names = colnames(testing_model_data)
#   organized_vars = acquire_variable_names(target_variable_prefix = response_prefix, 
#                                           variable_names = all_variable_names)
#   
#   
#   # Get target variable indices
#   target_variables_indices = which(all_variable_names %in% organized_vars$responses)
#   
#   # sdc_data_formula = sdc_form(target_variables = organized_vars$response, 
#   #                             predictor_variables = organized_vars$predictors)
#   
#   # Calculate MSE function
#   calc_mse = function(y_actual, y_hat)
#   {
#     mse = mean((y_actual - y_hat)^2)
#     return(mse)
#   }
#   
#   sim_cluster = makeCluster(cores)
#   registerDoSNOW(sim_cluster)
#   
#   results = foreach(samp_size_index = 1:length(n), .combine = "rbind", 
#                     .packages = c( "Matrix",  "ranger", 
#                                    "nnet", "e1071", "xgboost", "methods", "glmnet"),
#                     .inorder = FALSE) %:%
#     foreach( iteration = m_start:m, 
#              .combine = "rbind",.inorder = FALSE) %:%
#     foreach(response_index = 1:length(organized_vars$response), 
#             .combine = "rbind",
#             .inorder = FALSE) %dopar% {
#               
#               # setting the seed
#               dat.seed = iteration + n[samp_size_index]
#               set.seed(dat.seed)
#               
#               # Training dataset
#               # Generated with the same parameters sxcluding the seed
#               training_model_data = 
#                 model_generating_function(N = n[samp_size_index], 
#                                           rho = model_rho, 
#                                           max_pred = model_predictor_quant, 
#                                           dat.seed = dat.seed,
#                                           noise = model_noise)
#               
#               model_data = rbind(training_model_data, 
#                                  testing_model_data)
#               
#               # Performing SDC GADP method
#               privatized_dataframe = 
#                 perturbation_func(unprotected_data = model_data, 
#                                   target_indices = target_variables_indices, 
#                                   theta = theta_value, dat.seed = dat.seed)
#               
#               
#               privatized_responses = privatized_dataframe[, target_variables_indices]
#               
#               # Current response from the privatized data, 
#               # as indicated by the index for the response matrix
#               current_response = privatized_responses[,response_index]
#               
#               privatized_dataframe = model_data[,organized_vars$predictors]
#               
#               # Get the sample from privatized df
#               drawing = 1:n[samp_size_index]
#               
#               # Sample data after gapd process
#               sample_private_data = privatized_dataframe[drawing,]
#               
#               # Sample data without gapd process
#               sample_unprotected_data = model_data[drawing,]
#               
#               # x predictors for test set
#               # Adding intercept term for linear regression
#               x_dat_holdout = cbind(1,as.matrix(privatized_dataframe)[-drawing,organized_vars$predictors])
#               
#               # for the population
#               orig_data_response =
#                 as.matrix(model_data)[,
#                                       organized_vars$responses[response_index]]
#               
#               # x data does not change after cgapd, but preparing for case when both x and y are randomized
#               x_dat = as.matrix(cbind(1,
#                                       sample_private_data[,organized_vars$predictors]))
#               
#               
#               # Get the predictors for the training set for unprotected data
#               x_dat_unprotected = 
#                 as.matrix(cbind(1, sample_unprotected_data[,organized_vars$predictors]))
#               
#               # Get the predictors for the testing set for unprotected data
#               x_dat_holdout_unprotected = 
#                 as.matrix(cbind(1, model_data[-drawing,organized_vars$predictors]))
#               
#               
#               
#               ### CGAPD Used----
#               
#               # Linear Regression on data after cgapd
#               betas_linear = solve(t(x_dat) %*% x_dat) %*% t(x_dat) %*% 
#                 as.matrix(current_response[drawing])
#               
#               
#               linear_reg_pred = as.vector(x_dat_holdout %*% betas_linear)
#               
#               # Get the predictions for the training data
#               
#               training_lin_reg_pred = as.vector(x_dat %*% betas_linear)
#               
#               
#               # Get the mse for the training data
#               
#               
#               training_lin_reg_mse = calc_mse(as.matrix(current_response[drawing]),
#                                               training_lin_reg_pred)
#               
#               # Get the mse for the test data
#               
#               
#               linear_reg_mse = calc_mse(orig_data_response[-drawing],
#                                         linear_reg_pred)
#               
#               # LASSO on data after cgapd ---------------------
#               
#               # Build the LASSO regression model
#               # using cross validation
#               lasso_regress_model = cv.glmnet( 
#                 x = as.matrix(sample_private_data[,organized_vars$predictors]),
#                 y = current_response[drawing])
#               
#               
#               lasso_pred = predict(lasso_regress_model, 
#                                    as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]),
#                                    s = "lambda.min")
#               
#               # get predictions on the training data
#               training_lasso_pred = predict(lasso_regress_model, 
#                                             as.matrix(privatized_dataframe[drawing,organized_vars$predictors]),
#                                             s = "lambda.min")
#               
#               # Get the MSE for the training data
#               training_lasso_mse = 
#                 calc_mse(orig_data_response[drawing],
#                          training_lasso_pred)
#               
#               # # MSE for the testing data
#               
#               lasso_mse = calc_mse(orig_data_response[-drawing],
#                                    lasso_pred)
#               
#               # Neural Net Regression on data after cgapd
#               cgapd_nnet = nnet(x = x_dat[,-1], y = current_response[drawing], 
#                                 size = 3,
#                                 maxit = 200, linout = TRUE)
#               
#               
#               cgapd_nnet_reg_pred = predict(cgapd_nnet,x_dat_holdout[,-1])
#               
#               # training data predictions
#               training_cgapd_nnet_reg_pred = predict(cgapd_nnet,x_dat[,-1])
#               
#               # training data MSE
#               training_cgapd_nnet_reg_mse = 
#                 calc_mse(as.matrix(current_response[drawing]),
#                          training_cgapd_nnet_reg_pred)
#               
#               
#               # testing data MSE
#               cgapd_nnet_reg_mse = calc_mse(orig_data_response[-drawing],
#                                             cgapd_nnet_reg_pred)
#               
#               ## SVM Runs for CGADP -------
#               
#               # SVM Radial Regression on data after cgapd
#               cgapd_radial_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
#                                                gamma = seq(0.001,0.1,length.out = 2), 
#                                                cost = seq(0.1,1,length.out = 2),
#                                                
#                                                kernel = c("radial"))
#               
#               cgapd_radial_svm = cgapd_radial_svm_tune$best.model
#               
#               cgapd_radial_svm_reg_pred = predict(cgapd_radial_svm,x_dat_holdout[,-1])
#               
#               # Get the training predictions of the radial svm
#               training_cgapd_radial_svm_reg_pred =
#                 predict(cgapd_radial_svm, x_dat[,-1])
#               
#               # Get the training mse for radial svm
#               training_cgapd_radial_svm_reg_mse = 
#                 calc_mse(as.matrix(current_response[drawing]),
#                          training_cgapd_radial_svm_reg_pred)
#               
#               # Testing data MSE
#               cgapd_radial_svm_reg_mse = calc_mse(orig_data_response[-drawing],
#                                                   cgapd_radial_svm_reg_pred)
#               
#               
#               
#               # SVM Linear Regression on data after cgapd
#               cgapd_linear_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
#                                               
#                                                cost = seq(0.1,1,length.out = 2),
#                                              
#                                                kernel = c("linear"))
#               
#               cgapd_linear_svm = cgapd_linear_svm_tune$best.model
#               
#               # Predictions for the training data for Linear SVM
#               training_cgapd_linear_svm_reg_pred =
#                 predict(cgapd_linear_svm,x_dat_holdout[,-1])
#               
#               # Training MSE for Linear SVM
#               training_cgapd_linear_svm_reg_mse =
#                 calc_mse(as.matrix(current_response[drawing]),
#                          training_cgapd_linear_svm_reg_pred)
#               
#               # Predictions for the test dataset
#               cgapd_linear_svm_reg_pred = predict(cgapd_linear_svm,x_dat_holdout[,-1])
#               
#               # Testing data MSE
#               cgapd_linear_svm_reg_mse = calc_mse(orig_data_response[-drawing],
#                                                   cgapd_linear_svm_reg_pred)
#               
#               # SVM Polynomial Regression on data after gapd
#               cgapd_poly_svm_tune = tune.svm(x = x_dat[,-1], y = current_response[drawing],
#                                              gamma = seq(0.001,0.1,length.out = 2), 
#                                              cost = seq(0.1,1,length.out = 2),
#                                              degree = seq(1,3, length.out=2),
#                                              kernel = c("polynomial"))
#               
#               cgapd_poly_svm = cgapd_poly_svm_tune$best.model
#               
#               # Training data polynomial SVM predictions
#               training_cgapd_poly_svm_reg_pred = predict(cgapd_poly_svm,x_dat[,-1])
#               
#               # Training data MSE polynomial SVM
#               training_cgapd_poly_svm_reg_mse =
#                 calc_mse(as.matrix(current_response[drawing]),
#                          training_cgapd_poly_svm_reg_pred)
#               
#               # Predictions for testing dataset
#               cgapd_poly_svm_reg_pred = predict(cgapd_poly_svm,x_dat_holdout[,-1])
#               
#               # MSE for testing dataset for polynomial kernel SVM
#               cgapd_poly_svm_reg_mse = calc_mse(orig_data_response[-drawing],
#                                                 cgapd_poly_svm_reg_pred)
#               
#               ## Random Forest Run ---------------------------
#               
#               # Random forest on data after cgapd
#               num_x_vars = length(organized_vars$predictors)
#               
#               ranger_rf_regress_model = ranger( 
#                 x = as.matrix(sample_private_data[,organized_vars$predictors]),
#                 y = current_response[drawing], 
#                 num.trees = 500,
#                 num.threads = 1,
#                 mtry = ceiling(1 / 3 * num_x_vars))
#               
#               # Training dataset predicions for the random forest
#               training_ranger_rf_pred = 
#                 predict(ranger_rf_regress_model, 
#                         privatized_dataframe[drawing,organized_vars$predictors])$predictions
#               
#               # training dataset MSE for the radom forest
#               training_ranger_rf_mse =
#                 calc_mse(as.matrix(current_response[drawing]),
#                          training_ranger_rf_pred)
#               
#               
#               # Testing dataset prediction for random forest
#               ranger_rf_pred = predict(ranger_rf_regress_model, 
#                                        privatized_dataframe[-drawing,organized_vars$predictors])$predictions
#               
#               # Testing dataset MSE for random forest
#               ranger_rf_mse = calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
#                                        ranger_rf_pred)
#               
#               # XGBT on data after cgapd
#               xgbt_regress_model = xgboost( 
#                 data = as.matrix(sample_private_data[,organized_vars$predictors]),
#                 label = current_response[drawing], 
#                 nrounds = 40,
#                 nthread = 1)
#               
#               # Training data predictions for the XGB tree
#               training_xgbt_pred = 
#                 predict(xgbt_regress_model,
#                         as.matrix(privatized_dataframe[drawing,organized_vars$predictors]))
#               
#               # training data MSE for the XGB tree
#               training_xgbt_mse = 
#                 calc_mse(as.matrix(current_response[drawing]) ,
#                          training_xgbt_pred)
#               
#               # Testing data predicitons for the XGB tree
#               xgbt_pred = predict(xgbt_regress_model, 
#                                   as.matrix(privatized_dataframe[-drawing,organized_vars$predictors]))
#               
#               # Testing data MSE for the XGB tree
#               xgbt_mse = calc_mse(orig_data_response[-drawing],
#                                   xgbt_pred)
#               
#               ##
#               ## Before CGAPD (On original data) -----------------------------------
#               # Linear Regression on data before gapd (unprotected original data)
#               
#               sample_orig_data_response = 
#                 as.matrix(model_data)[drawing,organized_vars$responses[response_index]]
#               
#               
#               # Beta coefficients for the original data
#               orig_data_betas_linear = solve(t(x_dat_unprotected) %*% 
#                                                x_dat_unprotected) %*% t(x_dat_unprotected) %*% 
#                 as.matrix(sample_orig_data_response)
#               
#               # Predictions on the training data no GADP
#               training_lin_reg_unprivat_pred = 
#                 as.vector(x_dat_unprotected %*% orig_data_betas_linear)
#               
#               
#               # MSE for training data for linear regression, no GADP
#               training_linear_reg_mse_orig_data =
#                 calc_mse(orig_data_response[drawing],
#                          training_lin_reg_unprivat_pred)
#               
#               # Testing data MSE for linear regression no ADP
#               linear_reg_unprivatized_pred = as.vector(x_dat_holdout_unprotected %*% 
#                                                          orig_data_betas_linear)
#               
#               linear_reg_mse_orig_data =
#                 calc_mse(orig_data_response[-drawing],
#                          linear_reg_unprivatized_pred)
#               
#               # LASSO on data before gapd (original data)
#               
#               
#               lasso_regress_model_orig_data =
#                 cv.glmnet( x = as.matrix(model_data[drawing,organized_vars$predictors]), # Get rid of intercept term
#                            y = as.matrix(sample_unprotected_data[,organized_vars$responses[response_index]]))
#               
#               # Training data predictions for LASSO
#               
#               training_lasso_pred_orig_data = 
#                 predict(lasso_regress_model_orig_data,
#                         as.matrix(model_data[drawing,organized_vars$predictors]), 
#                         s = "lambda.min")
#               
#               # Training data MSE for LASSO
#               training_lasso_mse_orig_data =
#                 calc_mse(as.matrix(model_data)[drawing,organized_vars$responses[response_index]],
#                          training_lasso_pred_orig_data)
#               
#               # Predictions for the test dataset for the LASSO model
#               lasso_pred_orig_data = predict(lasso_regress_model_orig_data,
#                                              as.matrix(model_data[-drawing,organized_vars$predictors]), 
#                                              s = "lambda.min")
#               # MSE for the LASSO model for the test data
#               lasso_mse_orig_data =
#                 calc_mse(as.matrix(model_data)[-drawing,organized_vars$responses[response_index]],
#                          lasso_pred_orig_data)
#               
#               
#               # Neural Net before GAPD 
#               
#               no_cgapd_nnet = nnet(x = x_dat_unprotected[,-1], 
#                                    y = sample_orig_data_response, 
#                                    size = 3,
#                                    maxit = 200, linout = TRUE)
#               
#               # The predictions of the training set for neural network
#               training_nnet_reg_unprivatized_pred = predict(no_cgapd_nnet, 
#                                                             x_dat_unprotected[,-1])
#               
#               # The MSE for the training set for the neural network
#               training_nnet_reg_unprivatized_mse = 
#                 calc_mse(orig_data_response[drawing],
#                          training_nnet_reg_unprivatized_pred)
#               
#               # Predicitons on testing set for neural net
#               nnet_reg_unprivatized_pred = predict(no_cgapd_nnet, 
#                                                    x_dat_holdout_unprotected[,-1])
#               
#               # MSE for testing set for neural network
#               nnet_reg_mse_orig_data =
#                 calc_mse(orig_data_response[-drawing],
#                          nnet_reg_unprivatized_pred)
#               
#               # SVM Radial before GAPD
#               no_cgapd_radial_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
#                                                   y = sample_orig_data_response,
#                                                   gamma = seq(0.001,0.1,length.out = 2), 
#                                                   cost = seq(0.1,1,length.out = 2),
#                                                   
#                                                   kernel = c("radial"))
#               
#               # Get the best model for Radial SVM
#               no_cgapd_radial_svm = no_cgapd_radial_svm_tune$best.model
#               
#               # Predictions on the training set for radial SVM
#               training_radial_svm_reg_unprivatized_pred = 
#                 predict(no_cgapd_radial_svm, x_dat_unprotected[,-1])
#               
#               # The MSE for the training set for radial SVM
#               training_radial_svm_reg_mse_orig_data =
#                 calc_mse(orig_data_response[drawing],
#                          training_radial_svm_reg_unprivatized_pred)
#               
#               # Get the predictions of Radial SVM for the testing set
#               radial_svm_reg_unprivatized_pred = 
#                 predict(no_cgapd_radial_svm, x_dat_holdout_unprotected[,-1])
#               
#               # Get the MSE of the Radial SVM for the testing set
#               radial_svm_reg_mse_orig_data =
#                 calc_mse(orig_data_response[-drawing],
#                          radial_svm_reg_unprivatized_pred)
#               
#               # Linear SVM before GAPD
#               no_cgapd_linear_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
#                                                   y = sample_orig_data_response,
#                                                   cost = seq(0.1,1,length.out = 2),
#                                                   
#                                                   kernel = c("linear"))
#               # The Linear SVM best tuned model
#               no_cgapd_linear_svm = no_cgapd_linear_svm_tune$best.model
#               
#               # The predictions on the training set, Linear SVM
#               training_linear_svm_reg_unprivatized_pred = 
#                 predict(no_cgapd_linear_svm, x_dat_unprotected[,-1])
#               
#               # the MSE on the training set, Linear SVM
#               training_linear_svm_reg_mse_orig_data =
#                 calc_mse(orig_data_response[drawing],
#                          training_linear_svm_reg_unprivatized_pred)
#               
#               # The predictions for the testing set Linear SVM
#               linear_svm_reg_unprivatized_pred = 
#                 predict(no_cgapd_linear_svm, x_dat_holdout_unprotected[,-1])
#               
#               # The MSE for the testing set Linear SVM
#               linear_svm_reg_mse_orig_data =
#                 calc_mse(orig_data_response[-drawing],
#                          linear_svm_reg_unprivatized_pred)
#               
#               
#               # Polynomial SVM before GAPD
#               no_cgapd_poly_svm_tune = tune.svm(x = x_dat_unprotected[,-1], 
#                                                 y = sample_orig_data_response,
#                                                 gamma = seq(0.001,0.1,length.out = 2), 
#                                                 cost = seq(0.1,1,length.out = 2),
#                                                 degree = seq(1,3, length.out=2),
#                                                 kernel = c("polynomial"))
#               # Get the best tuned polynomial SVM
#               no_cgapd_poly_svm = no_cgapd_poly_svm_tune$best.model
#               
#               # The training set predicitons for the polynomial SVM
#               training_poly_svm_reg_unprivatized_pred = 
#                 predict(no_cgapd_poly_svm, x_dat_unprotected[,-1])
#               
#               # the training set MSE for the polynomial SVM
#               training_poly_svm_reg_mse_orig_data =
#                 calc_mse(orig_data_response[drawing],
#                          training_poly_svm_reg_unprivatized_pred)
#               
#               # The testing set predicitons using the polynomial SVM
#               poly_svm_reg_unprivatized_pred = 
#                 predict(no_cgapd_poly_svm, x_dat_holdout_unprotected[,-1])
#               
#               # The testing set MSE for the polynomial SVM
#               poly_svm_reg_mse_orig_data =
#                 calc_mse(orig_data_response[-drawing],
#                          poly_svm_reg_unprivatized_pred)
#               
#               # Random forest on data before cgapd
#               ranger_rf_model_orig_data = ranger( 
#                 x = model_data[drawing,organized_vars$predictors],
#                 y = orig_data_response[drawing], 
#                 num.trees = 500,
#                 num.threads = 1,
#                 mtry = ceiling(1 / 3 * num_x_vars))
#               
#               # The training set predictions from random forest
#               training_ranger_rf_pred_orig_data = 
#                 predict(ranger_rf_model_orig_data,
#                         model_data[drawing,organized_vars$predictors])$predictions
#               
#               # The training set MSE for the random forest
#               training_ranger_rf_mse_orig_data =
#                 calc_mse(orig_data_response[drawing],
#                          training_ranger_rf_pred_orig_data)
#               
#               # Testing set predictions for random forest
#               ranger_rf_pred_orig_data = predict(ranger_rf_model_orig_data,
#                                                  model_data[-drawing,organized_vars$predictors])$predictions
#               # Testing set MSE for random forest
#               ranger_rf_mse_orig_data =
#                 calc_mse(orig_data_response[-drawing],
#                          ranger_rf_pred_orig_data)
#               
#               # XGBT on data before cgapd
#               xgbt_regress_model_orig_data =
#                 xgboost(data = as.matrix(sample_unprotected_data[,organized_vars$predictors]),
#                         label = orig_data_response[drawing],
#                         nrounds = 40,
#                         nthread = 1)
#               
#               # The training set predictions for XGB tree
#               training_xgbt_pred_orig_data = 
#                 predict(xgbt_regress_model_orig_data,
#                         as.matrix(sample_unprotected_data[,organized_vars$predictors]))
#               # Training set MSE for XGB tree for 
#               training_xgbt_mse_orig_data =
#                 calc_mse(orig_data_response[drawing],
#                          training_xgbt_pred_orig_data)
#               
#               # The testing set XGB tree predictions
#               xgbt_pred_orig_data = 
#                 predict(xgbt_regress_model_orig_data,
#                         as.matrix(model_data[-drawing,organized_vars$predictors]))
#               
#               # The testing set MSE for XGB tree
#               xgbt_mse_orig_data =
#                 calc_mse(orig_data_response[-drawing],
#                          xgbt_pred_orig_data)
#               
#               
#               #print("Constructing the stacked ensemble")
#               ###
#               ### Stacked Ensemble Equal Weighting ----
#               # CGAPD
#               avg_stacked_ensemble_predictions_cgapd = (linear_reg_pred  + 
#                                                           lasso_pred  +  cgapd_nnet_reg_pred + 
#                                                           cgapd_radial_svm_reg_pred  + 
#                                                           cgapd_linear_svm_reg_pred  + 
#                                                           cgapd_poly_svm_reg_pred  +
#                                                           ranger_rf_pred  + xgbt_pred) / 8
#               
#               avg_stacked_ensemble_cgapd_mse =
#                 calc_mse(orig_data_response[-drawing],
#                          avg_stacked_ensemble_predictions_cgapd)
#               # No GAPD
#               avg_stacked_ensemble_pred_no_cgapd = 
#                 (linear_reg_unprivatized_pred  + 
#                    lasso_pred_orig_data  + 
#                    nnet_reg_unprivatized_pred  + 
#                    radial_svm_reg_unprivatized_pred  + 
#                    linear_svm_reg_unprivatized_pred  + 
#                    poly_svm_reg_unprivatized_pred  +
#                    ranger_rf_pred_orig_data  + xgbt_pred_orig_data) / 8
#               
#               avg_stacked_ensemble_no_cgapd_mse =
#                 calc_mse(orig_data_response[-drawing],
#                          avg_stacked_ensemble_pred_no_cgapd)
#               
#               
#               ### Stacked Ensemble Equal Weighting by Proportion of variance Explained ----------------
#               # under CGADP
#               # in this order:
#               # linear regression, LASSO, neural net, radial SVM, linear SVM,
#               # poly SVM, RF, XGB Tree
#               training_set_mses_cgapd = 
#                 c(training_lin_reg_mse, training_lasso_mse, 
#                   training_cgapd_nnet_reg_mse, training_cgapd_radial_svm_reg_mse,
#                   training_cgapd_linear_svm_reg_mse, 
#                   training_cgapd_poly_svm_reg_mse, training_ranger_rf_mse,
#                   training_xgbt_mse)
#               
#               # The models with the lowest MSE get the highest weight
#               prop_var_exp_cgapd_weighting = 
#                 (training_set_mses_cgapd / sum(training_set_mses_cgapd))^(-1)
#               
#               prop_var_exp_cgapd_weighting = prop_var_exp_cgapd_weighting / 
#                 sum(prop_var_exp_cgapd_weighting)
#               
#               
#               
#               # Get the predictions for the stacked ensemble for the prop
#               # of variance explained
#               prop_var_stacked_ens_cgapd_pred = 
#                 linear_reg_pred * prop_var_exp_cgapd_weighting[1] +
#                 lasso_pred * prop_var_exp_cgapd_weighting[2] + 
#                 cgapd_nnet_reg_pred * prop_var_exp_cgapd_weighting[3] + 
#                 cgapd_radial_svm_reg_pred * prop_var_exp_cgapd_weighting[4] + 
#                 cgapd_linear_svm_reg_pred * prop_var_exp_cgapd_weighting[5] + 
#                 cgapd_poly_svm_reg_pred * prop_var_exp_cgapd_weighting[6] +
#                 ranger_rf_pred * prop_var_exp_cgapd_weighting[7] + 
#                 xgbt_pred * prop_var_exp_cgapd_weighting[8]
#               
#               
#               
#               # Get the MSE
#               prop_var_stacked_ens_cgapd_mse = 
#                 calc_mse(orig_data_response[-drawing],
#                          prop_var_stacked_ens_cgapd_pred)
#               
#               
#               # Get the Stacked Ensemble (Prop. of Var. explained) with No GADP 
#               
#               # in this order:
#               # linear regression, LASSO, neural net, radial SVM, linear SVM,
#               # poly SVM, RF, XGB Tree
#               training_set_mses_no_cgapd = 
#                 c(training_linear_reg_mse_orig_data, 
#                   training_lasso_mse_orig_data, 
#                   training_nnet_reg_unprivatized_mse, 
#                   training_radial_svm_reg_mse_orig_data,
#                   training_linear_svm_reg_mse_orig_data, 
#                   training_poly_svm_reg_mse_orig_data, 
#                   training_ranger_rf_mse_orig_data,
#                   training_xgbt_mse_orig_data)
#               
#               # weighting for the SE
#               prop_var_exp_no_cgapd_weighting = 
#                 (training_set_mses_no_cgapd / sum(training_set_mses_no_cgapd))^(-1)
#               
#               prop_var_exp_no_cgapd_weighting = prop_var_exp_no_cgapd_weighting /
#                 sum(prop_var_exp_no_cgapd_weighting)
#               
#               
#               # get the predictions
#               prop_var_stacked_ens_no_cgapd_pred = 
#                 linear_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[1] +
#                 lasso_pred_orig_data * prop_var_exp_no_cgapd_weighting[2] + 
#                 nnet_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[3] + 
#                 radial_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[4] + 
#                 linear_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[5] + 
#                 poly_svm_reg_unprivatized_pred * prop_var_exp_no_cgapd_weighting[6] +
#                 ranger_rf_pred_orig_data * prop_var_exp_no_cgapd_weighting[7] + 
#                 xgbt_pred_orig_data * prop_var_exp_no_cgapd_weighting[8]
#               
#               # Get the MSE
#               prop_var_stacked_ens_no_cgapd_mse = 
#                 calc_mse(orig_data_response[-drawing],
#                          prop_var_stacked_ens_no_cgapd_pred)
#               
#               
#               # Compile MSEs into a vector
#               cgapd_mse_vector = 
#                 c(linear_reg_mse, lasso_mse, cgapd_nnet_reg_mse,
#                   cgapd_radial_svm_reg_mse, cgapd_linear_svm_reg_mse,
#                   cgapd_poly_svm_reg_mse, ranger_rf_mse,
#                   xgbt_mse, avg_stacked_ensemble_cgapd_mse, 
#                   prop_var_stacked_ens_cgapd_mse)
#               
#               no_cgapd_mse_vector =
#                 c(linear_reg_mse_orig_data, lasso_mse_orig_data,
#                   nnet_reg_mse_orig_data, radial_svm_reg_mse_orig_data,
#                   linear_svm_reg_mse_orig_data, poly_svm_reg_mse_orig_data,
#                   ranger_rf_mse_orig_data, xgbt_mse_orig_data,
#                   avg_stacked_ensemble_no_cgapd_mse, 
#                   prop_var_stacked_ens_no_cgapd_mse)
#               
#               #print("Constructing the final results")
#               
#               # Dataframe names
#               data_frame_names = c("seed", "iteration", "training_size","testing_size", "y", 
#                                    "LR", "LASSO", "NNET", "Radial_SVM",
#                                    "Linear_SVM", "Poly_SVM", "RF", "XGBTree",
#                                    "Stacked_equal", "Stacked_prop","cgadp", "theta")
#               
#               
#               # construct the dataframe of results
#               # With CGADP
#               
#               estimator_reg_df = matrix( rep(0, length(data_frame_names) * 2), nrow = 2, ncol = length(data_frame_names), byrow = TRUE)
#               estimator_reg_df[1,] = c(dat.seed,iteration, n[samp_size_index], testing_n, 
#                                        response_index, cgapd_mse_vector, 1, theta_value)
#               
#               # Adding results for LASSO without CGAPD
#               estimator_reg_df[2,] = c(dat.seed, iteration, 
#                                        n[samp_size_index], testing_n,
#                                        response_index, no_cgapd_mse_vector, 0, theta_value)
#               
#               
#               colnames(estimator_reg_df) = data_frame_names
#               
#               return(estimator_reg_df)
#             }
#   #### Stopping Cluster -------------
#   stopCluster(sim_cluster)
#   
#   results = data.frame(results)
#   
#   # Add name of the estimators
#   
#   return(results)
# }


# source(paste0(here::here(), "/Code/Model-Generation-Code.R"))
# 
# model_set_1_rho_0.2 = model_set_1$rho_0.2 
# 
# model_set_1_rho_0.4 = model_set_1$rho_0.4
# 
# model_set_1_rho_0.6 = model_set_1$rho_0.6 
# 
# model_set_1_rho_0.8 = model_set_1$rho_0.8 
# 
# var_names = colnames(model_set_1_rho_0.4)

# retrieve the target variables and predictor variables

# 
# organized_vars = get_variables(variable_names = var_names)
# 
# 
# 
# test_formula = sdc_formula_generation(organized_vars$responses, organized_vars$predictors)
# 
# # correlation of 0.1 does not work. Correlation of 0.2 does not work
# aa = shuffle(model_set_1_rho_0.4, test_formula)


