#' Targets R for SDC GADP and CGADP comparisons
#' Thomas Johnson III
#' 10-10-2023
#' 
#' 
library(targets)
library(tidyverse)
# Uses doFuture and future.callr for parallelization
library(doFuture)
library(future.callr)

# Directory search function to safegauard against code running outside of
# the project directory. This is not foolproof. It will only search from the 
# home directory or "~" unless altered.
directory_search = function(start_directory = "~/")
{
  
  # check for the necessary files
  check_files = function(project_directory)
  {
    file_bool_rproj = file.exists(paste0(start_directory,"/SDC-Comparisons.Rproj"))
    # Prepare to search for code files
    code_dir = "/Code/"
    necessary_code_files = c("Basic-Funcs.R", "Estimator-Evaluation.R",
                             "CE-data-cleaning.R", "GADP-from-scratch.R",
                             "Iterative-Model-Gen-Code.R")
    # Prepare to search for report files
    reports_dir = "/Code/Reports/"
    necessary_report_files = c("More-tuned-SDC-GAPD-CGAPD-Comparisons-01-29-2025.qmd",
                               "CE-Real-Data-Tables-fincbtax-salaryx-tuned-sims.qmd")
    
    # check the code files are present
    code_files_bool = logical(length(necessary_code_files))
    # for loop to iterate for code files
    for(index_bool in 1:length(necessary_code_files))
    {
      if(file.exists(paste0(project_directory, code_dir, 
                            necessary_code_files[index_bool])))
      {
        code_files_bool[index_bool] = TRUE
      }
    }
    if(any(code_files_bool) == FALSE)
    {
      print("Code files were not detected.")
      print("Please manually specify the project directory (`base_dir` = path/to/SDC-Comparisons-Submission/).")
    }
    
    # logical vector for managing detection of report files
    report_files_bool = logical(length(necessary_report_files))
    
    # for loop to iterate for report files
    for(index_bool in 1:length(necessary_report_files))
    {
      if(file.exists(paste0(project_directory, reports_dir,
                            necessary_report_files[index_bool])))
      {
        report_files_bool[index_bool] = TRUE
      }
      
    }
    if(any(report_files_bool) == FALSE)
    {
      print("Code files were not detected.")
      print("Please manually specify the project directory (`base_dir` = path/to/SDC-Comparisons-Submission/).")
    }
    # boolean to determine if the files were not found
    # TRUE if all files are found, FLASE otherwise
    missing_files_bool = all(c(code_files_bool, report_files_bool, file_bool_rproj) == TRUE)
    
    return(missing_files_bool)
  }
  # Attempt to get the project directory via a recursive search
  obtain_project_directory = function(suggested_directory)
  {
    potential_directories =list.dirs(suggested_directory)
    # indices of the directories that belong to the project
    project_directory_index = which(stringr::str_detect(potential_directories, "SDC-Comparisons-Submission$") == TRUE)
    # Get the project directory
    project_directory = potential_directories[project_directory_index]
    
    # ensure that the necessary files are present
    file_presence_bool = check_files(project_directory)
    
    if(file_presence_bool == TRUE)
    {
      # Do nothing project directory will be returned
    }
    else
    {
      print("Could not find directory for the project. Please manually specify in `base_dir`.")
      quit(save = "no")
    }
    return(project_directory)
  }
  
  
  if (is.na(start_directory) == TRUE)
  {
    # If nothing specified, start in the home directory
    start_directory = "~/"
  }
  
  found_project_directory = NA
  
  # If the substring of the project directory is found
  if(stringr::str_detect(start_directory, "SDC-Comparisons-Submission"))
  {
    files_all_present_bool = check_files(start_directory)
    
    if(files_all_present_bool == TRUE)
    {
      # When starting directory is project directory
      print("Relevant files found.")
      found_project_directory = start_directory
      return(found_project_directory)
    }
    else
    {
      print("Relevant files are not found.")
      # start in home directory
      found_project_directory = obtain_project_directory("~")
      
      return(found_project_directory)
    }
  }
  # If not in anything that could be considered the project directory
  else
  {
    found_project_directory = obtain_project_directory(start_directory)
    print(found_project_directory)
    
    if(is.na(found_project_directory) == TRUE | is.character(found_project_directory) == FALSE)
    {
      found_project_directory = obtain_project_directory("~/")
    }
    
    return(found_project_directory)
  }
  print("An unexpected situation has emerged. None of the previous cases were met.")
  print("Attempting to find project directory. If the project directory is not found")
  print("it will have to be manually specified in `base_dir`.")
  found_project_directory = obtain_project_directory(start_directory)
  
  if(is.na(found_project_directory) == TRUE | is.character(found_project_directory) == FALSE)
  {
    found_project_directory = obtain_project_directory("~/")
  }
  
  return(found_project_directory)
}




# Assumes here package can accurately locate the project directory. This assumes
# you can open up the R project to run the code.
# If here package cannot locate the project directory or you cannot 
# open up the R project, assign
# base_dir the path to the R project. 
base_dir = directory_search(here::here())

print("Please verify that the code is running in the project directory: SDC-Comparisons-Submission/")
print("Your current working directory is: ")
print(paste0(base_dir))

setwd(base_dir)

source(paste0(base_dir, "/Code/Basic-Funcs.R"))
source(paste0(base_dir, "/Code/Estimator-Evaluation.R"))
source(paste0(base_dir, "/Code/Iterative-Model-Gen-Code.R"))

source(paste0(base_dir, "/Code/CE-data-cleaning.R"))

model_sim_runs_directory = paste0(base_dir, "/Data/tuned-sims/")

if(dir.exists(model_sim_runs_directory) == FALSE)
{
  dir.create(model_sim_runs_directory)
}


model_1_sim_runs_directory = paste0(model_sim_runs_directory, "Model-1/")

if(dir.exists(model_1_sim_runs_directory) == FALSE)
{
  dir.create(model_1_sim_runs_directory)
}

model_2_sim_runs_directory = paste0(model_sim_runs_directory, "Model-2/")

if(dir.exists(model_2_sim_runs_directory) == FALSE)
{
  dir.create(model_2_sim_runs_directory)
}


model_3_sim_runs_directory = paste0(model_sim_runs_directory, "Model-3/")

if(dir.exists(model_3_sim_runs_directory) == FALSE)
{
  dir.create(model_3_sim_runs_directory)
}

CE_sim_runs_directory = paste0(model_sim_runs_directory, "CE/")

if(dir.exists(CE_sim_runs_directory) == FALSE)
{
  dir.create(CE_sim_runs_directory)
}

# training set sizes
## Test the pipeline
# n_sizes = 500 
## Use training set sizes 500, 1000, 2000 for simulation models full results
n_sizes =  c(500, 1000, 2000)

# Allcocate logical processors/threads/cores
# edit to ensure computing system does not deadlock. You can modify
# based on your own experiences and resources.
core_alloc = parallelly::availableCores(omit = 2) # parallelly::availableCores(omit = 1)

# number of simulations
# test pipeline
# m_sims = 2
# 1000 simulations for obtaining results
m_sims = 1000

list(
  # Model 1 GADP ------------
  ## 10 Predictors --------------
  ### rho = 0.2 ------------
  tar_target(model_1_rho_02_theta_02_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
             ),
  tar_target(model_1_rho_02_theta_02_pred_10_save,
             produce_csv(
              created_object= model_1_rho_02_theta_02_pred_10, 
              base_str = model_1_sim_runs_directory, 
              file_name_str = "model-1-gadp-rho-0_2-theta-0_2-pred-10.csv")),
  
  tar_target(model_1_rho_02_theta_04_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_02_theta_04_pred_10_save,
             produce_csv(
               created_object= model_1_rho_02_theta_04_pred_10, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_2-theta-0_4-pred-10.csv")),
  
  tar_target(model_1_rho_02_theta_06_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_02_theta_06_pred_10_save,
             produce_csv(
               created_object= model_1_rho_02_theta_06_pred_10, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_2-theta-0_6-pred-10.csv")),
  
  tar_target(model_1_rho_02_theta_08_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_02_theta_08_pred_10_save,
             produce_csv(
               created_object= model_1_rho_02_theta_08_pred_10, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_2-theta-0_8-pred-10.csv")),
  
  ### rho = 0.4 -----------------
  tar_target(model_1_rho_04_theta_02_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_04_theta_02_pred_10_save,
             produce_csv(
               created_object= model_1_rho_02_theta_02_pred_10, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_4-theta-0_2-pred-10.csv")),
  
  tar_target(model_1_rho_04_theta_04_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_04_theta_04_pred_10_save,
             produce_csv(
               created_object= model_1_rho_02_theta_04_pred_10, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_4-theta-0_4-pred-10.csv")),
  
  tar_target(model_1_rho_04_theta_06_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_04_theta_06_pred_10_save,
             produce_csv(
               created_object= model_1_rho_02_theta_06_pred_10, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_4-theta-0_6-pred-10.csv")),
  
  
  tar_target(model_1_rho_04_theta_08_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_04_theta_08_pred_10_save,
             produce_csv(
               created_object= model_1_rho_02_theta_08_pred_10, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_4-theta-0_8-pred-10.csv")),
  
  ### rho = 0.6 -------------------------
  
  tar_target(model_1_rho_06_theta_02_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_06_theta_02_pred_10_save,
             produce_csv(
               created_object= model_1_rho_06_theta_02_pred_10, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_6-theta-0_2-pred-10.csv")),
  
  tar_target(model_1_rho_06_theta_04_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_06_theta_04_pred_10_save,
             produce_csv(
               created_object= model_1_rho_06_theta_04_pred_10, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_6-theta-0_4-pred-10.csv")),
  
  tar_target(model_1_rho_06_theta_06_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_06_theta_06_pred_10_save,
             produce_csv(
               created_object= model_1_rho_06_theta_06_pred_10, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_6-theta-0_6-pred-10.csv")),
  
  tar_target(model_1_rho_06_theta_08_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_06_theta_08_pred_10_save,
             produce_csv(
               created_object= model_1_rho_06_theta_08_pred_10, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_6-theta-0_8-pred-10.csv")),
  
  
  
  ## 50 Predictors ------------
  ### rho = 0.2 -------------------
  tar_target(model_1_rho_02_theta_02_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_02_theta_02_pred_50_save,
             produce_csv(
               created_object= model_1_rho_02_theta_02_pred_50, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_2-theta-0_2-pred-50.csv")),
  
  tar_target(model_1_rho_02_theta_04_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_02_theta_04_pred_50_save,
             produce_csv(
               created_object= model_1_rho_02_theta_04_pred_50, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_2-theta-0_4-pred-50.csv")),
  
  tar_target(model_1_rho_02_theta_06_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_02_theta_06_pred_50_save,
             produce_csv(
               created_object= model_1_rho_02_theta_06_pred_50, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_2-theta-0_6-pred-50.csv")),
  
  tar_target(model_1_rho_02_theta_08_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_02_theta_08_pred_50_save,
             produce_csv(
               created_object= model_1_rho_02_theta_08_pred_50, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_2-theta-0_8-pred-50.csv")),
  
  ### rho = 0.4 -------------------------
  
  tar_target(model_1_rho_04_theta_02_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_04_theta_02_pred_50_save,
             produce_csv(
               created_object= model_1_rho_04_theta_02_pred_50, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_4-theta-0_2-pred-50.csv")),
  
  tar_target(model_1_rho_04_theta_04_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_04_theta_04_pred_50_save,
             produce_csv(
               created_object= model_1_rho_04_theta_04_pred_50, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_4-theta-0_4-pred-50.csv")),
  
  tar_target(model_1_rho_04_theta_06_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_04_theta_06_pred_50_save,
             produce_csv(
               created_object= model_1_rho_04_theta_06_pred_50, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_4-theta-0_6-pred-50.csv")),
  
  tar_target(model_1_rho_04_theta_08_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_04_theta_08_pred_50_save,
             produce_csv(
               created_object= model_1_rho_04_theta_08_pred_50, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_4-theta-0_8-pred-50.csv")),
  
  ### rho = 0.6 -----
  tar_target(model_1_rho_06_theta_02_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_06_theta_02_pred_50_save,
             produce_csv(
               created_object= model_1_rho_06_theta_02_pred_50, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_6-theta-0_2-pred-50.csv")),
  
  tar_target(model_1_rho_06_theta_04_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_06_theta_04_pred_50_save,
             produce_csv(
               created_object= model_1_rho_06_theta_04_pred_50, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_6-theta-0_4-pred-50.csv")),
  
  tar_target(model_1_rho_06_theta_06_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_06_theta_06_pred_50_save,
             produce_csv(
               created_object= model_1_rho_06_theta_06_pred_50, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_6-theta-0_6-pred-50.csv")),
  
  tar_target(model_1_rho_06_theta_08_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_06_theta_08_pred_50_save,
             produce_csv(
               created_object= model_1_rho_06_theta_08_pred_50, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_6-theta-0_8-pred-50.csv")),
  
  ## 100 Predictors -------------
  ### rho = 0.2 -------
  tar_target(model_1_rho_02_theta_02_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_02_theta_02_pred_100_save,
             produce_csv(
               created_object= model_1_rho_02_theta_02_pred_100, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_2-theta-0_2-pred-100.csv")),
  
  tar_target(model_1_rho_02_theta_04_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_02_theta_04_pred_100_save,
             produce_csv(
               created_object= model_1_rho_02_theta_04_pred_100, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_2-theta-0_4-pred-100.csv")),
  
  tar_target(model_1_rho_02_theta_06_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_02_theta_06_pred_100_save,
             produce_csv(
               created_object= model_1_rho_02_theta_06_pred_100, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_2-theta-0_6-pred-100.csv")),
  
  tar_target(model_1_rho_02_theta_08_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_02_theta_08_pred_100_save,
             produce_csv(
               created_object= model_1_rho_02_theta_08_pred_100, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_2-theta-0_8-pred-100.csv")),
  
  ### rho = 0.4 -------------------------
  
  tar_target(model_1_rho_04_theta_02_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_04_theta_02_pred_100_save,
             produce_csv(
               created_object= model_1_rho_04_theta_02_pred_100, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_4-theta-0_2-pred-100.csv")),
  
  tar_target(model_1_rho_04_theta_04_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_04_theta_04_pred_100_save,
             produce_csv(
               created_object= model_1_rho_04_theta_04_pred_100, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_4-theta-0_4-pred-100.csv")),
  
  tar_target(model_1_rho_04_theta_06_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_04_theta_06_pred_100_save,
             produce_csv(
               created_object= model_1_rho_04_theta_06_pred_100, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_4-theta-0_6-pred-100.csv")),
  
  tar_target(model_1_rho_04_theta_08_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_04_theta_08_pred_100_save,
             produce_csv(
               created_object= model_1_rho_04_theta_08_pred_100, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_4-theta-0_8-pred-100.csv")),
  
  ### rho = 0.6 -----
  tar_target(model_1_rho_06_theta_02_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_06_theta_02_pred_100_save,
             produce_csv(
               created_object= model_1_rho_06_theta_02_pred_100, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_6-theta-0_2-pred-100.csv")),
  
  tar_target(model_1_rho_06_theta_04_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_06_theta_04_pred_100_save,
             produce_csv(
               created_object= model_1_rho_06_theta_04_pred_100, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_6-theta-0_4-pred-100.csv")),
  
  tar_target(model_1_rho_06_theta_06_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_06_theta_06_pred_100_save,
             produce_csv(
               created_object= model_1_rho_06_theta_06_pred_100, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_6-theta-0_6-pred-100.csv")),
  
  tar_target(model_1_rho_06_theta_08_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_1_rho_06_theta_08_pred_100_save,
             produce_csv(
               created_object= model_1_rho_06_theta_08_pred_100, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-gadp-rho-0_6-theta-0_8-pred-100.csv")),
  
  # Model 1 CGADP ------------
  ## 10 Predictors --------------
  ### rho = 0.2 ------------
  tar_target(model_1_rho_02_theta_02_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_02_theta_02_pred_10.cgadp_save,
             produce_csv(
               created_object= model_1_rho_02_theta_02_pred_10.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_2-theta-0_2-pred-10.csv")),
  
  tar_target(model_1_rho_02_theta_04_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_02_theta_04_pred_10.cgadp_save,
             produce_csv(
               created_object= model_1_rho_02_theta_04_pred_10.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_2-theta-0_4-pred-10.csv")),
  
  tar_target(model_1_rho_02_theta_06_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_02_theta_06_pred_10.cgadp_save,
             produce_csv(
               created_object= model_1_rho_02_theta_06_pred_10.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_2-theta-0_6-pred-10.csv")),
  
  tar_target(model_1_rho_02_theta_08_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_02_theta_08_pred_10.cgadp_save,
             produce_csv(
               created_object= model_1_rho_02_theta_08_pred_10.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_2-theta-0_8-pred-10.csv")),
  
  ### rho = 0.4 ----------------
  tar_target(model_1_rho_04_theta_02_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_04_theta_02_pred_10.cgadp_save,
             produce_csv(
               created_object= model_1_rho_04_theta_02_pred_10.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_4-theta-0_2-pred-10.csv")),
  
  tar_target(model_1_rho_04_theta_04_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_04_theta_04_pred_10.cgadp_save,
             produce_csv(
               created_object= model_1_rho_04_theta_04_pred_10.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_4-theta-0_4-pred-10.csv")),
  
  tar_target(model_1_rho_04_theta_06_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_04_theta_06_pred_10.cgadp_save,
             produce_csv(
               created_object= model_1_rho_04_theta_06_pred_10.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_4-theta-0_6-pred-10.csv")),
  
  tar_target(model_1_rho_04_theta_08_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_04_theta_08_pred_10.cgadp_save,
             produce_csv(
               created_object= model_1_rho_04_theta_08_pred_10.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_4-theta-0_8-pred-10.csv")),
  
  ### rho = 0.6 ---------------------------------
  tar_target(model_1_rho_06_theta_02_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_06_theta_02_pred_10.cgadp_save,
             produce_csv(
               created_object= model_1_rho_06_theta_02_pred_10.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_6-theta-0_2-pred-10.csv")),
  
  tar_target(model_1_rho_06_theta_04_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_06_theta_04_pred_10.cgadp_save,
             produce_csv(
               created_object= model_1_rho_06_theta_04_pred_10.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_6-theta-0_4-pred-10.csv")),
  
  tar_target(model_1_rho_06_theta_06_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_06_theta_06_pred_10.cgadp_save,
             produce_csv(
               created_object= model_1_rho_06_theta_06_pred_10.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_6-theta-0_6-pred-10.csv")),
  
  tar_target(model_1_rho_06_theta_08_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_06_theta_08_pred_10.cgadp_save,
             produce_csv(
               created_object= model_1_rho_06_theta_08_pred_10.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_6-theta-0_8-pred-10.csv")),
  
  ## 50 Predictors -----------------------------
  ### rho = 0.2 ------------
  tar_target(model_1_rho_02_theta_02_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_02_theta_02_pred_50.cgadp_save,
             produce_csv(
               created_object= model_1_rho_02_theta_02_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_2-theta-0_2-pred-50.csv")),
  
  tar_target(model_1_rho_02_theta_04_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_02_theta_04_pred_50.cgadp_save,
             produce_csv(
               created_object= model_1_rho_02_theta_04_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_2-theta-0_4-pred-50.csv")),
  
  tar_target(model_1_rho_02_theta_06_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_02_theta_06_pred_50.cgadp_save,
             produce_csv(
               created_object= model_1_rho_02_theta_06_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_2-theta-0_6-pred-50.csv")),
  
  tar_target(model_1_rho_02_theta_08_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_02_theta_08_pred_50.cgadp_save,
             produce_csv(
               created_object= model_1_rho_02_theta_08_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_2-theta-0_8-pred-50.csv")),
  
  ### rho = 0.4 ----------------
  tar_target(model_1_rho_04_theta_02_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_04_theta_02_pred_50.cgadp_save,
             produce_csv(
               created_object= model_1_rho_04_theta_02_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_4-theta-0_2-pred-50.csv")),
  
  tar_target(model_1_rho_04_theta_04_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_04_theta_04_pred_50.cgadp_save,
             produce_csv(
               created_object= model_1_rho_04_theta_04_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_4-theta-0_4-pred-50.csv")),
  
  tar_target(model_1_rho_04_theta_06_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_04_theta_06_pred_50.cgadp_save,
             produce_csv(
               created_object= model_1_rho_04_theta_06_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_4-theta-0_6-pred-50.csv")),
  
  tar_target(model_1_rho_04_theta_08_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_04_theta_08_pred_50.cgadp_save,
             produce_csv(
               created_object= model_1_rho_04_theta_08_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_4-theta-0_8-pred-50.csv")),
  
  ### rho = 0.6 ---------------------------------
  tar_target(model_1_rho_06_theta_02_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_06_theta_02_pred_50.cgadp_save,
             produce_csv(
               created_object= model_1_rho_06_theta_02_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_6-theta-0_2-pred-50.csv")),
  
  tar_target(model_1_rho_06_theta_04_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_06_theta_04_pred_50.cgadp_save,
             produce_csv(
               created_object= model_1_rho_06_theta_04_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_6-theta-0_4-pred-50.csv")),
  
  tar_target(model_1_rho_06_theta_06_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_06_theta_06_pred_50.cgadp_save,
             produce_csv(
               created_object= model_1_rho_06_theta_06_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_6-theta-0_6-pred-50.csv")),
  
  tar_target(model_1_rho_06_theta_08_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_06_theta_08_pred_50.cgadp_save,
             produce_csv(
               created_object= model_1_rho_06_theta_08_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_6-theta-0_8-pred-50.csv")),
  
  ## 100 Predictors -----------------------------
  ### rho = 0.2 ------------
  tar_target(model_1_rho_02_theta_02_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_02_theta_02_pred_100.cgadp_save,
             produce_csv(
               created_object= model_1_rho_02_theta_02_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_2-theta-0_2-pred-100.csv")),
  
  tar_target(model_1_rho_02_theta_04_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_02_theta_04_pred_100.cgadp_save,
             produce_csv(
               created_object= model_1_rho_02_theta_04_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_2-theta-0_4-pred-100.csv")),
  
  tar_target(model_1_rho_02_theta_06_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_02_theta_06_pred_100.cgadp_save,
             produce_csv(
               created_object= model_1_rho_02_theta_06_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_2-theta-0_6-pred-100.csv")),
  
  tar_target(model_1_rho_02_theta_08_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_02_theta_08_pred_100.cgadp_save,
             produce_csv(
               created_object= model_1_rho_02_theta_08_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_2-theta-0_8-pred-100.csv")),
  
  ### rho = 0.4 ----------------
  tar_target(model_1_rho_04_theta_02_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_04_theta_02_pred_100.cgadp_save,
             produce_csv(
               created_object= model_1_rho_04_theta_02_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_4-theta-0_2-pred-100.csv")),
  
  tar_target(model_1_rho_04_theta_04_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_04_theta_04_pred_100.cgadp_save,
             produce_csv(
               created_object= model_1_rho_04_theta_04_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_4-theta-0_4-pred-100.csv")),
  
  tar_target(model_1_rho_04_theta_06_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_04_theta_06_pred_100.cgadp_save,
             produce_csv(
               created_object= model_1_rho_04_theta_06_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_4-theta-0_6-pred-100.csv")),
  
  tar_target(model_1_rho_04_theta_08_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_04_theta_08_pred_100.cgadp_save,
             produce_csv(
               created_object= model_1_rho_04_theta_08_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_4-theta-0_8-pred-100.csv")),
  
  ### rho = 0.6 ---------------------------------
  tar_target(model_1_rho_06_theta_02_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_06_theta_02_pred_100.cgadp_save,
             produce_csv(
               created_object= model_1_rho_06_theta_02_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_6-theta-0_2-pred-100.csv")),
  
  tar_target(model_1_rho_06_theta_04_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_06_theta_04_pred_100.cgadp_save,
             produce_csv(
               created_object= model_1_rho_06_theta_04_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_6-theta-0_4-pred-100.csv")),
  
  tar_target(model_1_rho_06_theta_06_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_06_theta_06_pred_100.cgadp_save,
             produce_csv(
               created_object= model_1_rho_06_theta_06_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_6-theta-0_6-pred-100.csv")),
  
  tar_target(model_1_rho_06_theta_08_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.1_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_1_rho_06_theta_08_pred_100.cgadp_save,
             produce_csv(
               created_object= model_1_rho_06_theta_08_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-1-cgadp-rho-0_6-theta-0_8-pred-100.csv")),
  
  # Model 2 GADP -----------
  ## 10 Predictors --------------
  ### rho = 0.2 ------------
  tar_target(model_2_rho_02_theta_02_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_02_theta_02_pred_10_save,
             produce_csv(
               created_object= model_2_rho_02_theta_02_pred_10, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_2-theta-0_2-pred-10.csv")),
  
  tar_target(model_2_rho_02_theta_04_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_02_theta_04_pred_10_save,
             produce_csv(
               created_object= model_2_rho_02_theta_04_pred_10, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_2-theta-0_4-pred-10.csv")),
  
  tar_target(model_2_rho_02_theta_06_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_02_theta_06_pred_10_save,
             produce_csv(
               created_object= model_2_rho_02_theta_06_pred_10, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_2-theta-0_6-pred-10.csv")),
  
  tar_target(model_2_rho_02_theta_08_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_02_theta_08_pred_10_save,
             produce_csv(
               created_object= model_2_rho_02_theta_08_pred_10, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_2-theta-0_8-pred-10.csv")),
  
  ### rho = 0.4 -----------------
  tar_target(model_2_rho_04_theta_02_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_04_theta_02_pred_10_save,
             produce_csv(
               created_object= model_2_rho_04_theta_02_pred_10, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_4-theta-0_2-pred-10.csv")),
  
  tar_target(model_2_rho_04_theta_04_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_04_theta_04_pred_10_save,
             produce_csv(
               created_object= model_2_rho_04_theta_04_pred_10, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_4-theta-0_4-pred-10.csv")),
  
  tar_target(model_2_rho_04_theta_06_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_04_theta_06_pred_10_save,
             produce_csv(
               created_object= model_2_rho_04_theta_06_pred_10, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_4-theta-0_6-pred-10.csv")),
  
  
  tar_target(model_2_rho_04_theta_08_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_04_theta_08_pred_10_save,
             produce_csv(
               created_object= model_2_rho_04_theta_08_pred_10, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_4-theta-0_8-pred-10.csv")),
  
  ### rho = 0.6 -------------------------
  
  tar_target(model_2_rho_06_theta_02_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_06_theta_02_pred_10_save,
             produce_csv(
               created_object= model_2_rho_06_theta_02_pred_10, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_6-theta-0_2-pred-10.csv")),
  
  tar_target(model_2_rho_06_theta_04_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_06_theta_04_pred_10_save,
             produce_csv(
               created_object= model_2_rho_06_theta_04_pred_10, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_6-theta-0_4-pred-10.csv")),
  
  tar_target(model_2_rho_06_theta_06_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_06_theta_06_pred_10_save,
             produce_csv(
               created_object= model_2_rho_06_theta_06_pred_10, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_6-theta-0_6-pred-10.csv")),
  
  tar_target(model_2_rho_06_theta_08_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_06_theta_08_pred_10_save,
             produce_csv(
               created_object= model_2_rho_06_theta_08_pred_10, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_6-theta-0_8-pred-10.csv")),
  
  
  
  ## 50 Predictors ------------
  ### rho = 0.2 -------------------
  tar_target(model_2_rho_02_theta_02_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_02_theta_02_pred_50_save,
             produce_csv(
               created_object= model_2_rho_02_theta_02_pred_50, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_2-theta-0_2-pred-50.csv")),
  
  tar_target(model_2_rho_02_theta_04_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_02_theta_04_pred_50_save,
             produce_csv(
               created_object= model_2_rho_02_theta_04_pred_50, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_2-theta-0_4-pred-50.csv")),
  
  tar_target(model_2_rho_02_theta_06_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_02_theta_06_pred_50_save,
             produce_csv(
               created_object= model_2_rho_02_theta_06_pred_50, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_2-theta-0_6-pred-50.csv")),
  
  tar_target(model_2_rho_02_theta_08_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_02_theta_08_pred_50_save,
             produce_csv(
               created_object= model_2_rho_02_theta_08_pred_50, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_2-theta-0_8-pred-50.csv")),
  
  ### rho = 0.4 -------------------------
  
  tar_target(model_2_rho_04_theta_02_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_04_theta_02_pred_50_save,
             produce_csv(
               created_object= model_2_rho_04_theta_02_pred_50, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_4-theta-0_2-pred-50.csv")),
  
  tar_target(model_2_rho_04_theta_04_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_04_theta_04_pred_50_save,
             produce_csv(
               created_object= model_2_rho_04_theta_04_pred_50, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_4-theta-0_4-pred-50.csv")),
  
  tar_target(model_2_rho_04_theta_06_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_04_theta_06_pred_50_save,
             produce_csv(
               created_object= model_2_rho_04_theta_06_pred_50, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_4-theta-0_6-pred-50.csv")),
  
  tar_target(model_2_rho_04_theta_08_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_04_theta_08_pred_50_save,
             produce_csv(
               created_object= model_2_rho_04_theta_08_pred_50, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_4-theta-0_8-pred-50.csv")),
  
  ### rho = 0.6 -----
  tar_target(model_2_rho_06_theta_02_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_06_theta_02_pred_50_save,
             produce_csv(
               created_object= model_2_rho_06_theta_02_pred_50, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_6-theta-0_2-pred-50.csv")),
  
  tar_target(model_2_rho_06_theta_04_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_06_theta_04_pred_50_save,
             produce_csv(
               created_object= model_2_rho_06_theta_04_pred_50, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_6-theta-0_4-pred-50.csv")),
  
  tar_target(model_2_rho_06_theta_06_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_06_theta_06_pred_50_save,
             produce_csv(
               created_object= model_2_rho_06_theta_06_pred_50, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_6-theta-0_6-pred-50.csv")),
  
  tar_target(model_2_rho_06_theta_08_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_06_theta_08_pred_50_save,
             produce_csv(
               created_object= model_2_rho_06_theta_08_pred_50, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_6-theta-0_8-pred-50.csv")),
  
  ## 100 Predictors -------------
  ### rho = 0.2 -------
  tar_target(model_2_rho_02_theta_02_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_02_theta_02_pred_100_save,
             produce_csv(
               created_object= model_2_rho_02_theta_02_pred_100, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_2-theta-0_2-pred-100.csv")),
  
  tar_target(model_2_rho_02_theta_04_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_02_theta_04_pred_100_save,
             produce_csv(
               created_object= model_2_rho_02_theta_04_pred_100, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_2-theta-0_4-pred-100.csv")),
  
  tar_target(model_2_rho_02_theta_06_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_02_theta_06_pred_100_save,
             produce_csv(
               created_object= model_2_rho_02_theta_06_pred_100, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_2-theta-0_6-pred-100.csv")),
  
  tar_target(model_2_rho_02_theta_08_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_02_theta_08_pred_100_save,
             produce_csv(
               created_object= model_2_rho_02_theta_08_pred_100, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_2-theta-0_8-pred-100.csv")),
  
  ### rho = 0.4 -------------------------
  
  tar_target(model_2_rho_04_theta_02_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_04_theta_02_pred_100_save,
             produce_csv(
               created_object= model_2_rho_04_theta_02_pred_100, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_4-theta-0_2-pred-100.csv")),
  
  tar_target(model_2_rho_04_theta_04_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_04_theta_04_pred_100_save,
             produce_csv(
               created_object= model_2_rho_04_theta_04_pred_100, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_4-theta-0_4-pred-100.csv")),
  
  tar_target(model_2_rho_04_theta_06_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_04_theta_06_pred_100_save,
             produce_csv(
               created_object= model_2_rho_04_theta_06_pred_100, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_4-theta-0_6-pred-100.csv")),
  
  tar_target(model_2_rho_04_theta_08_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_04_theta_08_pred_100_save,
             produce_csv(
               created_object= model_2_rho_04_theta_08_pred_100, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_4-theta-0_8-pred-100.csv")),
  
  ### rho = 0.6 -----
  tar_target(model_2_rho_06_theta_02_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_06_theta_02_pred_100_save,
             produce_csv(
               created_object= model_2_rho_06_theta_02_pred_100, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_6-theta-0_2-pred-100.csv")),
  
  tar_target(model_2_rho_06_theta_04_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_06_theta_04_pred_100_save,
             produce_csv(
               created_object= model_2_rho_06_theta_04_pred_100, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_6-theta-0_4-pred-100.csv")),
  
  tar_target(model_2_rho_06_theta_06_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_06_theta_06_pred_100_save,
             produce_csv(
               created_object= model_2_rho_06_theta_06_pred_100, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_6-theta-0_6-pred-100.csv")),
  
  tar_target(model_2_rho_06_theta_08_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_2_rho_06_theta_08_pred_100_save,
             produce_csv(
               created_object= model_2_rho_06_theta_08_pred_100, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-gadp-rho-0_6-theta-0_8-pred-100.csv")),
  
  
  # Model 2 CGADP ----------
  ## 10 Predictors --------------
  ### rho = 0.2 ------------
  tar_target(model_2_rho_02_theta_02_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_02_theta_02_pred_10.cgadp_save,
             produce_csv(
               created_object= model_2_rho_02_theta_02_pred_10.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_2-theta-0_2-pred-10.csv")),
  
  tar_target(model_2_rho_02_theta_04_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_02_theta_04_pred_10.cgadp_save,
             produce_csv(
               created_object= model_2_rho_02_theta_04_pred_10.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_2-theta-0_4-pred-10.csv")),
  
  tar_target(model_2_rho_02_theta_06_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_02_theta_06_pred_10.cgadp_save,
             produce_csv(
               created_object= model_2_rho_02_theta_06_pred_10.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_2-theta-0_6-pred-10.csv")),
  
  tar_target(model_2_rho_02_theta_08_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_02_theta_08_pred_10.cgadp_save,
             produce_csv(
               created_object= model_2_rho_02_theta_08_pred_10.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_2-theta-0_8-pred-10.csv")),
  
  ### rho = 0.4 ----------------
  tar_target(model_2_rho_04_theta_02_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_04_theta_02_pred_10.cgadp_save,
             produce_csv(
               created_object= model_2_rho_04_theta_02_pred_10.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_4-theta-0_2-pred-10.csv")),
  
  tar_target(model_2_rho_04_theta_04_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_04_theta_04_pred_10.cgadp_save,
             produce_csv(
               created_object= model_2_rho_04_theta_04_pred_10.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_4-theta-0_4-pred-10.csv")),
  
  tar_target(model_2_rho_04_theta_06_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_04_theta_06_pred_10.cgadp_save,
             produce_csv(
               created_object= model_2_rho_04_theta_06_pred_10.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_4-theta-0_6-pred-10.csv")),
  
  tar_target(model_2_rho_04_theta_08_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_04_theta_08_pred_10.cgadp_save,
             produce_csv(
               created_object= model_2_rho_04_theta_08_pred_10.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_4-theta-0_8-pred-10.csv")),
  
  ### rho = 0.6 ---------------------------------
  tar_target(model_2_rho_06_theta_02_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_06_theta_02_pred_10.cgadp_save,
             produce_csv(
               created_object= model_2_rho_06_theta_02_pred_10.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_6-theta-0_2-pred-10.csv")),
  
  tar_target(model_2_rho_06_theta_04_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_06_theta_04_pred_10.cgadp_save,
             produce_csv(
               created_object= model_2_rho_06_theta_04_pred_10.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_6-theta-0_4-pred-10.csv")),
  
  tar_target(model_2_rho_06_theta_06_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_06_theta_06_pred_10.cgadp_save,
             produce_csv(
               created_object= model_2_rho_06_theta_06_pred_10.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_6-theta-0_6-pred-10.csv")),
  
  tar_target(model_2_rho_06_theta_08_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_06_theta_08_pred_10.cgadp_save,
             produce_csv(
               created_object= model_2_rho_06_theta_08_pred_10.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_6-theta-0_8-pred-10.csv")),
  
  ## 50 Predictors -----------------------------
  ### rho = 0.2 ------------
  tar_target(model_2_rho_02_theta_02_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_02_theta_02_pred_50.cgadp_save,
             produce_csv(
               created_object= model_2_rho_02_theta_02_pred_50.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_2-theta-0_2-pred-50.csv")),
  
  tar_target(model_2_rho_02_theta_04_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_02_theta_04_pred_50.cgadp_save,
             produce_csv(
               created_object= model_2_rho_02_theta_04_pred_50.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_2-theta-0_4-pred-50.csv")),
  
  tar_target(model_2_rho_02_theta_06_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_02_theta_06_pred_50.cgadp_save,
             produce_csv(
               created_object= model_2_rho_02_theta_06_pred_50.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_2-theta-0_6-pred-50.csv")),
  
  tar_target(model_2_rho_02_theta_08_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_02_theta_08_pred_50.cgadp_save,
             produce_csv(
               created_object= model_2_rho_02_theta_08_pred_50.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_2-theta-0_8-pred-50.csv")),
  
  ### rho = 0.4 ----------------
  tar_target(model_2_rho_04_theta_02_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_04_theta_02_pred_50.cgadp_save,
             produce_csv(
               created_object= model_2_rho_04_theta_02_pred_50.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_4-theta-0_2-pred-50.csv")),
  
  tar_target(model_2_rho_04_theta_04_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_04_theta_04_pred_50.cgadp_save,
             produce_csv(
               created_object= model_2_rho_04_theta_04_pred_50.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_4-theta-0_4-pred-50.csv")),
  
  tar_target(model_2_rho_04_theta_06_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_04_theta_06_pred_50.cgadp_save,
             produce_csv(
               created_object= model_2_rho_04_theta_06_pred_50.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_4-theta-0_6-pred-50.csv")),
  
  tar_target(model_2_rho_04_theta_08_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_04_theta_08_pred_50.cgadp_save,
             produce_csv(
               created_object= model_2_rho_04_theta_08_pred_50.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_4-theta-0_8-pred-50.csv")),
  
  ### rho = 0.6 ---------------------------------
  tar_target(model_2_rho_06_theta_02_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_06_theta_02_pred_50.cgadp_save,
             produce_csv(
               created_object= model_2_rho_06_theta_02_pred_50.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_6-theta-0_2-pred-50.csv")),
  
  tar_target(model_2_rho_06_theta_04_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_06_theta_04_pred_50.cgadp_save,
             produce_csv(
               created_object= model_2_rho_06_theta_04_pred_50.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_6-theta-0_4-pred-50.csv")),
  
  tar_target(model_2_rho_06_theta_06_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_06_theta_06_pred_50.cgadp_save,
             produce_csv(
               created_object= model_2_rho_06_theta_06_pred_50.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_6-theta-0_6-pred-50.csv")),
  
  tar_target(model_2_rho_06_theta_08_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_06_theta_08_pred_50.cgadp_save,
             produce_csv(
               created_object= model_2_rho_06_theta_08_pred_50.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_6-theta-0_8-pred-50.csv")),
  
  ## 100 Predictors -----------------------------
  ### rho = 0.2 ------------
  tar_target(model_2_rho_02_theta_02_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_02_theta_02_pred_100.cgadp_save,
             produce_csv(
               created_object= model_2_rho_02_theta_02_pred_100.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_2-theta-0_2-pred-100.csv")),
  
  tar_target(model_2_rho_02_theta_04_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_02_theta_04_pred_100.cgadp_save,
             produce_csv(
               created_object= model_2_rho_02_theta_04_pred_100.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_2-theta-0_4-pred-100.csv")),
  
  tar_target(model_2_rho_02_theta_06_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_02_theta_06_pred_100.cgadp_save,
             produce_csv(
               created_object= model_2_rho_02_theta_06_pred_100.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_2-theta-0_6-pred-100.csv")),
  
  tar_target(model_2_rho_02_theta_08_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_02_theta_08_pred_100.cgadp_save,
             produce_csv(
               created_object= model_2_rho_02_theta_08_pred_100.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_2-theta-0_8-pred-100.csv")),
  
  ### rho = 0.4 ----------------
  tar_target(model_2_rho_04_theta_02_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_04_theta_02_pred_100.cgadp_save,
             produce_csv(
               created_object= model_2_rho_04_theta_02_pred_100.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_4-theta-0_2-pred-100.csv")),
  
  tar_target(model_2_rho_04_theta_04_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_04_theta_04_pred_100.cgadp_save,
             produce_csv(
               created_object= model_2_rho_04_theta_04_pred_100.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_4-theta-0_4-pred-100.csv")),
  
  tar_target(model_2_rho_04_theta_06_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_04_theta_06_pred_100.cgadp_save,
             produce_csv(
               created_object= model_2_rho_04_theta_06_pred_100.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_4-theta-0_6-pred-100.csv")),
  
  tar_target(model_2_rho_04_theta_08_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_04_theta_08_pred_100.cgadp_save,
             produce_csv(
               created_object= model_2_rho_04_theta_08_pred_100.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_4-theta-0_8-pred-100.csv")),
  
  ### rho = 0.6 ---------------------------------
  tar_target(model_2_rho_06_theta_02_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_06_theta_02_pred_100.cgadp_save,
             produce_csv(
               created_object= model_2_rho_06_theta_02_pred_100.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_6-theta-0_2-pred-100.csv")),
  
  tar_target(model_2_rho_06_theta_04_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_06_theta_04_pred_100.cgadp_save,
             produce_csv(
               created_object= model_2_rho_06_theta_04_pred_100.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_6-theta-0_4-pred-100.csv")),
  
  tar_target(model_2_rho_06_theta_06_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_06_theta_06_pred_100.cgadp_save,
             produce_csv(
               created_object= model_2_rho_06_theta_06_pred_100.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_6-theta-0_6-pred-100.csv")),
  
  tar_target(model_2_rho_06_theta_08_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_2_cgadp_ver_1, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.2_p_ver_1, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_2_rho_06_theta_08_pred_100.cgadp_save,
             produce_csv(
               created_object= model_2_rho_06_theta_08_pred_100.cgadp, 
               base_str = model_2_sim_runs_directory, 
               file_name_str = "model-2-cgadp-rho-0_6-theta-0_8-pred-100.csv")),
  
  
  # Model 3 GADP -----------
  ## 10 Predictors --------------
  ### rho = 0.2 ------------
  tar_target(model_3_rho_02_theta_02_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_02_theta_02_pred_10_save,
             produce_csv(
               created_object= model_3_rho_02_theta_02_pred_10, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_2-theta-0_2-pred-10.csv")),
  
  tar_target(model_3_rho_02_theta_04_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_02_theta_04_pred_10_save,
             produce_csv(
               created_object= model_3_rho_02_theta_04_pred_10, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_2-theta-0_4-pred-10.csv")),
  
  tar_target(model_3_rho_02_theta_06_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_02_theta_06_pred_10_save,
             produce_csv(
               created_object= model_3_rho_02_theta_06_pred_10, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_2-theta-0_6-pred-10.csv")),
  
  tar_target(model_3_rho_02_theta_08_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_02_theta_08_pred_10_save,
             produce_csv(
               created_object= model_3_rho_02_theta_08_pred_10, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_2-theta-0_8-pred-10.csv")),
  
  ### rho = 0.4 -----------------
  tar_target(model_3_rho_04_theta_02_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_04_theta_02_pred_10_save,
             produce_csv(
               created_object= model_3_rho_04_theta_02_pred_10, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_4-theta-0_2-pred-10.csv")),
  
  tar_target(model_3_rho_04_theta_04_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_04_theta_04_pred_10_save,
             produce_csv(
               created_object= model_3_rho_04_theta_04_pred_10, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_4-theta-0_4-pred-10.csv")),
  
  tar_target(model_3_rho_04_theta_06_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_04_theta_06_pred_10_save,
             produce_csv(
               created_object= model_3_rho_04_theta_06_pred_10, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_4-theta-0_6-pred-10.csv")),
  
  
  tar_target(model_3_rho_04_theta_08_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_04_theta_08_pred_10_save,
             produce_csv(
               created_object= model_3_rho_04_theta_08_pred_10, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_4-theta-0_8-pred-10.csv")),
  
  ### rho = 0.6 -------------------------
  
  tar_target(model_3_rho_06_theta_02_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_06_theta_02_pred_10_save,
             produce_csv(
               created_object= model_3_rho_06_theta_02_pred_10, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_6-theta-0_2-pred-10.csv")),
  
  tar_target(model_3_rho_06_theta_04_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_06_theta_04_pred_10_save,
             produce_csv(
               created_object= model_3_rho_06_theta_04_pred_10, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_6-theta-0_4-pred-10.csv")),
  
  tar_target(model_3_rho_06_theta_06_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_06_theta_06_pred_10_save,
             produce_csv(
               created_object= model_3_rho_06_theta_06_pred_10, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_6-theta-0_6-pred-10.csv")),
  
  tar_target(model_3_rho_06_theta_08_pred_10,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_06_theta_08_pred_10_save,
             produce_csv(
               created_object= model_3_rho_06_theta_08_pred_10, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_6-theta-0_8-pred-10.csv")),
  
  
  
  ## 50 Predictors ------------
  ### rho = 0.2 -------------------
  tar_target(model_3_rho_02_theta_02_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_02_theta_02_pred_50_save,
             produce_csv(
               created_object= model_3_rho_02_theta_02_pred_50, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_2-theta-0_2-pred-50.csv")),
  
  tar_target(model_3_rho_02_theta_04_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_02_theta_04_pred_50_save,
             produce_csv(
               created_object= model_3_rho_02_theta_04_pred_50, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_2-theta-0_4-pred-50.csv")),
  
  tar_target(model_3_rho_02_theta_06_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_02_theta_06_pred_50_save,
             produce_csv(
               created_object= model_3_rho_02_theta_06_pred_50, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_2-theta-0_6-pred-50.csv")),
  
  tar_target(model_3_rho_02_theta_08_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_02_theta_08_pred_50_save,
             produce_csv(
               created_object= model_3_rho_02_theta_08_pred_50, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_2-theta-0_8-pred-50.csv")),
  
  ### rho = 0.4 -------------------------
  
  tar_target(model_3_rho_04_theta_02_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_04_theta_02_pred_50_save,
             produce_csv(
               created_object= model_3_rho_04_theta_02_pred_50, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_4-theta-0_2-pred-50.csv")),
  
  tar_target(model_3_rho_04_theta_04_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_04_theta_04_pred_50_save,
             produce_csv(
               created_object= model_3_rho_04_theta_04_pred_50, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_4-theta-0_4-pred-50.csv")),
  
  tar_target(model_3_rho_04_theta_06_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_04_theta_06_pred_50_save,
             produce_csv(
               created_object= model_3_rho_04_theta_06_pred_50, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_4-theta-0_6-pred-50.csv")),
  
  tar_target(model_3_rho_04_theta_08_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_04_theta_08_pred_50_save,
             produce_csv(
               created_object= model_3_rho_04_theta_08_pred_50, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_4-theta-0_8-pred-50.csv")),
  
  ### rho = 0.6 -----
  tar_target(model_3_rho_06_theta_02_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_06_theta_02_pred_50_save,
             produce_csv(
               created_object= model_3_rho_06_theta_02_pred_50, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_6-theta-0_2-pred-50.csv")),
  
  tar_target(model_3_rho_06_theta_04_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_06_theta_04_pred_50_save,
             produce_csv(
               created_object= model_3_rho_06_theta_04_pred_50, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_6-theta-0_4-pred-50.csv")),
  
  tar_target(model_3_rho_06_theta_06_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_06_theta_06_pred_50_save,
             produce_csv(
               created_object= model_3_rho_06_theta_06_pred_50, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_6-theta-0_6-pred-50.csv")),
  
  tar_target(model_3_rho_06_theta_08_pred_50,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_06_theta_08_pred_50_save,
             produce_csv(
               created_object= model_3_rho_06_theta_08_pred_50, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_6-theta-0_8-pred-50.csv")),
  
  ## 100 Predictors -------------
  ### rho = 0.2 -------
  tar_target(model_3_rho_02_theta_02_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_02_theta_02_pred_100_save,
             produce_csv(
               created_object= model_3_rho_02_theta_02_pred_100, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_2-theta-0_2-pred-100.csv")),
  
  tar_target(model_3_rho_02_theta_04_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_02_theta_04_pred_100_save,
             produce_csv(
               created_object= model_3_rho_02_theta_04_pred_100, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_2-theta-0_4-pred-100.csv")),
  
  tar_target(model_3_rho_02_theta_06_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_02_theta_06_pred_100_save,
             produce_csv(
               created_object= model_3_rho_02_theta_06_pred_100, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_2-theta-0_6-pred-100.csv")),
  
  tar_target(model_3_rho_02_theta_08_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_02_theta_08_pred_100_save,
             produce_csv(
               created_object= model_3_rho_02_theta_08_pred_100, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_2-theta-0_8-pred-100.csv")),
  
  ### rho = 0.4 -------------------------
  
  tar_target(model_3_rho_04_theta_02_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_04_theta_02_pred_100_save,
             produce_csv(
               created_object= model_3_rho_04_theta_02_pred_100, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_4-theta-0_2-pred-100.csv")),
  
  tar_target(model_3_rho_04_theta_04_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_04_theta_04_pred_100_save,
             produce_csv(
               created_object= model_3_rho_04_theta_04_pred_100, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_4-theta-0_4-pred-100.csv")),
  
  tar_target(model_3_rho_04_theta_06_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_04_theta_06_pred_100_save,
             produce_csv(
               created_object= model_3_rho_04_theta_06_pred_100, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_4-theta-0_6-pred-100.csv")),
  
  tar_target(model_3_rho_04_theta_08_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_04_theta_08_pred_100_save,
             produce_csv(
               created_object= model_3_rho_04_theta_08_pred_100, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_4-theta-0_8-pred-100.csv")),
  
  ### rho = 0.6 -----
  tar_target(model_3_rho_06_theta_02_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_06_theta_02_pred_100_save,
             produce_csv(
               created_object= model_3_rho_06_theta_02_pred_100, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_6-theta-0_2-pred-100.csv")),
  
  tar_target(model_3_rho_06_theta_04_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_06_theta_04_pred_100_save,
             produce_csv(
               created_object= model_3_rho_06_theta_04_pred_100, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_6-theta-0_4-pred-100.csv")),
  
  tar_target(model_3_rho_06_theta_06_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_06_theta_06_pred_100_save,
             produce_csv(
               created_object= model_3_rho_06_theta_06_pred_100, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_6-theta-0_6-pred-100.csv")),
  
  tar_target(model_3_rho_06_theta_08_pred_100,
             estimator_eval_stacked_ensemb_iterative_cv(# m_start =1, # start of portion simulations, default is 1
               m = m_sims, # end of portion of simulations,
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = gadp_func_ver_2, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )
  ),
  tar_target(model_3_rho_06_theta_08_pred_100_save,
             produce_csv(
               created_object= model_3_rho_06_theta_08_pred_100, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-gadp-rho-0_6-theta-0_8-pred-100.csv")),
  
  
  
  
  # Model 3 CGADP ----------
  ## 10 Predictors --------------
  ### rho = 0.2 ------------
  tar_target(model_3_rho_02_theta_02_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_02_theta_02_pred_10.cgadp_save,
             produce_csv(
               created_object= model_3_rho_02_theta_02_pred_10.cgadp, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_2-theta-0_2-pred-10.csv")),
  
  tar_target(model_3_rho_02_theta_04_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_02_theta_04_pred_10.cgadp_save,
             produce_csv(
               created_object= model_3_rho_02_theta_04_pred_10.cgadp, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_2-theta-0_4-pred-10.csv")),
  
  tar_target(model_3_rho_02_theta_06_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_02_theta_06_pred_10.cgadp_save,
             produce_csv(
               created_object= model_3_rho_02_theta_06_pred_10.cgadp, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_2-theta-0_6-pred-10.csv")),
  
  tar_target(model_3_rho_02_theta_08_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_02_theta_08_pred_10.cgadp_save,
             produce_csv(
               created_object= model_3_rho_02_theta_08_pred_10.cgadp, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_2-theta-0_8-pred-10.csv")),
  
  ### rho = 0.4 ----------------
  tar_target(model_3_rho_04_theta_02_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_04_theta_02_pred_10.cgadp_save,
             produce_csv(
               created_object= model_3_rho_04_theta_02_pred_10.cgadp, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_4-theta-0_2-pred-10.csv")),
  
  tar_target(model_3_rho_04_theta_04_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_04_theta_04_pred_10.cgadp_save,
             produce_csv(
               created_object= model_3_rho_04_theta_04_pred_10.cgadp, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_4-theta-0_4-pred-10.csv")),
  
  tar_target(model_3_rho_04_theta_06_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_04_theta_06_pred_10.cgadp_save,
             produce_csv(
               created_object= model_3_rho_04_theta_06_pred_10.cgadp, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_4-theta-0_6-pred-10.csv")),
  
  tar_target(model_3_rho_04_theta_08_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_04_theta_08_pred_10.cgadp_save,
             produce_csv(
               created_object= model_3_rho_04_theta_08_pred_10.cgadp, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_4-theta-0_8-pred-10.csv")),
  
  ### rho = 0.6 ---------------------------------
  tar_target(model_3_rho_06_theta_02_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_06_theta_02_pred_10.cgadp_save,
             produce_csv(
               created_object= model_3_rho_06_theta_02_pred_10.cgadp, 
               base_str = model_3_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_6-theta-0_2-pred-10.csv")),
  
  tar_target(model_3_rho_06_theta_04_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_06_theta_04_pred_10.cgadp_save,
             produce_csv(
               created_object= model_3_rho_06_theta_04_pred_10.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_6-theta-0_4-pred-10.csv")),
  
  tar_target(model_3_rho_06_theta_06_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_06_theta_06_pred_10.cgadp_save,
             produce_csv(
               created_object= model_3_rho_06_theta_06_pred_10.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_6-theta-0_6-pred-10.csv")),
  
  tar_target(model_3_rho_06_theta_08_pred_10.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 10, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_06_theta_08_pred_10.cgadp_save,
             produce_csv(
               created_object= model_3_rho_06_theta_08_pred_10.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_6-theta-0_8-pred-10.csv")),
  
  ## 50 Predictors -----------------------------
  ### rho = 0.2 ------------
  tar_target(model_3_rho_02_theta_02_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_02_theta_02_pred_50.cgadp_save,
             produce_csv(
               created_object= model_3_rho_02_theta_02_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_2-theta-0_2-pred-50.csv")),
  
  tar_target(model_3_rho_02_theta_04_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_02_theta_04_pred_50.cgadp_save,
             produce_csv(
               created_object= model_3_rho_02_theta_04_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_2-theta-0_4-pred-50.csv")),
  
  tar_target(model_3_rho_02_theta_06_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_02_theta_06_pred_50.cgadp_save,
             produce_csv(
               created_object= model_3_rho_02_theta_06_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_2-theta-0_6-pred-50.csv")),
  
  tar_target(model_3_rho_02_theta_08_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_02_theta_08_pred_50.cgadp_save,
             produce_csv(
               created_object= model_3_rho_02_theta_08_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_2-theta-0_8-pred-50.csv")),
  
  ### rho = 0.4 ----------------
  tar_target(model_3_rho_04_theta_02_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_04_theta_02_pred_50.cgadp_save,
             produce_csv(
               created_object= model_3_rho_04_theta_02_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_4-theta-0_2-pred-50.csv")),
  
  tar_target(model_3_rho_04_theta_04_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_04_theta_04_pred_50.cgadp_save,
             produce_csv(
               created_object= model_3_rho_04_theta_04_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_4-theta-0_4-pred-50.csv")),
  
  tar_target(model_3_rho_04_theta_06_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_04_theta_06_pred_50.cgadp_save,
             produce_csv(
               created_object= model_3_rho_04_theta_06_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_4-theta-0_6-pred-50.csv")),
  
  tar_target(model_3_rho_04_theta_08_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_04_theta_08_pred_50.cgadp_save,
             produce_csv(
               created_object= model_3_rho_04_theta_08_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_4-theta-0_8-pred-50.csv")),
  
  ### rho = 0.6 ---------------------------------
  tar_target(model_3_rho_06_theta_02_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_06_theta_02_pred_50.cgadp_save,
             produce_csv(
               created_object= model_3_rho_06_theta_02_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_6-theta-0_2-pred-50.csv")),
  
  tar_target(model_3_rho_06_theta_04_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_06_theta_04_pred_50.cgadp_save,
             produce_csv(
               created_object= model_3_rho_06_theta_04_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_6-theta-0_4-pred-50.csv")),
  
  tar_target(model_3_rho_06_theta_06_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_06_theta_06_pred_50.cgadp_save,
             produce_csv(
               created_object= model_3_rho_06_theta_06_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_6-theta-0_6-pred-50.csv")),
  
  tar_target(model_3_rho_06_theta_08_pred_50.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 50, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_06_theta_08_pred_50.cgadp_save,
             produce_csv(
               created_object= model_3_rho_06_theta_08_pred_50.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_6-theta-0_8-pred-50.csv")),
  
  ## 100 Predictors -----------------------------
  ### rho = 0.2 ------------
  tar_target(model_3_rho_02_theta_02_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_02_theta_02_pred_100.cgadp_save,
             produce_csv(
               created_object= model_3_rho_02_theta_02_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_2-theta-0_2-pred-100.csv")),
  
  tar_target(model_3_rho_02_theta_04_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_02_theta_04_pred_100.cgadp_save,
             produce_csv(
               created_object= model_3_rho_02_theta_04_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_2-theta-0_4-pred-100.csv")),
  
  tar_target(model_3_rho_02_theta_06_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_02_theta_06_pred_100.cgadp_save,
             produce_csv(
               created_object= model_3_rho_02_theta_06_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_2-theta-0_6-pred-100.csv")),
  
  tar_target(model_3_rho_02_theta_08_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.2, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_02_theta_08_pred_100.cgadp_save,
             produce_csv(
               created_object= model_3_rho_02_theta_08_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_2-theta-0_8-pred-100.csv")),
  
  ### rho = 0.4 ----------------
  tar_target(model_3_rho_04_theta_02_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_04_theta_02_pred_100.cgadp_save,
             produce_csv(
               created_object= model_3_rho_04_theta_02_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_4-theta-0_2-pred-100.csv")),
  
  tar_target(model_3_rho_04_theta_04_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_04_theta_04_pred_100.cgadp_save,
             produce_csv(
               created_object= model_3_rho_04_theta_04_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_4-theta-0_4-pred-100.csv")),
  
  tar_target(model_3_rho_04_theta_06_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_04_theta_06_pred_100.cgadp_save,
             produce_csv(
               created_object= model_3_rho_04_theta_06_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_4-theta-0_6-pred-100.csv")),
  
  tar_target(model_3_rho_04_theta_08_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.4, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_04_theta_08_pred_100.cgadp_save,
             produce_csv(
               created_object= model_3_rho_04_theta_08_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_4-theta-0_8-pred-100.csv")),
  
  ### rho = 0.6 ---------------------------------
  tar_target(model_3_rho_06_theta_02_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.2, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_06_theta_02_pred_100.cgadp_save,
             produce_csv(
               created_object= model_3_rho_06_theta_02_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_6-theta-0_2-pred-100.csv")),
  
  tar_target(model_3_rho_06_theta_04_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.4, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_06_theta_04_pred_100.cgadp_save,
             produce_csv(
               created_object= model_3_rho_06_theta_04_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_6-theta-0_4-pred-100.csv")),
  
  tar_target(model_3_rho_06_theta_06_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.6, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_06_theta_06_pred_100.cgadp_save,
             produce_csv(
               created_object= model_3_rho_06_theta_06_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_6-theta-0_6-pred-100.csv")),
  
  tar_target(model_3_rho_06_theta_08_pred_100.cgadp, 
             estimator_evaluation_stacked_ensemb_cgadp_cv(# m_start =param_df$sim_start[row_index], # start of portion simulations
               m = m_sims, # end of portion of simulations
               samp_n = n_sizes, # sample sizes to be used for training set
               cores = core_alloc, # Number of CPU threads/ logical processors to use
               response_prefix = "y", # prefix for response variables
               acquire_variable_names = get_variables, #Function for getting the variables
               perturbation_func = model_1_cgadp, # CGADP or GADP function
               theta_value = 0.8, # privacy parameter value
               model_generating_function = correlate_uniform_m.3_p_ver_2, # function to generate the model
               testing_data_seed = 30000, # Seed for the testing set
               model_predictor_quant = 100, # Quantity of the predictors for the model
               model_rho = 0.6, # correlation of predictors in the model
               model_noise = 0.2, # Noise to be used in the model
               testing_n = 500 # size of testing set
             )),
  tar_target(model_3_rho_06_theta_08_pred_100.cgadp_save,
             produce_csv(
               created_object= model_3_rho_06_theta_08_pred_100.cgadp, 
               base_str = model_1_sim_runs_directory, 
               file_name_str = "model-3-cgadp-rho-0_6-theta-0_8-pred-100.csv")),
  
  
  # CE ------------
  # All CE data is ran with training set size set to 1000.
  tar_target(CE_data_gadp_and_cgadp, get_CE_data()),
  
  ## GADP -----
  tar_target(CE_gadp_theta_02,
             estimator_eval_stacked_ensemb_cv_stag(model_data = CE_data_gadp_and_cgadp, # the dataset
                                                   m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                   m = m_sims, # Number of Monte Carlo simulations
                                                   samp_n = c(1000), # sample sizes to be used
                                                   cores = core_alloc,
                                                   response_prefix = "y",
                                                   acquire_variable_names = get_variables,
                                                   
                                                   perturbation_func = gadp_func_ver_2,
                                                   theta_value = 0.2)),
  tar_target(CE_gadp_theta_04,
             estimator_eval_stacked_ensemb_cv_stag(model_data = CE_data_gadp_and_cgadp, # the dataset
                                                   m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                   m = m_sims, # Number of Monte Carlo simulations
                                                   samp_n = c(1000), # sample sizes to be used
                                                   cores = core_alloc,
                                                   response_prefix = "y",
                                                   acquire_variable_names = get_variables,
                                                   
                                                   perturbation_func = gadp_func_ver_2,
                                                   theta_value = 0.4)),
  
  tar_target(CE_gadp_theta_06,
             estimator_eval_stacked_ensemb_cv_stag(model_data = CE_data_gadp_and_cgadp, # the dataset
                                                   m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                   m = m_sims, # Number of Monte Carlo simulations
                                                   samp_n = c(1000), # sample sizes to be used
                                                   cores = core_alloc,
                                                   response_prefix = "y",
                                                   acquire_variable_names = get_variables,
                                                   
                                                   perturbation_func = gadp_func_ver_2,
                                                   theta_value = 0.6)),
  
  tar_target(CE_gadp_theta_08,
             estimator_eval_stacked_ensemb_cv_stag(model_data = CE_data_gadp_and_cgadp, # the dataset
                                                   m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                   m = m_sims, # Number of Monte Carlo simulations
                                                   samp_n = c(1000), # sample sizes to be used
                                                   cores = core_alloc,
                                                   response_prefix = "y",
                                                   acquire_variable_names = get_variables,
                                                   
                                                   perturbation_func = gadp_func_ver_2,
                                                   theta_value = 0.8)),
  
  tar_target(CE_GADP_save,
             produce_csv(
               created_object= bind_rows(list(CE_gadp_theta_02,
                                              CE_gadp_theta_04,
                                              CE_gadp_theta_06,
                                              CE_gadp_theta_08)), 
               base_str = CE_sim_runs_directory, 
               file_name_str = "gadp-sims-CE-outlier-rem-ver-1.csv")),
  
  ## CGADP ----------
  tar_target(CE_cgadp_theta_02,
             estimator_evaluation_stacked_ensemb_cgadp_cv_stag(model_data = CE_data_gadp_and_cgadp, # the dataset
                                                   m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                   m = m_sims, # Number of Monte Carlo simulations
                                                   samp_n = c(1000), # sample sizes to be used
                                                   cores = core_alloc,
                                                   response_prefix = "y",
                                                   acquire_variable_names = get_variables,
                                                   
                                                   perturbation_func = real_CE_cgadp_ver_2,
                                                   theta_value = 0.2)),
  tar_target(CE_cgadp_theta_04,
             estimator_eval_stacked_ensemb_cv_stag(model_data = CE_data_gadp_and_cgadp, # the dataset
                                                   m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                   m = m_sims, # Number of Monte Carlo simulations
                                                   samp_n = c(1000), # sample sizes to be used
                                                   cores = core_alloc,
                                                   response_prefix = "y",
                                                   acquire_variable_names = get_variables,
                                                   
                                                   perturbation_func = real_CE_cgadp_ver_2,
                                                   theta_value = 0.4)),
  
  tar_target(CE_cgadp_theta_06,
             estimator_eval_stacked_ensemb_cv_stag(model_data = CE_data_gadp_and_cgadp, # the dataset
                                                   m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                   m = m_sims, # Number of Monte Carlo simulations
                                                   samp_n = c(1000), # sample sizes to be used
                                                   cores = core_alloc,
                                                   response_prefix = "y",
                                                   acquire_variable_names = get_variables,
                                                   
                                                   perturbation_func = real_CE_cgadp_ver_2,
                                                   theta_value = 0.6)),
  
  tar_target(CE_cgadp_theta_08,
             estimator_eval_stacked_ensemb_cv_stag(model_data = CE_data_gadp_and_cgadp, # the dataset
                                                   m_start = 1, # Primarily used for debugging determines which iteration to start from
                                                   m = m_sims, # Number of Monte Carlo simulations
                                                   samp_n = c(1000), # sample sizes to be used
                                                   cores = core_alloc,
                                                   response_prefix = "y",
                                                   acquire_variable_names = get_variables,
                                                   
                                                   perturbation_func = real_CE_cgadp_ver_2,
                                                   theta_value = 0.8)),
  tar_target(CE_CGADP_save,
             produce_csv(
               created_object= bind_rows(list(CE_cgadp_theta_02,
                                              CE_cgadp_theta_04,
                                              CE_cgadp_theta_06,
                                              CE_cgadp_theta_08)), 
               base_str = CE_sim_runs_directory, 
               file_name_str = "cgadp-sims-CE-outlier-rem-ver-1-n-1000.csv"))
  
  # End ---------------
)